{
        Program Name : DBBinary
        Package Name : LB
        Purpose      : Database implemented as a binary tree
        Institution  : Conroy & Conroy Co.
        Date Written : 11-April-2002
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2002 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit DBBinary ;

interface

uses Database ; { TDatabase }

const Binary_DatabaseErr_Facility = 51 ;
      Binary_DatabaseErr_Success = 0 ;
      Binary_DatabaseErr_Record_Not_Found = 1 ;

type PBinary_Record = ^TBinary_Record ;
     TBinary_Database = object( TISAM_DataBase )
                            private { Instance data... }
                                Balance : boolean ; { True if we should keep the tree balanced }
                                _Root : PBinary_Record ;
                                Last_Visited_Node : PBinary_Record ;
                                Need_Rebuild : array[ 0..255 ] of boolean ;

                            public { Constructors and destructors... }
                                constructor Init ;

                            private { Internal utility routines... }
                                procedure Build_Index( Index : longint ) ;
(*
                                procedure LL_Rotation( Rec : PBinary_Record ) ;
                                procedure LR_Rotation( Rec : PBinary_Record ) ;
                                procedure RR_Rotation( Rec : PBinary_Record ) ;
                                procedure RL_Rotation( Rec : PBinary_Record ) ;
*)

                            protected { Property handlers... }
                                function Get_Root( Index : integer ) : PBinary_Record ;
                                    virtual ;
                                    
                                procedure Set_Root( Index : integer ;
                                    Value : PBinary_Record ) ; virtual ;

                            public { API... }
                                function Add( Rec : PDatabase_Record ) : PDatabase_Record ;
                                    virtual ;
                                { Adds the record to the database.  If a record
                                  with the matching key is found and duplicates
                                  are not allowed, it returns the existing
                                  record.  Otherwise, it returns the passed
                                  record. }

                                procedure Delete( Index : longint ;
                                    const Key : string ) ; virtual ;

                                function Get( Index : longint ;
                                    const Key : string ) : PDatabase_Record ;
                                    virtual ;

                                function Get_GE( Index : longint ;
                                    const Key : string ) : PDatabase_Record ;
                                    virtual ;

                                function Get_LE( Index : longint ;
                                    const Key : string ) : PDatabase_Record ;
                                    virtual ;

                                function Next( Index : longint ;
                                    Rec : PDatabase_Record ) : PDatabase_Record ;
                                    virtual ;

                                function Previous( Index : longint ;
                                    Rec : PDatabase_Record ) : PDatabase_Record ;
                                    virtual ;

                                procedure Zero ; virtual ;

                                property Root[ index : integer ] : PBinary_Record
                                             read Get_Root
                                             write Set_Root ;
                        end ; { TBinary_Database }
     PBinary_Database = ^TBinary_Database ;
     TBinary_Record = object( TDatabase_Record )
                          private { Instance data... }
                              Height : longint ; { Height of subtree }
                              Balance : array[ 0..2 ] of integer ;
                              Left, Right, Parent : array[ 0..2 ] of PBinary_Record ;
                      end ;

implementation

{ TBinary_Database methods... }

{ Constructors and destructors... }

constructor TBinary_Database.Init ;

begin
    inherited Init ;

    Root[ 0 ] := nil ;
    Balance := False ;
    fillchar( Need_Rebuild, sizeof( Need_Rebuild ), 0 ) ;
end ;


procedure AdjustHeight( Rec : PBinary_Record ) ;

begin
end ;


{ Internal utility routines... }

procedure TBinary_Database.Build_Index( Index : longint ) ;

    procedure _Build( Rec, Parent : PBinary_Record ) ;

    var L, R : PBinary_Record ;

    begin
        Rec^.Parent[ Index ] := Parent ;
        L := Rec^.Left[ Index ] ;
        R := Rec^.Right[ Index ] ;
        if( L <> nil ) then
        begin
            _Build( L, Rec ) ;
        end ;
        if( R <> nil ) then
        begin
            _Build( R, Rec ) ;
        end ;
    end ;

begin
    Need_Rebuild[ Index ] := False ;
    _Build( Root[ Index ], nil ) ;
end ;


function TBinary_Database.Get_Root( Index : integer ) : PBinary_Record ;

begin
    Get_Root := _Root ;
end ;


procedure TBinary_Database.Set_Root( Index : integer ; Value : PBinary_Record ) ;

begin
    _Root := Value ;
end ;


(*
procedure TBinary_Database.LL_Rotation( Rec : PBinary_Record ) ;

{var Temp : PBinary_Record ;}

begin
    if( Rec = nil ) then
    begin
        exit ;
    end ;
    Temp := Rec^.Left ;
    Rec^.Left := Temp^.Right ;
    Temp^.Right := Rec ;
    if( Root = Rec ) then
    begin
        Root := Temp ;
    end else
    if( Rec^.Parent^.Left = Rec ) then
    begin
        Rec^.Parent^.Left := Temp ;
    end else
    begin
        Rec^.Parent^.Right := Temp ;
    end ;
    AdjustHeight( Rec^.Right ) ;
end ;


procedure TBinary_Database.LR_Rotation( Rec : PBinary_Record ) ;

begin
    if( Rec = nil ) then
    begin
        exit ;
    end ;
    RR_Rotation( Rec^.Left ) ;
    LL_Rotation( Rec ) ;
end ;


procedure TBinary_Database.RR_Rotation( Rec : PBinary_Record ) ;

{var Temp, Temp_R : PBinary_Record ;}

begin
    if( Rec = nil ) then
    begin
        exit ;
    end ;
    Temp := Rec^.Right ;
    Temp_R := Temp^.Right ;
    Temp^.Right := Rec ;
    Rec^.Right := Temp_R ;
    if( Root = Rec ) then
    begin
        Root := Temp ;
    end else
    if( Rec^.Parent^.Left = Rec ) then
    begin
        Rec^.Parent^.Left := Temp ;
    end else
    begin
        Rec^.Parent^.Right := Temp ;
    end ;
    AdjustHeight( Rec^.Right ) ;
end ;


procedure TBinary_Database.RL_Rotation( Rec : PBinary_Record ) ;

begin
    if( Rec = nil ) then
    begin
        exit ;
    end ;
    LL_Rotation( Rec^.Right ) ;
    RR_Rotation( Rec ) ;
end ;
*)


{ API... }

{ Adds the record to the database.  If a record with the matching key is found
  and duplicates are not allowed, it returns the existing record.  Otherwise, it
  returns the passed record. }
function TBinary_Database.Add( Rec : PDatabase_Record ) : PDatabase_Record ;

var Index : longint ;
    Found_Duplicate : boolean ;
    _Rec : PBinary_Record ;
    Rec_Key : string ; { Key of passed record }
    This_Key : string ; { Key of current node }

(*
    procedure Retrieve_Heights( _Rec : PBinary_Record ;
        var Left_Height, Right_Height : longint ) ;

    begin
        if( _Rec^.Left = nil ) then
        begin
            Left_Height := 1 ;
        end else
        begin
            Left_Height := _Rec^.Left^.Height ;
        end ;
        if( _Rec^.Right = nil ) then
        begin
            Right_Height := 1 ;
        end else
        begin
            Right_Height := _Rec^.Right^.Height ;
        end ;
    end ;

var _Rec : PBinary_Record ;
    Factor : integer ;
    This_Key : string ;
    Left_Height, Right_Height : longint ;

begin
    { Setup... }
    Index := 0 ;
    Set_Error( 0, 0 ) ;
    _Rec := PBinary_Record( Rec ) ;

    { Add the record... }
    if( Root = nil ) then { First record }
    begin
        Root := _Rec ;
        _Rec^.Left := nil ;
        _Rec^.Right := nil ;
	_Rec^.Parent := nil ;
        _Rec^.Height := 1 ;
        exit ;
    end ;

    This_Key := Rec^.Get_Key( Index ) ;
    _Rec := PBinary_Record( Get( Index, This_Key ) ) ;
    if( _Rec <> nil ) then { Duplicate key }
    begin
	{ Not yet implemented }
        exit ;
    end ;

    PBinary_Record( Rec )^.Left := nil ;
    PBinary_Record( Rec )^.Right := nil ;
    PBinary_Record( Rec )^.Database := @self ;

    _Rec := Last_Visited_Node ;
    if( This_Key > _Rec^.Get_Key( Index ) ) then
    begin
        _Rec^.Right := PBinary_Record( Rec ) ;
    end else
    begin
        _Rec^.Left := PBinary_Record( Rec ) ;
    end ;
    PBinary_Record( Rec )^.Parent := _Rec ;
    PBinary_Record( Rec )^.Height := 1 ;
    if( Balance ) then
    begin
        if( ( _Rec^.Left <> nil ) and ( _Rec^.Right <> nil ) ) then
        begin
            exit ; { Didn't change height of subtree }
        end ;

        { Adjust heights... }
        inc( _Rec^.Height ) ;
        while( _Rec^.Parent <> nil ) do
        begin
            _Rec := _Rec^.Parent ;
            Retrieve_Heights( _Rec, Left_Height, Right_Height ) ;
            Factor := Left_Height - Right_Height ;
            if( Factor > 0 ) then { Left subtree is deeper than right subtree }
            begin
                Retrieve_Heights( _Rec^.Left, Left_Height, Right_Height ) ;
                if( Left_Height <> Right_Height ) then
                begin
                    if( Left_Height > Right_Height ) then
                    begin
                        LL_Rotation( _Rec ) ;
                    end else
                    begin
                        LR_Rotation( _Rec ) ;
                    end ;
		break ;
                end ;
            end else
            if( Factor < 0 ) then
            begin
                Retrieve_Heights( _Rec^.Right, Left_Height, Right_Height ) ;
                if( Left_Height <> Right_Height ) then
                begin
                    if( Left_Height > Right_Height ) then
                    begin
                        RL_Rotation( _Rec ) ;
                    end else
                    begin
                        RR_Rotation( _Rec ) ;
                    end ;
                    break ;
                end ;
            end ;
            inc( _Rec^.Height ) ;
        end ;
    end ;
*)
    procedure _Add( var p : PBinary_Record ; var h : boolean ) ;

    var p1, p2 : PBinary_Record ;

    begin
        if( p = nil ) then
        begin { Record is not in tree; insert it }
            h := true ;
            p := _Rec ;
            p^.Height := 1 ;
            p^.Left[ Index ] := nil ;
            p^.Right[ Index ] := nil ;
            p^.Balance[ Index ] := 0 ;
        end else
        begin
            This_Key := p^.Get_Key( Index ) ;
            if( Rec_Key < This_Key ) then
            begin
                _Add( P^.Left[ Index ], h ) ;
                if( Found_Duplicate ) then
                begin
                    exit ; { Early out }
                end ;
{$IFDEF DEBUG}
                This_Key := p^.Get_Key( Index ) ;
                if( This_Key < P^.Left[ Index ]^.Get_Key( Index ) ) then
                begin
                    halt ;
                end ;
                if( ( P^.Right[ Index ] <> nil ) and ( This_Key > P^.Right[ Index ]^.Get_Key( Index ) ) ) then
                begin
                    halt ;
                end ;
{$ENDIF}
                if( h ) then { Left branch has grown higher }
                begin
                    case p^.Balance[ Index ] of
                        1 : begin
                                p^.Balance[ Index ] := 0 ;
                                h := false ;
                            end ;
                        0 : p^.Balance[ index ] := -1 ;
                       -1 : begin { Rebalance }
                               p1 := p^.Left[ Index ] ;
                               if( p1^.Balance[ Index ] = -1 ) then
                               begin { Single LL rotation }
                                   p^.Left[ Index ] := p1^.Right[ Index ] ;
                                   p1^.Right[ Index ] := p ;
                                   p^.Balance[ Index ] := 0 ;
                                   p := p1 ;
                               end else
                               begin { Double LR rotation }
                                   p2 := p1^.Right[ Index ] ;
                                   p1^.Right[ Index ] := p2^.Left[ Index ] ;
                                   p2^.Left[ Index ] := p1 ;
                                   p^.Left[ Index ] := p2^.Right[ Index ] ;
                                   p2^.Right[ Index ] := p ;
                                   if( p2^.Balance[ Index ] = -1 ) then
                                   begin
                                       p^.Balance[ Index ] := 1 ;
                                   end else
                                   begin
                                       p^.Balance[ Index ] := 0 ;
                                   end ;
                                   if( p2^.Balance[ Index ] = 1 ) then
                                   begin
                                       p1^.Balance[ Index ] := -1 ;
                                   end else
                                   begin
                                       p1^.Balance[ Index ] := 0 ;
                                   end ;
                                   p := p2 ;
                               end ; { if( p1^.Balance[ Index ] = -1 ) }
                               p^.Balance[ Index ] := 0 ;
                               h := false ;
                            end ; { -1 }
                    end ; { case p^.Balance[ Index ] }
                end ; { if( h ) }
            end else
            if( Rec_Key > This_Key ) then
            begin
                _Add( P^.Right[ Index ], h ) ;
                if( Found_Duplicate ) then
                begin
                    exit ; { Early out }
                end ;
{$IFDEF DEBUG}
                This_Key := p^.Get_Key( Index ) ;
                if( This_Key > P^.Right[ Index ].Get_Key( Index ) ) then
                begin
                    halt ;
                end ;
                if( ( P^.Left[ Index ] <> nil ) and ( This_Key < P^.Left[ Index ]^.Get_Key( Index ) ) ) then
                begin
                    halt ;
                end ;
{$ENDIF}
                if( h ) then { Right branch has grown higher }
                begin
                    case p^.Balance[ Index ] of
                        -1 : begin
                                 p^.Balance[ Index ] := 0 ;
                                 h := false ;
                             end ;
                         0 : p^.Balance[ Index ] := 1 ;
                         1 : begin { Rebalance }
                                p1 := p^.Right[ Index ] ;
                                if( p1^.Balance[ Index ] = 1 ) then
                                begin { Single RR rotation }
                                    p^.Right[ Index ] := p1^.Left[ Index ] ;
                                    p1^.Left[ Index ] := p ;
                                    p^.Balance[ Index ] := 0 ;
                                    p := p1 ;
                                end else
                                begin { Double RL rotation }
                                   p2 := p1^.Left[ Index ] ;
                                   p1^.Left[ Index ] := p2^.Right[ Index ] ;
                                   p2^.Right[ Index ] := p1 ;
                                   p^.Right[ Index ] := p2^.Left[ Index ] ;
                                   p2^.Left[ Index ] := p ;
                                   if( p2^.Balance[ Index ] = 1 ) then
                                   begin
                                       p^.Balance[ Index ] := -1 ;
                                   end else
                                   begin
                                       p^.Balance[ Index ] := 0 ;
                                   end ;
                                   if( p2^.Balance[ Index ] = -1 ) then
                                   begin
                                       p1^.Balance[ Index ] := 1 ;
                                   end else
                                   begin
                                       p1^.Balance[ Index ] := 0 ;
                                   end ;
                                   p := p2 ;
                                end ;
                                p^.Balance[ Index ] := 0 ;
                                h := false ;
                             end ; { 1 }
                    end ; { case p^.Balance[ Index ] }
                end ; { if( h ) }
            end else
            begin { They are equal }
                if( not Duplicates_Allowed( Index ) ) then
                begin
                    Add := p ;
                    Found_Duplicate := True ;
                end else
                begin
                    p^.Height := p^.Height + 1 ;
                    h := false ;
                end ;
            end ;
        end ;
    end ; { _Add }

var B : boolean ;
    R : PBinary_Record ;

begin
    _Rec := PBinary_Record( Rec ) ;
    Add := _Rec ; { Assume a new record }
    for Index := 0 to Key_Count - 1 do { Add to indexes }
    begin
        _Rec^.Left[ Index ] := nil ;
        _Rec^.Right[ Index ] := nil ;
        _Rec^.Parent[ Index ] := nil ;
        Rec_Key := Rec^.Get_Key( Index ) ;
        R := Root[ Index ] ;
        Found_Duplicate := False ;
        _Add( R, B ) ;
        if( not Found_Duplicate ) then
        begin
            Root[ Index ] := R ;
            Need_Rebuild[ Index ] := True ;
        end ;
    end ; { for Index := 0 to Key_Count - 1 }
end ; { TBinary_Database.Add }


procedure TBinary_Database.Delete( Index : longint ; const Key : string ) ;

var X : PBinary_Record ;
    Rec_Key : string ;

(*
var Rec, _Rec : PBinary_Record ;

begin
    { Find the record... }
    Rec := PBinary_Record( Get( Index, Key ) ) ;
    if( Rec = nil ) then
    begin
        exit ;
    end ;

    Set_Error( 0, 0 ) ;
    if( Rec^.Right = nil ) then { Simple case }
    begin
        if( Rec = Root ) then
        begin
            Root := Rec^.Left ;
        end else
        if( Rec^.Parent^.Left = Rec ) then
        begin
            Rec^.Parent^.Left := Rec^.Left ;
        end else
        begin
            Rec^.Parent^.Right := Rec^.Left ;
        end ;
        dispose( Rec, Done ) ;
        exit ;
    end ;

    if( Rec = Root ) then
    begin
        Root := Rec^.Right ;
    end else
    if( Rec^.Parent^.Left = Rec ) then
    begin
        Rec^.Parent^.Left := Rec^.Right ;
    end else
    begin
        Rec^.Parent^.Right := Rec^.Right ;
    end ;
    if( Rec^.Left <> nil ) then { Need to attach left branch }
    begin
        _Rec := Rec^.Right ;
        while( _Rec^.Left <> nil ) do
        begin
            _Rec := _Rec^.Left ;
        end ;
        _Rec^.Left := Rec^.Left ;
    end ;

    dispose( Rec, Done ) ;
*)
    procedure _Delete( P : PBinary_Record ; var h : boolean ) ;

    var q : PBinary_Record ;

        procedure Balance1( var p : PBinary_Record ; var h : boolean ) ;

        var p1, p2 : PBinary_Record ;
            b1, b2 : integer ;

        begin { h = true, left branch has become less high }
           case p^.Balance[ Index ] of
               -1 : p^.Balance[ Index ] := 0 ;
                0 : begin
                        p^.Balance[ Index ] := 1 ;
                        h := false ;
                    end ;
                1 : begin { Rebalance }
                       p1 := p^.Right[ Index ] ;
                       b1 := p1^.Balance[ Index ] ;
                       if( b1 >= 0 ) then
                       begin  { Single RR rotation }
                           p^.Right[ Index ] := p1^.Left[ Index ] ;
                           p1^.Left[ Index ] := p ;
                           if( b1 = 0 ) then
                           begin
                               p1^.Balance[ Index ] := 1 ;
                               p1^.Balance[ Index ] := -1 ;
                               h := false ;
                           end else
                           begin
                               p^.Balance[ Index ] := 0 ;
                               p1^.Balance[ Index ] := 0 ;
                           end ;
                           p := p1 ;
                       end else
                       begin { Double RL rotation }
                           p2 := p1^.Left[ Index ] ;
                           b2 := p2^.Balance[ Index ] ;
                           p1^.Left[ Index ] := p2^.Right[ Index ] ;
                           p2^.Right[ Index ] := p1 ;
                           p^.Right[ Index ] := p2^.Left[ Index ] ;
                           p2^.Left[ Index ] := p ;
                           if( b2 = 1 ) then
                           begin
                               p^.Balance[ Index ] := -1 ;
                           end else
                           begin
                               p^.Balance[ Index ] := 0 ;
                           end ;
                           if( b2 = -1 ) then
                           begin
                               p1^.Balance[ Index ] := 1 ;
                           end else
                           begin
                               p1^.Balance[ Index ] := 0 ;
                           end ;
                           p := p2 ;
                           p2^.Balance[ Index ] := 0 ;
                       end ;
                    end ;
           end
        end ; { balance 1 }


        procedure Balance2( var p : PBinary_Record ; var h : boolean ) ;

        var p1, p2 : PBinary_Record ;
            b1, b2 : integer ;

        begin { h = true, right branch has become less high }
           case p^.Balance[ Index ] of
              1 : p^.Balance[ Index ] := 0 ;
              0 : begin
                      p^.Balance[ Index ] := -1 ;
                      h := false ;
                  end ;
             -1 : begin { Rebalance }
                     p1 := p^.Left[ Index ] ;
                     b1 := p1^.Balance[ Index ] ;
                     if( b1 <= 0 ) then
                     begin { Single LL rotation }
                        p^.Left[ Index ] := p1^.Right[ Index ] ;
                        p1^.Right[ Index ] := p ;
                        if( b1 = 0 ) then
                        begin
                            p^.Balance[ Index ] := -1 ;
                            p1^.Balance[ Index ] := 1 ;
                            h := false ;
                        end else
                        begin
                            p^.Balance[ Index ] := 0 ;
                            p1^.Balance[ Index ] := 0 ;
                        end ;
                        p := p1 ;
                     end else
                     begin { Double LR rotation }
                         p2 := p1^.Right[ Index ] ;
                         b2 := p2^.Balance[ Index ] ;
                         p1^.Right[ Index ] := p2^.Left[ Index ] ;
                         p2^.Left[ Index ] := p1 ;
                         p^.Left[ Index ] := p2^.Right[ Index ] ;
                         p2^.Right[ Index ] := p ;
                         if( b2 = -1 ) then
                         begin
                             p^.Balance[ Index ] := 1 ;
                         end else
                         begin
                             p^.Balance[ Index ] := 0 ;
                         end ;
                         if( b2 = 1 ) then
                         begin
                             p1^.Balance[ Index ] := -1 ;
                         end else
                         begin
                             p1^.Balance[ Index ] := 0 ;
                         end ;
                         p := p2 ;
                         p2^.Balance[ Index ] := 0 ;
                     end ;
                  end ;
           end ; { case }
        end ; { Balance2 }


        procedure Del( var r : PBinary_Record ; var h : boolean ) ;

        var _R : PBinary_Record ;

        begin
            if( r^.Right[ Index ] <> nil ) then
            begin
                _R := R^.Right[ Index ] ;
                del( _R, h ) ;
                R^.Right[ Index ] := _R ;
                if( h ) then
                begin
                    balance2( r, h ) ;
                end ;
            end else
            begin
                {q^.Get_key( Index ) := r^.Get_key( Index ) ;}
                q^.Height := r^.Height ;
                r := r^.Left[ Index ] ;
                h := true ;
            end ;
        end ;

    var L : PBinary_Record ;

    begin { _Delete }
        if( Rec_Key < p^.Get_key( Index ) ) then
        begin
            _Delete( p^.left[ Index ], h ) ;
            if( h ) then
            begin
                balance1( p, h ) ;
            end ;
        end else
        if( Rec_Key > p^.Get_key( Index ) ) then
        begin
            _Delete( p^.right[ Index ], h ) ;
            if h then
            begin
                balance2( p, h ) ;
            end ;
        end else
        begin
            q := p ;
            if( q^.Right[ Index ] = nil ) then
            begin
                p := q^.Left[ Index ] ;
                h := true ;
            end else
            if( q^.Left[ Index ] = nil ) then
            begin
                p := q^.Right[ Index ] ;
                h := true ;
            end else
            begin
                L := Q^.Left[ Index ] ;
                del( L, h ) ;
                Q^.Left[ Index ] := L ;
                if h then
                begin
                    balance1( p, h ) ;
                end ;
            end ;
        end ;
    end ; { TBinary_Database.Delete._Delete }

var B : boolean ;

begin { TBinary_Database.Delete }
    X := PBinary_Record( Get( Index, Key ) ) ;
    if( X = nil ) then
    begin
        exit ;
    end ;
    Set_Error( 0, 0 ) ;
    for Index := 0 to Key_Count - 1 do
    begin
        Rec_Key := X^.Get_Key( Index ) ;
        _Delete( Root[ Index ], B ) ;
        Need_Rebuild[ Index ] := True ;
    end ;
end ; { TBinary_Database.Delete }


function TBinary_Database.Get( Index : longint ;
    const Key : string ) : PDatabase_Record ;

var Current : PBinary_Record ;
    This_Key : string ;

begin
    Current := Root[ Index ] ;
    while( Current <> nil ) do
    begin
        This_Key := Current^.Get_Key( Index ) ;
        if( This_Key = Key ) then
        begin
            Set_Error( 0, 0 ) ;
            Get := PDatabase_Record( Current ) ;
            exit ;
        end ;
        Last_Visited_Node := Current ;
        if( This_Key > Key ) then
        begin
            Current := Current^.Left[ Index ] ;
        end else
        begin
            Current := Current^.Right[ Index ] ;
        end ;
    end ;

    { Lookup failed... }
    Get := nil ;
    Set_Error( Binary_DatabaseErr_Facility, Binary_DatabaseErr_Record_Not_Found ) ;
end ; { TBinary_Database.Get }


function TBinary_Database.Get_GE( Index : longint ;
    const Key : string ) : PDatabase_Record ;

var Current : PBinary_Record ;

begin
    Current := Root[ Index ] ;
    while( Current <> nil ) do
    begin
        if( Current^.Get_Key( Index ) >= Key ) then
        begin
            Set_Error( 0, 0 ) ;
            Get_GE := PDatabase_Record( Current ) ;
            exit ;
        end ;
        Last_Visited_Node := Current ;
        Current := Current^.Right[ Index ];
    end ;

    { Lookup failed... }
    Get_GE := nil ;
    Set_Error( Binary_DatabaseErr_Facility, Binary_DatabaseErr_Record_Not_Found ) ;
end ;


function TBinary_Database.Get_LE( Index : longint ;
    const Key : string ) : PDatabase_Record ;

var Current : PBinary_Record ;

begin
    Current := Root[ Index ] ;
    while( Current <> nil ) do
    begin
        if( Current^.Get_Key( Index ) <= Key ) then
        begin
            Set_Error( 0, 0 ) ;
            Get_LE := PDatabase_Record( Current ) ;
            exit ;
        end ;
        Last_Visited_Node := Current ;
        Current := Current^.Left[ Index ] ;
    end ;

    { Lookup failed... }
    Get_LE := nil ;
    Set_Error( Binary_DatabaseErr_Facility, Binary_DatabaseErr_Record_Not_Found ) ;
end ;


function TBinary_Database.Next( Index : longint ;
    Rec : PDatabase_Record ) : PDatabase_Record ;

var L, Last : PBinary_Record ;
    _Rec : PBinary_Record ;

begin
    _Rec := PBinary_Record( Rec ) ;
    if( _Rec = nil ) then { Find first }
    begin
        _Rec := Root[ Index ] ;
        if( _Rec = nil ) then { No records in database }
        begin
            Next := nil ;
            exit ;
        end ;
        L := _Rec^.Left[ Index ] ;
        while( L <> nil ) do
        begin
            _Rec := L ;
            L := _Rec^.Left[ Index ] ;
        end ;
        Next := _Rec ;
        exit ;
    end ;
    if( _Rec^.Right[ Index ] <> nil ) then
    begin
        _Rec := _Rec^.Right[ Index ] ;
        L := _Rec^.Left[ Index ] ;
        while( L <> nil ) do
        begin
            _Rec := L ;
            L := _Rec^.Left[ Index ] ;
        end ;
    end else
    begin
        if( Need_Rebuild[ Index ] ) then
        begin
            Build_Index( Index ) ;
        end ;
        while( True ) do
        begin
            Last := _Rec ;
            _Rec := _Rec^.Parent[ Index ] ;
            if( _Rec = nil ) then
            begin
                Next := nil ;
                exit ;
            end ;
	    if( _Rec^.Left[ Index ] = Last ) then { Came from left }
            begin
                Next := _Rec ;
                exit ;
            end ;
        end ; { while( True ) }
    end ;
    Next := PDatabase_Record( _Rec ) ;
end ; { TBinary_Database.Next }


function TBinary_Database.Previous( Index : longint ;
    Rec : PDatabase_Record ) : PDatabase_Record ;

var Last : PBinary_Record ;
    R, _Rec : PBinary_Record ;

begin
    _Rec := PBinary_Record( Rec ) ;
    if( _Rec = nil ) then { Find first }
    begin
        _Rec := Root[ Index ] ;
        if( Rec = nil ) then { No records in database }
        begin
            Previous := nil ;
            exit ;
        end ;
        R := _Rec^.Right[ Index ] ;
        while( _Rec^.Right[ Index ] <> nil ) do
        begin
            _Rec := R ;
            R := _Rec^.Right[ Index ] ;
        end ;
        Previous := _Rec ;
        exit ;
    end ;
    if( _Rec^.Left[ Index ] <> nil ) then
    begin
        _Rec := _Rec^.Left[ Index ] ;
        while( _Rec^.Right[ Index ] <> nil ) do
        begin
            _Rec := _Rec^.Right[ Index ] ;
        end ;
    end else
    begin
        if( Need_Rebuild[ Index ] ) then
        begin
            Build_Index( Index ) ;
        end ;
        while( True ) do
        begin
            Last := _Rec ;
            _Rec := _Rec^.Parent[ Index ] ;
            if( _Rec = nil ) then
            begin
                Previous := nil ;
                exit ;
            end ;
	    if( _Rec^.Right[ Index ] = Last ) then { Came from right }
            begin
                Previous := _Rec ;
                exit ;
            end ;
        end ; { while( True ) }
    end ;
    Previous := Rec ;
end ; { TBinary_Database.Previous }


procedure TBinary_Database.Zero ;

var Back, Current : PBinary_Record ;
    Index : integer ;

begin
    Current := PBinary_Record( Next( 0, nil ) ) ;
    while( Current <> nil ) do
    begin
        while( Current^.Left[ 0 ] <> nil ) do
        begin
            Current := Current^.Left[ 0 ] ;
        end ;
        if( Current^.Right[ 0 ] <> nil ) then
        begin
            Current := Current^.Right[ 0 ] ;
        end else
        begin
            Back := Current^.Parent[ 0 ] ;
	    dispose( Current, Done ) ;
            if( Back <> nil ) then
            begin
                if( Back^.Left[ 0 ] = Current ) then
                begin
                    Back^.Left[ 0 ] := nil ;
                end else
                begin
                    Back^.Right[ 0 ] := nil ;
                end ;
            end ;
            Current := Back ;
        end ;
    end ;
    for Index := 0 to Key_Count - 1 do
    begin
        Root[ Index ] := nil ;
    end ;
    fillchar( Need_Rebuild, sizeof( Need_Rebuild ), 0 ) ;
end ; { TBinary_Database.Zero }



end.
