{
        Program Name : English
        Package Name : LB
        Purpose      : English language support
        Institution  : Conroy & Conroy Co.
        Date Written : 27-August-2001
        Written By   : Alan Conroy
        Version      : 1.0A

	Copyright (C) 2001 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  30-Nov-2003   EAC         Handle special prefixes in     *
        *                            Normalize.                     *   
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit contains functions to help with supporting English
	natural language processing.
}

unit English ;

interface

{$I EDEFINES.INC}

uses
{$IFDEF DELPHI}
     { Borland }
     Classes ; { TStringList }
{$ELSE}
     Collect ; { TStringList }
{$ENDIF}


{ API... }

{ Returns the passed word with each vowel (or group of contiguous vowels)
  replaced with asterisks (*), multiple consonants replaced with single
  consonents, and other minor substitutions. }
function Normalize_Word( W : string ) : string ;

{ Turns passed singular noun into a plural form. }
function Pluralize( S : string ) : string ;

{ Turns passed plural noun into a singular form. }
function Singularize( S : string ) : string ;

{ Fills the passed list with suggested alternates to the passed unknown word W.
  Note that the suggestions are not necessarily valid words and should be
  checked against a dictionary before actual use. }
procedure Suggestions( W : string ; var List : TStringList ) ;


implementation

uses { Layer 0... }
     CommonUt ; { Edit }

{ Returns the passed word with each vowel (or group of contiguous vowels)
  replaced with asterisks (*). }
function Normalize_Word( W : string ) : string ;

var C : Char ;
    Len, Loop : integer ;
    S : string ;

begin
    { Setup... }
    S := '' ;
    Len := 0 ;
    W := Edit( W, 32 ) ;

    { Do certain minor replacements... }
    { "WR*" to "R*" }
    if( copy( W, 1, 2 ) = 'WR' ) then
    begin
        W := copy( W, 2, length( W ) ) ;
    end ;
    { "TC*" to "C*" }
    if( copy( W, 1, 2 ) = 'TC' ) then
    begin
        W := copy( W, 2, length( W ) ) ;
    end ;
    { "PF*" to "F*" }
    if( copy( W, 1, 2 ) = 'PF' ) then
    begin
        W := copy( W, 2, length( W ) ) ;
    end ;
    { "TP*" to "P*" }
    if( copy( W, 1, 2 ) = 'TP' ) then
    begin
        W := copy( W, 2, length( W ) ) ;
    end ;
    { "KN" to "N" }
    Loop := pos( 'KN', W ) ;
    while( Loop > 0 ) do
    begin
        W := copy( W, 1, Loop - 1 ) + copy( W, Loop + 1, length( W ) ) ;
        Loop := pos( 'KN', W ) ;
    end ;
    { "CK" to "C" }
    Loop := pos( 'CK', W ) ;
    while( Loop > 0 ) do
    begin
        W := copy( W, 1, Loop - 1 ) + copy( W, Loop + 1, length( W ) ) ;
        Loop := pos( 'CK', W ) ;
    end ;
    { "PH" to "F" }
    Loop := pos( 'PH', W ) ;
    while( Loop > 0 ) do
    begin
        W := copy( W, 1, Loop - 1 ) + 'F' + copy( W, Loop + 2, length( W ) ) ;
        Loop := pos( 'PH', W ) ;
    end ;

    for Loop := 1 to length( W ) do
    begin
        if( W[ Loop ] = 'Z' ) then
        begin
            W[ Loop ] := 'S' ; { "Z" to "S" }
        end ;
    end ;

    { Replace vowels with "*" and remove duplicate consonents }
    for Loop := 1 to length( W ) do
    begin
        if( pos( W[ Loop ], 'AEIOUY' ) > 0 ) then { Vowel }
        begin
            if( copy( S, length( S ), 1 ) <> '*' ) then
            begin
                S := S + '*' ;
                inc( Len ) ;
            end ;
        end else
        begin
            C := W[ Loop ] ;
            if( ( Len = 0 ) or ( S[ Len ] <> C ) ) then { Not a duplicated consonent }
            begin
                S := S + C ;
                inc( Len ) ;
            end ;
        end ;
    end ;

    Normalize_Word := S ;
end ;


{ Turns passed singular noun into a plural form. }
function Pluralize( S : string ) : string ;

var Work : string ;
    Last_Three : string[ 3 ] ;
    Last_Two : string[ 2 ] ;

begin
    if( length( S ) = 0 ) then
    begin
        Pluralize := '' ;
        exit ;
    end ;
    Work := Edit( S, 32 or 8 or 128 ) ; { Uppercase, no leading/trailing spaces }
    Last_Three := copy( Work, length( Work ) - 2, 3 ) ;
    Last_Two := copy( Last_Three, 2, 2 ) ;
    if( Last_Three = 'IUM' ) then { "IUM" to "IA" }
    begin
        S := copy( Work, 1, length( Work ) - 2 ) + 'A' ;
    end else
    if( Last_Three = 'MAN' ) then { "MAN" to "MEN" }
    begin
        S := copy( Work, 1, length( Work ) - 2 ) + 'EN' ;
    end else
    if( Work[ length( Work ) ] = 'Y' ) then { "Y" to "IES" }
    begin
        Work := copy( Work, 1, length( Work ) - 1 ) + 'IES' ;
    end else
    if( Last_Two = 'LF' ) then { "LF" to "LVES" }
    begin
        Work := copy( Work, 1, length( Work ) - 1 ) + 'VES' ;
    end else
    if( Work[ length( Work ) ] = 'S' ) then
    begin
        Work := Work + 'ES' ;
    end else
    begin
        Work := Work + 'S' ;
    end ;
    Pluralize := Work ;
end ;


{ Turns passed plural noun into a singular form. }
function Singularize( S : string ) : string ;

var Work : string ;
    Last_Four : string[ 4 ] ;
    Last_Three : string[ 3 ] ;
    Last_Two : string[ 2 ] ;

begin
    if( length( S ) = 0 ) then
    begin
        Singularize := '' ;
        exit ;
    end ;
    Work := Edit( S, 32 or 8 or 128 ) ; { Uppercase, no leading/trailing spaces }
    Last_Four := copy( Work, length( Work ) - 3, 4 ) ;
    Last_Three := copy( Last_Four, 2, 3 ) ;
    Last_Two := copy( Last_Three, 2, 2 ) ;
    if( Work[ length( Work ) ] = 'I' ) then { "I" to "UM" }
    begin
        S := copy( Work, 1, length( Work ) - 1 ) + 'UM' ;
    end else
    if( Last_Two = 'IA' ) then { "IA" to "UM" }
    begin
        S := copy( Work, 1, length( Work ) - 1 ) + 'UM' ;
    end else
    if( Work[ length( Work ) ] = 'A' ) then { "A" to "UM" }
    begin
        S := copy( Work, 1, length( Work ) - 1 ) + 'UM' ;
    end else
    if( Last_Three = 'IES' ) then { "IES" to "Y" }
    begin
        Work := copy( Work, 1, length( Work ) - 3 ) + 'Y' ;
    end else
    if( Last_Three = 'MEN' ) then { "MEN" to "MAN" }
    begin
        Work := copy( Work, 1, length( Work ) - 3 ) + 'AN' ;
    end else
    if( Last_Four = 'LVES' ) then { "LVES" to "LF" }
    begin
        Work := copy( Work, 1, length( Work ) - 3 ) + 'F' ;
    end else
    if( Last_Two = 'ES' ) then { Remove trailing ES }
    begin
        Work := copy( Work, 1, length( Work ) - 2 ) ;
    end else
    if( Work[ length( Work ) ] = 'S' ) then { Remove trailing S }
    begin
        Work := copy( Work, 1, length( Work ) - 1 ) ;
    end ;
    if( Work = 'CHILDREN' ) then
    begin
        Work := 'CHILD' ;
    end ;
    Singularize := Work ;
end ; { Singularize }


{ Fills the passed list with suggested alternates to the passed unknown word W.
  Note that the suggestions are not necessarily valid words and should be
  checked against a dictionary before actual use. }
procedure Suggestions( W : string ; var List : TStringList ) ;

var Dummy : integer ;
    Last_Two, Last_Three, Last_Four : string ;
    Loop : integer ;
    
begin
    List.Clear ; { Start with a clean slate }
    if( length( W ) < 2 ) then
    begin
        exit ; { No suggestions for single letters or null strings }
    end ;
    W := Edit( W, 32 ) ;
    List.Add( W[ 2 ] + W[ 1 ] + copy( W, 3, length( W ) ) ) ; { Swap first two letters }
    if( length( W ) = 2 ) then
    begin
        exit ;
    end ;
    List.Add( copy( W, 1, length( W ) - 2 ) + W[ length( W ) ] + W[ length( W ) - 1 ] ) ; { Swap last two letters}
    if( length( W ) = 3 ) then
    begin
        exit ;
    end ;
    List.Add( W[ 2 ] + W[ 1 ] + copy( W, 3, length( W ) - 4 ) + W[ length( W ) ] + W[ length( W ) - 1 ]  ) ; { Swap first and last letter pairs }
    if( length( W ) = 4 ) then
    begin
        List.Add( W[ 1 ] + W[ 3 ] + W[ 2 ] + W[ 4 ] ) ; { Swap middle two letters }
    end ;

    { For performance... }
    Last_Two := copy( W, length( W ) - 2, 2 ) ;
    Last_Three := copy( W, length( W ) - 3, 3 ) ;
    Last_Four := copy( W, length( W ) - 4, 4 ) ;

    { Common minor substitution errors... }
    if( ( Last_Two = 'FS' ) and ( W[ length( W ) - 3 ] <> 'VES' ) ) then
    begin
        List.Add( copy( W, 1, length( W ) - 2 ) + 'LLY' ) ;
    end ;
    if( ( Last_Two = 'LY' ) and ( W[ length( W ) - 3 ] <> 'L' ) ) then
    begin
        List.Add( copy( W, 1, length( W ) - 2 ) + 'LLY' ) ;
    end ;
    if( Last_Two = 'NT' ) then
    begin
        List.Add( copy( W, 1, length( W ) - 2 ) + 'N'#39'T' ) ;
    end ;
    if( Last_Two = 'OR' ) then
    begin
        List.Add( copy( W, 1, length( W ) - 2 ) + 'OUR' ) ;
    end ;
    if( Last_Two = 'IT' ) then
    begin
        List.Add( W + 'E' ) ;
    end ;
    if( Last_Three = 'OUR' ) then
    begin
        List.Add( copy( W, 1, length( W ) - 3 ) + 'OR' ) ;
    end ;
    if( Last_Three = 'NCE' ) then
    begin
        List.Add( copy( W, 1, length( W ) - 3 ) + 'NSE' ) ;
    end ;
    if( Last_Three = 'NSE' ) then
    begin
        List.Add( copy( W, 1, length( W ) - 3 ) + 'NCE' ) ;
    end ;
    if( Last_Four = 'ABLE' ) then
    begin
        List.Add( copy( W, 1, length( W ) - 4 ) + 'IBLE' ) ;
    end ;
    if( Last_Four = 'IBLE' ) then
    begin
        List.Add( copy( W, 1, length( W ) - 4 ) + 'ABLE' ) ;
    end ;
    if( copy( W, 1, 2 ) = 'AQ' ) then
    begin
        List.Add( 'ACQ' + copy( W, 3, length( W ) ) ) ;
    end ;
    Dummy := pos( 'EI', W ) ;
    if( Dummy > 0 ) then
    begin
        List.Add( copy( W, 1, Dummy - 1 ) + 'IE' + copy( W, Dummy + 2, length( W ) ) ) ;
    end ;
    Dummy := pos( 'IE', W ) ;
    if( Dummy > 0 ) then
    begin
        List.Add( copy( W, 1, Dummy - 1 ) + 'EI' + copy( W, Dummy + 2, length( W ) ) ) ;
    end ;
    Dummy := pos( 'YED', W ) ;
    if( Dummy > 0 ) then
    begin
        List.Add( copy( W, 1, Dummy - 1 ) + 'I' + copy( W, Dummy + 1, length( W ) ) ) ;
    end ;
    Dummy := pos( 'F', W ) ;
    if( Dummy > 0 ) then
    begin
        List.Add( copy( W, 1, Dummy - 1 ) + 'PH' + copy( W, Dummy + 1, length( W ) ) ) ;
    end ;
    Dummy := pos( 'PH', W ) ;
    if( Dummy > 0 ) then
    begin
        List.Add( copy( W, 1, Dummy - 1 ) + 'F' + copy( W, Dummy + 2, length( W ) ) ) ;
    end ;
    Dummy := pos( 'K', W ) ;
    if( Dummy > 0 ) then
    begin
        List.Add( copy( W, 1, Dummy - 1 ) + 'CH' + copy( W, Dummy + 1, length( W ) ) ) ;
    end ;
    Dummy := pos( 'Q', W ) ;
    if( ( Dummy > 0 ) and ( Dummy < length( W ) ) and ( copy( W, Dummy + 1, 1 ) <> 'U' ) ) then
    begin
        List.Add( copy( W, 1, Dummy ) + 'U' + copy( W, Dummy + 1, length( W ) ) ) ;
    end ;

    { Try making double letters into singles }
    for Loop := 1 to length( W ) - 1 do
    begin
        if( W[ Loop ] = W[ Loop + 1 ] ) then
        begin
            List.Add( copy( W, 1, Loop ) + copy( W, Loop + 1, length( W ) ) ) ;
        end ;
    end ;
end ; { Suggestions}


end.
