{$N+}
{
              Program Name : Files
              Package Name : LB
              Purpose      : File routines
              Institution  : Conroy & Conroy Co.
              Date Written : 30-Nov-88
              Written By   : Alan Conroy
              Version      : 2.0D

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        Copyright (C) 1988-2002 by Conroy & Conroy.  Released to the public domain.

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *  20-Jan-89    EAC            Fix imcompatibility with *
        *                              DOS V3.n APPEND feature. *
        *                                                       *
        *  22-Oct-97	EAC	       Added Explicit.		*
        *                                                       *
        *  10-Aug-98    EAC            Added TWin32_File.       *
        *                                                       *
        *  26-Apr-00    EAC            Added Get_Temp_File_Name.*
        *                                                       *
        *  23-Aug-2002  EAC            Allow DOS FS findfirst   *
        *                              to find non-wildcard     *
        *                              directories.             *
        *                              Include device in lookups*
        *                              Don't throw away the     *
        *                              separator in Find_Record *
        *                              in Lookup.               *
        *  29-Oct-2002  EAC            Fix bug with lookups on  *
        *			       DOS files.               *
	*			       Use TObject from Common  *
	*			       and Boolean_To_String	*
	*			       from ustrings.		*
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit defines the TSFile, TFile_System, TWin32_File, and
        TWin32_File_System objects, and various file functions.
}

unit Files ;

interface

{$I EDEFINES.INC}

{$IFDEF DELPHI}
{$DEFINE WINDOWS}
{$ENDIF}

uses Common, { TBase_Object }
     _DebugIn, { PDebug_Interface }
     DebugInt,
     Stores, { TSBase_Store }
     _Stores, { TStore_Address }
     TypeDefs, { TRecord_Size }
     _UEHDefs, { TUEC }
{$IFDEF WINDOWS}
     Collect, { TCollection }
  {$IFDEF DELPHI}
     SysUtils ;
  {$ELSE}
     Windos,
     strings,
     Objects;
  {$ENDIF}

type FileRec = TFileRec ;

{$ELSE}
     Dos,
     Windos,
     Objects,
     Strings,
     Collect, { TCollection }
     UEHDefs ; { TUEC }
{$ENDIF}

const File_SystemErr_Facility = 59 ;
      File_SystemErr_Success = 0 ;
      File_SystemErr_Invalid_Handle = 1 ;
      File_SystemErr_FileNotOpen = 2 ;
      File_SystemErr_Unsupported_Operation = 3 ;
      File_SystemErr_EndOfFile = 4 ;
      File_SystemErr_IllegalOperationType = 5 ;
      File_SystemErr_Internal_Corruption = 6 ;
      File_SystemErr_Data_Format_Error = 7 ;
      File_SystemErr_LineTooLong = 8 ;
      File_SystemErr_Invalid_Record_Size = 9 ;
      File_SystemErr_FileName = 10 ;
      File_SystemErr_InvalidDevice = 11 ;
      File_SystemErr_ProtectionViolation = 12 ;
      File_SystemErr_Device_In_Use = 13 ;
      File_SystemErr_FileNotFound = 14 ;
      
const { Open modes... }
      { Mode 0 indicates to open file if there, create it if not. }
      FSOM_Append = 1 ; { Append mode (file must already exist) }
      FSOM_Input = 2 ; { File must already exist }
      FSOM_Output = 3 ; { File is created (deleted if existing) }
      FSOM_Text = 4 ; { File is text mode }
      { None of the following is currently supported... }
      FSOM_Read_Only = 8 ; { File cannot be written to }
      FSOM_Share_Read = 16 ; { File can be opened read-only by others }
      FSOM_Share_Write = 32 ; { File can be opened writable by others }
      { FS Open Mode 64+ are reserved for future use}

      { File System File Flags... }
      FSFF_Directory = 1 ; { A directory file }
      FSFF_Journaling = 2 ; { A journalled file }
      FSFF_File_System = 4 ; { A file-system file }
      FSFF_Compressed = 8 ; { A compressed file }
      FSFF_Erase = 16 ; { Erase file upon deletion }
      FSFF_Statistics = 32 ; { Keep file access statistics }
      FSFF_Read_Verify = 64 ; { Verify read accesses }
      FSFF_Write_Verify = 128 ; { Verify write accesses }
      FSFF_Hidden = 256 ; { A hidden file }
      FSFF_Contiguous = 512 ; { File is contiguous on file system device }
      FSFF_Deleted = 1024 ;
      { File is deleted but still open (marked for deletion) }
      FSFF_Protected = 2048 ; { for DOS compatibility only }
      FSFF_System = 4096 ; { for DOS compatibility only }
      FSFF_VolumeID = 8192 ; { for DOS compatibility only }
      FSFF_Archive = 16384 ; { for DOS compatibility only }
      { FS File Flags 32768+ are reserved for future use }

{ Used to hold results of Filespec_String_Scan }
type TFSSS = record
                 Device : string ; { Device, server, or protocol specification, including colon }
                 Path : string ; { Path, including final separator }
                 Filename : string ; { Full filename }
                 Name : string ; { Name, minus (last) extension }
                 Extension : string ; { (Last) extension (not including dot) }
                 Password : string ; { Embedded password }
                 Switches : string ; { Any switches at end of filename }
                 Flags : longint ; { Result flags (see FSSS_*) }
             end ;

{ Flags for TFSSS.Flags }
const FSSS_Null_Extension = 1 ; { Filename had no dot for an extension }

type FS_Header_Information = record
                                Header_Version : smallint ; { Version of FS_Header_Information }
                                ID : string[ 127 ] ; { Label for this file system device }
                                Root : TStore_Address ; { Pointer to root of file system/directories }
                                Version : smallint ; { File system structure version }
                                Default_Cluster_Size : smallint ; { Default cluster size for allocations (0 is invalid) }
                                Flags : smallint ; { FSD_* }
                            end ;

     PWildcard = ^TWildcard ;

     PFile_System = ^TFile_System ;

     TFind_Record = record
                        Initialized : longint ; { 0 if this record was initialized }
                        Context : PWildcard ; { Top-level Wildcard object }
                        Separator : char ; { File list separator (NULL means none) }
                        Specification : string ; { Original specification }
			File_Name : string ; { Full name and path of found file }
                        File_List : string ; { comma-delimited list of paths }
                        FS : PFile_System ; { File system }
                        Wildcard_Context : string ; { Used by FindFirst/FindNext }
                        Wildcard_Attributes : integer ; { Used by FindFirst/FindNext }
                        Last_Name : string ; { Used by FindFirst/FindNext }
                        Found_Name : string ; { Name (not path) of found file }
                        Found_Attributes : longint ; { Attributes of found file }
{$IFDEF DEBUG}
                        Sanity : string[ 13 ] ;
                        Serial_Number : integer ;
{$ENDIF}
			Data : array[ 0..511 ] of char ; { Context information - file system dependant }
		    end ;
     PFind_Record = ^TFind_Record ;

     TWildcard = object( TBase_Object )
                     private { Instance data... }
{$IFDEF DEBUG}
                         Sanity : string[ 10 ] ;
{$ENDIF}
                         Child : PWildcard ; { Child who is looking through sub-directories }
                         Parent : PWildcard ; { Parent wildcard object }
                         Current_Name : string ; { Name of current file }
                         Current_Spec : string ; { Current lookup specification }
                         Remaining_Spec : string ; { Part of lookup specification that applies to directory levels below ours }
                         Remaining_Path : string ; { Remaining_Spec minus "\...\" }
                         Doing_Directories : boolean ; { True while searching subdirectories for \...\ }
                         Find_Record : TFind_Record ; { Our current context }
                         Serial_Number : integer ;

                     public { API... }
                         File_System : PFile_System ; { File system we are doing lookup in }
                         Case_Sensitive : boolean ; { True if filenames are case-sensitive }

                     public { Constructors and destructors... }
                         constructor Init( FS : PFile_System
                             ; _Parent : PWildcard ) ;
                         constructor Create( FS : PFile_System ) ;
                         destructor Done ; virtual ;

                     public { API... }
                         function Lookup( const Name, Path : string ;
                             var Context : TFind_Record ) : string ;
                 end ; { TWildcard }

     TSFile = object( TSBase_Store )
                 private { Instance data... }
                     Current_Line : string ; { Current line for Read_* }
                     Current_Position : TStore_Address ; { Position in file }
                     Handle : longint ; { TFile_System handle for this file }
{$IFDEF Debug}
                     Sanity : string[ 6 ] ;
{$ENDIF}

                 public
                     BufSize : longint ; { Size of file buffer }
                     BufPos : longint ; { Position in buffer }
                     BufEnd : longint ; { Current end of buffer }
                     BufPtr : PChar ; { Pointer to file buffer }
                     Text_Buffer : array[ 0..127 ] of char ; { Default file buffer }
                     _N : char ; { terminating null for Text_Buffer }
                     Text_File : Boolean ; { True if accessed as a text file }
                     Record_Size : TRecord_Size ; { Size of record, in bytes, for non-text files }
                     IO_Error : tUEC ; { Result of last operation }
                     Extendable : boolean ; { True if we can extend the file }
                     File_Mode : integer ; { Mode that the file is opened in }
                     File_System : PFile_System ; { Pointer to file system that manages this file }
                     Serial_Number : longint ;

                 public { Constructors and destructors... }
                     constructor Init( FS : PFile_System ; TF : Boolean ;
                         Size : longint ) ;
                     destructor Done ; virtual ;

                 public { Overrides... }
                      function Extend( Amount : TStore_Address ) : TStore_Address ;
                          virtual ;

                      function Get_Name : string ; virtual ;

                      function Read_Data( var Data ; Address, _Size : TStore_Address ;
                          var UEC : TUEC ) : TStore_Address ; virtual ;

                      function Write_Data( var Data ;
                          Address, _Size : TStore_Address ;
                          var UEC : TUEC ) : TStore_Address ; virtual ;

                      function Max_Storage : TStore_Size ; virtual ;

                      function Min_Storage : TStore_Address ; virtual ;

                      function Contiguous_Store : boolean ; virtual ;

                      procedure Set_Max_Storage( Value : TStore_Address ;
                          var Res : TUEC ) ; virtual ;

                 public { API... }
                     procedure Reset( const Name : String ) ; virtual ;
                     procedure Rewrite( const Name : String ) ; virtual ;
                     procedure Append( const Name : String ) ; virtual ;
                     procedure Close ; virtual ;
                     procedure Seek( Pos : TStore_Address ) ; virtual ;

                     function Read_Integer : TStore_Address ; virtual ;

                     function Read_Float : Extended ; virtual ;

                     procedure Read( var Buffer ) ; virtual ;
                     procedure Write( var Buffer ) ; virtual ;
                     procedure Readln( var Buffer : String ) ; virtual ;
                     procedure Writeln( Buffer : String ) ; virtual ;
                     procedure Blockread( var Buf ; Count : TRecord_Size ;
                         var Res : TRecord_Size ) ; virtual ;
                     procedure Blockwrite( var Buf ; Count : TRecord_Size ;
                         var Res : TRecord_Size ) ; virtual ;
                     function EOF : boolean ; virtual ;
                     function EOLn : boolean ; virtual ;
                     function FilePos : TStore_Size ; virtual ;
                     function Get_Size : TStore_Size ; virtual ;
                     procedure Set_Size( New_Size : TStore_Size ) ; virtual ;
                     procedure Erase( const Name : String ) ; virtual ;
                     procedure Flush ; virtual ;
                     function FileName : string ; virtual ;
                     function SeekEOF : Boolean ; virtual ;
                     function SeekEOLn : Boolean ; virtual ;
                     procedure GetFTime( var _Time : Longint ) ; virtual ;
                     procedure SetFTime( _Time : longint ) ; virtual ;
                     procedure SetTextBuf( var buf ; size : Word ) ;
                         virtual ;
                     procedure Set_Extend( State : boolean ) ; virtual ;
                     procedure Truncate( Size : TStore_Address ) ; virtual ;
                     procedure Initialize ; virtual ;
                     function Debugger : PSDebug_Interface ; virtual ;
                     function Get_Attributes : longint ; virtual ;
                     procedure Set_Attributes( Value : longint ) ; virtual ;
                     function Facility : longint ; virtual ;

                     function Version : longint ; virtual ;

                     function Facility_Name : PChar ; virtual ;
                     function Opened : boolean ; virtual ;
                     function Get_Time : longint ; virtual ;
                     procedure Set_Time( Value : longint ) ; virtual ;
{$IFDEF DELPHI}
                     property Attributes : longint
                                  read Get_Attributes
                                  write Set_Attributes ;
                     property Time : longint
                                  read Get_Time
                                  write Set_Time ;
{$ENDIF}

                 private { Internal utility routines... }
                     function Read_Text_Buffer( Eat_Delimiters : boolean ) : string ;
                     procedure Reset_Buffer ;

{$IFDEF DELPHI}
		 protected { Instance support... }
{$ELSE}
		 public
{$ENDIF}
                     procedure GetFAttr( var Attr : Word ) ;
                         virtual ;
                     procedure SetFAttr( Attr : Word ) ; virtual ;
                     procedure Set_Error( F, X : integer ) ;
             end ;
     PSFile = ^TSFile ;

     TFile_Record = record
                        _File : file ;
                    end ;
     PFile_Record = ^TFile_Record ;

     TFile_System = object( TBase_Object )
                        private { Instance data... }
{$IFDEF DEBUG}
                            Sanity : string[ 15 ] ;
{$ENDIF}
                            Serial_Number : longint ;

{$IFDEF DELPHI}
                        protected { Instance data... }
{$ELSE}
			public
{$ENDIF}
                            File_List : TCollection ;
                            IO_Error : TUEC ;

                        public { Constructors and destructors... }
                            constructor Init ;
                            destructor Done ; virtual ;

                        private { Internal utility routines... }
    			    function Get_File_Rec( Handle : integer ) : PFile_Record ;

{$IFDEF DELPHI}
			protected { Instance support... }
{$ELSE}
			public { These routines should NOT be called from outside this unit! }
{$ENDIF}
                            procedure Filespec_String_Scan( const Name, _Separator : string ;
                                var Info : TFSSS ) ; virtual ;
                            procedure Set_Error( F, X : integer ) ;

                        public { API... }
                            { Increment reference count to file system. }
                            procedure Attach ; virtual ;

                            { Decrement reference count to file system. }
                            procedure Detach ; virtual ;

                            { Close an open file via handle. }
                            procedure Close( Handle : integer ) ; virtual ;

                            { Close an open find record. }
                            procedure Close_Find( var Find_Record : TFind_Record ) ;
                                virtual ;

                            { Create file and open with specified mode.  If
                              file already exists, it is deleted and re-created.
                              Returns the file handle (0 if an error). }
                            function Create( const Name : string ;
                                Mode : integer ) : integer ; virtual ;

                            { Returns a debugger interface to the file system }
                            function Debugger : PSDebug_Interface ; virtual ;

                            { Deletes the specified file. }
                            procedure Erase( const Name : String ) ; virtual ;

                            { Returns name associated with specified file. }
                            function File_Name( Handle : integer ) : string ;
                                 virtual ;

                            function First_Directory( const Name : string ) : integer ;
                                virtual ;

                            { Flush any buffered output for specified file. }
                            procedure Flush( Handle : integer ) ; virtual ;

                            { Returns the attributes of the specified file. }
                            function Get_Attributes( Handle : integer ) : integer ;
                                 virtual ;

                            function Get_File_Object( Mode : integer ) : PSFile ;
                                virtual ;

                            { Returns size of specified file, in bytes. }
                            function Get_Size( Handle : integer ) : TStore_Address ;
                                 virtual ;

                            function Get_Time( Handle, Typ : integer ) : Longint ;
                                 virtual ;

                            { Normalize the passed path: (device:)\path }
                            function Normalize_Path( Name : string ) : string ;
                                virtual ;

                            { Open existing file with specified mode.  Returns
                              the file handle (0 if an error). }
                            function Open( const Name : string ;
                                Mode : integer ) : integer ; virtual ;

                            { Read from specified file at specified position
                              into Buffer, up to Len bytes.  Actual bytes read
                              is returned in Count.  Count is ignored on call. }
                            procedure Read( Handle : integer ;
                                Position : TStore_Address ; Len : TStore_Address ;
                                var Buffer ; var Count : TStore_Address ) ; virtual ;

                            { Renames Old file to New file name. }
                            procedure Rename( const Old, New : String ) ;
                                 virtual ;

                            procedure Set_Attributes( Handle : integer ;
                                Attr : integer ) ; virtual ;

                            procedure Set_Time( Handle : integer ;
                                Typ, Time : Longint ) ; virtual ;

                            { Sets the size of the specified file to the
                              specified size, in bytes. }
                            procedure Set_Size( Handle : integer ;
                                New_Size : TStore_Address ) ; virtual ;

                            { Write to specified file at specified position from
                              Buffer, up to Len bytes.  Actual bytes written is
                              returned in Count.  Count is ignored on call. }
                            procedure Write( Handle : integer ;
                                Position : TStore_Address ; Len : TStore_Address ;
                                var Buffer ; var Count : TStore_Address ) ;
                                virtual ;

                            { Returns last error code. }
                            procedure Last_Error( var Res : TUEC ) ; virtual ;

                            function DeviceSize( Nam : string ) : TStore_Address ;
                                virtual ;

                            function DeviceFree( Nam : string ) : TStore_Address ;
                                virtual ;

                            function MkDir( Nam : string ) : longint ;
                                virtual ;

                            procedure Get_Header( _Path : string ;
                                var Header : FS_Header_Information ;
                                 var Res : TUEC ) ; virtual ;

                            procedure Set_Header( _Path : string ;
                                var Header : FS_Header_Information ;
                                var Res : TUEC ) ; virtual ;

                            { Returns the attributes of the specified file. }
                            function Get_File_Attributes( const Name : string ) : longint ;
                                 virtual ;

                            { Returns the attributes of the specified file. }
                            procedure Set_File_Attributes( const Name : string ;
                                Value : longint ) ; virtual ;

                            function FindFirst( Nam : string ; Attributes : longint ;
                                var Find_Record : TFind_Record ) : longint ;
                                virtual ;

                            function FindNext( var Find_Record : TFind_Record ) : longint ;
                                virtual ;

                            procedure Initialize( Nam, ID : string ;
                                Flags : longint ; var UEC : tUEC ) ; virtual ;

                            procedure Initialize_Store( S : PSStore ; ID : string ;
                                Flags : longint ; var UEC : tUEC ) ; virtual ;

                            procedure Verify( Name : string ; var Res : TUEC ) ;
                                virtual ;

                            function Path : string ; virtual ;

                            { What character or string is used to separate
                              paths from each other and the filename.  If the
                              file system doesn't support paths, this returns a
                              null string. }
                            function Separator : string ; virtual ;

                            function Facility : longint ; virtual ;

                            function Version : longint ; virtual ;

                            { Locks the file system for synchronous I/O.
                              Returns True if this process was the previously
                              locking process. }
                            function Lock( Nam : string ) : boolean ; virtual ;

                            { Unlocks the file system. }
                            procedure Unlock( Nam : string ) ; virtual ;
                     end ; { TFile_System }

const TFileErr_Facility = 43 ;

const TBlock_DeviceErr_Facility = 43 ;

type TBlock_Device = object( TSFile )
                         private
			     Class_Name : string[ 13 ] ;

                         public { Constructors and destructors }
                             constructor Init ;

                         public { API... }
                             Drive  : integer ; { Drive number }
                             Position : int64 ; { Current position on drive (byte offset) }

                             procedure Reset( const Name : String ) ; virtual ;
                             procedure Rewrite( const Name : String ) ;
                                 virtual ;
                             procedure Append( const Name : String ) ; virtual ;
                             procedure Close ; Virtual ;
                             procedure Seek( Pos : TStore_Address ) ; virtual ;
                             procedure Read( var Buffer ) ; virtual ;
                             procedure Write( var Buffer ) ; virtual ;
                             procedure Readln( var Buffer : String ) ; virtual ;
                             procedure Writeln( Buffer : String ) ;
                                 virtual ;
                             procedure Blockread( var Buf ; Count : TRecord_Size ;
                                 var Res : TRecord_Size ) ; virtual ;
                             procedure Blockwrite( var Buf ; Count : TRecord_Size ;
                                 var Res : TRecord_Size ) ; virtual ;
                             function EOF : boolean ; virtual ;
                             function EOLn : boolean ; virtual ;
                             function Filepos : TStore_Size ; virtual ;
                             function Get_Size : TStore_Size ; virtual ;
                             procedure Set_Size( New_Size : TStore_Size ) ; virtual ;
                             procedure Flush ; virtual ;
                             procedure Getfattr( var Attr : Word ) ; virtual ;
                             procedure Erase( const Name : String ) ; virtual ;
                             function SeekEOF : Boolean ; virtual ;
                             function SeekEOLn : Boolean ; virtual ;
                             procedure Setfattr( Attr : Word ) ; virtual ;
                             procedure Getftime( var _Time : Longint ) ;
                                 virtual ;
                             procedure Setftime( _Time : longint ) ; virtual ;
                             procedure Settextbuf( var buf ; size : Word ) ;
                                 virtual ;
                             procedure Set_Extend( State : boolean ) ; virtual ;
                             procedure Truncate( Size : TStore_Address ) ; virtual ;
                             procedure Initialize ; virtual ;
                             function Facility : longint ; virtual ;

                             function Version : longint ; virtual ;

                             function Facility_Name : PChar ; virtual ;
                     end ; { TBlock_Device }
     PBlock_Device = ^TBlock_Device ;

type TFile_System_Server = object( TBase_Object )
                               public { API... }
                                   function Get_File_System_For_Store( Store : PSStore ;
                                       Own : boolean ) : PFile_System ;
                                       virtual ;
                                   { Creates and returns a file system for the
                                     passed store.  If there is no known file
                                     system on the store, nil is returned. }

                                   function Get_File_System( const Nam : string ) : PFile_System ;
                                       virtual ;
                                   { Creates and returns a file system for the
                                     passed file specification.  If there is no
                                     known file system on the store, nil is
                                     returned. }

                                   procedure Initialize( const Nam, ID : string ;
                                       Flags : integer ; var UEC : TUEC ) ;
                                       virtual ;
                                   { Initializes a file system on the specified
                                     device, with the specified ID and flags.
                                     Returns status in UEC. }
     			   end ; { TFile_System_Server }
     PFile_System_Server = ^TFile_System_Server ;

     TStore_Server = object( TBase_Object )
                         public { API... }
                             function Get_Store_ID( Name : string ) : string ;
                                 virtual ;
                             { Returns the portion of Name which is the ID of
                               the store.  For instance, the ID might be the
                               device name (eg C:) for a device. }

                             function Get_Store( var Name : string ) : PSStore ;
                                 virtual ;
                             { Creates and returns a TStore object for the
                               passed specification.  On return, Name has the
                               device ID stripped from it.  If there is no store
                               available, nil is returned. }
                     end ; { TStore_Server }
     PStore_Server = ^TStore_Server ;

const { File System Flags... }
      FSF_Fixed_Size = 1 ; { The file system is on a fixed-size device }
      FSF_Read_Only = 2 ; { The file system is opened in read-only mode }
      FSF_NoOwnStore = 4 ; { The file system doesn't own the store - so don't close it when file system closes }
      { 8-32768 = reserved }

const TWin32_FileErr_Facility = 45 ;

function Normalize_Hierarchy( Name, Separator, Self_Reference, Parent_Reference,
    Prefix : string ) : string ;
{ Returns a normalized version of Name.  Parameters:
      Parameter           Meaning
      ---------           -------
      Name                Name to normalize
      Separator           String that separates nodes in the hierarcy
      Self_Reference      String representing a self-reference (null if none)
      Parent_Reference    String representing a parent reference (null if none)
      Prefix              String representing the end of any initial non-hierarchical prefix (null if none)
}

function Explicit( X : string ) : string ;
function FileName( X : String ) : string ;
function Filnam( const X : string ; Literal : boolean ) : string ;
function Find_File_In_Path( const Nam : string ; Env : string ) : string ;
{ Given a semi-colon-delimited set of paths, find file Nam, and return path + Nam }

function Get_Temp_File_Name( const Prefix : string ) : string ;
procedure Initialize_Find_Record( var F : TFind_Record ) ;
procedure Uninitialize_Find_Record( var F : TFind_Record ) ;

function Open( const Name : string ; Mode : integer ; var UEC : TUEC ) : PSFile ;
{ Open a file and return a pointer to the file object.  Parameters:
      Name = Name of file to open
      Mode = Open mode (see FSOM_* constants)
      UEC = open status }

function Lookup( Filename : string;
    var Find_Record : TFind_Record ) : string ;

procedure Initialize_File_System( Nam, ID : string ; Flags : longint ;
    var UEC : tUEC ) ;
{ Initialize a file system on the device specified with Nam.  Give it the
  specified ID.  Returns condition code in UEC.  Flags: and 1 = Force fixed size
                                                        and 2+ = reserved }

procedure Register_File_System_Server( FSS : PFile_System_Server ) ;

procedure Register_Store_Server( SS : PStore_Server ) ;

function DeviceSize( const Nam : string ) : int64 ;

function DeviceFree( const Nam : string ) : int64 ;

function MKDir( const Nam : string ) : longint ;

function Open_File_System( const Name : string ; Flags : integer ;
    var Res : TUEC ) : PFile_System ;
    
procedure Close_File_System( const Name : string ) ;

{ Convert from TFile_System file mode to Windows/DOS file mode }
function Convert_File_Mode( File_Mode : integer ) : integer ;


{ Scan a file specification and fill the Info record.  Separator is the path
  separator to use in parsing. }
procedure Filespec_String_Scan( const Name, Separator : string ;
    var Info : TFSSS ) ;

procedure Terminate ;

var DOS_FS : TFile_System ;

implementation

uses _App, { Application_Manager }
     ArrayInt, { TCOM_Array_Interface }
     ASCIIDef,
     Standard,
     Num1s,
{$IFDEF DELPHI}
     Windows,
{$ENDIF}
     CommonUt, { Boolean_To_String }
     Compatib, { FindFirst }
     Disk,
     ErrMastr, { Error_Code }
     O_S, { OS }
     _Types, { Small_String }
     UStrings, { Boolean_To_String }
     XLate ; { WC_Match }

{$IFDEF DELPHI}
type TWin32_File_System = object( TFile_System )
                              private { Instance data... }
                                  Handles : TInteger_Collection ;
                                  Names : TString_Collection ;

                              protected { Instance support... }
                                  procedure Set_Last_Error ;

                              public { Constructors and destructors... }
                                  constructor Init ;
                                  destructor Done ; virtual ;

                              public { API... }
                                  { Close an open file via handle. }
                                  procedure Close( Handle : integer ) ; virtual ;

                                  { Create file and open with specified mode.  If
                                    file already exists, it is deleted and re-created.
                                    Returns the file handle (0 if an error). }
                                  function Create( const Name : string ;
                                      Mode : integer ) : integer ; virtual ;

                                  { Returns a debugger interface to the file system }
                                  function Debugger : PSDebug_Interface ; virtual ;

                                  { Returns name associated with specified file. }
                                  function File_Name( Handle : integer ) : string ;
                                       virtual ;

                                  { Flush any buffered output for specified file. }
                                  procedure Flush( Handle : integer ) ; virtual ;

                                  { Returns size of specified file, in bytes. }
                                  function Get_Size( Handle : integer ) : TStore_Address ;
                                       virtual ;

                                  function Get_Time( Handle, Typ : integer ) : Longint ;
                                       virtual ;

                                  { Open existing file with specified mode.  Returns
                                    the file handle (0 if an error). }
                                  function Open( const Name : string ;
                                      Mode : integer ) : integer ; virtual ;

                                  { Read from specified file at specified position
                                    into Buffer, up to Len bytes.  Actual bytes read
                                    is returned in Count.  Count is ignored on call. }
                                  procedure Read( Handle : integer ;
                                      Position : TStore_Address ; Len : TStore_Address ;
                                      var Buffer ; var Count : TStore_Address ) ; virtual ;

                                  procedure Set_Time( Handle : integer ;
                                      Typ, Time : Longint ) ; virtual ;

                                  { Sets the size of the specified file to the
                                    specified size, in bytes. }
                                  procedure Set_Size( Handle : integer ;
                                      New_Size : TStore_Address ) ; virtual ;

                                  { Write to specified file at specified position from
                                    Buffer, up to Len bytes.  Actual bytes written is
                                    returned in Count.  Count is ignored on call. }
                                  procedure Write( Handle : integer ;
                                      Position : TStore_Address ; Len : TStore_Address ;
                                      var Buffer ; var Count : TStore_Address ) ;
                                      virtual ;

                                  { Returns the attributes of the specified file. }
                                  function Get_File_Attributes( const Name : string ) : longint ;
                                       virtual ;

                                  { Returns the attributes of the specified file. }
                                  procedure Set_File_Attributes( const Name : string ;
                                      Value : integer ) ; virtual ;

                                  function Facility : longint ; virtual ;

                                  function Version : longint ; virtual ;
                          end ; { TWin32_File_System }

type TWin32_File = object( TSFile )
                 public { Constructors and destructors... }
                     constructor Init( FS : PFile_System ; Tf : Boolean ;
                         Size : Word ) ;

                 public { API... }
                     procedure Initialize ; virtual ;
                     
                     function Facility : longint ; virtual ;

                     function Version : longint ; virtual ;

                     function Facility_Name : PChar ; virtual ;
             end ;
     PWin32_File = ^TWin32_File ;
{$ENDIF}

type PFS_Record = ^TFS_Record ;
     TFS_Record = record
                        Store_Name : string ; { Name of store }
                        FS : PFile_System ; { File system object controlling this file system file }
                        Users : longint ; { Reference count }
                        Next : PFS_Record ; { Next item in list }
                  end ;

type PFile_Processor = ^TFile_Processor ;
     TFile_Processor = object( TBase_Object )
		      private
                          IO_Error : TUEC ;

                          { The FS_Root variable is the head of a SLL which
                            contains one record per DOS file which is opened as
                            a FS file. }
                          FS_Root : PFS_Record ;

                          { List of file system servers... }
                          FSS_List : TCollection ;

                          { List of store servers... }
                          SS_List : TCollection ;

                      private { Internal utility routines... }
                          procedure Determine_Parts( Nam : string ;
                              var D, N, DN : string ) ;

		      public { API... }
			constructor Init ;
			destructor Done ; virtual ;
			function Lookup( Filename : string;
			    var Find_Record : TFind_Record ) : string ;
			    virtual ;
			function Open( const Name : string ; Mode : integer ;
                            var UEC : TUEC ) : PSFile ; virtual ;
			procedure Initialize( const Nam, ID : string ;
                            Flags : longint ; var UEC : tUEC ) ; virtual ;
			function DeviceSize( const Nam : string ) : int64 ;
			    virtual ;
			function DeviceFree( const Nam : string ) : int64 ;
			    virtual ;
			function MKDir( const Nam : string ) : longint ;
			    virtual ;
                        procedure Close_File_System( Name : string ) ;
                            virtual ;
                        function Open_File_System( Name : string ;
                            Flags : integer ;
                            var Res : TUEC ) : PFile_System ; virtual ;
                        function Debugger : PSDebug_Interface ; virtual ;
                        function File_System_Open( Nam : string ) : boolean ;
                            virtual ;
                        { Scan a file specification and fill the Info record.  Separator is the path
                          separator to use in parsing. }
                        procedure Filespec_String_Scan( const Name, Separator : string ;
                            var Info : TFSSS ) ;
                        procedure Register_File_System_Server( FSS : PFile_System_Server ) ;
                            virtual ;
                        procedure Register_Store_Server( SS : PStore_Server ) ;
                            virtual ;
		       end ; { TFile_Processor }

var File_Processor : TFile_Processor ;

procedure Initialize_Find_Record( var F : TFind_Record ) ;

begin
{$IFDEF DELPHI}
    F.Initialized := 0 ;
    F.Context := nil ;
    F.FS := nil ;
    fillchar( F.Data, sizeof( F.Data ), 0 ) ;
    System.Initialize( F.Specification ) ;
    System.Initialize( F.File_Name ) ;
    System.Initialize( F.File_List ) ;
    System.Initialize( F.Wildcard_Context ) ;
    System.Initialize( F.Last_Name ) ;
    System.Initialize( F.Found_Name ) ;
{$ELSE}
    fillchar( F, sizeof( F ), 0 ) ;
{$ENDIF}
    F.Separator := ',' ; { Default file separator }
{$IFDEF DEBUG}
    F.Sanity := 'TFind_Record' ;
    F.Serial_Number := Get_Serial_Number ;
{$ENDIF}
end ;


{ This is global because when defined locally it causes stack/memory corruption.
  Apparently the compiler inserts some code to close the context when the
  variable goes out of scope.  By making it global, this close operation doesn't
  happen.  Certainly, the problem goes away when this is gloabl instead of
  local. }
var FR : TSearchRec ;

procedure Uninitialize_Find_Record( var F : TFind_Record ) ;

begin
    if( F.Context <> nil ) then
    begin
{$IFDEF DEBUG}
        if( F.Context.Sanity <> 'TWildcard' ) then
        begin
            halt ;
        end ;
{$ENDIF}
        dispose( F.Context, Done ) ;
        F.Context := nil ;
    end ;
    if( F.FS <> nil ) then
    begin
        F.FS^.Close_Find( F ) ;
    end ;
    F.FS := nil ;
{$IFDEF DEBUG}
    F.Sanity := ' ' ;
{$ENDIF}
end ; { Uninitialize_Find_Record }


{ Returns True if Name matches Wildcard }
function WC_Path_Match( Wildcard, Name : string ; Case_Sensitive : boolean ) : boolean ;

var Count, Dummy : integer ;
    Wildcard_Position, Name_Position : integer ;

begin
    if( not Case_Sensitive ) then
    begin
        Wildcard := uppercase( Wildcard ) ;
        Name := uppercase( Name ) ;
    end ;

    { Setup... }
    if( copy( Wildcard, 1, 1 ) <> '\' ) then
    begin
        Wildcard := '\' + Wildcard ;
    end ;
    if( copy( Name, 1, 1 ) <> '\' ) then
    begin
        Name := '\' + Name ;
    end ;
    Dummy := pos( '...', Wildcard ) ;
    if( copy( Wildcard, 1, Dummy - 1 ) <> copy( Name, 1, Dummy - 1 ) ) then
    begin
        WC_Path_Match := False ;
        exit ;
    end ;
    if( Dummy > 0 ) then { \...\ specified }
    begin
        Wildcard := copy( Wildcard, Dummy + 3, length( Wildcard ) ) ;
	Count := 0 ;
        for Dummy := 1 to length( Wildcard ) do
        begin
            if( Wildcard[ Dummy ] = '\' ) then
            begin
                inc( Count ) ;
            end ;
        end ;
        Dummy := length( Name ) ;
        while( ( Dummy > 0 ) and ( ( Count > 1 ) or ( pos( Name[ Dummy ], ':\' ) = 0 ) ) ) do
        begin
            if( Name[ Dummy ] = '\' ) then
            begin
                dec( Count ) ;
            end ;
            dec( Dummy ) ;
        end ;
        Name := copy( Name, Dummy, length( Name ) ) ;
    end ;

    while( length( Name ) > 0 ) do
    begin
        Wildcard_Position := length( Wildcard ) ;
        while(
               ( Wildcard_Position > 0 )
               and
               ( Wildcard[ Wildcard_Position ] <> '\' )
             ) do
        begin
            dec( Wildcard_Position ) ;
        end ;
        Name_Position := length( Name ) ;
        while(
               ( Name_Position > 0 )
               and
               ( Name[ Name_Position ] <> '\' )
             ) do
        begin
            dec( Name_Position ) ;
        end ;

        if( not WC_Match( copy( Wildcard, Wildcard_Position, length( Wildcard ) ), 
            copy( Name, Name_Position, length( Name ) ) ) ) then
        begin
            WC_Path_Match := False ;
            exit ;
        end ;
        Wildcard := copy( Wildcard, 1, Wildcard_Position - 1 ) ;
        Name := copy( Name, 1, Name_Position - 1 ) ;
    end ; { while( length( Name ) > 0 ) }
    WC_Path_Match := ( length( Wildcard ) = 0 ) ;
end ; { WC_Path_Match }



{ TWildcard methods... }

{ Constructors and destructors... }

constructor TWildcard.Init( FS : PFile_System ; _Parent : PWildcard ) ;

begin
    Serial_Number := Get_Serial_Number ;
    Child := nil ;
    File_System := FS ;
    Doing_Directories := False ;
{$IFDEF DELPHI}
    Initialize( Current_Name ) ;
    Initialize( Current_Spec ) ;
    Initialize( Remaining_Spec ) ;
    Initialize( Remaining_Path ) ;
{$ELSE}
    Current_Name := '' ;
    Current_Spec := '' ;
    Remaining_Spec := '' ;
    Remaining_Path := '' ;
{$ENDIF}
    Case_Sensitive := False ; { Default to non-case-sensitive }
    Parent := _Parent ;
    if( FS <> nil ) then
    begin
        FS^.Attach ;
    end ;
    Initialize_Find_Record( Find_Record ) ;
{$IFDEF DEBUG}
    Sanity := 'TWildcard' ;
{$ENDIF}
end ;


constructor TWildcard.Create( FS : PFile_System ) ;

begin
    Init( FS, nil ) ;
end ;


destructor TWildcard.Done ;

begin
    if( Child <> nil ) then
    begin
        dispose( Child, Done ) ;
        Child := nil ;
    end ;

    if( File_System <> nil ) then
    begin
        File_System^.Detach ;
    end ;
    Uninitialize_Find_Record( Find_Record ) ;

{$IFDEF DEBUG}
    Sanity := ' ' ;
{$ENDIF}

    inherited Done ;
end ;


{ API... }

function TWildcard.Lookup( const Name, Path : string ;
    var Context : TFind_Record ) : string ;

    function _Lookup( Name : string ) : string ;

    var A : integer ;
        Dummy{, Dummy1} : integer ;
        _Path, Work : string ;
        FSSS : TFSSS ;

    label Loop ;

    begin
        if( length( Name ) > 0 ) then { Starting new lookup }
        begin
            if( Child <> nil ) then
            begin
                dispose( Child, Done ) ;
                Child := nil ;
            end ;

            Name := File_System.Normalize_Path( Name ) ;
            if( copy( Name, 1, 1 ) = '\' ) then
            begin
                Name := copy( Name, 2, length( Name ) ) ;
            end ;

            Dummy := File_System.First_Directory( Name ) ;
            Current_Spec := copy( Name, 1, Dummy - 1 ) ;
            Remaining_Spec := copy( Name, Dummy + 1, length( Name ) ) ;
            Current_Name := '' ;
            Doing_Directories :=
                ( Current_Spec = '...' ) or ( length( Remaining_Spec ) > 0 ) ;
            if( Current_Spec = '...' ) then
            begin
                Remaining_Spec := '\...\' + Remaining_Spec ; { Make sure child does ellipse search }
            end ;
            if( copy( Remaining_Spec, 1, 5 ) = '\...\' ) then
            begin
                Remaining_Path := copy( Remaining_Spec, 5, length( Remaining_Spec ) ) ;
            end else
            begin
                Remaining_Path := Remaining_Spec ;
            end ;
            Dummy := length( Remaining_Path ) ;
            while( ( Dummy > 0 ) and ( Remaining_Path[ Dummy ] <> '\' ) ) do
            begin
                dec( Dummy ) ;
            end ;
            Remaining_Path := copy( Remaining_Path, 1, Dummy ) ;
        end ; { if( length( Name ) > 0 ) }

        if( Child <> nil ) then
        begin
            Work := Child^.Lookup( '', Path + '\' + Current_Name, Context ) ; { Have child lookup next match }
            if( length( Work ) > 0 ) then
            begin
                Context.File_Name := Current_Name + '\' + Work ;
                Dummy := pos( '\', Current_Spec ) ;
                if( Dummy > 0 ) then
                begin
                    Context.File_Name :=
                        copy( Current_Spec, 1, Dummy ) + Context.File_Name ;
                end ;
                _Lookup := Context.File_Name ;
                exit ;
            end ;

            { Child found nothing - done with it... }
            dispose( Child, Done ) ;
            Child := nil ;
        end ;
        { (Child will always be nil at this point) }

        { Find next match at our level... }
Loop:
        if(
            ( pos( '*', Current_Spec ) + pos( '?', Current_Spec ) + pos( '...', Current_Spec ) > 0 )
            or
            ( length( Current_Name ) = 0 )
          ) then
        begin
            if( Doing_Directories ) then
            begin
                A := FSFF_Directory ; { Look only in directories }
            end else
            begin
{                A := not FSFF_Directory ; { Look for everything but directories }
                A := -1 ; { Look for everything }
            end ;
            if( length( Current_Name ) = 0 ) then { Our first lookup }
            begin
                if( Current_Spec = '...' ) then
                begin
                    Work := '*' ;
                end else
                begin
                    Work := Current_Spec ;
                end ;
                Filespec_String_Scan( Work, '\', FSSS ) ;
                _Path := FSSS.Device ;
                if( length( Path ) > 0 ) then
                begin
                    Dummy := File_System^.FindFirst( Path + '\' + Work, A, Find_Record ) ;
                end else
                begin
                    Dummy := File_System^.FindFirst( Work, A, Find_Record ) ;
                end ;
                if( Dummy <> 0 ) then
                begin
                    _Lookup := '' ;
                    exit ;
                end ;
            end else
            begin
                if( File_System^.FindNext( Find_Record ) <> 0 ) then { No more found }
                begin
                    _Lookup := '' ;
                    exit ;
                end ;
            end ;
            Current_Name := Find_Record.Found_Name ;
        end else
        begin
            _Lookup := '' ;
            exit ;
        end ;
        if( ( Find_Record.Found_Attributes and FSFF_Directory ) = 0 ) then { Not a directory }
        begin
            if( Doing_Directories ) then
            begin
                goto Loop ;
            end ;
        end ;

        { We found a match for our current spec.  See if sub-directories to search }
        if(
            Doing_Directories
            and
            ( length( Remaining_Spec ) > 0 )
            and
            ( Current_Name <> '.' )
            and
            ( Current_Name <> '..' ) { No iterating on relative directories (would cause an infinite loop) }
          ) then
        begin
            Child := new( PWildcard, Init( File_System, @self ) ) ;
            Work := Child^.Lookup( Remaining_Spec, Path + _Path + '\' + Current_Name, Context ) ;
            if( length( Work ) > 0 ) then
            begin
                Work := Current_Name + '\' + Work ;
                Context.File_Name := Work ;
                Dummy := pos( ':', Current_Spec ) ;
                if(
                    ( Dummy > 0 )
                    and
                    ( copy( Current_Spec, Dummy + 1, 1 ) = '\' )
                  ) then
                begin
                    Dummy := Dummy + 1 ;
                end ;
                _Lookup := copy( Current_Spec, 1, Dummy ) + Context.File_Name ;
                exit ;
            end ;
            dispose( Child, Done ) ;
            Child := nil ;
            goto Loop ;
        end ; { if }
        _Lookup := Current_Name ;
{$IFDEF DEBUG}
        if( Context.Sanity <> 'TFind_Record' ) then
        begin
            halt ;
        end ;
{$ENDIF}
{        Context.Data := Find_Record.Data ;}
        Context.Found_Attributes := Find_Record.Found_Attributes ;
        Context.Wildcard_Attributes := Find_Record.Wildcard_Attributes ;
        Context.Found_Name := Find_Record.Found_Name ;
        Context.File_Name := Current_Name ;
        Context.Last_Name := Current_Name ;
        Context.Wildcard_Context := Find_Record.Wildcard_Context ;
    end ; { TWildcard.Lookup._Lookup }

var Res, Work : string ;

label Try_Again ;

begin
{$IFDEF DEBUG}
    if( Context.Sanity <> 'TFind_Record' ) then
    begin
        halt ;
    end ;
{$ENDIF}
    Res := _Lookup( Name ) ;
Try_Again:
    if( ( length( Res ) = 0 ) and Doing_Directories ) then
    begin
        if( ( Current_Spec = '...' ) or ( length( Remaining_Spec ) = 0 ) ) then
        begin
            Doing_Directories := False ;
            Current_Name := '' ; { Start search over for non-directories }
            Res := _Lookup( '' ) ;
        end ;
    end ;
    if(
        ( Parent = nil ) { We are the top level }
        and
        ( length( Res ) > 0 ) { ...and something was found }
        and
        ( pos( '...', Current_Spec + Remaining_Spec ) > 0 ) { Using ... construct }
      ) then
    begin
        if( Current_Spec = '...' ) then
        begin
            Work := Remaining_Spec ;
        end else
        begin
            Work := Current_Spec + '\' + Remaining_Spec ;
        end ;
	if( not WC_Path_Match( Work, Res, Case_Sensitive ) ) then
        begin
            Res := _Lookup( '' ) ;
            goto Try_Again ;
        end ;
    end ;
    Lookup := Res ;
end ; { TWildcard.Lookup }



{$IFDEF DELPHI2UP}
type TCOM_File_Array_Interface = class( TCOM_Array_Interface )
                                     private { Instance data... }
                                         _File : PSFile ;

                                     public { API... }
                                           function Get_Byte( Index : int64 ) : byte ;
                                               override ;

                                           procedure Set_Byte( Index : int64 ;
                                               Value : byte ) ;
                                               override ;

                                           function Read_Only : boolean ;
                                               override ;

                                           function Low_Bound( Subscript : integer ) : int64 ;
                                               override ;

                                           function High_Bound( Subscript : integer ) : int64 ;
                                               override ;

                                           function Subscripts : integer ;
                                               override ;
                                 end ;

function TCOM_File_Array_Interface.Get_Byte( Index : int64 ) : byte ;

var B : byte ;
    UEC : TUEC ;

begin
    _File.Read_Data( B, Index, 1, UEC ) ;
    Result := B ;
end ;


procedure TCOM_File_Array_Interface.Set_Byte( Index : int64 ; Value : byte ) ;

begin
end ;


function TCOM_File_Array_Interface.Read_Only : boolean ;

begin
    Result := True ;
end ;


function TCOM_File_Array_Interface.Low_Bound( Subscript : integer ) : int64 ;

begin
    Result := 0 ;
end ;


function TCOM_File_Array_Interface.High_Bound( Subscript : integer ) : int64 ;

begin
    Result := _File.Get_Size ;
end ;


function TCOM_File_Array_Interface.Subscripts : integer ;

begin
    Result := 1 ;
end ;
{$ENDIF}


type TSFile_Contents_Debugger = object( TSText_Debugger )
                                    private
                                        _Instance : PSFile ;
                                        _Context, _Handle : integer ;
{$IFDEF DELPHI2UP}
                                        _Array : TCOM_File_Array_Interface ;
{$ENDIF}

                                    public
                                        constructor Init( I : PSFile ) ;

                                        destructor Done ; virtual ;

                                        procedure Activate ; virtual ;

                                        function Count : longint ; virtual ;

                                        function Child( Ordinal : longint ) : PSDebug_Interface ;
                                            virtual ;
                                end ;
     PSFile_Contents_Debugger = ^TSFile_Contents_Debugger ;


constructor TSFile_Contents_Debugger.Init( I : PSFile ) ;

begin
    inherited Init ;

    _Instance := I ;
end ;


destructor TSFile_Contents_Debugger.Done ;

{$IFDEF DELPHI2UP}
var Free_Viewer : procedure( Context : longint ) ; stdcall ;
{$ENDIF}

begin
{$IFDEF DELPHI2UP}
    if( _Context <> 0 ) then
    begin
        Free_Viewer := GetProcAddress( _Handle, 'Free_Viewer' ) ;
        if( assigned( Free_Viewer ) ) then
        begin
            Free_Viewer( _Context ) ;
        end ;
    end ;
    if( _Handle = 0 ) then
    begin
        FreeLibrary( _Handle ) ;
        _Handle := 0 ;
    end ;
    if( _Array <> nil ) then
    begin
        _Array.Free ;
        _Array := nil ;
    end ;
{$ENDIF}

    inherited Done ;
end ;


procedure TSFile_Contents_Debugger.Activate ;

{$IFDEF DELPHI2UP}
var Make_Viewer : function( P : TCOM_Array_Interface ; Title : PChar ;
        Default_Size, Default_Base, Flags : longint ;
        Default_Home : int64 ) : longint ; stdcall ;
{$ENDIF}

begin
{$IFDEF DELPHI2UP}
    if( _Handle = 0 ) then
    begin
        _Handle := LoadLibrary( 'BinaryDi.dll' ) ;
        if( _Handle = 0 ) then
        begin
            exit ;
        end ;
    end ;
    Make_Viewer := GetProcAddress( _Handle, 'Make_Viewer' ) ;
    if( assigned( Make_Viewer ) ) then
    begin
        _Array := TCOM_File_Array_Interface.Create ;
        _Array._File := _Instance ;
        _Context := Make_Viewer( _Array, PChar( _Instance^.Get_Name ),
            1, 10, 0, 0 ) ;
    end ;
{$ENDIF}
end ;


function TSFile_Contents_Debugger.Count : longint ;

begin
    Count := 0
end ;


function TSFile_Contents_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

begin
    Child := nil ;
end ;



{ TSFile_Debugger... }

type PSFile_Debugger = ^TSFile_Debugger ;
     TSFile_Debugger = object( TSText_Debugger )
                                private
                                    _Instance : PSFile ;
                                    S : Small_String ;

                                public
                                    constructor Init( I : PSFile ) ;

                                    function Count : longint ; virtual ;

                                    function Child( Ordinal : longint ) : PSDebug_Interface ;
                                        virtual ;
                            end ;

constructor TSFile_Debugger.Init( I : PSFile ) ;

begin
    inherited Init ;

    _Instance := I ;
end ;


function TSFile_Debugger.Count : longint ;

begin
    Count := 9 ;
end ;


function File_Mode_To_String( X : integer ) : string ;

var S : string ;

begin
    S := '[' ;
    if( ( X and FSOM_Append ) <> 0 ) then
    begin
        S := S + 'FSOM_Append' ;
    end ;
    if( ( X and FSOM_Input ) <> 0 ) then
    begin
        if( length( S ) > 0 ) then
        begin
            S := S + ',' ;
        end ;
        S := S + 'FSOM_Input' ;
    end ;
    if( ( X and FSOM_Output ) <> 0 ) then
    begin
        if( length( S ) > 0 ) then
        begin
            S := S + ',' ;
        end ;
        S := S + 'FSOM_Output' ;
    end ;
    if( ( X and FSOM_Text ) <> 0 ) then
    begin
        if( length( S ) > 0 ) then
        begin
            S := S + ',' ;
        end ;
        S := S + 'FSOM_Text' ;
    end ;
    if( ( X and FSOM_Read_Only ) <> 0 ) then
    begin
        if( length( S ) > 0 ) then
        begin
            S := S + ',' ;
        end ;
        S := S + 'FSOM_Read_Only' ;
    end ;
    if( ( X and FSOM_Share_Read ) <> 0 ) then
    begin
        if( length( S ) > 0 ) then
        begin
            S := S + ',' ;
        end ;
        S := S + 'FSOM_Share_Read' ;
    end ;
    if( ( X and FSOM_Share_Write ) <> 0 ) then
    begin
        if( length( S ) > 0 ) then
        begin
            S := S + ',' ;
        end ;
        S := S + 'FSOM_Share_Write' ;
    end ;
    File_Mode_To_String := S + ']' ;
end ;


function TSFile_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

var Dummy : integer ;
    R : PSText_Debugger ;
    R1 : PSFile_Contents_Debugger ;

begin
    new( R, Init ) ;
    case Ordinal of
        0 : S := 'Extendable = ' + Boolean_To_String( _Instance^.Extendable ) ;
        1 : S := 'File_Mode = ' + File_Mode_To_String( _Instance^.File_Mode ) ;
        2 : S := 'File_Type = ' + strpas( _Instance^.Facility_Name ) ;
        3 : S := 'IO_Error = ' + Num1( _Instance^.IO_Error.Facility ) + ', ' + Num1( _Instance^.IO_Error.Code ) ;
        4 : S := 'Name = "' + _Instance^.FileName + '"' ;
        5 : S := 'Record_Size = ' + Num1( _Instance^.Record_Size ) ;
        6 : begin
                Set_Length( S, 128 ) ;
                move( _Instance^.Text_Buffer[ 0 ], S[ 1 ], 128 ) ;
                for Dummy := 1 to 128 do
                begin
                    if( ( ord( S[ Dummy ] ) < 32 ) or ( ord( S[ Dummy ] ) > 126 ) ) then
                    begin
                        S[ Dummy ] := '.' ;
                    end ;
                end ;
                S := 'Text_Buffer = "' + S + '"' ;
            end ;
        7 : S := 'Text_File = ' + Boolean_To_String( _Instance^.Text_File ) ;
        8 : begin
                new( R1, Init( _Instance ) ) ;
                dispose( R, Done ) ;
                R := R1 ;
                S := 'Contents' ;
            end ;
    end ;
    S := S + #0 ;
    R^.Set_Title( @S[ 1 ] ) ;
    Child := PSDebug_Interface( R ) ;
end ; { TFile_Debugger.Child }


{ Unit utility routines... }

function Convert_File_Mode( File_Mode : integer ) : integer ;

var X : integer ;

begin
    X := 0 ;
    if( ( File_Mode and FSOM_Read_Only ) = 0 ) then
    begin
	X := 2 ;
    end ;
(*
    case ( File_Mode and ( FSOM_Share_Write or FSOM_Share_Read ) ) of
	0 : X := X or $10 ;
	FSOM_Share_Write: X := X or $30 ;
	FSOM_Share_Read: X := X or $20 ;
	else X := X or $40 ;
    end ;
*)
    Convert_File_Mode := X ;
end ;



{ TFile_System methods... }

{ Constructors and destructors... }

constructor TFile_System.Init ;

begin
    File_List.Init( 0, 16 ) ;
    Serial_Number := Get_Serial_Number ;
{$IFDEF Debug}
    Sanity := 'TFile_System' ;
{$ENDIF}
end ;


destructor TFile_System.Done ;

var Loop : integer ;

begin
    { Close all files... }
    for Loop := File_List.Count - 1 downto 0 do
    begin
        Close( Loop + 1 ) ;
    end ;
    File_List.Done ;
{$IFDEF Debug}
    Sanity := ' ' ;
{$ENDIF}

    inherited Done ;
end ;


{ Internal utility routines... }

procedure TFile_System.Filespec_String_Scan( const Name, _Separator : string ;
    var Info : TFSSS ) ;

begin
    { This routine left intentionally blank.
      We do no additional processing other than what the file processor does. }
end ;


function TFile_System.Get_File_Rec( Handle : integer ) : PFile_Record ;

var File_Rec : PFile_Record ;

begin
    { Check for errors... }
    dec( Handle ) ; { Handles are zero-based }
    if( ( Handle < 0 ) or ( Handle >= File_List.Count ) ) then
    begin
        Get_File_Rec := nil ;
        Set_Error( File_SystemErr_Facility, File_SystemErr_Invalid_Handle ) ;
        exit ;
    end ;
    File_Rec := PFile_Record( File_List.At( Handle ) ) ;
    Get_File_Rec := File_Rec ;
    if( File_Rec = nil ) then
    begin
        Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
        exit ;
    end ;
end ;


procedure TFile_System.Set_Error( F, X : integer ) ;

var
{$IFDEF DELPHI}
    R : tUEC ;
{$ELSE}
    R : double ;
{$ENDIF}
    UEC : tUEC ;

begin
    IO_Error.Facility := F ;
    IO_Error.Code := X ;
    if( X <> 0 ) then
    begin
	R := Error_Code( F, X ) ;
	UEC := tUEC( R ) ;
	Log_Error( UEC, '', 3 ) ;
    end ;
end ;


{ API... }

{ Increment reference count to file system. }
procedure TFile_System.Attach ;

begin
    { Because this is an abstract object, we do nothing here. }
end ;


{ Decrement reference count to file system. }
procedure TFile_System.Detach ;

begin
    { Because this is an abstract object, we do nothing here. }
end ;


{ Close an open file via handle. }
procedure TFile_System.Close( Handle : integer ) ;

var File_Rec : PFile_Record ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    { Close the file... }
    Clear_IOResult ; { Clear any pending errors }
    {$I-}
    System.Close( File_Rec^._File ) ;
    {$I+}
    Set_Error( 2, IOResult ) ;
    if( IO_Error.Code = 0 ) then
    begin
        dispose( PFile_Record( File_List.At( Handle - 1 ) ) ) ;
        File_List.AtPut( Handle - 1, nil ) ;
    end ;

    { Reclaim unused handles at the end of the file list }
    while(
           ( File_List.Count > 0 )
           and
           ( File_List.At( File_List.Count - 1 ) = nil )
         ) do
    begin
        File_List.AtDelete( File_List.Count - 1 ) ;
    end ;
end ;


{ Close an open find record. }
procedure TFile_System.Close_Find( var Find_Record : TFind_Record ) ;

begin
{$IFDEF DELPHI}
    if( @self = @DOS_FS ) then
    begin
        if( Find_Record.Data[ 0 ] <> #0 ) then
        begin
            Find_Record.Data[ 0 ] := #0 ;
            move( Find_Record.Data[ 1 ], FR, sizeof( FR ) ) ;
            try
                IOResult ; { Clear any pending errors }
                {$I-}
                SysUtils.FindClose( FR ) ;
                {$I+}
                IOResult ; { Ignore any error on clone }
            except
            end ;
        end ;
    end ;
{$ENDIF}
end ;


{ Create file and open with specified mode.  If file already exists, it is
  deleted and re-created.  Returns the file handle (0 if an error). }
function TFile_System.Create( const Name : string ; Mode : integer ) : integer ;

var F : file ;

begin
    Clear_IOResult ; { Clear any pending errors }
    assign( F, Name ) ;
    {$I-}
    System.Rewrite( F ) ;
    {$I+}
    Set_Error( 2, IOResult ) ;
    if( IO_Error.Code <> 0 ) then
    begin
        Create := 0 ; { Creation failure }
    end else
    begin
        System.close( F ) ;
        Create := Open( Name, Mode ) ;
    end ;
end ;


{ Returns a debugger interface to the file system. }
function TFile_System.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ; { Not yet supported }
end ;


{ Deletes the specified file. }
procedure TFile_System.Erase( const Name : String ) ;

var F : File ;

begin
    Clear_IOResult ; { Clear any pending errors }
    assign( F, Name ) ;
    {$I-}
    System.Erase( F ) ;
    {$I+}
    Set_Error( 2, IOResult ) ;
end ;


{ Returns name associated with specified file. }
function TFile_System.File_Name( Handle : integer ) : string ;

var File_Rec : PFile_Record ;
    Temp : FileRec ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    move( File_Rec^, Temp, sizeof( Temp ) ) ;
    File_Name := strpas( PChar( @Temp.Name ) ) ;
end ;


function TFile_System.First_Directory( const Name : string ) : integer ;

var Dummy1 : integer ;

begin
    Dummy1 := pos( ':', Name ) ;
    if( Dummy1 > 0 ) then { Device included }
    begin
        First_Directory := Dummy1 + 1 +
            pos( '\', copy( Name, Dummy1 + 2, length( Name ) ) + '\' ) ;
    end else
    begin
        First_Directory := pos( '\', Name + '\' ) ;
    end ;
end ;


{ Flush any buffered output for specified file. }
procedure TFile_System.Flush( Handle : integer ) ;

var File_Rec : PFile_Record ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;
end ;


{ Returns the attributes of the specified file. }
function TFile_System.Get_Attributes( Handle : integer ) : integer ;

var File_Rec : PFile_Record ;

begin
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        Get_Attributes := 0 ;
        exit ;
    end ;

    Get_Attributes := Get_File_Attributes( File_Name( Handle ) ) ;
end ;


function TFile_System.Get_File_Object( Mode : integer ) : PSFile ;

begin
    Get_File_Object := new( PSFile, Init( @self, ( ( Mode and FSOM_Text ) <> 0 ), 1 ) ) ;
end ;


{ Returns size of specified file, in bytes. }
function TFile_System.Get_Size( Handle : integer ) : TStore_Address ;

var File_Rec : PFile_Record ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        Get_Size := -1 ;
        exit ;
    end ;

    Clear_IOResult ; { Clear any pending errors }
    {$I-}
    Get_Size := System.Filesize( File_Rec^._File ) ;
    {$I+}
    Set_Error( 2, IOResult ) ;
end ;


function TFile_System.Get_Time( Handle, Typ : integer ) : Longint ;

var File_Rec : PFile_Record ;
    Time : longint ;

begin
    { Setup... }
    Time := 0 ;
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        Get_Time := Time ;
        exit ;
    end ;

    Clear_IOResult ; { Clear any pending errors }
    {$I-}
{$IFDEF DELPHI}
    Time := FileAge( File_Name( Handle ) ) ;
{$ELSE}
    {$IFDEF WINDOWS}
        Windos.Getftime( File_Rec^._File, Time ) ;
    {$ELSE}
        Dos.Getftime( File_Rec^._File, Time ) ;
    {$ENDIF}
{$ENDIF}
    {$I+}
    Set_Error( 2, IOResult ) ;
    Get_Time := Time ;
end ;


function TFile_System.Normalize_Path( Name : string ) : string ;

begin
    Normalize_Path := Normalize_Hierarchy( Name, '\', '.', '..', ':' ) ;
end ;


{ Open existing file with specified mode.  Returns the file handle (0 if an
  error). }
function TFile_System.Open( const Name : string ; Mode : integer ) : integer ;

var Save : integer ;
    File_Rec : PFile_Record ;

begin
    new( File_Rec ) ;
    Save := FileMode ;
    FileMode := Convert_File_Mode( Mode ) ;
    Clear_IOResult ; { Clear any pending errors }
    assign( File_Rec^._File, Name ) ;
    {$I-}
    System.Reset( File_Rec^._File, 1 ) ;
    {$I+}
    Set_Error( 2, IOResult ) ;
    FileMode := Save ;
    if( IO_Error.Code <> 0 ) then
    begin
        Open := 0 ; { Open failure }
        dispose( File_Rec ) ;
    end else
    begin
        File_List.Insert( File_Rec ) ;
        Open := File_List.Count ;
    end ;
end ;


{ Read from specified file at specified position into Buffer, up to Len bytes.
  Actual bytes read is returned in Count.  Count is ignored on call. }
procedure TFile_System.Read( Handle : integer ; Position : TStore_Address ;
    Len : TStore_Address ; var Buffer ; var Count : TStore_Address ) ;

var Dummy : integer ;
    File_Rec : PFile_Record ;
    R : TRecord_Size ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    { Position... }
    Clear_IOResult ; { Clear any pending errors }
    {$I-}
{$IFDEF DELPHI}
    System.Seek( File_Rec^._File, Position ) ;
{$ELSE}
    System.Seek( File_Rec^._File, trunc( Position ) ) ;
{$ENDIF}
    {$I+}
    Dummy := IOResult ;
    if( Dummy <> 0 ) then
    begin
        Set_Error( 2, Dummy ) ;
        Count := 0 ;
        exit ;
    end ;

    { Read the data... }
    {$I-}
    System.BlockRead( File_Rec^._File, Buffer, Len, R ) ;
    {$I+}
    Set_Error( 2, IOResult ) ;
    Count := R ;
end ; { TFile_System.Read }


{ Renames Old file to New file name. }
procedure TFile_System.Rename( const Old, New : String ) ;

var F : File ;

begin
    Clear_IOResult ; { Clear any pending errors }
    assign( F, Old ) ;
    {$I-}
    System.Rename( F, New ) ;
    {$I+}
    Set_Error( 2, IOResult ) ;
end ;


procedure TFile_System.Set_Attributes( Handle : integer ; Attr : integer ) ;

var File_Rec : PFile_Record ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    Set_File_Attributes( File_Name( Handle ), Attr ) ;
end ;


procedure TFile_System.Set_Time( Handle : integer ; Typ, Time : Longint ) ;

var File_Rec : PFile_Record ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    Clear_IOResult ; { Clear any pending errors }
{$IFDEF DELPHI}
(*
    Set_Error( 0, FileSetDate( File_Rec^.TF, Time ) ) ;
*)
{$ELSE}
    {$I-}
    {$IFDEF WINDOWS}
        Windos.Setftime( File_Rec^._File, Time ) ;
    {$ELSE}
        Dos.Setftime( File_Rec^._File, Time ) ;
    {$ENDIF}
        {$I+}
        Set_Error( 2, IOResult ) ;
{$ENDIF}
end ;


{ Sets the size of the specified file to the specified size, in bytes. }
procedure TFile_System.Set_Size( Handle : integer ; New_Size : TStore_Address ) ;

var File_Rec : PFile_Record ;

{$IFDEF DELPHI}
var Buffer : array[ 0..32767 ] of char ;
    R : TStore_Address ;
    Size : int64 ;
{$ELSE}

label Finish ;

var Reg : TRegisters ;
    _Handle : integer ;
    Name : string ;
{$ENDIF}

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

{$IFDEF DELPHI}
    Size := Get_Size( Handle ) ;
    if( New_Size > Size ) then { Expand file size }
    begin
        while( New_Size > Size ) do
        begin
            if( New_Size - Size > 32768 ) then
            begin
                Write( Handle, Size, 32768, Buffer, R ) ;
                if( R <> 32768 ) then
                begin
                    exit ;
                end ;
            end else
            begin
                Write( Handle, Size, New_Size - Size, Buffer, R ) ;
                if( R <> New_Size - Size ) then
                begin
                    exit ;
                end ;
            end ;
            Size := Size + R ;
        end ;
        exit ;
    end ;

    Seek( File_Rec^._File, New_Size ) ;
    Truncate( File_Rec^._File ) ;
{$ELSE}
    Name := File_Name( Handle ) + #0 ;
    Reg.AX := $3D02 ;
    Reg.DS := seg( Name ) ;
    Reg.DX := ofs( Name[ 1 ] ) ;
    intr( $21, Reg ) ; { Open file }
    _Handle := Reg.AX ;
    if( ( Reg.Flags and 1 ) = 1 ) then
    begin
	Set_Error( 1, Reg.AX ) ;
	exit ;
    end ;

    Reg.AX := $4200 ;
    Reg.BX := _Handle ;
    Reg.CX := New_Size shr 16 ;
    Reg.DX := New_Size and $FFFF ;
    intr( $21, Reg ) ; { Set position }
    if( ( Reg.Flags and 1 ) = 1 ) then
    begin
	Set_Error( 1, Reg.AX ) ;
	goto Finish ;
    end ;

    Reg.AX := $4000 ;
    Reg.BX := _Handle ;
    Reg.CX := 0 ;
    Reg.DS := DSeg ;
    Reg.DX := ofs( Reg ) ;
    intr( $21, Reg ) ; { Truncate/expand }
    if( ( Reg.Flags and 1 ) = 1 ) then
    begin
	Set_Error( 1, Reg.AX ) ;
    end ;

Finish:
    Reg.AX := $3E00 ;
    Reg.BX := _Handle ;
    intr( $21, Reg ) ; { Close file }
    if( ( Reg.Flags and 1 ) = 1 ) then
    begin
	if( IO_Error.Code <> 0 ) then
	begin
	    Set_Error( 1, Reg.AX ) ;
	end ;
    end ;
{$ENDIF}
end ; { TFile_System.Set_Size }


{ Write to specified file at specified position from Buffer, up to Len bytes.
  Actual bytes written is returned in Count.  Count is ignored on call. }
procedure TFile_System.Write( Handle : integer ; Position : TStore_Address ;
    Len : TStore_Address ; var Buffer ; var Count : TStore_Address ) ;

var File_Rec : PFile_Record ;
    R : TRecord_Size ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    { Position... }
    Clear_IOResult ; { Clear any pending errors }
    {$I-}
{$IFDEF DELPHI}
    System.Seek( File_Rec^._File, Position ) ;
{$ELSE}
    System.Seek( File_Rec^._File, trunc( Position ) ) ;
{$ENDIF}
    {$I+}
    Set_Error( 2, IOResult ) ;
    if( IO_Error.Code <> 0 ) then
    begin
        Count := 0 ;
        exit ;
    end ;

    { Write the data... }
    {$I-}
    System.BlockWrite( File_Rec^._File, Buffer, Len, R ) ;
    {$I+}
    Set_Error( 2, IOResult ) ;
    Count := R ;
end ;


{ Returns last error code. }
procedure TFile_System.Last_Error( var Res : TUEC ) ;

begin
    Res := IO_Error ;
end ;


function TFile_System.DeviceSize( Nam : string ) : TStore_Address ;

begin
    if( ( length( Nam ) < 2 ) or ( Nam[ 2 ] <> ':' ) ) then
    begin
        DeviceSize := DiskSize( 0 ) ;
        exit ;
    end ;
    DeviceSize := DiskSize( ord( Nam[ 1 ] ) - 64 ) ;
end ;


function TFile_System.DeviceFree( Nam : string ) : TStore_Address ;

begin
    if( ( length( Nam ) < 2 ) or ( Nam[ 2 ] <> ':' ) ) then
    begin
        DeviceFree := -1 ;
        exit ;
    end ;
    DeviceFree := DiskFree( ord( Nam[ 1 ] ) - 64 ) ;
end ;


function TFile_System.MkDir( Nam : string ) : longint ;

begin
    Clear_IOResult ; { Clear any pending errors }
    {$I-}
    System.MkDir( Nam ) ;
    {$I+}
    IO_Error.Facility := 1 ;
    IO_Error.Code := IOResult ;
    MkDir := IO_Error.Code ;
end ;


procedure TFile_System.Get_Header( _Path : string ;
    var Header : FS_Header_Information ; var Res : TUEC ) ;

begin
    Set_Error( 0, 0 ) ;
    {TODO}
    Res := IO_Error ;
end ;


procedure TFile_System.Set_Header( _Path : string ;
    var Header : FS_Header_Information ; var Res : TUEC ) ;

begin
    Set_Error( 0, 0 ) ;
    {TODO}
    Res := IO_Error ;
end ;


function Convert_DOS_Attributes( Attr : longint ) : longint ;

var A : longint ;

begin
    A := 0 ;
{$IFDEF DELPHI}
  {$WARNINGS OFF}
{$ENDIF}
    if( ( Attr and faReadOnly ) <> 0 ) then
    begin
	A := A or FSFF_Protected ;
    end ;
    if( ( Attr and faHidden ) <> 0 ) then
    begin
	A := A or FSFF_Hidden ;
    end ;
    if( ( Attr and faSysFile ) <> 0 ) then
    begin
	A := A or FSFF_System ;
    end ;
    if( ( Attr and faVolumeID ) <> 0 ) then
    begin
	A := A or FSFF_VolumeID ;
    end ;
    if( ( Attr and faDirectory ) <> 0 ) then
    begin
	A := A or FSFF_Directory ;
    end ;
    if( ( Attr and faArchive ) <> 0 ) then
    begin
	A := A or FSFF_Archive ;
    end ;
{$IFDEF DELPHI}
  {$WARNINGS ON}
{$ENDIF}
    Convert_DOS_Attributes := A ;
end ; { Convert_DOS_Attributes }


{ Returns the attributes of the specified file. }
function TFile_System.Get_File_Attributes( const Name : string ) : longint ;

{$IFDEF DELPHI}
var Attr : integer ;
{$ELSE}
var Attr : word ;
    F : File ;
{$ENDIF}

begin
    Clear_IOResult ; { Clear any pending errors }
{$IFDEF DELPHI}
    Attr := GetFileAttributes( PChar( Name ) ) ;
{$ELSE}
    assign( F, Name ) ;
    {$I-}
  {$IFDEF WINDOWS}
    Windos.GetFAttr( F, Attr ) ;
  {$ELSE}
    Dos.GetFAttr( F, Attr ) ;
  {$ENDIF}
    {$I+}
    Set_Error( 2, IOResult ) ;
{$ENDIF}
    Get_File_Attributes := Convert_DOS_Attributes( Attr ) ;
end ; { TFile_System.Get_File_Attributes }


function Convert_Attributes( Value : longint ) : longint ;

var Attr : integer ;

begin
    Attr := 0 ;
{$IFDEF DELPHI}
  {$WARNINGS OFF}
{$ENDIF}
    if( ( Value and FSFF_Protected ) <> 0 ) then
    begin
	Attr := Attr or faReadOnly ;
    end ;
    if( ( Value and FSFF_Hidden ) <> 0 ) then
    begin
	Attr := Attr or faHidden ;
    end ;
    if( ( Value and FSFF_System ) <> 0 ) then
    begin
	Attr := Attr or faSysFile ;
    end ;
    if( ( Value and FSFF_VolumeID ) <> 0 ) then
    begin
	Attr := Attr or faVolumeID ;
    end ;
    if( ( Value and FSFF_Directory ) <> 0 ) then
    begin
	Attr := Attr or faDirectory ;
    end ;
    if( ( Value and FSFF_Archive ) <> 0 ) then
    begin
	Attr := Attr or faArchive ;
    end ;
{$IFDEF DELPHI}
  {$WARNINGS ON}
{$ENDIF}
    Convert_Attributes := Attr ;
end ; { Convert_Attributes }


{ Returns the attributes of the specified file. }
procedure TFile_System.Set_File_Attributes( const Name : string ;
    Value : longint ) ;

{$IFDEF DELPHI}
var Attr : integer ;
{$ELSE}
var Attr : word ;
    F : File ;
{$ENDIF}

begin
    Attr := Convert_Attributes( Value ) ;
    Clear_IOResult ; { Clear any pending errors }
{$IFDEF DELPHI}
    if( SetFileAttributes( PChar( Name ), Attr ) ) then
    begin
        Set_Error( File_SystemErr_Facility, 0 ) ;
    end else
    begin
        Set_Error( 0, GetLastError ) ;
    end ;
{$ELSE}
    assign( F, Name ) ;
    {$I-}
  {$IFDEF WINDOWS}
    Windos.Setfattr( F, Attr ) ;
  {$ELSE}
    Dos.Setfattr( F, Attr ) ;
  {$ENDIF}
    {$I+}
    Set_Error( 2, IOResult ) ;
{$ENDIF}
end ;


procedure Invalid_Compile ;

var UEC : TUEC ;

begin
    fillchar( UEC, sizeof( UEC ), 0 ) ;
    Log_Error( UEC, 'Files unit was not compiled correctly', -1 ) ;
    halt ;
end ;


procedure Copy_Find_Info( var Find_Record : TFind_Record ;
    const FR : TSearchRec ) ;

var Dummy : integer ;

begin
    if( sizeof( FR ) > sizeof( Find_Record.Data ) - 1 ) then
    begin
        Invalid_Compile ;
    end ;
    move( FR, Find_Record.Data[ 1 ], sizeof( FR ) ) ;
    Find_Record.Last_Name := FR.Name ;
    Dummy := length( FR.Name ) ;
    while( ( Dummy > 0 ) and ( FR.Name[ Dummy ] <> '\' ) ) do
    begin
        dec( Dummy ) ;
    end ;
    Find_Record.Found_Name := copy( FR.Name, Dummy + 1, length( FR.Name ) ) ;
    Find_Record.Found_Attributes := Convert_DOS_Attributes( FR.Attr ) ;
end ; { Copy_Find_Info }


function TFile_System.FindFirst( Nam : string ; Attributes : longint ;
    var Find_Record : TFind_Record ) : longint ;

var Attr : integer ;
    Dummy : integer ;
    F : file ;

begin
    if( Find_Record.Initialized <> 0 ) then
    begin
        Initialize_Find_Record( Find_Record ) ;
    end ;
    Attr := Convert_Attributes( Attributes ) ;
    IO_Error.Facility := 1 ;
    IO_Error.Code := Compatib.FindFirst( Nam, Attr, FR ) ;
    Find_Record.Data[ 0 ] := #1 ;
    Copy_Find_Info( Find_Record, FR ) ;
    if(
        ( IO_Error.Code = 18 ) { No more files }
        and
        ( pos( '*', Nam ) + pos( '?', Nam ) = 0 ) { No wildcard }
        and
        Compatib.FileExists( Nam )
      ) then
    begin
        IO_Error.Code := 0 ;
        Dummy := length( Nam ) ;
        while( ( Dummy > 0 ) and ( pos( Nam[ Dummy ], ':\' ) = 0 ) ) do
        begin
            dec( Dummy ) ;
        end ;
        Find_Record.Found_Name := copy( Nam, Dummy + 1, length( Nam ) ) ;
        Find_Record.Found_Attributes := 0 ;
        Clear_IOResult ; { Clear any pending errors }
        assign( F, Nam ) ;
        {$I-}
        reset( F ) ;
        {$I+}
        if( IOResult = 5 ) then { Access violation }
        begin
            Find_Record.Found_Attributes := FSFF_Directory ;
        end else
        begin
            System.close( F ) ;
        end ;
    end ;
    FindFirst := IO_Error.Code ;
end ; { TFile_System.FindFirst }


function TFile_System.FindNext( var Find_Record : TFind_Record ) : longint ;

begin
    move( Find_Record.Data[ 1 ], FR, sizeof( FR ) ) ;
    IO_Error.Facility := 1 ;
    IO_Error.Code := Compatib.FindNext( FR ) ;
    FindNext := IO_Error.Code ;
    Copy_Find_Info( Find_Record, FR ) ;
end ;


procedure TFile_System.Initialize( Nam, ID : string ; Flags : longint ;
    var UEC : tUEC ) ;

begin
    IO_Error.Facility := File_SystemErr_Facility ;
    IO_Error.Code := File_SystemErr_Unsupported_Operation ;
    Last_Error( UEC ) ;
end ;


procedure TFile_System.Initialize_Store( S : PSStore ; ID : string ;
    Flags : longint ; var UEC : tUEC ) ;

begin
    IO_Error.Facility := File_SystemErr_Facility ;
    IO_Error.Code := File_SystemErr_Unsupported_Operation ;
    Last_Error( UEC ) ;
end ;


procedure TFile_System.Verify( Name : string ; var Res : TUEC ) ;

begin
    IO_Error.Facility := File_SystemErr_Facility ;
    IO_Error.Code := File_SystemErr_Unsupported_Operation ;
    Last_Error( Res ) ;
end ;


function TFile_System.Separator : string ;

begin
    Separator := '\' ; { This is the default }
end ;


function TFile_System.Path : string ;

begin
    Path := '' ;
end ;


function TFile_System.Facility : longint ;

begin
    Facility := 1 ; { DOS }
end ;


function TFile_System.Version : longint ;

begin
    Version := 10 ; { Version 1.0 }
end ;


function TFile_System.Lock( Nam : string ) : boolean ;

begin
    Lock := False ;
end ;


procedure TFile_System.Unlock( Nam : string ) ;

begin
    { This routine left intentionally blank }
end ;


var F : Text ;

function Find_Logical( const X : string ) : string ;
{ Try to resolve a logical to a valid path or device }

var Env : string ;

begin
    Env := OS^.Get_Symbol( X ) ; { Get logical definition }
    if( length( Env ) = 0 ) then { Not defined }
    begin
	if( length( X ) = 1 ) then { A valid device }
	begin
	    Find_Logical := X ;
	end else
	begin
	    Find_Logical := '' ;
	end ;
	exit ;
    end ;

    Env := Find_File_In_Path( 'X.X', Env ) ;
    Env := copy( Env, 1, length( Env ) - 3 ) ; { Subtract off "X.X" }
    Find_Logical := Env ;
end ; { Find_Logical }


const Max_Find_File_Nest = 8 ;

var Nested : integer ;

function Find_File_In_Path( const Nam : string ; Env : string ) : string ;

var Dummy, Index : integer ;
    This_Path, XX : string ;

begin
    if( Nested > Max_Find_File_Nest ) then
    begin
	Find_File_In_Path := '' ; { Nested too deep }
	exit ;
    end ;

    while( length( Env ) > 0 ) do { Process all items in list }
    begin
	{ Extract next path in list...}
        Index := pos( ';', Env ) ; { End of next item }
	if( Index = 0 ) then
	begin
	    Index := length( Env ) + 1 ;
	end ;
        This_Path := copy( Env, 1, Index - 1 ) ;
	Env := copy( Env, Index + 1, 254 ) ;

	{ See if we can find a file... }
        if( length( This_Path ) > 0 ) then { Non-null path }
        begin
            if( This_Path[ 1 ] = '_' ) then { Use literal device }
            begin
                Clear_IOResult ; { Clear any pending errors }
                assign( F, copy( This_Path, 2, length( This_Path ) ) ) ;
                {$I-}
                reset( F ) ;
                {$I+}
                Dummy := IOResult ;
                if( Dummy = 0 ) then
                begin
                    {$I-}
                    close( F ) ;
                    {$I+}
                    Clear_IOResult ;
                    XX := copy( This_Path, 2, length( This_Path ) ) ;
                    if( XX[ length( XX ) ] = '\' ) then
                    begin
                        Find_File_In_Path := XX + Nam ;
                    end else
                    begin
                        Find_File_In_Path := XX + '\' + Nam ;
                    end ;
                    exit ;
                end ; {if( IOResult = 0 ) then}
            end ; {if( B[ 1 ] = '_' ) then}
            inc( Nested ) ;
            XX := Find_Logical( This_Path ) ; { Recursively translate the logical }
            dec( Nested ) ;
            if( length( XX ) = 0 ) then
            begin
                XX := This_Path ;
            end ;
            if( length( XX ) > 0 ) then
            begin
                Clear_IOResult ; { Clear any pending errors }
                if( XX[ length( XX ) ] = '\' ) then
                begin
                    assign( F, XX + Nam ) ;
                end else
                begin
                    assign( F, XX + '\' + Nam ) ;
                end ;
                {$I-}
                reset( F ) ;
                {$I+}
                Dummy := IOResult ;
                if( Dummy = 0 ) then
                begin
                    {$I-}
                    close( F ) ;
                    {$I+}
                    Clear_IOResult ;
                end else
                if( Dummy = 3 ) then { Path not found or invalid }
                begin
                    XX := '' ;
                end ;
            end ; { if( length( XX ) > 0 ) }
            if( length( XX ) <> 0 ) then { Found a valid path }
            begin
                if( XX[ length( XX ) ] = '\' ) then
                begin
                    Find_File_In_Path := XX + Nam ;
                end else
                begin
                    Find_File_In_Path := XX + '\' + Nam ;
                end ;
                exit ;
            end ;
        end ;
    end ;
    Find_File_In_Path := '' ; { Nothing valid was found }
end ; { Find_File_In_Path }


function Find_File( X : String ; Literal : boolean ) : String ;

    function Combine( const Path, Work : string ) : string ;

    begin
        if( ( copy( Path, length( Path ), 1 ) = '\' ) and ( copy( Work, 1, 1 ) = '\' ) ) then
        begin
            Combine := Path + copy( Work, 2, length( Work ) ) ;
        end else
        begin
            Combine := Path + Work ;
        end ;
    end ;

label Skip ;

var A, Dummy : Integer ;
    B, Env, Path, Work : String ;
    F : Text ;
    Saved : integer ;

Begin
    { First, see if we are nested more than eight (8) deep.  If so, we
      automatically fail.  This prevents any infinite loop due to recursive
      symbol definitions. }
    inc( Nested ) ; { Nested one more level }
    if( Nested = 8 ) then { Nested too deep }
    begin
	Find_File := '' ; { Not found }
	exit ;
    end ;

    { Now try to find the file }

    while length( X ) > 0 do { Check all items in list }
    begin
	{ Get next item in list }

	if( Literal ) then
        begin
            A := 0 ;
        end else
        begin
            A := pos( ';', X ) ;
        end ;
	if( A = 0 ) then
        begin
            A := length( X ) + 1 ;
        end ;
	B := copy( X, 1, A - 1 ) ; {Get list item}
	X := copy( X, A + 1, 254 ) ; {What remains}

	{ Check for literal device name }

	if( B[ 1 ] = '_' ) then {Use literal device}
	begin
            assign( F, copy( B, 2, 254 ) ) ;
            Saved := FileMode ;
            FileMode := 0 ;
            Clear_IOResult ; { Clear any pending errors }
	    {$I-}
	    reset( F ) ;
	    {$I+}
            FileMode := Saved ;
            Dummy := IOResult ;
	    if( Dummy = 0 ) then
	    begin
		close( F ) ;
		Find_File := copy( B, 2, 254 ) ;
		exit ;
	    end ; {if( IOResult = 0 ) then}
	end ; {if( B[ 1 ] = '_' ) then}

	{Handle case where there is no device/logical specified}

	A := pos( ':', B ) ;
	if( A = 0 ) then {No device/logical}
	begin
            Clear_IOResult ; { Clear any pending errors }
            assign( F, B ) ;
            Saved := FileMode ;
            FileMode := 0 ;
	    {$I-}
	    reset( F ) ;
	    {$I+}
            FileMode := Saved ;
            Dummy := IOResult ;
	    if( Dummy = 0 ) then
	    begin
		close( F ) ;
		Find_File := B ;
		exit ;
	    end ; {if( IOResult = 0 ) then}
	    continue ;
	end ; {if( A = 0 ) then}

	{ Try opening with symbol as a logical... }

	Path := '' ;
	Env := OS^.Get_Symbol( copy( B, 1, A - 1 ) ) ; { Get device }
	Work := copy( B, A + 1, 254 ) ; {Get only file name/extension}
	while length( Env ) > 0 Do {Try all specifications}
	begin
	    A := Pos( ';', Env ) ;
	    if( A = 0 )
		then
		    A := length( Env ) + 1 ;
	    Path := Copy( Env, 1, A - 1 ) ; {Get next specification}
	    Env := Copy( Env, A + 1, Length( Env ) - 1 ) ;
	    if( length( Path ) > 0 ) then
	    begin
		if( Path[ Length( Path ) ] = ':' ) then {Nested logical}
		begin
		    Path := Find_File( Combine( Path, Work ), False ) ;
		    if( Path <> '' ) then {Found it}
		    begin
			Find_File := Path ;
			exit ;
		    end ;
		    goto Skip ;
		end ;
		if( Path[ Length( Path ) ] <> '\' )
		    then
			Path := Path + '\' ;
	    end ;
            Clear_IOResult ; { Clear any pending errors }
	    assign( F, Combine( Path, Work ) ) ;
            Saved := FileMode ;
            FileMode := 0 ;
	    {$I-}
	    reset( F ) ;
	    {$I+}
            FileMode := Saved ;
            Dummy := IOResult ;
	    if( Dummy = 0 ) Then
	    begin
		close( F ) ;
		Find_File := Combine( Path, Work ) ;
		exit ;
	    end ;
	end ; {while length( Env ) > 0 do}

Skip:
        { Now try device as a real device name... }

        Clear_IOResult ; { Clear any pending errors }
	assign( F , B ) ;
        Saved := FileMode ;
        FileMode := 0 ;
	{$I-}
	reset( F ) ;
	{$I+}
        FileMode := Saved ;
        Dummy := IOResult ;
	if( Dummy = 0 ) then
	begin
	    close( F ) ;
	    Find_File := B ;
	    exit ;
	end ; { if( IOResult = 0 ) }
    end ;

    Find_File := '' ; { File was not found }
end ; { Find_File }


function Normalize_Hierarchy( Name, Separator, Self_Reference, Parent_Reference,
    Prefix : string ) : string ; { UNTESTED }

var Dummy, Dummy1 : integer ;
    S : string ;

begin
    { Handle initial prefix }
    if( Prefix <> '' ) then
    begin
        Dummy1 := pos( Prefix, Name ) ;
        if( ( copy( Name, Dummy1 + 1, 1 ) <> Separator ) and ( Dummy1 > 0 ) ) then
        begin
            Name := copy( Name, 1, Dummy1 ) + Separator +
                copy( Name, Dummy1 + 1, length( Name ) ) ;
        end ;
    end ;

    { Strip out self-references... }
    if( Self_Reference <> '' ) then
    begin
        S := Separator + Self_Reference + Separator ;
        Dummy1 := pos( S, Name + Separator ) ;
        while( Dummy1 <> 0 ) do
        begin
            Name := copy( Name, 1, Dummy1 - 1 ) + copy( Name, Dummy1 + length( S ), length( Name ) ) ;
            Dummy1 := pos( S, Name + Separator ) ;
        end ;
        S := Self_Reference + Separator ;
        while( copy( Name, 1, length( S ) ) = S ) do
        begin
            Name := copy( Name, length( S ) + 1, length( Name ) ) ;
        end ;
    end ;

    { Process parent references... }
    if( Parent_Reference <> '' ) then
    begin
        S := Separator + Parent_Reference + Separator ;
        Dummy1 := pos( S, Name + Separator ) ;
        while( Dummy1 <> 0 ) do
        begin
            Dummy := Dummy1 - 1 ;
            while( ( Dummy > 0 ) and ( copy( Name, Dummy, length( Separator ) ) <> Separator ) ) do
            begin
                dec( Dummy ) ;
            end ;
            if(
                ( Dummy > 0 )
                and
                ( Prefix <> '' )
                and
                ( Dummy < pos( Prefix, Name ) )
              ) then // Indicates a prefix immediately followed by a parent reference
            begin
                Name := copy( Name, 1, Dummy1 - 1 ) + copy( Name, Dummy1 + length( S ), length( Name ) ) ;
            end else
            begin
                Name := copy( Name, 1, Dummy ) + copy( Name, Dummy1 + length( S ), length( Name ) ) ;
            end ;
            Dummy1 := pos( S, Name + Separator ) ;
        end ;
    end ;

    Normalize_Hierarchy := Name ;
end ; { Normalize_Hierarchy }


function Explicit( X : string ) : string ;

var Dummy, Temp : integer ;
    Work : string ;

begin
    { Handle missing drive and current directory }
    if( copy( X, 1, 2 ) <> '\\' ) then { Don't process network specification }
    begin
	if( copy( X, 2, 1 ) <> ':' ) then { Add drive, if needed }
	begin
	    GetDir( 0, Work ) ;
	    X := copy( Work, 1, 2 ) + X ;
	end ;
	if( copy( X, 3, 1 ) <> '\' ) then { Add current path, if needed }
	begin
	    GetDir( ord( X[ 1 ] ) - 64, Work ) ;
            if( copy( Work, length( Work ), 1 ) <> '\' ) then
            begin
                Work := Work + '\' ;
            end ;
	    X := Work + copy( X, 3, length( X ) ) ;
	end ;
    end ;

    { Remove \. }
    Dummy := pos( '\.\', X + '\' ) ;
    while( Dummy > 0 ) do
    begin
	X := copy( X, 1, Dummy - 1 ) + copy( X, Dummy + 2, length( X ) ) ;
	Dummy := pos( '\.\', X + '\' ) ;
    end ;

    { Now process \.. }
    Dummy := pos( '\..\', X + '\' ) ;
    while( Dummy <> 0 ) do
    begin
	X := copy( X, 1, Dummy - 1 ) + copy( X, Dummy + 3, length( X ) ) ;
	Temp := Dummy - 1 ;
	while( ( Temp > 2 ) and ( X[ Temp ] <> '\' ) ) do
	begin
	    dec( Temp ) ; { Backup to beginning of previous directory in string }
	end ;
	if( X[ Temp ] = '\' ) then { Found one }
	begin
	    X := copy( X, 1, Temp - 1 ) + copy( X, Dummy, length( X ) ) ;
	end ;
	Dummy := pos( '\..\', X + '\' ) ;
    end ;

    { Return Res }
    Explicit := X ;
end ;


function FileName( X : String ) : string ;
{Same as FILNAM, except the passed name, with translated logicals, is returned
 if no file is found}

var B, Dummy : integer ;
    Env, XX : string ;

begin
    XX := FILNAM( X, False ) ;
    if( length( XX ) = 0 ) then
    begin
	{Reduce list, if one, to first file}
	Dummy := pos( ';', X ) ;
	if( Dummy > 0 ) then
	begin
	    X := copy( X, 1, Dummy - 1 ) ;
	end ;

	B := pos( ':', X ) ;
	if( B > 0 ) then {A possible logical}
	begin
	    if( X[ 1 ] = '_' ) then {Force physical}
	    begin
		X := copy( X, 2, 254 ) ; {Remove underscore}
	    end else
	    begin
		Env := copy( X, 1, B - 1 ) ; {Get device/logical}
		X := copy( X, B + 1, 254 ) ; {Get only file name/extension}
		Nested := 0 ;
		X := Find_Logical( Env ) + X ;
	    end ;
	end ;

	XX := X ;
    end ;
    FileName := XX ;
end ; { FileName }


function FILNAM( const X : string ; Literal : boolean ) : string ;
{ Find a file given a file name.  Name can contain logicals (symbols), and
  search lists (;).  If no device/directory is specified and file is not
  found with name given, it is searched for in PATH:.  Literal devices are
  specified with underscores (_).  If a logical translation fails, the
  "logical" is tried as if a literal device name.  Literal is true if the
  characters of the filename are to be treated literally (no special handling
  for ;). }

var Dummy : integer ;
    S, Work : string ;

begin
    {Pass 1 : Try as passed}
    S := X ;
    while( length( S ) > 0 ) do { Try each listed item }
    begin
	if( Literal ) then
        begin
            Dummy := 0 ;
        end else
        begin
            Dummy := pos( ';', S ) ;
        end ;
	if( Dummy = 0 ) then
        begin
	    Dummy := length( S ) + 1 ;
        end ;
	Work := copy( S, 1, Dummy - 1 ) ;
	S := copy( S, Dummy + 1, 254 ) ;
	Nested := 0 ;
	Work := Find_File( Work, Literal ) ;
	if( Work <> '' ) then { Found file as passed }
	begin
	    FILNAM := Work ;
	    exit ;
	end ;
    end ;


    {Pass 2 : Try it with PATH:}

    S := X ;
    while length( S ) > 0 do {Try each listed item}
    begin
	if( Literal ) then
        begin
            Dummy := 0 ;
        end else
        begin
            Dummy := pos( ';', S ) ;
        end ;
	if( Dummy = 0 )
	    then
		Dummy := length( S ) + 1 ;
	Work := copy( S, 1, Dummy - 1 ) ;
	S := copy( S, Dummy + 1, 254 ) ;
	if( pos( ':', Work ) + pos( '\', Work ) = 0 ) then { No device/directory }
	begin
	    Nested := 0 ;
	    Work := Find_File( 'PATH:' + Work, Literal ) ;
	    if( Work <> '' ) then {Found file}
	    begin
		FILNAM := Work ;
		exit ;
	    end ; {if( Work <> '' ) then}
	end ; {if( pos( ':', Work ) + pos( '\', Work ) = 0 ) then}
    end ; {while length( S ) > 0 do}
    FILNAM := '' ; {Not found at all}
end ; { FILNAM }


function Get_Temp_File_Name( const Prefix : string ) : string ;

{$IFDEF DELPHI}
var Dummy : integer ;
    S, S1 : string ;

begin
    System.setlength( S, MAX_PATH + 1 ) ;
    Dummy := GetTempPath( MAX_PATH, PChar( S ) ) ;
    if( Dummy = 0 ) then
    begin
        S := '' ;
    end else
    begin
        System.Setlength( S, Dummy ) ;
    end ;
    System.setlength( S1, MAX_PATH + 1 ) ;
    if( GetTempFileName( PChar( S ), PChar( Prefix ), 0, PChar( S1 ) ) = 0 ) then
    begin
        S1 := Prefix + '.TMP' ;
    end else
    begin
        System.setlength( S1, strlen( PChar( S1 ) ) ) ;
    end ;
    Result := S1 ;
{$ELSE}

begin
    Get_Temp_File_Name := FileName( 'TEMP:' + Prefix + '.TMP' ) ;
{$ENDIF}
end ; { Get_Temp_File_Name }



{ TSFile methods... }

{ Constructors and destructors... }

constructor TSFile.Init( FS : PFile_System ; Tf : Boolean ; Size : longint ) ;

begin
    inherited Init ;

{$IFDEF Debug}
    Sanity := 'TSFile' ;
{$ENDIF}
    Handle := 0 ; { File starts out closed }
    Text_File := Tf ;
    Record_Size := Size ;
    Extendable := True ;
    Serial_Number := Get_Serial_Number ;
    if( FS = nil ) then
    begin
        FS := @DOS_FS ;
    end ;
    File_System := FS ;
    FS^.Attach ;
    BufSize := sizeof( Text_Buffer ) ;
    Reset_Buffer ;
    _N := #0 ;
{$IFDEF DELPHI}
    System.Initialize( Current_Line ) ;
{$ELSE}
    Current_Line := '' ;
{$ENDIF}
end ;


destructor TSFile.Done ;

begin
    if( Handle <> 0 ) then
    begin
        Close ; { Make sure file is closed }
    end ;
    if( File_System <> nil ) then
    begin
        File_System^.Detach ;
        File_System := nil ;
    end ;
{$IFDEF Debug}
    Sanity := '     ' ;
{$ENDIF}

    inherited Done ;
end ;


{ Overrides... }

function TSFile.Extend( Amount : TStore_Address ) : TStore_Address ;

var Old_Size : TStore_Address ;

begin
    if( Amount = 0 ) then
    begin
        if( Extendable ) then
        begin
            Extend := 0 ;
        end else
        begin
            Extend := -1 ;
        end ;
        exit ;
    end ;
    Old_Size := Get_Size ;
    Set_Size( Old_Size + Amount ) ;
    Extend := Get_Size - Old_Size ;
end ;


function TSFile.Get_Name : string ;

begin
    Get_Name := FileName ;
end ;


function TSFile.Read_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : TUEC ) : TStore_Address ;

var R : TStore_Address ;

begin
    File_System^.Read( Handle, Address, _Size, Data, R ) ;
    File_System^.Last_Error( IO_Error ) ;
    Read_Data := R ;
    UEC := IO_Error ;
end ;


function TSFile.Write_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : TUEC ) : TStore_Address ;

var R : TRecord_Size ;

begin
    Seek( Address ) ;
    Blockwrite( Data, _Size, R ) ;
    Write_Data := R ;
    UEC := IO_Error ;
end ;


function TSFile.Max_Storage : TStore_Size ;

begin
    Max_Storage := Get_Size ;
end ;


function TSFile.Min_Storage : TStore_Address ;

begin
    Min_Storage := Record_Size ;
end ;


function TSFile.Contiguous_Store : boolean ;

begin
    Contiguous_Store := True ;
end ;


procedure TSFile.Set_Max_Storage( Value : TStore_Address ; var Res : TUEC ) ;

begin
    Set_Size( Value ) ;
    Res := IO_Error ;
end ;


{ API... }

function TSFile.Get_Time : longint ;

var Temp : longint ;

begin
    GetFTime( Temp ) ;
    Get_Time := Temp ;
end ;


procedure TSFile.Set_Time( Value : longint ) ;

begin
    SetFTime( Value ) ;
end ;


procedure TSFile.Set_Error( F, X : integer ) ;

var
{$IFDEF DELPHI}
    R : tUEC ;
{$ELSE}
    R : double ;
{$ENDIF}
    UEC : tUEC ;

begin
    IO_Error.Facility := F ;
    IO_Error.Code := X ;
    if( X <> 0 ) then
    begin
	R := Error_Code( F, X ) ;
	UEC := tUEC( R ) ;
	Log_Error( UEC, '', 3 ) ;
    end ;
end ;


procedure TSFile.Reset( const Name : String ) ;

begin
    if( length( Name ) = 0 ) then
    begin
        Seek( 0 ) ;
        exit ;
    end ;
    if( Handle <> 0 ) then
    begin
        Close ;
    end ;
    Handle := File_System^.Open( Name, File_Mode ) ;
    File_System^.Last_Error( IO_Error ) ;
    Reset_Buffer ;
end ;


procedure TSFile.Rewrite( const Name : String ) ;

begin
    if( Handle <> 0 ) then
    begin
        Close ;
    end ;
    Handle := File_System^.Create( Name, File_Mode ) ;
    File_System^.Last_Error( IO_Error ) ;
    Reset_Buffer ;
end ;


procedure TSFile.Append( const Name : String ) ;

begin
    Reset( Name ) ;
    if( Handle <> 0 ) then
    begin
        Seek( Get_Size ) ;
    end ;
end ;


procedure TSFile.Close ;

begin
    Flush ; { Make sure buffered output gets written before close }
    File_System^.Close( Handle ) ;
    File_System^.Last_Error( IO_Error ) ;
    Handle := 0 ;
    Current_Line := '' ;
end ;


procedure TSFile.Seek( Pos : TStore_Address ) ;

begin
    if( Pos > Get_Size ) then
    begin
        Set_Error( File_SystemErr_Facility, File_SystemErr_EndOfFile ) ;
        exit ;
    end ;
    Current_Position := Record_Size ;
    Current_Position := Current_Position * Pos ;
    BufPos := BufSize + 1 ;
    Set_Error( 2, 0 ) ;
end ;


procedure TSFile.Read( var Buffer ) ;

var Dummy : TRecord_Size ;

begin
    if( Text_File ) then
    begin
        if( not Opened ) then
        begin
            Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
            exit ;
        end ;
        string( Buffer ) := Read_Text_Buffer( False ) ;
        Current_Line := '' ;
    end else
    begin
        Blockread( Buffer, Record_Size, Dummy ) ;
    end ;
end ;


procedure TSFile.Write( var Buffer ) ;

var Dummy : TRecord_Size ;

begin
    if( Text_File ) then
    begin
        if( not Opened ) then
        begin
            Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
            exit ;
        end ;
        Blockwrite( string( Buffer )[ 1 ], length( string( Buffer ) ), Dummy ) ; { Write string }
    end else
    begin
        Blockwrite( Buffer, Record_Size, Dummy ) ;
    end ;
end ;


procedure TSFile.Readln( var Buffer : String ) ;

begin
    if( not Opened ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
	exit ;
    end ;
{
    if( not Text_File ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
	exit ;
    end ;
}
    Buffer := Read_Text_Buffer( True ) ;
    Current_Line := '' ;
end ; { TSFile.Readln }


procedure TSFile.Writeln( Buffer : String ) ;

var S : string[ 2 ] ;
    W : TRecord_Size ;

begin
    if( not Opened ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
	exit ;
    end ;
    if( not Text_File ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
	exit ;
    end ;
    Blockwrite( Buffer[ 1 ], length( Buffer ), W ) ; { Write string }
    S := CRLF ;
    Blockwrite( S[ 1 ], 2, W ) ; { Write CRLF }
    BufPos := BufSize + 1 ;
end ;


procedure TSFile.BlockRead( var Buf ; Count : TRecord_Size ;
    var Res : TRecord_Size ) ;

var UEC : TUEC ;

begin
    Res := Read_Data( Buf, Current_Position, Count, UEC ) ;
    Current_Position := Current_Position + Res ;
    BufPos := BufSize + 1 ;
    Current_Line := '' ;
end ;


procedure TSFile.Blockwrite( var Buf ; Count : TRecord_Size ;
    var Res : TRecord_Size ) ;

var R : TStore_Address ;

begin
    File_System^.Write( Handle, Current_Position, Count, Buf, R ) ;
    Res := R ;
    File_System^.Last_Error( IO_Error ) ;
    Current_Position := Current_Position + Res ;
    BufPos := BufSize + 1 ;
    Current_Line := '' ;
end ;


function TSFile.EOF : boolean ;

begin
    if( Text_File ) then
    begin
        if( not Opened ) then
        begin
            Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
            EOF := True ;
            exit ;
        end ;
        EOF := ( Current_Position >= Get_Size ) and ( BufPos >= BufEnd ) ;
        if( BufPos <= BufEnd ) then
        begin
            if( BufPtr[ BufPos ] = #26 ) then
            begin
                EOF := True ; { EOF if sitting at Control-Z }
            end ;
        end ;
    end else
    begin
        EOF := ( Current_Position >= Get_Size ) ;
    end ;
end ;


function TSFile.EOLn : boolean ;

var R : TRecord_Size ;

begin
    EOLn := True ;
    if( not Opened ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
	exit ;
    end ;
    if( not Text_File ) then
    begin
        Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
        exit ;
    end ;
    if( ( BufPos > BufSize + 1 ) or ( BufEnd > BufSize ) ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_Internal_Corruption ) ;
        exit ;
    end ;

    if( BufPos > BufEnd ) then
    begin { Need to fill buffer }
	Blockread( BufPtr^, BufSize, R ) ;
        BufEnd := R ;
        if( ( BufPos > BufSize + 1 ) or ( BufEnd > BufSize ) ) then
        begin
            Set_Error( File_SystemErr_Facility, File_SystemErr_Internal_Corruption ) ;
            exit ;
        end ;
	if( BufEnd = 0 ) then
	begin
            EOLn := True ;
	    exit ;
	end else
	begin
	    dec( BufEnd ) ;
	end ;
	BufPos := 0 ;
    end ;
    EOLn := (
              ( BufPtr[ BufPos + 1 ] <> #26 )
              and
              ( BufPtr[ BufPos + 1 ] <> CR )
              and
              ( BufPtr[ BufPos + 1 ] <> LF )
            ) ;
    Set_Error( 2, 0 ) ;
end ; { TSFile.EOLn }


function TSFile.FilePos : TStore_Size ;

begin
    if( not Opened ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
        FilePos := -1 ;
        exit ;
    end ;
{$IFDEF DELPHI}
    FilePos := Current_Position div Record_Size ;
{$ELSE}
    FilePos := trunc( Current_Position / Record_Size ) ;
{$ENDIF}
    Set_Error( 2, 0 ) ;
end ;


function TSFile.Get_Size : TStore_Size ;

begin
    Get_Size := File_System^.Get_Size( Handle ) div Record_Size ;
    File_System^.Last_Error( IO_Error ) ;
end ;


procedure TSFile.Set_Size( New_Size : TStore_Size ) ;

begin
    if( New_Size = -1 ) then
    begin
        New_Size := FilePos ;
    end ;
    File_System^.Set_Size( Handle, New_Size * Record_Size ) ;
    File_System^.Last_Error( IO_Error ) ;
    if( Current_Position > New_Size + 1 ) then
    begin
        Current_Position := New_Size + 1 ; { Cannot be positioned past EOF }
    end ;
end ;


procedure TSFile.Flush ;

begin
    File_System^.Flush( Handle ) ;
    File_System^.Last_Error( IO_Error ) ;
end ;


procedure TSFile.GetFAttr( var Attr : Word ) ;

begin
    Attr := File_System^.Get_Attributes( Handle ) ;
    File_System^.Last_Error( IO_Error ) ;
end ;


procedure TSFile.Erase( const Name : String ) ;

begin
    if( length( Name ) = 0 ) then
    begin
        if( not Opened ) then
        begin
            Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
            exit ;
        end ;
        File_System^.Erase( FileName ) ;
    end else
    begin
        File_System^.Erase( Name ) ;
    end ;
    File_System^.Last_Error( IO_Error ) ;
end ;


function TSFile.FileName : string ;

begin
    FileName := File_System^.File_Name( Handle ) ;
    File_System^.Last_Error( IO_Error ) ;
end ;


function TSFile.Read_Integer : TStore_Address ;

var Dummy : integer ;
    Res : TStore_Address ;
    S : string ;

begin
    Read_Integer := 0 ;
    if( not Opened ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
        exit ;
    end ;

    if( not Text_File ) then
    begin
        Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
        exit ;
    end ;

    if( length( Current_Line ) = 0 ) then
    begin
        Readln( Current_Line ) ;
        if( IO_Error.Code <> 0 ) then
        begin
            exit ;
        end ;
    end ;

    FEdit( Current_Line, Current_Line, 8 ) ; { Remove leading spaces }
    Dummy := pos( ' ', Current_Line + ' ' ) ;
    S := copy( Current_Line, 1, Dummy - 1 ) ;
    Current_Line := copy( Current_Line, Dummy, length( Current_Line ) ) ;
    val( S, Res, Dummy ) ;
    if( Dummy <> 0 ) then
    begin
        Set_Error( File_SystemErr_Facility, File_SystemErr_Data_Format_Error) ;
    end ;
    Read_Integer := Res ;
end ; { TSFile.Read_Integer }


function TSFile.Read_Float : Extended ;

var Dummy : integer ;
    Res : Extended ;
    S : string ;

begin
    Read_Float := 0 ;
    if( not Opened ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
        exit ;
    end ;

    if( not Text_File ) then
    begin
        Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
        exit ;
    end ;

    if( length( Current_Line ) = 0 ) then
    begin
        Readln( Current_Line ) ;
        if( IO_Error.Code <> 0 ) then
        begin
            exit ;
        end ;
    end ;

    FEdit( Current_Line, Current_Line, 8 ) ; { Remove leading spaces }
    Dummy := pos( ' ', Current_Line + ' ' ) ;
    S := copy( Current_Line, 1, Dummy - 1 ) ;
    Current_Line := copy( Current_Line, Dummy, length( Current_Line ) ) ;
    val( S, Res, Dummy ) ;
    if( Dummy <> 0 ) then
    begin
        Set_Error( File_SystemErr_Facility, File_SystemErr_Data_Format_Error) ;
    end ;
    Read_Float := Res ;
end ; { TSFile.Read_Float }


function TSFile.SeekEOF : Boolean ;

begin
    SeekEOF := True ;
    if( not Opened ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
        exit ;
    end ;

    if( not Text_File ) then
    begin
        Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
        exit ;
    end ;

    SeekEOF := ( length( Current_Line ) = 0 ) and EOF ;
end ;


function TSFile.SeekEOLn : Boolean ;

begin
    SeekEOLN := True ;
    if( not Opened ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
        exit ;
    end ;

    if( not Text_File ) then
    begin
        Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
        exit ;
    end ;

    SeekEOLn := ( length( Current_Line ) = 0 ) ;
end ;


procedure TSFile.SetFAttr( Attr : Word ) ;

begin
    File_System^.Set_Attributes( Handle, Attr ) ;
    File_System^.Last_Error( IO_Error ) ;
end ;


procedure TSFile.GetFTime( var _Time : Longint ) ;

begin
    _Time := File_System^.Get_Time( Handle, 0 ) ;
    File_System^.Last_Error( IO_Error ) ;
end ;


procedure TSFile.SetFTime( _Time : Longint ) ;

begin
    File_System^.Set_Time( Handle, 0, _Time ) ;
    File_System^.Last_Error( IO_Error ) ;
end ;


procedure TSFile.SetTextBuf( var Buf ; Size : Word ) ;

begin
    if( Size < 1 ) then { Bad buffer size }
    begin
        Set_Error( 2, File_SystemErr_Invalid_Record_Size ) ;
        exit ;
    end ;
    BufPtr := PChar( @Buf ) ;
    BufSize := Size ;
    BufPos := BufSize + 1 ;
    Set_Error( 2, 0 ) ;
end ;


procedure TSFile.Set_Extend( State : boolean ) ;

begin
    Set_Error( File_SystemErr_Facility, 0 ) ;
    Extendable := State ;
end ;


procedure TSFile.Truncate( Size : TStore_Address ) ;

begin
    if( not Opened ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
        exit ;
    end ;

    if( Get_Size > Size ) then
    begin
        File_System^.Set_Size( Handle, Size ) ;
    end ;
    File_System^.Last_Error( IO_Error ) ;
end ;


procedure TSFile.Initialize ;

begin
    Rewrite( '' ) ;
end ;


function TSFile.Debugger : PSDebug_Interface ;

var R : PSFile_Debugger ;

begin
    new( R, Init( @self ) ) ;
    Debugger := R ;
end ;


function TSFile.Get_Attributes : longint ;

var Attr : word ;

begin
    GetFAttr( Attr ) ;
    Get_Attributes := Attr ;
end ;


procedure TSFile.Set_Attributes( Value : longint ) ;

var Attr : word ;

begin
    Attr := Value ;
    Setfattr( Attr ) ;
end ;


function TSFile.Facility : longint ;

begin
    Facility := TFileErr_Facility ;
end ;


function TSFile.Version : longint ;

begin
    Version := 10 ; { Version 1.0 }
end ;


const TFile_Facility_Name : string[ 7 ] = 'TSFile'#0 ;

function TSFile.Facility_Name : PChar ;

begin
    Facility_Name := @TFile_Facility_Name[ 1 ] ;
end ;


function TSFile.Opened : boolean ;

begin
    Opened := ( Handle <> 0 ) ;
end ;



{ Internal utility routines... }

procedure TSFile.Reset_Buffer ;

begin
    BufPos := BufSize + 1 ;
    BufEnd := 0 ;
    BufPtr := @Text_Buffer ;
end ;


function TSFile.Read_Text_Buffer( Eat_Delimiters : boolean ) : string ;

label Next_Char ;

var S : string ;
    R : TRecord_Size ;

begin
    if( ( BufPos > BufSize + 1 ) or ( BufEnd > BufSize ) ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_Internal_Corruption ) ;
        exit ;
    end ;
    S := '' ;
Next_Char:
    if( BufPos > BufEnd ) then
    begin { Need to fill buffer }
	Blockread( BufPtr[ 0 ], BufSize, R ) ;
        BufEnd := R ;
        if( ( BufPos > BufSize + 1 ) or ( BufEnd > BufSize ) ) then
        begin
            Set_Error( File_SystemErr_Facility, File_SystemErr_Internal_Corruption ) ;
            exit ;
        end ;
	if( BufEnd = 0 ) then
	begin
            if( length( S ) > 0 ) then { There was something at the end of the file }
            begin
                Read_Text_Buffer := S ;
                Set_Error( File_SystemErr_Facility, 0 ) ;
            end else
            begin
	        Set_Error( File_SystemErr_Facility, File_SystemErr_EndOfFile ) ;
            end ;
	    exit ;
	end else
	begin
	    dec( BufEnd ) ;
	end ;
	BufPos := 0 ;
    end ; { if( BufPos > BufEnd ) }
    S := S + BufPtr[ BufPos ] ;
    inc( BufPos ) ;
    if(
        ( S[ length( S ) ] <> #26 )
        and
	( S[ length( S ) ] <> CR )
        and
	( S[ length( S ) ] <> LF )
      ) then { No delimeter yet }
    begin
	if( length( S ) = Max_String_Length ) then
	begin
	    Set_Error( File_SystemErr_Facility, File_SystemErr_LineTooLong ) ; { TEST: Untested }
	    exit ;
	end ;
	goto Next_Char ;
    end ;
    if( Eat_Delimiters ) then
    begin
        if( S[ length( S ) ] = CR ) then { Swallow CRLF }
        begin
            if( BufPos > BufEnd ) then
            begin { Need to fill buffer }
                Blockread( BufPtr^, BufSize, R ) ;
                BufEnd := R ;
                if( BufEnd > 0 ) then
                begin
                    BufPos := 0 ;
                    dec( BufEnd ) ;
                end ;
            end ;
            if( BufPtr[ BufPos ] = LF ) then
            begin
                inc( BufPos ) ; { Eat LF }
            end ;
        end ;
    end else
    begin
        if( ( S[ length( S ) ] = CR ) or ( S[ length( S ) ] = CR ) ) then
        begin
            dec( BufPos ) ; { Don't advance past delimiter }
        end ;
    end ;
    if( S[ length( S ) ] = #26 ) then
    begin
        dec( BufPos ) ; { Don't advance past control-Z }
    end ;
    Read_Text_Buffer := copy( S, 1, length( S ) - 1 ) ; { Return string }
    Set_Error( File_SystemErr_Facility, 0 ) ;
end ; { TSFile.Read_Text_Buffer }


{ TBlock_Device methods }

constructor TBlock_Device.Init ;

begin
    inherited Init( nil, False, 256 ) ;
    Drive := 0 ;
    Extendable := False ;
    Class_Name := 'TBlock_Device' ;
end ;


procedure TBlock_Device.Reset( const Name : String ) ;

var C : char ;
{$IFNDEF DELPHI}
var Reg : TRegisters ;
{$ENDIF}

begin
    if( copy( Name, 2, 1 ) <> ':' ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileName ) ;
	exit ;
    end ;
    C := upcase( Name[ 1 ] ) ;
    if( ( C < 'A' ) or ( C > 'Z' ) ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_InvalidDevice ) ;
	exit ;
    end ;
    Drive := ord( C ) - 64 ;
    if( Disksize( Drive ) = -1 ) then
    begin
	Drive := 0 ;
	Set_Error( File_SystemErr_Facility, File_SystemErr_InvalidDevice ) ;
        exit ;
    end ;
{$IFDEF DELPHI}
    Set_Error( File_SystemErr_Facility, File_SystemErr_InvalidDevice ) ;
{$ELSE}
    Position := 0 ;
    Reg.AX := $1C00 ;
    Reg.DL := Drive ;
    intr( $21, Reg ) ;
    Record_Size := Reg.CX ;
{$ENDIF}
end ;


procedure TBlock_Device.Rewrite( const Name : String ) ;

begin
    Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


procedure TBlock_Device.Append( const Name : String ) ;

begin
    Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


procedure TBlock_Device.Close ;

begin
    { Do nothing }
end ;


procedure TBlock_Device.Seek( Pos : TStore_Address ) ;

begin
    if( Drive = 0 ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
	exit ;
    end ;
    if( Pos > DiskSize( Drive ) ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_EndOfFile ) ;
	exit ;
    end ;
    Position := Pos ;
end ;


procedure TBlock_Device.Read( var Buffer ) ;

var Res : TRecord_Size ;

begin
    Blockread( Buffer, Record_Size, Res ) ;
end ;


procedure TBlock_Device.Write( var Buffer ) ;

var Res : TRecord_Size ;

begin
    Blockwrite( Buffer, Record_Size, Res ) ;
end ;


procedure TBlock_Device.Readln( var Buffer : String ) ;

begin
    Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


procedure TBlock_Device.Writeln( Buffer : String ) ;

begin
    Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


procedure TBlock_Device.Blockread( var Buf ; Count : TRecord_Size ;
    var Res : TRecord_Size ) ;

{$IFDEF DELPHI}
var R, Flags : smallint ;
{$ELSE}
var R, Flags : integer ;
{$ENDIF}

begin
    if( Drive = 0 ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
	exit ;
    end ;
{$IFDEF DELPHI}
    try
        Disk_Read( Drive, Position div Record_Size, Count div Record_Size, Buf,
	    R, Flags ) ;
    except
        Set_Error( File_SystemErr_Facility, File_SystemErr_ProtectionViolation ) ;
        exit ;
    end ;
{$ELSE}
    Disk_Read( Drive, trunc( Position ) div Record_Size, Count div Record_Size, Buf,
	R, Flags ) ;
{$ENDIF}
    if( ( Flags and 1 ) = 1 ) then
    begin
	Set_Error( 1, R ) ;
	Res := 0 ;
    end else
    begin
	Set_Error( File_SystemErr_Facility, 0 ) ;
	Res := Count ;
    end ;
end ;


procedure TBlock_Device.Blockwrite( var Buf ; Count : TRecord_Size ;
    var Res : TRecord_Size ) ;

{$IFDEF DELPHI}
var R, Flags : smallint ;
{$ELSE}
var R, Flags : integer ;
{$ENDIF}

begin
    if( Drive = 0 ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
	exit ;
    end ;
{$IFDEF DELPHI}
    Disk_Write( Drive, Position div Record_Size, Count div Record_Size, Buf,
	R, Flags ) ;
{$ELSE}
    Disk_Write( Drive, trunc( Position ) div Record_Size, Count div Record_Size, Buf,
	R, Flags ) ;
{$ENDIF}
    if( ( Flags and 1 ) = 1 ) then
    begin
	Set_Error( 1, R ) ;
	Res := 0 ;
    end else
    begin
	Set_Error( File_SystemErr_Facility, 0 ) ;
	Res := Count ;
    end ;
end ;


function TBlock_Device.Eof : boolean ;

begin
    if( Drive = 0 ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
        EOF := True ;
	exit ;
    end ;
    if( DiskSize( Drive ) = Position ) then
    begin
	EOF := True ;
    end else
    begin
	EOF := False ;
    end ;
end ;


function TBlock_Device.Eoln : boolean ;

begin
    EOLn := True ;
    Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


function TBlock_Device.Filepos : TStore_Size ;

begin
    if( Drive = 0 ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
        FilePos := 0 ;
	exit ;
    end ;
{$IFDEF DELPHI}
    FilePos := Position ;
{$ELSE}
    FilePos := trunc( Position ) ;
{$ENDIF}
end ;


function TBlock_Device.Get_Size : TStore_Size ;

begin
    if( Drive = 0 ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
        Get_Size := 0 ;
	exit ;
    end ;
    Get_Size := DiskSize( Drive ) ;
end ;


procedure TBlock_Device.Set_Size( New_Size : TStore_Size ) ;

begin
    Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


procedure TBlock_Device.Flush ;

begin
    { Do nothing }
end ;


procedure TBlock_Device.GetFAttr( var Attr : Word ) ;

begin
   Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


procedure TBlock_Device.Erase( const Name : String ) ;

begin
   Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


function TBlock_Device.Seekeof : Boolean ;

begin
    if( Drive = 0 ) then
    begin
	Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
        SeekEOF := False ;
	exit ;
    end ;
    Position := DiskSize( Drive ) ;
    SeekEOF := True ;
end ;


function TBlock_Device.Seekeoln : Boolean ;

begin
    SeekEOLn := False ;
    Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


procedure TBlock_Device.SetFAttr( Attr : Word ) ;

begin
   Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


procedure TBlock_Device.GetFTime( var _Time : Longint ) ;

begin
   Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


procedure TBlock_Device.SetFTime( _Time : Longint ) ;

begin
   Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


procedure TBlock_Device.Settextbuf( var buf ; size : Word ) ;

begin
    Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ;
end ;


procedure TBlock_Device.Set_Extend( State : boolean ) ;

begin
    Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ; { Disks cannot be extended }
end ;


procedure TBlock_Device.Truncate( Size : TStore_Address ) ;

begin
    Set_Error( File_SystemErr_Facility, File_SystemErr_IllegalOperationType ) ; { Disks cannot be extended }
end ;


procedure TBlock_Device.Initialize ;

begin
    { Does nothing for now }
end ;


function TBlock_Device.Facility : longint ;

begin
    Facility := TBlock_DeviceErr_Facility ;
end ;


function TBlock_Device.Version : longint ;

begin
    Version := 10 ; { Version 1.0 }
end ;


const TBlock_Device_Name : string[ 14 ] = 'TBlock_Device'#0 ;

function TBlock_Device.Facility_Name : PChar ;

begin
    Facility_Name := @TBlock_Device_Name[ 1 ] ;
end ;



{$IFDEF DELPHI}
{ TWin32_File_System methods... }

{ Constructors and destructors... }

constructor TWin32_File_System.Init ;

begin
    inherited Init ;

    Handles.Init( 0, 4 ) ;
    Names.Init( 0, 4 ) ;
end ;


destructor TWin32_File_System.Done ;

var Dummy : integer ;

begin
    for Dummy := Handles.Count - 1 downto 0 do
    begin
        Close( Handles.At( Dummy ) ) ;
    end ;
    Handles.Done ;
    Names.Done ;

    inherited Done ;
end ;


{ Instance support... }

procedure TWin32_File_System.Set_Last_Error ;

begin
    IO_Error.Code := GetLastError ;
    IO_Error.Facility := 0 ;
end ;


{ API... }

procedure TWin32_File_System.Close( Handle : integer ) ;

begin
    if( Handle <> 0 ) then
    begin
        CloseHandle( Handle ) ;
        Handle := Handles.IndexOf( Handle ) ;
        if( Handle <> -1 ) then
        begin
            Handles.AtDelete( Handle ) ;
            Names.AtDelete( Handle ) ;
        end ;
        Set_Last_Error ;
    end else
    begin
        inherited Close( Handle ) ;
    end ;
end ;


function TWin32_File_System.Create( const Name : string ;
    Mode : integer ) : integer ;

var Handle : integer ;

begin {TODO: Handle Mode}
    Handle := FileCreate( Name ) ;
    if( Handle < 0 ) then
    begin
        if( GetLastError = 2 ) then
        begin
            Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotFound ) ;
        end else
        begin
            Set_Error( 0, Handle ) ;
        end ;
        Create := 0 ;
        exit ;
    end ;

    Set_Error( 0, 0 ) ;
    Create := Handle ;
    Handles.Insert( Handle ) ;
    Names.Insert( Name ) ;
end ;


function TWin32_File_System.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ; { Not yet supported }
end ;


function TWin32_File_System.File_Name( Handle : integer ) : string ;

begin
    Handle := Handles.IndexOf( Handle ) ;
    if( Handle <> -1 ) then
    begin
        File_Name := Names.At( Handle ) ;
    end else
    begin
        File_Name := inherited File_Name( Handle ) ;
    end ;
end ;


procedure TWin32_File_System.Flush( Handle : integer ) ;

begin
    FlushFileBuffers( Handle ) ;
    Set_Last_Error ;
end ;


function TWin32_File_System.Get_Size( Handle : integer ) : TStore_Address ;

begin
    Get_Size := GetFileSize( Handle, nil ) ;
    Set_Last_Error ;
end ;


function TWin32_File_System.Get_Time( Handle, Typ : integer ) : Longint ;

begin
    Get_Time := FileGetDate( Handle ) ;
    Set_Last_Error ;
end ;


function TWin32_File_System.Open( const Name : string ;
    Mode : integer ) : integer ;

var Handle : integer ;
    OFStruct : TOFStruct ;

begin {TODO: Handle Mode}
    fillchar( OFStruct, sizeof( OFStruct ), 0 ) ;
    OFStruct.cBytes := sizeof( OFStruct ) ;
    if( ( Mode and FSOM_Read_Only ) <> 0 ) then
    begin
        Handle := openfile( PChar( Name ), OFStruct, OF_READ or OF_SHARE_COMPAT ) ;
    end else
    begin
        Handle := openfile( PChar( Name ), OFStruct, OF_READWRITE or OF_SHARE_COMPAT ) ;
    end ;
    if( Handle < 0 ) then
    begin
        Handle := GetLastError ;
        if( Handle = 2 ) then
        begin
            Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotFound ) ;
        end else
        begin
            Set_Error( 0, Handle ) ;
        end ;
        Handle := 0 ;
    end else
    begin
        Set_Error( 0, 0 ) ;
        Handles.Insert( Handle ) ;
        Names.Insert( Name ) ;
    end ;
    Open := Handle ;
end ; { TWin32_File_System.Open }


procedure TWin32_File_System.Read( Handle : integer ; Position : TStore_Address ;
    Len : TStore_Address ; var Buffer ; var Count : TStore_Address ) ;

var C : DWord ;

begin
    if( FileSeek( Handle, Position, 0 ) < 0 ) then
    begin
        Set_Last_Error ;
        exit ;
    end ;
    if( not ReadFile( Handle, Buffer, Len, C, nil ) ) then
    begin
        Count := 0 ;
    end else
    begin
        Count := C ;
    end ;
    Set_Last_Error ;
end ;


procedure TWin32_File_System.Set_Time( Handle : integer ; Typ, Time : Longint ) ;

begin
{$IFDEF DELPHI}
  {$WARNINGS OFF}
{$ENDIF}
    FileSetDate( Handle, Time ) ;
{$IFDEF DELPHI}
  {$WARNINGS ON}
{$ENDIF}
    Set_Last_Error ;
end ;


procedure TWin32_File_System.Set_Size( Handle : integer ; New_Size : TStore_Address ) ;

begin
    if( New_Size < Get_Size( Handle ) ) then { Truncating }
    begin
        if( FileSeek( Handle, New_Size, 0 ) < 0 ) then
        begin
            Set_Last_Error ;
            exit ;
        end ;
        SetEndOfFile( Handle ) ;
        Set_Last_Error ;
    end else
    begin
        inherited Set_Size( Handle, New_Size ) ;
    end ;
end ;


procedure TWin32_File_System.Write( Handle : integer ; Position : TStore_Address ;
    Len : TStore_Address ; var Buffer ; var Count : TStore_Address ) ;

begin
    if( FileSeek( Handle, Position, 0 ) < 0 ) then
    begin
        Set_Last_Error ;
        exit ;
    end ;
    Count := FileWrite( Handle, Buffer, Len ) ;
    if( Count < 0 ) then
    begin
        Set_Error( 2, 101 ) ;
        Count := 0 ;
    end else
    begin
	Set_Error( 0, 0 ) ;
    end ;
end ;


function TWin32_File_System.Get_File_Attributes( const Name : string ) : longint ;

var Attr : longint ;

begin
{$IFDEF DELPHI}
  {$WARNINGS OFF}
{$ENDIF}
    Attr := FileGetAttr( Name ) ;
{$IFDEF DELPHI}
  {$WARNINGS ON}
{$ENDIF}
    Get_File_Attributes := Convert_DOS_Attributes( Attr ) ;
    Set_Last_Error ;
end ;


procedure TWin32_File_System.Set_File_Attributes( const Name : string ;
    Value : longint ) ;

var Attr : longint ;

begin
    Attr := Convert_Attributes( Value ) ;
{$IFDEF DELPHI}
  {$WARNINGS OFF}
{$ENDIF}
    FileSetAttr( Name, Attr ) ;
{$IFDEF DELPHI}
  {$WARNINGS ON}
{$ENDIF}
    Set_Last_Error ;
end ;


function TWin32_File_System.Facility : longint ;

begin
    Facility := 46 ;
end ;


function TWin32_File_System.Version : longint ;

begin
    Version := 10 ; { Version 1.0 }
end ;


{ TWin32_File methods... }

constructor TWin32_File.Init( FS : PFile_System ; TF : Boolean ; Size : Word ) ;

begin
    Text_File := Tf ;
    Record_Size := Size ;
    Extendable := True ;
    Handle := 0 ;
    File_System := FS ;
end ;


procedure TWin32_File.Initialize ;

begin
    Rewrite( 'CON' ) ;
end ;


function TWin32_File.Facility : longint ;

begin
    Facility := TWin32_FileErr_Facility ;
end ;


function TWin32_File.Version : longint ;

begin
    Version := 10 ; { Version 1.0 }
end ;


const TWin32_File_Name : string[ 12 ] = 'TWin32_File'#0 ;

function TWin32_File.Facility_Name : PChar ;

begin
    Facility_Name := @TWin32_File_Name[ 1 ] ;
end ;
{$ENDIF}



procedure Close_File_System( const Name : string ) ;

begin
    File_Processor.Close_File_System( Name ) ;
end ;


function Open_File_System( const Name : string ; Flags : integer ;
    var Res : TUEC ) : PFile_System ;

begin
    Open_File_System := File_Processor.Open_File_System( Name, Flags, Res ) ;
end ;


function Open( const Name : string ; Mode : integer ; var UEC : TUEC ) : PSFile ;

begin
    Open := File_Processor.Open( Name, Mode, UEC ) ;
end ;


function _Open( const Name : string ; Mode : integer ; var UEC : TUEC ) : PSFile ;

begin
    _Open := Open( Name, Mode, UEC ) ;
end ;


function Lookup( Filename : string ; var Find_Record : TFind_Record ) : string ;

begin
    Lookup := File_Processor.Lookup( Filename, Find_Record ) ;
end ;


procedure Register_File_System_Server( FSS : PFile_System_Server ) ;

begin
    File_Processor.Register_File_System_Server( FSS ) ;
end ;


procedure Register_Store_Server( SS : PStore_Server ) ;

begin
    File_Processor.Register_Store_Server( SS ) ;
end ;


procedure Initialize_File_System( Nam, ID : string ; Flags : longint ;
    var UEC : tUEC ) ;

begin
    File_Processor.Initialize( Nam, ID, Flags, UEC ) ;
end ;


function DeviceSize( const Nam : string ) : int64 ;

begin
    DeviceSize := File_Processor.DeviceSize( Nam ) ;
end ;


function DeviceFree( const Nam : string ) : int64 ;

begin
    DeviceFree := File_Processor.DeviceFree( Nam ) ;
end ;


function MKDir( const Nam : string ) : longint ;

begin
    MKDir := File_Processor.MKDir( Nam ) ;
end ;


procedure Filespec_String_Scan( const Name, Separator : string ;
    var Info : TFSSS ) ;

begin
    File_Processor.Filespec_String_Scan( Name, Separator, Info ) ;
end ; { Filespec_String_Scan }



{ TFile_System_Server methods (abstract class)... }

{ API... }

function TFile_System_Server.Get_File_System_For_Store( Store : PSStore ;
    Own : boolean ) : PFile_System ;

begin
    Get_File_System_For_Store := nil ;
end ;


function TFile_System_Server.Get_File_System( const Nam : string ) : PFile_System ;

begin
    Get_File_System := nil ;
end ;


procedure TFile_System_Server.Initialize( const Nam, ID : string ;
    Flags : integer ; var UEC : TUEC ) ;

begin
    UEC.Code := 1 ;
end ;


type TDOS_File_System_Server = object( TFile_System_Server )
                                   public { API... }
                                       function Get_File_System_For_Store( Store : PSStore ;
                                           Own : boolean ) : PFile_System ;
                                           virtual ;

                                       procedure Initialize( const Nam, ID : string ;
                                           Flags : integer ; var UEC : TUEC ) ;
                                           virtual ;
                               end ; { TDOS_File_System_Server }

function TDOS_File_System_Server.Get_File_System_For_Store( Store : PSStore ;
    Own : boolean ) : PFile_System ;

begin
    Get_File_System_For_Store := nil ; { Assume failure }
    if( ( Store = nil ) or ( PBlock_Device( Store )^.Class_Name = 'TBlock_Device' ) ) then
    begin
        Get_File_System_For_Store := @DOS_FS ;
    end ;
end ;


procedure TDOS_File_System_Server.Initialize( const Nam, ID : string ;
    Flags : integer ; var UEC : TUEC ) ;

begin
    UEC.Code := 1 ;
end ;


var DOS_FSS : TDOS_File_System_Server ;



{ TStore_Server methods... }

{ API... }

function TStore_Server.Get_Store_ID( Name : string ) : string ;

begin
    Get_Store_ID := '' ;
end ;


function TStore_Server.Get_Store( var Name : string ) : PSStore ;

begin
    Get_Store := nil ;
end ;



type TDOS_SS = object( TStore_Server )
                   public
                       function Get_Store_ID( Name : string ) : string ;
                           virtual ;
                       function Get_Store( var Name : string ) : PSStore ;
                           virtual ;
               end ;

{ API... }

function TDOS_SS.Get_Store_ID( Name : string ) : string ;

var D, N, DN, Nam : string ;
    Dummy, Dummy1 : integer ;
    F : file ;

begin
    Get_Store_ID := '' ; { Assume failure }
    if( length( Name ) = 0 ) then
    begin
        exit ;
    end ;

    { First, find first wildcard... }
    Dummy1 := pos( '*', Name + '*' ) ;
    Dummy := pos( '?', Name + '?' ) ;
    if( Dummy < Dummy1 ) then
    begin
        Dummy1 := Dummy ;
    end ;
    Dummy := pos( '...', Name + '...' ) ;
    if( Dummy < Dummy1 ) then
    begin
        Dummy1 := Dummy ;
    end ;
    dec( Dummy1 ) ;
    Nam := Edit( copy( Name, 1, Dummy1 ), 4 or 8 or 128 ) ;

    { Now find beginning of part of path with wildcard... }
    while( ( Dummy1 > 0 ) and ( Nam[ Dummy1 ] <> '\' ) and ( Nam[ Dummy1 ] <> ':' ) ) do
    begin
        Dummy1 := Dummy1 - 1 ;
    end ;

    { See if any part is a DOS file... }
    D := copy( Nam, 1, Dummy1 ) ;
    N := '' ;
    DN := FILNAM( Nam, False ) ;
    if( length( DN ) = 0 ) then
    begin
        DN := Nam ;
    end ;

    { Special case for root DOS directories }
    if( ( length( DN ) = 2 ) and ( DN[ 2 ] = ':' ) ) then
    begin
        DN := DN + '\' ;
    end ;
    if( ( length( DN ) = 3 ) and ( DN[ 2 ] = ':' ) and ( DN[ 3 ] = '\' ) ) then
    begin
        Get_Store_ID := copy( Name, 1, 3 ) ;
        Name := copy( Name, 4, length( Name ) ) ;
        exit ;
    end ;

    Dummy := 1 ; { Force loop to execute at least once }
    while( Dummy <> 0 ) do { Until we can open a file }
    begin
        if( length( DN ) = 0 ) then
        begin
            Dummy := 1 ;
        end else
        begin
            Clear_IOResult ; { Clear any pending errors }
            assign( F, DN ) ;
            {$I-}
            reset( F ) ;
            {$I+}
            Dummy := -IOResult ;
        end ;
        if( Dummy = 0 ) then
        begin
            Clear_IOResult ; { Clear any pending errors }
            {$I-}
            System.close( F ) ;
            {$I+}
            Dummy := IOResult ;
        end else
        begin
            { Backup one directory... }
            Dummy1 := length( DN ) ;
            while( ( Dummy1 > 0 ) and ( DN[ Dummy1 ] <> '\' ) ) do
            begin
                Dummy1 := Dummy1 - 1 ;
            end ;
            N := copy( DN, Dummy1, 254 ) + N ;
            if( Dummy1 = 0 ) then { Done - no more directories }
            begin
                Dummy := 0 ; { end loop }
                DN := '' ;
            end else
            begin
                DN := copy( DN, 1, Dummy1 - 1 ) ; { DOS file to try next }
            end ; { if( Dummy1 = 0 ) }
        end ; { if( Dummy <> 0 ) }
    end ; { while( Dummy <> 0 ) }
    Name := copy( Name, length( DN ) + 1, length( Name ) ) ;
    Get_Store_ID := DN ;
end ; { TDOS_SS.Get_Store_ID }


function TDOS_SS.Get_Store( var Name : string ) : PSStore ;

var BD : PBlock_Device ;
    D : string ;
    Res : TUEC ;

begin
    D := Get_Store_ID( Name ) ;
    Name := copy( Name, length( D ) + 1, length( Name ) ) ;
    Get_Store := nil ;
    if( length( D ) > 0 ) then
    begin
        if( pos( D[ 1 ], 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz' ) > 0 ) then
        begin
            if(
                ( D = Name ) { Only the device name }
                and
                (
                  ( ( length( D ) = 2 ) and ( D[ 2 ] = ':' ) )
                  or
                  ( ( length( D ) = 3 ) and ( D[ 2 ] = ':' ) and ( D[ 3 ] = '\' ) )
                )
              ) then
            begin
                BD := new( PBlock_Device, Init ) ;
                BD^.Drive := ord( upcase( D[ 1 ] ) ) - 64 ;
                Get_Store := BD ;
            end else
            begin
                Get_Store := Open( D, 0, Res ) ;
            end ;
        end ;
    end ;
end ; { TDOS_SS.Get_Store }


var DOS_SS : TDOS_SS ;

{ TFile_Processor methods... }

var File_Processor_Exists : boolean ;

{ Constructors and destructors... }

constructor TFile_Processor.Init ;

begin
    if( File_Processor_Exists ) then
    begin
        Application_Manager^.Fatal_Exception( 'TFile_Processor is a singleton object' ) ;
    end ;
    inherited Init ;
    FS_Root := nil ;
    File_Processor_Exists := True ;
    FSS_List.Init( 0, 4 ) ;
    SS_List.Init( 0, 4 ) ;
end ;


destructor TFile_Processor.Done ;

var Current : PFS_Record ;

begin
    while( FS_Root <> nil ) do
    begin
        Current := FS_Root^.Next ;
        dispose( FS_Root^.FS, Done ) ;
        dispose( FS_Root ) ;
        FS_Root := Current ;
    end ;
    FS_Root := nil ;
    FSS_List.Done ;
    SS_List.Done ;
    
    inherited Done ;

    File_Processor_Exists := False ;
end ; { TFile_Processor.Done }


{ Internal utility routines... }

var Saved_DN, _Saved_DN : string ;

procedure TFile_Processor.Determine_Parts( Nam : string ;
    var D, N, DN : string ) ;

var Current : PFS_Record ;

begin
    Current := FS_Root ;
    while( Current <> nil ) do { Until a match is found or EOL }
    begin
        if(
            ( uppercase( copy( Nam, 1, length( Current^.Store_Name ) ) ) = uppercase( Current^.Store_Name ) )
            and
            (
              ( length( Nam ) = length( Current^.Store_Name ) )
              or
              ( pos( ':', Nam ) = length( Current^.Store_Name ) + 1 )
              or
              ( copy( Nam, length( Current^.Store_Name ) + 1, 1 ) = '\' )
            )
          ) then { Found match }
        begin
            DN := Current^.Store_Name ;
            N := copy( Nam, length( DN ) + 1, length( Nam ) ) ;
            Saved_DN := Edit( DN, 32 ) ;
            _Saved_DN := DN ;
            exit ;
        end else
        begin
            Current := Current^.Next ; { Follow link to next }
        end ; { if( Current^.DOS_Name = Name ) }
    end ; { while( Current <> nil ) }
end ; { TFile_Processor.Determine_Parts }


{ API... }

function TFile_Processor.Lookup( Filename : string;
    var Find_Record : TFind_Record ) : string ;

var A : Integer ;
    D, N, DN : string ;
    Dummy : integer ;
    _FS : PFile_System ;
    Original : string ;
    S : PSStore ;
    Saved : char ;
    Temp : string ;
    UEC : TUEC ;

begin
    N := '' ;
    S := nil ;
    _FS := nil ;
    if( length( Filename ) > 0 ) then
    begin
        Determine_Parts( Filename, D, N, DN ) ; { Check all file systems }
        if( length( DN ) = 0 ) then { Totally not found }
        begin
            Original := Filename ;
            for Dummy := 0 to SS_List.Count - 1 do
            begin
                S := PStore_Server( SS_List.At( Dummy ) )^.Get_Store( Filename ) ;
                if( S <> nil ) then
                begin
                    DN := PStore_Server( SS_List.At( Dummy ) )^.Get_Store_ID( Filename ) ;
                    N := '' ;
                    S^.Attach ;
                    break ;
                end ;
            end ; { for Dummy := 0 to SS_List.Count - 1 }
            if( S = nil ) then { See if we can find a file system that knows about this file }
            begin
                _FS := nil ;
                for Dummy := 0 to FSS_List.Count - 1 do
                begin
                    _FS := PFile_System_Server( FSS_List.At( Dummy ) )^.Get_File_System( Filename ) ;
                    if( _FS <> nil ) then
                    begin
                        break ;
                    end ;
                end ;
                if( _FS = nil ) then { Try DOS as a last resort }
                begin
                    N := '' ;
                    Filename := Original ; { Is this correct? }
                    _FS := @DOS_FS ;
                end ;
            end else
            begin
                _FS := nil ;
                for Dummy := 0 to FSS_List.Count - 1 do
                begin
                    _FS := PFile_System_Server( FSS_List.At( Dummy ) )^.Get_File_System_For_Store( S, False ) ;
                    if( _FS <> nil ) then
                    begin
                        break ;
                    end ;
                end ;
                if( _FS = nil ) then
                begin
                    N := '' ;
                    Filename := Original ;
                    _FS := @DOS_FS ;
                end ;
            end ; { if( S = nil ) }
        end else
        begin
            if( length( N ) = 0 ) then { DOS }
            begin
                _FS := @DOS_FS ;
            end else
            begin
                _FS := Open_File_System( DN, FSF_Read_Only, UEC ) ;
                if( UEC.Code <> 0 ) then
                begin
                    Lookup := '' ;
                    exit ;
                end ;
                Filename := copy( Filename, length( DN ) + 1, length( Filename ) ) ;
            end ;
        end ; { if( length( DN ) = 0 ) }
    end ; { if( length( Filename ) > 0 ) }

    if( ( length( Filename ) > 0 ) and ( Find_Record.Initialized = 0 ) ) then
    begin
        Saved := Find_Record.Separator ;
        Uninitialize_Find_Record( Find_Record ) ;
        Initialize_Find_Record( Find_Record ) ;
        Find_Record.Separator := Saved ;
    end ;
    if( Find_Record.Initialized <> 0 ) then
    begin
        Initialize_Find_Record( Find_Record ) ;
    end ;
    with Find_Record do
    begin
        if( Length( Filename ) > 0 ) then { Starting new lookup }
        begin
            File_List := Filename ;
            Temp := '' ; { Assume failure of previous lookup }
            if( Context <> nil ) then
            begin
                dispose( Context, Done ) ;
                Context := nil ;
            end ;
            FS := _FS ;
        end else
        begin
            if( Context = nil ) then { Never started the search }
            begin
                Lookup := '' ;
                if( length( N ) <> 0 ) then { Not DOS }
                begin
                    Close_File_System( DN ) ;
                end ;
                exit ;
            end ;
            Temp := Context^.Lookup( '', '', Find_Record ) ;
            { Lookup next matching file }
            if( length( Temp ) > 0 ) then
            begin
                Dummy := pos( ':', Temp ) ;
                if( ( Dummy = 0 ) or ( Dummy > pos( '\', Temp ) ) ) then { No colon, or colon part of a filename }
                begin
                    if( length( FS^.Path ) > 0 ) then
                    begin
                        Temp := FS^.Path + '\' + Temp ;
                    end ;
                end ;
            end ;
        end ; { if( Length( Filename ) > 0 ) else }
        while( ( length( Temp ) = 0 ) and ( length( File_List ) > 0 ) ) do
        { while lookup failures }
        begin
            if( Context <> nil ) then
            begin
                dispose( Context, Done ) ;
            end ;
            Context := new( PWildcard, Create( _FS ) ) ;
            if( Separator = #0 ) then { No separator means no list }
            begin
                A := length( File_List ) + 1 ; { Take entire string as a single literal filename }
            end else
            begin
                A := pos( Separator, File_List + Separator ) ; { Find end of next file specification }
            end ;
            Temp := copy( File_List, 1, A - 1 ) ; { Get file specification }
            File_List := copy( File_List, A + 1, 255 ) ; { Strip from list }
(*
            Determine_Parts( Filename, D, N, DN ) ;
            if( ( length( DN ) = 0 ) and ( copy( Temp, 2, 1 ) = ':' ) ) then
            begin
                DN := copy( Temp, 1, 2 ) ;
            end ;
            A := pos( '\...', DN ) ;
            if( A > 0 ) then
            begin
                DN := copy( DN, 1, A - 1 ) ;
            end ;
            Temp := copy( Temp, length( DN ) + 1, length( Temp ) ) ; { Strip DOS name }
*)
            Temp := Context^.Lookup( Temp, '', Find_Record ) ; { Do the lookup }

            if( length( Temp ) > 0 ) then
            begin
                Dummy := pos( ':', Temp ) ;
                if( ( Dummy = 0 ) or ( Dummy > pos( '\', Temp ) ) ) then { No colon, or colon part of a filename }
                begin
                    if( length( _FS^.Path ) > 0 ) then
                    begin
                        Temp := _FS^.Path + '\' + Temp ;
                    end ;
                end ;
            end ;
        end ; { while... }
    end ; { with Find_Record do }
    Lookup := Temp ; { Return result }

    if( length( N ) <> 0 ) then { Opened with Open_File_System }
    begin
        Close_File_System( DN ) ;
    end else
    begin
        if( S <> nil ) then
        begin
            S^.Detach ;
        end ;
    end ;
end ; { TFile_Processor.Lookup }


function TFile_Processor.Open_File_System( Name : string ;
    Flags : integer ; var Res : TUEC ) : PFile_System ;

var Current, Previous : PFS_Record ;
    Dummy : integer ;
    FS : PFile_System ;
    S : PSStore ;
    Store_Name, Temp : string ;

begin
    fillchar( Res, sizeof( Res ), 0 ) ;
    if( length( Name ) = 0 ) then
    begin
        Open_File_System := @DOS_FS ;
        exit ;
    end ;

    { Setup... }
    if( pos( '\', Name ) + pos( ':', Name ) = 0 ) then
    begin
        GetDir( 0, Temp ) ;
        if( copy( Temp, length( temp ), 1 ) <> '\' ) then
        begin
            Temp := Temp + '\' ;
        end ;
        Name := Temp + Name ;
    end ;
    Name := Edit( Name, 4 or 8 or 16 or 32 or 128 ) ;
    {TODO: Normalize name if it contains .\ or ..\}

    { See if file system is already open... }
    Current := FS_Root ;
    Previous := nil ;
    while( Current <> nil ) do { Until a match is found or EOL }
    begin
        if( Current^.Store_Name = Name ) then { Found match }
        begin
            Current^.Users := Current^.Users + 1 ;
            Open_File_System := Current^.FS ;
            exit ;
        end else
        begin
            Previous := Current ;
            Current := Current^.Next ; { Follow link to next }
        end ; { if( Current^.DOS_Name = Name ) }
    end ; { while( Current <> nil ) }

    { See if a file system server serves up a file system for this
      specification... }
    Open_File_System := nil ;
    S := nil ;
    Store_Name := Name ;
    for Dummy := 0 to SS_List.Count - 1 do
    begin
        S := PStore_Server( SS_List.At( Dummy ) )^.Get_Store( Name ) ;
        if( S <> nil ) then
        begin
            Store_Name := copy( Store_Name, 1, length( Store_Name ) - length( Name ) ) ;
            break ;
        end ;
    end ; { for Dummy := 0 to SS_List.Count - 1 }
    if( S = nil ) then
    begin
        exit ;
    end ;

    { First open of this file system file... }
    FS := nil ;

    { See if a file system server serves up a file system for this
      specification... }
    for Dummy := 0 to FSS_List.Count - 1 do
    begin
        FS := PFile_System_Server( FSS_List.At( Dummy ) )^.Get_File_System_For_Store( S, True ) ;
        if( FS <> nil ) then
        begin
            break ;
        end ;
    end ; { for Dummy := 0 to FSS_List.Count - 1 }
    if( FS = nil ) then
    begin
        exit ;
    end ;

    new( Current ) ;
    if( Previous = nil ) then
    begin
        FS_Root := Current ;
    end else
    begin
        Previous^.Next := Current ;
    end ;

    Current^.FS := PFile_System( FS ) ;
    Current^.Store_Name := Store_Name ;
    Current^.Users := 1 ;
    Current^.Next := nil ;
    Open_File_System := FS ;
end ; { TFile_Processor.Open_File_System }


procedure TFile_Processor.Close_File_System( Name : string ) ;

var Current, Previous : PFS_Record ;

begin
    Current := FS_Root ;
    Previous := nil ;
    Name := Edit( Name, 4 or 8 or 16 or 32 or 128 ) ;
    while( Current <> nil ) do { Until a match is found or EOL }
    begin
        if( Name = Current^.Store_Name ) then { Found match }
        begin
            Current^.Users := Current^.Users - 1 ;
            if( Current^.Users < 1 ) then { No more references }
            begin
                { Remove file system from list }
                if( Previous = nil ) then { First item in list }
                begin
                    FS_Root := Current^.Next ;
                end else
                begin
                    Previous^.Next := Current^.Next ;
                end ;

                { Close file system... }
{$IFDEF DELPHI}
                try
{$ENDIF}
                dispose( Current^.FS, Done ) ;
{$IFDEF DELPHI}
                except { Why does this throw an exception? }
                end ;
{$ENDIF}
                dispose( Current ) ;
            end ; { if( Current^.Users < 1 ) }
            Current := nil ; { end loop }
        end else
        begin
            Previous := Current ;
            Current := Current^.Next ; { Follow link to next }
        end ; { if( Current^.DOS_Name = DOS_Name ) }
    end ; { while( Current <> nil ) }
end ; { TFile_Processor.Close_File_System }


function TFile_Processor.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ; { Not yet supported }
end ;


function TFile_Processor.File_System_Open( Nam : string ) : boolean ;

var Current : PFS_Record ;

begin
    Nam := Edit( Nam, 4 or 8 or 32 or 128 ) ;
    Current := FS_Root ;
    while( Current <> nil ) do { Until a match is found or EOL }
    begin
        if( Nam = Current^.Store_Name ) then { Found match }
        begin
	    File_System_Open := True ;
            exit ;
        end else
        begin
            Current := Current^.Next ; { Follow link to next }
        end ; { if( Current^.DOS_Name = DOS_Name ) }
    end ; { while( Current <> nil ) }
    File_System_Open := False ;
end ;


{ Scan a file specification and fill the Info record.  Separator is the path
  separator to use in parsing. }
procedure TFile_Processor.Filespec_String_Scan( const Name, Separator : string ;
    var Info : TFSSS ) ;

var Dummy : integer ;
    _Name : string ;
    FS : PFile_System ;

begin
    { Setup... }
    fillchar( Info, sizeof( Info ), 0 ) ;
    _Name := Name ;

    { Parse name }
    if( copy( _Name, 1, 2 * length( Separator ) ) = Separator + Separator ) then
    begin
        Dummy := 2 * length( Separator ) + 1 ;
        while(
               ( Dummy < length( _Name ) )
               and
               ( copy( _Name, Dummy, length( Separator ) ) <> Separator )
             ) do
        begin
            Dummy := Dummy + 1 ;
        end ;
        Dummy := Dummy - 1 ; { Device is //name }
    end else
    begin
        Dummy := pos( ':', _Name ) ;
    end ;
    Info.Device := copy( _Name, 1, Dummy ) ;
    _Name := copy( _Name, Dummy + 1, length( _Name ) ) ;

    Dummy := length( _Name ) ;
    while(
           ( Dummy > 0 )
           and
           ( copy( _Name, Dummy, length( Separator ) ) <> Separator )
         ) do
    begin
        dec( Dummy ) ;
    end ;
    Info.Filename := copy( _Name, Dummy + 1, length( _Name ) ) ;
    Info.Path := copy( _Name, 1, Dummy ) ;
    if( Separator <> '/' ) then
    begin
        Dummy := pos( '/', Info.Filename + '/' ) ;
        Info.Switches := copy( Info.Filename, Dummy, length( Info.Filename ) ) ;
        Info.Filename := copy( Info.Filename, 1, Dummy - 1 ) ;
    end ;
    Dummy := length( Info.Filename ) ;
    while( ( Dummy > 0 ) and ( Info.Filename[ Dummy ] <> '.' ) ) do
    begin
        dec( Dummy ) ;
    end ;
    if( Dummy = 0 ) then { No extension }
    begin
        Dummy := length( Info.Filename ) + 1 ;
        Info.Flags := Info.Flags or FSSS_Null_Extension ;
    end ;
    Info.Name := copy( Info.Filename, 1, Dummy - 1 ) ;
    Info.Extension := copy( Info.Filename, Dummy + 1, length( Info.Filename ) ) ;

    { Now give a file system a chance to modify the parsings... }
    for Dummy := 0 to FSS_List.Count - 1 do
    begin
        FS := PFile_System_Server( FSS_List.At( Dummy ) )^.Get_File_System( Name ) ;
        if( FS <> nil ) then
        begin
            FS^.Filespec_String_Scan( Name, Separator, Info ) ;
            break ;
        end ;
    end ;
end ; { TFile_Processor.Filespec_String_Scan }


procedure TFile_Processor.Register_File_System_Server( FSS : PFile_System_Server ) ;

begin
    FSS_List.Insert( FSS ) ;
end ;


procedure TFile_Processor.Register_Store_Server( SS : PStore_Server ) ;

begin
    SS_List.Insert( SS ) ;
end ;


function TFile_Processor.Open( const Name : string ; Mode : integer ;
    var UEC : TUEC ) : PSFile ;

var D, N, DN : string ;
    File_Object : PSFile ;
    FS : PFile_System ;

begin
    { Sanity check... }
    if( length( Name ) = 0 ) then { Null filename is invalid }
    begin
        UEC.Code := File_SystemErr_FileName ; { Invalid file name }
        UEC.Facility := File_SystemErr_Facility ;
        Open := nil ;
        exit ;
    end ;

    { Setup... }
    fillchar( UEC, sizeof( UEC ), 0 ) ;
    FS := nil ;
    Determine_Parts( Name, D, N, DN ) ;

    { Not an already-open file system... }
    if( ( length( N ) = 0 ) or ( length( DN ) = 0 ) ) then { A DOS file }
    begin
        File_Object := DOS_FS.Get_File_Object( Mode ) ;
        if( File_Object = nil ) then
        begin
            DOS_FS.Last_Error( UEC ) ;
            Open := nil ;
            exit ;
        end ;
        N := Name ;
    end else
    begin
        if( length( Name ) = 0 ) then { Null filename is invalid }
        begin
            UEC.Code := File_SystemErr_FileName ; { Invalid file name }
            UEC.Facility := File_SystemErr_Facility ;
            Open := nil ;
            exit ;
        end ;
        FS := Open_File_System( DN, 0, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Open := nil ;
            exit ;
        end ;
        if( length( N ) = 0 ) then
        begin
            N := '\' ;
        end ;

        { Create file object... }
        File_Object := FS^.Get_File_Object( Mode ) ;
        if( File_Object^.IO_Error.Code <> 0 ) then
        begin
            UEC := File_Object^.IO_Error ;
            dispose( File_Object, Done ) ;
            Close_File_System( DN ) ;
            Open := nil ;
            exit ;
        end ;
    end ;

    { Now open the file... }
    File_Object^.File_Mode := Mode ;
    if( ( Mode and 3 ) = FSOM_Append ) then
    begin
	File_Object^.Append( N ) ;
    end else
    if( ( Mode and 3 ) = FSOM_Input ) then
    begin
	File_Object^.Reset( N ) ;
    end else
    if( ( Mode and 3 ) = FSOM_Output ) then
    begin
	File_Object^.Rewrite( N ) ;
    end else
    if( ( Mode and 3 ) = 0 ) then { Open if there, create if not }
    begin
	File_Object^.Reset( N ) ;
	UEC := File_Object^.IO_Error ;
	if( UEC.Code = 2 ) then { code 2 always indicates file not found }
	begin
	    File_Object^.Rewrite( N ) ;
	end ;
    end ;
    UEC := File_Object^.IO_Error ;
    if( UEC.Code <> 0 ) then { not opened }
    begin
        dispose( File_Object, Done ) ;
        Open := nil ;
    end else
    begin
        Open := File_Object ;
    end ;
    if( FS <> nil ) then
    begin
        Close_File_System( DN ) ;
    end ;
end ; { TFile_Processor.Open }


procedure TFile_Processor.Initialize( const Nam, ID : string ; Flags : longint ;
    var UEC : tUEC ) ;

var D, N, DN : string ;
    Dummy : integer ;
    Current : PFS_Record ;

begin
    { Make sure there is no other access to this file... }
    Current := FS_Root ;
    while( Current <> nil ) do { Until done or found }
    begin
	if( Current^.Store_Name = Nam ) then
	begin
	    UEC.Facility := File_SystemErr_Facility ;
	    UEC.Code := File_SystemErr_Device_In_Use ;
            IO_Error := UEC ;
	    exit ;
	end ;
	Current := Current^.Next ;
    end ;

    fillchar( UEC, sizeof( UEC ), 0 ) ;
    Determine_Parts( Nam, D, N, DN ) ;
    if( ( length( DN ) > 0 ) and ( length( DN ) < 4 ) and ( length( N ) < 4 ) ) then { DOS }
    begin
	DOS_FS.Initialize( DN, ID, Flags, UEC ) ;
    end else
    begin
        for Dummy := 0 to FSS_List.Count - 1 do
        begin
            PFile_System_Server( FSS_List.At( Dummy ) )^.Initialize( Nam, ID, Flags, UEC ) ;
            if( UEC.Code = 0 ) then
            begin
                exit ;
            end ;
        end ; { for Dummy := 0 to FSS_List.Count - 1 }
    end ;
end ; { TFile_Processor.Initialize }


function TFile_Processor.DeviceSize( const Nam : string ) : int64 ;

var D, N, DN : string ;
    FS : PFile_System ;
    UEC : TUEC ;

begin
    Determine_Parts( Nam, D, N, DN ) ;
    if( length( DN ) = 0 ) then { Totally not found }
    begin
        DeviceSize := -1 ;
        exit ;
    end ;
    if( length( N ) = 0 ) then { DOS }
    begin
	DeviceSize := DOS_FS.DeviceSize( DN ) ;
    end else
    begin
        FS := Open_File_System( Nam, 0, UEC ) ;
        if( FS <> nil ) then
        begin
            DeviceSize := FS^.DeviceSize( Nam ) ;
            Close_File_System( Nam ) ;
        end else
        begin
            DeviceSize := -1 ;
            exit ;
        end ;
    end ;
end ;


function TFile_Processor.DeviceFree( const Nam : string ) : int64 ;

var D, N, DN : string ;
    FS : PFile_System ;
    UEC : TUEC ;

begin
    Determine_Parts( Nam, D, N, DN ) ;
    if( length( DN ) = 0 ) then { Totally not found }
    begin
        DeviceFree := -1 ;
        exit ;
    end ;
    if( length( N ) = 0 ) then { DOS }
    begin
	DeviceFree := DOS_FS.DeviceFree( DN ) ;
    end else
    begin
        FS := Open_File_System( Nam, 0, UEC ) ;
        if( FS <> nil ) then
        begin
            DeviceFree := FS^.DeviceFree( Nam ) ;
            Close_File_System( Nam ) ;
        end else
        begin
            DeviceFree := -1 ;
            exit ;
        end ;
    end ;
end ;


function TFile_Processor.MKDir( const Nam : string ) : longint ;

var D, N, DN : string ;
    FS : PFile_System ;
    UEC : TUEC ;

begin
    Determine_Parts( Nam, D, N, DN ) ;
    if( length( DN ) = 0 ) then { Totally not found }
    begin
        MKDir := -1 ;
        exit ;
    end ;
    if( length( N ) = 0 ) then { DOS }
    begin
	MKDir := DOS_FS.MKDir( DN ) ;
    end else
    begin
        FS := Open_File_System( DN, 0, UEC ) ;
        if( FS <> nil ) then
        begin
            MKDir := FS^.MKDir( N ) ;
            Close_File_System( DN ) ;
        end else
        begin
            MKDir := -1 ;
            exit ;
        end ;
    end ;
end ; { TFile_Processor.MKDir }


procedure Terminate ;

begin
    File_Processor.Done ;
    DOS_FS.Done ;
    DOS_SS.Done ;
    DOS_FSS.Done ;
end ;


{$IFDEF DELPHI}
initialization
{$ELSE}
begin
{$ENDIF}
    { Initialize data... }
    Saved_DN := '' ;
    File_Processor_Exists := False ;
    File_Processor.Init ;
    DOS_FS.Init ;
    DOS_SS.Init ;
    DOS_FSS.Init ;
    Register_Store_Server( @DOS_SS ) ;
    Register_File_System_Server( @DOS_FSS ) ;
    { Note: We do not destruct DOS_FS on exit since other units may shut down
      AFTER this one and they may access files opened via DOS_FS, which would
      result in the application blowing up.  Since DOS_FS is static, no memory
      leak should result (same with File_Processor and DOS_SS). }

{$IFDEF DELPHI}
finalization
  {$IFDEF CC_Auto_Free_Application_Manager}
      Terminate ;
  {$ENDIF}
{$ENDIF}
end.
