{
        Program Name : Menu_2D
        Package Name : LB
        Purpose      : 2-D GUI Menu
        Institution  : Conroy & Conroy Co.
        Date Written : 3-June-2002
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2002-2004 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  Mar-07-2006   RBJ         Cleaned up memory leaks.       *
        *  Nov-16-2006   EAC         Use T2D_Button.                *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          Implements a GUI menu in 2D.  If Align is alNone, this is a popup
        menu.  If Align is alClient, this is a popup within a right or left
        aligned main menu.  Otherwise, it is a "main menu" with the specified
        alignment.

          To allow proper operation, the application should route all messages
        to the top-level menu's Process_Message method.  Application messages
        can be captured via the TApplication.OnMessage.  Also, if the
        application is an MDI program, all MDI children must also route all
        messages to the top-level menu's Process_Message method.  MDI child
        messages can be captured via a descendant which overrides the
        DefaultHandler method.
}

unit Menu_2D ;

interface

uses // Borland...
     Windows, // TRect

     Classes, // TComponent
     Controls, // alTop
     Extctrls, // TPanel
     Forms, // TForm
     Graphics, // TColor
     Messages, // TWMLButtonDown

     // C&C...
     Button_2D, // T2D_Button
     Collect, // TInteger_List
     _DebugIn, // PDebug_Interface
     Decorations ; // TDecoration

type THighlight_Style = ( THS_Normal, // For horizontal menus use THS_Box, for vertical menus use THS_Color
                          THS_None, // Do no visible highlighting
                          THS_Color, // Use system menu highlight color
                          THS_Box // Box around highlighted item
                        ) ;
                        // NOTE: Currently THS_Box is not supported on vertical menus
                        //       and THS_Color is not supported on horizontal menus

type T2D_Menu = class( TPanel )
                    private // Instance data...
                        _Highlight_Style : THighlight_Style ; // How to visibly highlight
                        _Items : TList ;
                        Parent_Menu : T2D_Menu ;
                        Positions : TInteger_Collection ;
                        Highlighted_Menu_Index : integer ; // Sub menu over which the mouse is positioned
                        _Selected_Menu_Index : integer ; // Sub menu over which the mouse is positioned
                        In_Adjust_MDI_Buttons : boolean ; // True if in Adjust_MDI_Button routine (to prevent infinite recursion)
                        In_Process_Key : boolean ; // True if in Process_Key routine (to prevent infinite recursion)
                        In_RequestAlign : boolean ; // True if in requestalign to prevent infinite recursion
                        In_SetBounds : boolean ; // True if in setbounds routine (to prevent infinite recursion)
                        _Checked : boolean ; // True if checked
                        _RadioItem : boolean ; // True if a radio item
                        _GroupIndex : byte ;
                        _Caption : string ; // Ancestor has Caption too, but loses it when window handle is destroyed
                        _Image : TImage ; // Image to show
                        _Button : T2D_Button ; // Button to show (overrides Image)
                        _Visible : boolean ; // Using VCL's Visible property causes problems - this is our own
                        Last_Align : TAlign ;
                        _ShortCut : TShortCut ;
                        Min_Image_Width : integer ; // Minimum width to reserve for bitmap images on menu
                        _Hint : string ; // Override VCL's hint property so it doesn't automatically do stuff
                        _Last_Popup_Index : integer ;
                        _Translate_HTML : boolean ; // True to translate HTML on captions, false for literal text
                        _Decorations : TList ; // List of decorations for menu
                        _Style : TAlign ; // What style display (alNone = default for alignment)
                        Popup_Form : TForm ;
                        In_WMSetFocus : boolean ;
                        _Gradients : boolean ;
                        _Gradient_Bottom : TColor ;
                        _Gradient_Intermediate : TColor ;
                        _Gradient_Top : TColor ;

                        _MDI : boolean ; // True if we support MDI
                        MDI_Minimize : T2D_Button ;
                        MDI_Close : T2D_Button ;
                        MDI_Restore : T2D_Button ;
//                        MDI_Window : boolean ; // True if an MDI window menu ~~~ handle Window menu
                        // ~~~ handle MDI system menu (far left on main menu)
                        // ~~~ scrolling when too big for parent
                        // ~~~ support drag/drop
                        // ~~~ all menu text should gray out when a dialog pops up

                    public // Constructors and destructors...
                        constructor Create( AOwner : TComponent ) ; override ;
                        destructor Destroy ; override ;

                    protected // Property handlers...
                        function Get_Item( Index : integer ) : T2D_Menu ;
                            virtual ;
                        procedure Set_Button( Value : T2D_Button ) ; virtual ;
                        procedure Set_Caption( const Value : string ) ; virtual ;
                        procedure Set_Checked( Value : boolean ) ; virtual ;
                        procedure Set_GroupIndex( Value : byte ) ; virtual ;
                        procedure Set_Image( Value : TImage ) ; virtual ;
                        procedure Set_Item( Index : integer ;
                            Value : T2D_Menu ) ; virtual ;
                        procedure Set_MDI( Value : boolean ) ; virtual ;
                        procedure Set_RadioItem( Value : boolean ) ; virtual ;
                        procedure Set_Style( Value : TAlign ) ;
                        procedure Set_Translate_HTML( Value : boolean ) ; virtual ;

                    protected // Internal utility routines...
                        procedure Adjust_Alignment ;
                        procedure Adjust_MDI_Buttons ;
                        procedure Cancel_All_Popups ;
                        procedure CB_Minimize( Sender : TObject ) ;
                        procedure CB_MouseMove( Sender : TObject ;
                            Shift : TShiftState ; X, Y : Integer ) ;
                        procedure CB_Popup_Message( var Msg : TMsg ;
                            var Handled : Boolean ) ;
                        procedure CB_Close( Sender : TObject ) ;
                        procedure CB_Restore( Sender : TObject ) ;
                        function Check_Decorations : boolean ;
                        procedure Child_Highlighted( Child : T2D_Menu ) ;
                        procedure Child_Mouse_Move( Child : T2D_Menu ) ;
                        procedure CreateParams( var Params : TCreateParams ) ;
                            override ;
                        function Get_Decoration_Hint( out H : string ) : boolean ;
                        procedure Hide_MDI_Buttons ;
                        function Index_At( Point : TPoint ) : integer ;
                        procedure Invalidate_Index( Index : integer ) ;
                        function Item_Height( Index : integer ) : integer ;
                        procedure Mouse_Move ;
                        procedure Process_Decorations( _Canvas : TCanvas ;
                            X, Y, W, H : integer ) ;
                        function RectClientToScreen( const Position : TRect ) : TRect ;
                        procedure Setup_MDI_Buttons ;
                        function Text_Height : integer ; virtual ;
                        function Text_Width : integer ; virtual ;

                    public // Message handlers..
                        procedure CMCONTROLCHANGE( var Msg : TMessage ) ;
                            message CM_CONTROLCHANGE ;
                        procedure WmEraseBackground( var Msg : TMessage ) ;
                            message WM_ERASEBKGND ;
                        procedure WMLButtonDown( var Message : TWMLButtonDown ) ;
                            message WM_LBUTTONDOWN ;
                        procedure WMLButtonUp( var Message : TWMLButtonUp ) ;
                            message WM_LBUTTONUP ;
                        procedure WMRButtonUp( var Message : TWMRButtonUp ) ;
                            message WM_RBUTTONUP ;
                        procedure WMMouseMove( var Message : TWMMouseMove ) ;
                            message WM_MOUSEMOVE ;
                        procedure WMSetFocus( var Message : TWMSetFocus ) ;
                            message WM_SETFOCUS ;
                        procedure CMVisibleChanged( var Msg : TMessage ) ;
                            message CM_VisibleChanged ;

                    public // Overrides...
                        procedure DefaultHandler( var Message ) ; override ;
                        procedure Paint ; override ;
                        procedure RequestAlign ; override ;
                        procedure SetBounds( ALeft, ATop, AWidth, AHeight : Integer ) ;
                            override ;
                        procedure SetParent( AParent : TWinControl ) ;
                            override ;

                    public // MDI message handlers...
                        procedure Handle_MDICreate( Message : TMsg ) ; virtual ;
                        procedure Handle_Destroy( Message : TMsg ) ; virtual ;
                        procedure Handle_MDIMaximize( Message : TMsg ) ;
                            virtual ;
                        procedure Handle_Size( Message : TMsg ) ; virtual ;

                    public // API...
                        On_Property_Click : procedure( Sender : TObject ; var Keep_Active : boolean ) of object ;
                        On_Highlight : TNotifyEvent ;
                        On_Popup, On_Popped_Up : TNotifyEvent ;
                        On_Cancel_Popups : TNotifyEvent ;

                        procedure Activate( Key : word ; Up : boolean ) ;
                            virtual ;
                        procedure Add( M : T2D_Menu ) ; virtual ;
                        procedure Adjust_Size ; virtual ; // Resizes menu based on children
                        property Button : T2D_Button
                            read _Button
                            write Set_Button ;
                        property Caption : string
                            read _Caption
                            write Set_Caption ;
                        property Checked : Boolean
                            read _Checked
                            write Set_Checked ;
                        function Count : integer ; virtual ;
                        procedure Deactivate_Menus ; virtual ;
                        function Debugger : PDebug_Interface ; virtual ;
                        property Decorations : TList
                            read _Decorations
                            write _Decorations ;
                        procedure Deserialize( const Image : string ) ;
                        { Restore the component from a serialized string. }

                        function Display_Height : integer ;
                        procedure Do_Click ; virtual ;
                        property Gradients : boolean
                            read _Gradients
                            write _Gradients ;
                        property Gradient_Bottom : TColor
                            read _Gradient_Bottom
                            write _Gradient_Bottom ;
                        property Gradient_Intermediate : TColor
                            read _Gradient_Intermediate
                            write _Gradient_Intermediate ;
                        property Gradient_Top : TColor
                            read _Gradient_Top
                            write _Gradient_Top ;
                        property GroupIndex : byte
                            read _GroupIndex
                            write Set_GroupIndex ;
                        procedure _Hide ; virtual ;

                        function Highlighted_Index : integer ;
                        { Returns currently highlighted index. }

                        procedure Highlight_Index( Index : integer ;
                            Pop : boolean ) ; virtual ;
                        { Highlight the specified index.  If Pop is true, also
                          popup any child menu. }

                        property Highlight_Style : THighlight_Style
                            read _Highlight_Style
                            write _Highlight_Style ;

                        property Hint : string
                            read _Hint
                            write _Hint ;
                        property Image : TImage
                            read _Image
                            write Set_Image ;
                        function Index_Of( Menu : T2D_Menu ) : integer ;
                        function Index_Rectangle( Index : integer ) : TRect ;
                            virtual ;
                        procedure Invalidate ; override ;
                        function Item_Rectangle( Index : T2D_Menu ) : TRect ;
                            virtual ;

                        property Last_Popup_Index : integer
                            read _Last_Popup_Index
                            write _Last_Popup_Index ;

                        property MDI : boolean
                            read _MDI
                            write Set_MDI ;
                        function Popped_Up : boolean ; virtual ;
                        // True if we are visible via a call to Popup

                        procedure Popup( X, Y : integer ) ; virtual ;
                        procedure Popup_Index( Index : integer ) ; virtual ;

                        function Process_Accelerator( Key : integer ;
                            Shift : TShiftState ) : boolean ; virtual ;
                        procedure Process_Char( Key, Repeats : integer ) ;
                            virtual ;
                        procedure Process_Message( Message : TMsg ;
                            var Eat : boolean ) ; virtual ;
                        { Allows menu to monitor application messages and
                          respond appropriately.  If Eat is set to true, we have
                          processed the message and it should be thrown away. }

                        property RadioItem : Boolean
                                     read _RadioItem
                                     write Set_RadioItem ;

                        property Selected_Menu_Index : integer
                                     read _Selected_Menu_Index
                                     write _Selected_Menu_Index ;

                        function Serialize( State_Only : boolean ) : string ;
                        { Serialize the component.  If State_Only is true,
                          only serialize state information.  Otherwise, we
                          serialize our state and our full contents. }

                        property ShortCut : TShortCut
                                     read _Shortcut
                                     write _Shortcut ;

                        procedure _Show ; virtual ;

                        property Style : TAlign
                                     read _Style
                                     write Set_Style ;

                        function Top_Level_Menu : boolean ; virtual ;
                        { Returns True if we are a top-level menu (no parent). }

                        property Translate_HTML : boolean
                            read _Translate_HTML
                            write Set_Translate_HTML ;

                        property Items[ Index : integer ] : T2D_Menu
                            read Get_Item
                            write Set_Item ;
                end ; // T2D_Menu

implementation

uses // Borland...
     Menus, // Shortcut
     SysUtils, // inttostr

     // C&C...
     _App, // Application_Manager
     DebugInt, // TText_Debugger
     HTML, // Process_HTML_Text
     Menu_Popup_Form, // TPopup_Form
     MSWin, // XWM_Get_Hint_Exclusion_Rect
     CommonUt, // Edit
     Types, // Rect()
{$IFDEF Debug}
     UStrings, // Pointer_To_String
{$ENDIF}
     VCL_Std ; // Write_Text

{$IFDEF Debug}
type TMenu_2D_List = class( TText_Debugger )
                         public { API... }
                             List : TList ;

                             function Child( Index : longint ) : PDebug_Interface ;
                                 override ;

                             function Count : integer ; override ;
                     end ;

{ API... }

function TMenu_2D_List.Child( Index : longint ) : PDebug_Interface ;

var Iterator : TText_Debugger ;
    M : T2D_Menu ;

begin
    M := T2D_Menu( List[ Index ] ) ;
    Iterator := TText_Debugger( M.Debugger ) ;
    Iterator.Title := PChar( M.Caption ) ;
    Result := Iterator ;
end ;


function TMenu_2D_List.Count : integer ;

begin
    Count := List.Count + inherited Count ;
end ;


type T2D_Menu_Adjust_Size_Responder = class( TText_Debugger )
                                          public { API... }
                                              Menu : T2D_Menu ;

                                              procedure Activate ; override ;
                                              function Count : integer ;
                                                  override ;
                                      end ;

{ API... }

procedure T2D_Menu_Adjust_Size_Responder.Activate ;

begin
    Menu.Adjust_Size ;
end ;


function T2D_Menu_Adjust_Size_Responder.Count : integer ;

begin
    Count := 0 ;
end ;


type T2D_Menu_Debugger = class( TControl_Debug_Iterator )
                             public { API... }
                                 function Child( Index : integer ) : PDebug_Interface ;
                                     override ;

                                 function Count : integer ; override ;
                         end ;

{ API... }

function T2D_Menu_Debugger.Child( Index : integer ) : PDebug_Interface ;

var Iterator : TText_Debugger ;
    S : string ;

begin
    Result := nil ;
    Iterator := TText_Debugger.Create ;
    try
        case Index of
            0 : begin
                    Iterator.Free ;
                    Iterator := T2D_Menu_Adjust_Size_Responder.Create ;
                    T2D_Menu_Adjust_Size_Responder( Iterator ).Menu :=
                        T2D_Menu( Component ) ;
                    S := '!> Adjust_Size' ;
                end ;
            1 : begin
                    if( T2D_Menu( Component )._Items <> nil ) then
                    begin
                        Iterator.Free ;
                        Iterator := TMenu_2D_List.Create ;
                        TMenu_2D_List( Iterator ).List :=
                            T2D_Menu( T2D_Menu_Debugger )._Items ;
                    end ;
                    S := '_Items = ' +
                        Pointer_To_String( T2D_Menu( T2D_Menu_Debugger )._Items ) ;
                end ;
            else Result := inherited Child( Index - 2 ) ;
        end ; // case Index
    except
        on E : Exception do
            S := inttostr( Index ) + ' ' + E.Message ;
    end ;
    if( Result = nil ) then
    begin
        Iterator.Title := PChar( S ) ;
        Result := Iterator ;
    end ;
end ;


function T2D_Menu_Debugger.Count : integer ;

begin
    Count := 2 + inherited Count ;
end ;
{$ENDIF}


var Menus_Active : boolean = False ;
    Menu_Key_Mode : boolean = False ; // True if menu is in keyboard mode
    Mouse_Over : T2D_Menu = nil ; // Menu which is under mouse cursor
    Popped_Menus : TList = nil ;
    In_Constructor : boolean = false ;
    In_Destructor : integer = 0 ;
    Arrow_Bitmap : TBitmap ; // Menu arrow image
    Check_Bitmap : TBitmap ; // Menu check image
    Radio_Button_Bitmap : TBitmap ; // Menu dot image
    Off_Screen : integer ;
    _System_Menu : T2D_Menu = nil ; // Shared MDI system menu
    Dialog_Focused : boolean = false ; // True if a dialog is focused

function System_Menu : T2D_Menu ;

begin
    if( _System_Menu = nil ) then
    begin
        _System_Menu := T2D_Menu.Create( Application ) ;
        _System_Menu.Image := TImage.Create( _System_Menu ) ;
        _System_Menu.Image.Picture.Icon := TIcon.Create ;
        _System_Menu.Image.Picture.Icon.Handle := LoadIcon( 0, IDI_APPLICATION ) ;
    end ;
    Result := _System_Menu ;
end ; // System_Menu


procedure _Cancel_All_Popups ; // Cancels all popup menus, regardless

var Loop : integer ;
    This_Menu : T2D_Menu ;

begin
    for Loop := Popped_Menus.Count - 1 downto 0 do
    begin
        This_Menu := T2D_Menu( Popped_Menus[ Loop ] ) ;
        if(
            ( This_Menu.Parent_Menu = nil )
            or
            ( This_Menu.Align <> alClient )
          ) then
        begin
            if( not (
                  ( This_Menu.Align = alNone )
                  and
                  ( This_Menu.Parent_Menu <> nil )
                  and
                  (
                    ( This_Menu.Parent_Menu.Align = alLeft )
                    or
                    ( This_Menu.Parent_Menu.Align = alRight )
                  )
                )
              ) then
            begin
                This_Menu._Hide ;
            end ;
        end ;
    end ;
    Popped_Menus.Clear ;
end ;


procedure Cancel_Popups( Current : T2D_Menu ) ;
{ Cancels all popup menus which are children of Current, except children of
  left/right menus.  If Current is nil, then all popups are cancelled, except
  children of left/right menus. }

var Loop : integer ;
    This_Menu : T2D_Menu ;

begin
    if(
        ( Current <> nil )
        and
        ( Popped_Menus.IndexOf( Current ) = -1 )
        and
        ( Current.Parent_Menu <> nil )
      ) then
    begin
        exit ;
    end ;
    for Loop := Popped_Menus.Count - 1 downto 0 do
    begin
        This_Menu := T2D_Menu( Popped_Menus[ Loop ] ) ;
        if(
            ( This_Menu = Current )
            or
            (
              ( Loop = 0 )
              and
              ( This_Menu.Parent_Menu <> nil )
              and
              (
                ( This_Menu.Parent_Menu.Align = alLeft )
                or
                ( This_Menu.Parent_Menu.Align = alRight )
              )
            )
          ) then
        begin
            Popped_Menus.Count := Loop + 1 ;
            exit ;
        end ;
        This_Menu._Hide ;
    end ; // for Loop := Popped_Menus.Count - 1 downto 0
    Popped_Menus.Clear ;
end ; // Cancel_Popups


// T2D_Menu methods...

// Constructors and destructors...

constructor T2D_Menu.Create( AOwner : TComponent ) ;

begin
    In_Constructor := True ;

    inherited Create( AOwner ) ;

    _Items := TList.Create ;
    Positions.Init( 0, 4 ) ;
    Highlighted_Menu_Index := -1 ; // Nothing highlighted
    _Selected_Menu_Index := -1 ; // Nothing selected
    BevelOuter := bvNone ;
    Color := clMenu ;
    Align := alTop ;
    Last_Align := Align ;
    Height := GetSystemMetrics( SM_CYMENU ) + 2 ;
    Ctl3D := True ;
    MDI := True ;
    _Visible := True ;
    _Translate_HTML := True ;
    Decorations := TList.Create ;
    Gradients := False ; ;

    In_Constructor := False ;
end ;


destructor T2D_Menu.Destroy ;

var Loop : integer ;

begin
    inc( In_Destructor ) ;

    FreeAndNil( _Button ) ;

    _Items.Free ;
    _Items := nil ;
    Positions.Done ;
    for Loop := 0 to Decorations.Count - 1 do
    begin
        TDecoration( Decorations[ Loop ] ).Free ;
    end ;
    Decorations.Free ;
    Decorations := nil ;
    if( Popup_Form <> nil ) then
    begin
        TPopup_Form( Popup_Form ).On_Message := nil ;
        if( not Application_Manager^.Get_In_Shut_Down ) then
        begin
            Popup_Form.Release ;
        end ;
    end ;

    inherited Destroy ;

    dec( In_Destructor ) ;
end ;


// Property handlers...

function T2D_Menu.Get_Item( Index : integer ) : T2D_Menu ;

begin
    Result := T2D_Menu( _Items[ Index ] ) ;
end ;


procedure T2D_Menu.Set_Button( Value : T2D_Button ) ;

begin
    _Button := Value ;
    if( Parent_Menu = nil ) then
    begin
        Invalidate ;
    end else
    begin
        Parent_Menu.Invalidate ;
    end ;
end ;


procedure T2D_Menu.Set_Caption( const Value : string ) ;

begin
    if( _Caption <> Value ) then // Avoid excessive redraws
    begin
        _Caption := Value ;
        if( Parent_Menu = nil ) then
        begin
            Invalidate ;
        end else
        begin
            Parent_Menu.Invalidate ;
        end ;
    end ;
end ;


procedure T2D_Menu.Set_Checked( Value : boolean ) ;

var Loop : integer ;
    This_Menu : T2D_Menu ;

begin
    _Checked := Value ;
    Invalidate ;
    if( ( Parent_Menu <> nil ) and RadioItem and Value ) then // Checking a radio item
    begin
        for Loop := 0 to Parent_Menu.Count - 1 do
        begin
            This_Menu := Parent_Menu.Items[ Loop ] ;
            if(
                ( This_Menu <> self )
                and
                ( GroupIndex = This_Menu.GroupIndex )
                and
                This_Menu.RadioItem
              ) then
            begin
                This_Menu.Checked := False ;
            end ;
        end ;
    end ;
end ;


procedure T2D_Menu.Set_GroupIndex( Value : byte ) ;

begin
    _GroupIndex := Value ;
    Checked := Checked ; // Makes sure only one check per group and invalidates
end ;


procedure T2D_Menu.Set_Image( Value : TImage ) ;

begin
    _Image := Value ;
end ;


procedure T2D_Menu.Set_Item( Index : integer ; Value : T2D_Menu ) ;

begin
    _Items[ Index ] := Value ;
end ;


procedure T2D_Menu.Set_MDI( Value : boolean ) ;

begin
    _MDI := Value ;
    if( Application.MainForm <> nil ) then
    begin
        if( Value ) then
        begin
            Adjust_MDI_Buttons ;
        end else
        begin
            Hide_MDI_Buttons ;
        end ;
    end ;
end ;


procedure T2D_Menu.Set_RadioItem( Value : boolean ) ;

begin
    _RadioItem := Value ;
    Checked := Checked ; // Makes sure only one check per group and invalidates
end ;


procedure T2D_Menu.Set_Style( Value : TAlign ) ;

begin
    if( Value <> _Style ) then
    begin
        _Style := Value ;
        Invalidate ;
    end ;
end ;


procedure T2D_Menu.Set_Translate_HTML( Value : boolean ) ;

begin
    if( Value <> _Translate_HTML ) then
    begin
        _Translate_HTML := Value ;
        invalidate ;
    end ;
end ;


// Internal utility routines...

procedure T2D_Menu.Adjust_Alignment ;

var Loop : integer ;

begin
    Cancel_All_Popups ;
    Last_Align := Align ;
    _Selected_Menu_Index := -1 ;
    Menus_Active := False ;
    Highlighted_Menu_Index := -1 ;
    for Loop := 0 to Count - 1 do
    begin
        Items[ Loop ]._Hide ;
    end ;
    Paint ; // Recalculate positions
    Invalidate ;
end ;


procedure T2D_Menu.Adjust_MDI_Buttons ;

var Control : TControl ;
    Loop : integer ;
    Maxed : boolean ;

begin
    // Setup...
    if( In_Adjust_MDI_Buttons ) then
    begin
        exit ;
    end ;
    In_Adjust_MDI_Buttons := True ;
    Maxed := False ;
    for Loop := 0 to Application.MainForm.ControlCount - 1 do
    begin
        Control := Application.MainForm.Controls[ Loop ] ;
        if( Control is TForm ) then
        begin
            if( TForm( Control ).WindowState = wsMaximized ) then
            begin
                Maxed := True ;
                break ;
            end ;
        end ;
    end ;

    // Hide/show MDI buttons
    if(
        ( Application.MainForm.MDIChildCount = 0 )
        or
        ( not Maxed )
      ) then
    begin
        Hide_MDI_Buttons ;
    end else
    begin
        Setup_MDI_Buttons ;
    end ;
    In_Adjust_MDI_Buttons := False ;
end ;


procedure T2D_Menu.Cancel_All_Popups ;

begin
    if( assigned( On_Cancel_Popups ) ) then
    begin
        On_Cancel_Popups( self ) ;
    end ;
    _Cancel_All_Popups ;
end ;


procedure T2D_Menu.CB_Minimize( Sender : TObject ) ;

begin
    Application.MainForm.ActiveMDIChild.WindowState := wsMinimized ;
end ;


procedure T2D_Menu.CB_MouseMove( Sender : TObject ; Shift : TShiftState ;
    X, Y : Integer ) ;

begin
    Mouse_Move ;
end ;


procedure T2D_Menu.CB_Popup_Message( var Msg : TMsg ; var Handled : Boolean ) ;

begin
    Process_Message( Msg, Handled ) ;
end ;


procedure T2D_Menu.CB_Close( Sender : TObject ) ;

begin
    Application.MainForm.ActiveMDIChild.Close ;
end ;


procedure T2D_Menu.CB_Restore( Sender : TObject ) ;

begin
    if( Application.MainForm.ActiveMDIChild = nil ) then
    begin
        SendMessage( Application.MainForm.Handle, WM_MDIRESTORE, Application.MainForm.ActiveMDIChild.Handle, 0 ) ;
    end else
    begin
        Application.MainForm.ActiveMDIChild.WindowState := wsNormal ;
    end ;
end ;


function T2D_Menu.Get_Decoration_Hint( out H : string ) : boolean ;

var Decoration : TDecoration ;
    Loop : integer ;
    Point : TPoint ;
    This_Menu : T2D_Menu ;

begin
    Result := False ;
    Point := Mouse.CursorPos ;
    Point := ScreenToClient( Point ) ;
    Loop := Index_At( Point ) ;
    if( Loop <> -1 ) then // Found a match
    begin
        This_Menu := Items[ Loop ] ;
        for Loop := 0 to This_Menu.Decorations.Count - 1 do
        begin
            Decoration := TDecoration( This_Menu.Decorations[ Loop ] ) ;
            if( Decoration.Visible and Decoration.Enabled and ( Length( Decoration.Enabled_Hint ) > 0 ) ) then
            begin
                if( PtInRect( Decoration.Rect, Point ) ) then
                begin
                    H := Decoration.Enabled_Hint ;
                    Result := True ;
                    exit ;
                end ;
            end ;
        end ; // for Loop := 0 to This_Menu.Decorations.Count - 1
    end ; // if( Loop <> -1 )

    if( ( Count > 0 ) and ( Highlighted_Menu_Index >= 0 ) ) then // Check children
    begin
        Result := Items[ Highlighted_Menu_Index ].Get_Decoration_Hint( H ) ;
    end ;
end ; // T2D_Menu.Get_Decoration_Hint


function T2D_Menu.Check_Decorations : boolean ;

var Decoration : TDecoration ;
    Loop : integer ;
    Point : TPoint ;
    This_Menu : T2D_Menu ;

begin
    Result := False ;
    Point := ScreenToClient( Mouse.CursorPos ) ;
    Loop := Index_At( Point ) ;
    if( Loop <> -1 ) then // Found a match
    begin
        This_Menu := Items[ Loop ] ;
        for Loop := 0 to This_Menu.Decorations.Count - 1 do
        begin
            Decoration := TDecoration( This_Menu.Decorations[ Loop ] ) ;
            if( Decoration.Visible and Decoration.Enabled and assigned( Decoration.OnClick ) ) then
            begin
                if( PtInRect( Decoration.Rect, Point ) ) then
                begin
                    Decoration.OnClick( Decoration ) ;
                    Result := True ;
                    exit ;
                end ;
            end ;
        end ; // for Loop := 0 to This_Menu.Decorations.Count - 1
    end ; // if( Loop <> -1 )

    if( ( Count > 0 ) and ( Highlighted_Menu_Index >= 0 ) ) then // Check children
    begin
        Result := Items[ Highlighted_Menu_Index ].Check_Decorations ;
    end ;
end ; // T2D_Menu.Check_Decorations


procedure T2D_Menu.Child_Highlighted( Child : T2D_Menu ) ; // Notification of a child being highlighted

var Loop : integer ;

begin
    for Loop := 0 to Count - 1 do
    begin
        if( Child = Items[ Loop ] ) then
        begin
            if( Loop <> Highlighted_Menu_Index ) then
            begin
                Highlight_Index( Loop, False ) ;
                exit ;
            end ;
        end ;
    end ;
end ;


procedure T2D_Menu.Child_Mouse_Move( Child : T2D_Menu ) ;

begin
    if( inherited Hint <> '' ) then
    begin
        inherited Hint := '' ;
        Reset_Hint ;
    end ;
end ;


procedure T2D_Menu.CreateParams( var Params : TCreateParams ) ;

begin
    inherited CreateParams( Params ) ;

    if( Parent = nil ) then // Not the main menu or an embedded child
    begin
        Params.Style := WS_POPUPWINDOW and ( not ( WS_BORDER or WS_SYSMENU ) ) ;
        Params.ExStyle := WS_EX_TOOLWINDOW or WS_EX_TOPMOST ;
    end else
    begin
//        Params.Style := Params.Style and ( not WS_CLIPCHILDREN ) ;
    end ;
end ;


// Internal utility routines...

procedure T2D_Menu.Hide_MDI_Buttons ;

begin
    if( MDI_Minimize <> nil ) then
    begin
        MDI_Minimize.Free ;
        MDI_Close.Free ;
        MDI_Restore.Free ;
        MDI_Minimize := nil ;
        MDI_Close := nil ;
        MDI_Restore := nil ;
        _Items.Delete( 0 ) ; // Remove system menu
    end ;
end ;


function T2D_Menu.Index_At( Point : TPoint ) : integer ;

var Loop : integer ;

begin
    Result := -1 ; // Assume failure to find
    if( ( Align = alTop ) or ( Align = alBottom ) ) then
    begin
        for Loop := 0 to Positions.Count - 1 do
        begin
            if(
                ( Positions.At( Loop ) < Point.X )
                and
                ( ( Loop = Positions.Count - 1 ) or ( Positions.At( Loop + 1 ) > Point.X ) )
                and
                ( Loop < Count )
              ) then
            begin
                Result := Loop ;
                exit ;
            end ;
        end ;
    end else
    begin
        for Loop := 0 to Positions.Count - 1 do
        begin
            if(
                ( Positions.At( Loop ) < Point.Y )
                and
                ( ( Loop = Positions.Count - 1 ) or ( Positions.At( Loop + 1 ) > Point.Y ) )
                and
                ( Loop < Count )
              ) then
            begin
                Result := Loop ;
                exit ;
            end ;
        end ;
    end ;
end ; // T2D_Menu.Index_At


procedure T2D_Menu.Invalidate_Index( Index : integer ) ;

var Rect : TRect ;

begin
    if( Index < 0 ) then
    begin
        exit ; // Ignore invalid index
    end ;
    Rect := Index_Rectangle( Index ) ;
    InvalidateRect( Handle, @Rect, True ) ;
end ;


function T2D_Menu.Item_Height( Index : integer ) : integer ;

var DC : HDC ;
    Dummy, H : integer ;
    Saved : THandle ;
    TM : TTextMetric ;
    This_Menu : T2D_Menu ;

begin
    // Setup...
    DC := CreateDC( 'DISPLAY', nil, nil, nil ) ;
    Saved := SelectObject( DC, Font.Handle ) ;
    GetTextMetrics( DC, TM ) ;
    H := TM.tmHeight ;
    This_Menu := Items[ Index ] ;
    SelectObject( DC, Saved ) ;
    DeleteDC( DC ) ;

    // Calculate height
    if( This_Menu.Caption = '-' ) then
    begin
        if( ( Align = alLeft ) or ( Align = alRight ) ) then // left/right aligned menus have borders, so ignore lines
        begin
            Result := 0 ;
        end else
        begin
            Result := H div 2 ;
        end ;
    end else
    begin
        Dummy := H + 4 ;
        if( ( Align = alLeft ) or ( Align = alRight ) ) then
        begin
            if( Dummy < GetSystemMetrics( SM_CYMENU ) ) then
            begin
                Dummy := GetSystemMetrics( SM_CYMENU ) ; // Minimum height
            end ;
        end ;
        if( This_Menu.Button <> nil ) then
        begin
            if( ( This_Menu.Button.Height > Dummy ) and This_Menu.Button.Visible ) then
            begin
                Dummy := This_Menu.Button.Height ;
            end ;
        end else
        begin
            if( This_Menu.Image <> nil ) then
            begin
                if( ( This_Menu.Image.Height > Dummy ) and ( This_Menu.Image.Visible ) ) then
                begin
                    Dummy := This_Menu.Image.Height ;
                end ;
            end ;
        end ;
        Result := Dummy ;
    end ;
end ; // T2D_Menu.Item_Height


procedure T2D_Menu.Mouse_Move ;

var Loop : integer ;
    Point : TPoint ;
    H : string ;

begin
    try
        Point := Mouse.CursorPos ;
        Point := ScreenToClient( Point ) ;
        if(
            ( Point.X < 0 )
            or
            ( Point.Y < 0 )
            or
            ( Point.X > Width )
            or
            ( Point.Y > Height )
          ) then
        begin
            Highlight_Index( -1, True ) ;
        end else
        begin
            Loop := Index_At( Point ) ;
            if( Loop <> -1 ) then
            begin
                Highlight_Index( Loop, True ) ;
            end ;
        end ;

        if( Highlighted_Menu_Index = -1 ) then
        begin
            inherited Hint := '' ;
        end else
        begin
            if Get_Decoration_Hint( H ) then
            begin
                Hint := H ;
            end else
            begin
                Hint := Items[ Highlighted_Menu_Index ].Hint ;
                H := Hint ;
            end ;
            if( inherited Hint <> H ) then
            begin
                inherited Hint := H ;
                Reset_Hint ;
            end ;
        end ;
        if( Parent_Menu <> nil ) then
        begin
            Parent_Menu.Child_Mouse_Move( self ) ;
        end ;
    except
    end ;
end ; // T2D_Menu.Mouse_Move


procedure T2D_Menu.Process_Decorations( _Canvas : TCanvas ;
    X, Y, W, H : integer ) ;

var Decoration : TDecoration ;
    Loop, Offset : integer ;

begin
    // Do non-right-aligned decorations...
    for Loop := 0 to Decorations.Count - 1 do
    begin
        Decoration := TDecoration( Decorations[ Loop ] ) ;
        if( Decoration.Visible and ( Decoration.Align <> alRight ) ) then
        begin
            Decoration.Parent := self ;
            Offset := ( ( H - Decoration.ClientHeight( W, H ) ) div 2 ) - 1 ;
            Decoration.Paint( _Canvas, X, Y + Offset, W, H ) ;
            X := X + Decoration.ClientWidth( W, H ) ;
        end ;
    end ;

    // Do right-aligned decorations...
    X := Width - 1 ;
    for Loop := 0 to Decorations.Count - 1 do
    begin
        Decoration := TDecoration( Decorations[ Loop ] ) ;
        if( Decoration.Visible and ( Decoration.Align = alRight ) ) then
        begin
            Decoration.Parent := self ;
            X := X - Decoration.ClientWidth( W, H ) ;
            Offset := ( ( H - Decoration.ClientHeight( W, H ) ) div 2 ) - 1 ;
            Decoration.Paint( _Canvas, X, Y + Offset, W, H ) ;
        end ;
    end ;
end ;


function T2D_Menu.RectClientToScreen( const Position : TRect ) : TRect ;

var Point : TPoint ;

begin
    Point.X := Position.Left ;
    Point.Y := Position.Top ;
    Point := ClientToScreen( Point ) ;
    Result.Left := Point.X ;
    Result.Top := Point.Y ;
    Point.X := Position.Right ;
    Point.Y := Position.Bottom ;
    Point := ClientToScreen( Point ) ;
    Result.Right := Point.X ;
    Result.Bottom := Point.Y ;
end ;


procedure T2D_Menu.Setup_MDI_Buttons ;

var Bitmap : TBitmap ;

    procedure Setup_Button( Button : T2D_Button ; Image : integer ) ;

    var B : TBitmap ;
        Rect : TRect ;

    begin
        B := TBitmap.Create ;
        B.Height := Button.Height - 1 ; // Button height, minus border
        B.Width := B.Height ;
        Bitmap.Height := Button.Height ;
        Bitmap.Width := Button.Width ;
        B.Canvas.Brush.Color := clBtnFace ;
        Rect := Button.ClientRect ;
        Rect.Bottom := Rect.Bottom + 1 ;
        Rect.Right := Rect.Right + 1 ;
        B.Canvas.FillRect( Rect ) ;
        Rect.Top := 0 ;
        Rect.Left := 0 ;
        Rect.Bottom := Bitmap.Height - 1 ;
        Rect.Right := Bitmap.Width - 1 ;
        DrawFrameControl( Bitmap.Canvas.Handle, Rect,
            DFC_CAPTION, Image or DFCS_FLAT ) ;
        BitBlt( B.Canvas.Handle, 1, 0, B.Height - 1,
            B.Width - 1, Bitmap.Canvas.Handle, 1, 1, SRCAND ) ;
        Button.Glyph := B ;
    end ; // T2D_Menu.Setup_MDI_Buttons.Setup_Button

var W : integer ;

begin // T2D_Menu.Setup_MDI_Buttons
    // Setup MDI buttons...
    if( MDI_Restore = nil ) then
    begin
        MDI_Minimize := T2D_Button.Create( self ) ;
        MDI_Minimize.Parent := self ;
        MDI_Minimize.OnClick := CB_Minimize ;
        MDI_Minimize.Height := GetSystemMetrics( SM_CYSMSIZE ) ;
        MDI_Minimize.Width := GetSystemMetrics( SM_CXSMSIZE ) ;
        MDI_Close := T2D_Button.Create( self ) ;
        MDI_Close.Parent := self ;
        MDI_Close.OnClick := CB_Close ;
        MDI_Close.Height := MDI_Minimize.Height ;
        MDI_Close.Width := MDI_Minimize.Height ;
        MDI_Restore := T2D_Button.Create( self ) ;
        MDI_Restore.Parent := self ;
        MDI_Restore.OnClick := CB_Restore ;
        MDI_Restore.Height := MDI_Minimize.Height ;
        MDI_Restore.Width := MDI_Minimize.Height ;
        Bitmap := TBitmap.Create ;
        Setup_Button( MDI_Close, DFCS_CAPTIONCLOSE ) ;
        Setup_Button( MDI_Minimize, DFCS_CAPTIONMIN ) ;
        Setup_Button( MDI_Restore, DFCS_CAPTIONRESTORE ) ;
        Bitmap.Free ;
    end ;
    if( ( Align = alTop ) or ( Align = alBottom ) ) then
    begin
        W := BorderWidth + 2 ;
        if( Ctl3D ) then
        begin
            W := W + 1 ;
        end ;
        MDI_Close.Top := W ;
    end else
    begin
        if( ( Align = alLeft ) or ( Align = alRight ) ) then
        begin
            MDI_Close.Top := Parent.ClientHeight - BorderWidth - MDI_Close.Height - 1 ;
        end else
        begin
            MDI_Close.Top := ClientHeight - BorderWidth - MDI_Close.Height - 1 ;
        end ;
    end ;
    MDI_Minimize.Top := MDI_Close.Top ;
    MDI_Restore.Top := MDI_Close.Top ;
    if( ( ( Align = alTop ) or ( Align = alBottom ) ) and ( Parent <> nil ) ) then
    begin
        W := Parent.ClientWidth ;
    end else
    begin
        W := ClientWidth ;
    end ;
    MDI_Close.Left := W - BorderWidth - 2 - MDI_Close.Width ;
    MDI_Restore.Left := MDI_Close.Left - MDI_Restore.Width - 3 ;
    MDI_Minimize.Left := MDI_Restore.Left - MDI_Minimize.Width ;
    if( ( _Items.Count = 0 ) or ( Items[ 0 ] <> System_Menu ) ) then
    begin
        _Items.Insert( 0, System_Menu ) ;
    end ;
end ; // T2D_Menu.Setup_MDI_Buttons


function T2D_Menu.Text_Height : integer ;

var Loop : integer ;
    This_Menu : T2D_Menu ;

begin
    // Early out...
    if( ( ControlCount > 0 ) and ( Count = 0 ) ) then
    begin
        Result := Height ;
        exit ;
    end ;

    Result := BorderWidth * 2 ;
    if( Ctl3D ) then
    begin
        Result := Result + 2 ;
    end ;
    for Loop := 0 to Count - 1 do
    begin
        This_Menu := Items[ Loop ] ;
        if( This_Menu.Visible ) then
        begin
            Result := Result + Item_Height( Loop ) ;
        end ;
    end ;
end ; // T2D_Menu.Text_Height


function T2D_Menu.Text_Width : integer ;

var DC : HDC ;
    Size : TSize ;

    function Get_Width( const S : string ) : integer ;

    var Temp : integer ;
    begin
        if( Translate_HTML ) then
        begin
            Result := Process_HTML_Text( DC, Font, S, False, 0, 0, 32767, Temp, True ) ;
        end else
        begin
            GetTextExtentPoint32( DC, PChar( S ), length( S ), Size ) ;
            Result := Size.cx ;
        end ;
    end ;

var Loop : integer ;
    Margin : integer ;
    Min_Width, Min_Sub_Width : integer ;
    Saved : THandle ;
    This_Menu : T2D_Menu ;
    TM : TTextMetric ;
    W : integer ;

begin
    // Early out...
    if( ( ControlCount > 0 ) and ( Count = 0 ) ) then
    begin
        Result := Width ;
        exit ;
    end ;

    // Setup...
    DC := CreateDC( 'DISPLAY', nil, nil, nil ) ;
    Saved := SelectObject( DC, Font.Handle ) ;
    GetTextMetrics( DC, TM ) ;
    Margin := ( BorderWidth + GetSystemMetrics( SM_CXMENUCHECK ) );
    if( Margin < TM.tmHeight + 4 ) then
    begin
        Margin := TM.tmHeight + 4 ;
    end ;

    // Recalculate width...
    Min_Width := 0 ;
    Min_Sub_Width := 0 ;
    Min_Image_Width := 0 ;
    for Loop := 0 to Count - 1 do
    begin
        This_Menu := Items[ Loop ] ;
        if( This_Menu.Visible ) then
        begin
            W := Get_Width( This_Menu.Caption ) ;
            if( This_Menu.Shortcut <> 0 ) then
            begin
                W := W + Get_Width( ShortCutToText( This_Menu.Shortcut ) ) + Margin ;
            end else
            if( pos( #9, This_Menu.Caption ) > 0 ) then
            begin
                W := W + Margin - Get_Width( #9 ) ;
            end ;
            if( W > Min_Width ) then
            begin
                Min_Width := W ;
            end ;
            if(
                ( ( Align = alLeft ) or ( Align = alRight ) )
                and
                (
                  ( This_Menu.Count > 0 )
                  or
                  ( This_Menu.ControlCount > 0 )
                )
              ) then
            begin
                // Submenus must fit within left/right menus
                W := This_Menu.Text_Width ;//+ 2 ; // Removed since multiple calls to this routine caused the submenu to expand in width
                if( W > Min_Sub_Width ) then
                begin
                    Min_Sub_Width := W ;
                end ;
            end ;
            if( This_Menu.Button <> nil ) then
            begin
                if(
                    ( This_Menu.Button.Width div This_Menu.Button.NumGlyphs > Min_Image_Width )
                    and
                    This_Menu.Button.Visible
                  ) then
                begin
                    Min_Image_Width := This_Menu.Button.Width div This_Menu.Button.NumGlyphs ;
                end ;
            end else
            begin
                if( This_Menu.Image <> nil ) then
                begin
                    if(
                        ( This_Menu.Image.Width > Min_Image_Width )
                        and
                        This_Menu.Image.Visible
                      ) then
                    begin
                        Min_Image_Width := This_Menu.Image.Width div This_Menu.Button.NumGlyphs ;
                    end ;
                end ;
            end ;
        end ;
    end ; // for Loop := 0 to Count - 1
    SelectObject( DC, Saved ) ;
    DeleteDC( DC ) ;

    if( Min_Image_Width > 0 ) then
    begin
        Min_Image_Width := Min_Image_Width + 8 ;
    end ;
    if( Min_Image_Width > Margin ) then
    begin
        Margin := Margin + Min_Image_Width ;
    end else
    begin
        Margin := Margin * 2 ; // Need space on both sides
    end ;
    if( Min_Width + Margin > Min_Sub_Width ) then
    begin
        Result := Min_Width + Margin ;
    end else
    begin
        Result := Min_Sub_Width ;
    end ;

    // Add in the MDI buttons
    if(
        ( Application.MainForm <> nil )
        and
        ( Application.MainForm.MDIChildCount > 0 )
        and
        ( Application.MainForm.ActiveMDIChild.WindowState = wsMaximized )
        and
        MDI
        and
        ( ( Align = alTop ) or ( Align = alBottom ) )
      ) then
    begin
        Result := Result + MDI_Close.Width * 3 + 1 ;
    end ;
end ; // T2D_Menu.Text_Width


// Message handlers..

procedure T2D_Menu.CMCONTROLCHANGE( var Msg : TMessage ) ;
{ When a child is added to us, we snag its onmousemove so we can handle hint
  text properly.  Note that we treat all children controls as if they are
  TPanels.  The reason for this is that TPanels exposes the OnMouseMove event,
  whereas ancestors of that class do not.  The reason the code works is because
  TPanels and non-TPanels, alike, inherit the same OnMouseMove from the TControl
  ancestor so it matters not how we cast it.  Finally, note that
  TGraphicsControls are excluded because ControlCount causes an exception when
  used on them. }

    var Our_CB : TMouseMoveEvent ;

    procedure Reroute_Mouse_Moves( Panel : TPanel ) ;

    var Loop : integer ;

    begin
        if( Msg.LParam = 0 ) then // Removing child control
        begin
            Our_CB := CB_MouseMove ;
            if( @Panel.OnMouseMove = @Our_CB ) then
            begin
                Panel.OnMouseMove := nil ;
            end ;
            if( not ( TControl( Panel ) is TGraphicControl ) ) then
            begin
                for Loop := 0 to Panel.ControlCount - 1 do
                begin
                    Reroute_Mouse_Moves( TPanel( Panel.Controls[ Loop ] ) ) ;
                end ;
            end ;
        end else
        begin
            if( length( Panel.Hint ) = 0 ) then
            begin
                Panel.OnMouseMove := CB_MouseMove ; // Allows us to handle hint text better
                if( not ( TControl( Panel ) is TGraphicControl ) ) then
                begin
                    for Loop := 0 to Panel.ControlCount - 1 do
                    begin
                        Reroute_Mouse_Moves( TPanel( Panel.Controls[ Loop ] ) ) ;
                    end ;
                end ;
            end ;
        end ;
    end ; // T2D_Menu.CMCONTROLCHANGE.Reroute_Mouse_Moves

begin
    inherited ;

    if( not ( TControl( Msg.WParam ) is T2D_Menu ) ) then
    begin
        Reroute_Mouse_Moves( TPanel( Msg.WParam ) ) ;
    end ;
end ; // T2D_Menu.CMCONTROLCHANGE


procedure T2D_Menu.WmEraseBackground( var Msg : TMessage ) ;

begin
    if( ( ControlCount > 0 ) and ( Count = 0 ) ) then
    begin
        inherited ;
    end else
    begin
        Msg.Result := 1 ; // We do our own background
    end ;
end ;


procedure T2D_Menu.WMLButtonDown( var Message : TWMLButtonDown ) ;

begin
    if( Mouse_Over = nil ) then
    begin
        Deactivate_Menus ;
        inherited ;
        exit ;
    end ;

    inherited ;

    Activate( VK_LBUTTON, False ) ;
end ;


procedure T2D_Menu.WMLButtonUp( var Message : TWMLButtonUp ) ;

var Point : TPoint ;
    Saved : TNotifyEvent ;

begin
    if( Mouse_Over = nil ) then
    begin
        Deactivate_Menus ;
        inherited ;
        exit ;
    end ;

    Point.X := Message.XPos ;
    Point.Y := Message.YPos ;

    if(
        ( Point.X < 0 )
        or
        ( Point.Y < 0 )
        or
        ( Point.X > ClientWidth )
        or
        ( Point.Y > ClientHeight )
      ) then
    begin
        inherited ;
        exit ;
    end ;
    if( Menus_Active ) then
    begin
        Activate( VK_LBUTTON, True ) ;
    end else
    begin
        Check_Decorations ;
    end ;

    Saved := OnClick ; // Do this so VCL won't also do the OnClick
    OnClick := nil ;
    inherited ;
    OnClick := Saved ;
end ; // T2D_Menu.WMLButtonUp


procedure T2D_Menu.WMRButtonUp( var Message : TWMRButtonUp ) ;

begin
    Activate( VK_RBUTTON, True ) ;
end ;


procedure T2D_Menu.WMMouseMove( var Message : TWMMouseMove ) ;

(*
var C : TWinControl ;
    Handle : HWND ;
    M : T2D_Menu ;
    Point : TPoint ;
*)

begin
    inherited ;

(*
    Point.X := Message.XPos ;
    Point.Y := Message.YPos ;
    Point := ClientToScreen( Point ) ;
    Handle := WindowFromPoint( Point ) ;
    M := nil ;
    if( Handle <> 0 ) then
    begin
        C := FindControl( Handle ) ;
        if( ( C = nil ) or ( not ( C is T2D_Menu ) ) ) then
        begin
            Handle := 0 ;
            M := nil ;
        end else
        begin
            M := T2D_Menu( C ) ;
            if( ( M.ControlCount > 0 ) and ( M.Count = 0 ) ) then
            begin
                Handle := 0 ;
                M := nil ;
            end ;
            Point := C.ScreenToClient( Point ) ;
        end ;
    end ;

    if(
        ( Handle = 0 )
        or
        ( Point.X < 0 )
        or
        ( Point.Y < 0 )
        or
        ( Point.X > C.ClientWidth )
        or
        ( Point.Y > C.ClientHeight )
      ) then
    begin
        Mouse_Over := nil ;
        if( not Menus_Active ) then // No menu is active
        begin
            if( M <> nil ) then
            begin
                M.Highlight_Index( -1, True ) ;
            end ;
            Deactivate_Menus ;
        end ;
        exit ;
    end ;
    Mouse_Over := M ;
    M.Mouse_Move ;
*)
    Mouse_Over := self ;
    Mouse_Move ;
end ; // T2D_Menu.WMMouseMove


procedure T2D_Menu.WMSetFocus( var Message : TWMSetFocus ) ;

begin
    if( In_WMSetFocus ) then // Prevent infinite loops (happens in Windows 98 when code is compiled with Delphi 7)
    begin
        exit ;
    end ;
    In_WMSetFocus := True ;
    Windows.SetFocus( Message.FocusedWnd ) ; // Don't allow us to get the focus
    In_WMSetFocus := False ;
end ;


// Overrides...

procedure T2D_Menu.RequestAlign ;

var Dummy : integer ;

begin
    inherited RequestAlign ;

    if( not In_RequestAlign ) then
    begin
        In_RequestAlign := True ;
        if( ( Align = alTop ) or ( Align = alBottom ) ) then
        begin
            Dummy := GetSystemMetrics( SM_CYMENU ) + 2 ;
            if( Dummy <> Height ) then
            begin
                Height := Dummy ;
            end ;
        end else
        begin
            if( ControlCount = 0 ) then
            begin
                Dummy := Text_Width ;
                if( Dummy <> Width ) then
                begin
                    Width := Dummy ;
                end ;
            end ;
        end ;
        if( Align <> Last_Align ) then
        begin
            Adjust_Alignment ;
        end ;
        In_RequestAlign := False ;
    end ;
end ; // T2D_Menu.RequestAlign


procedure T2D_Menu.SetBounds( ALeft, ATop, AWidth, AHeight : Integer ) ;

var Loop, Old_Width, Old_Height : integer ;

begin
    Old_Width := Width ;
    Old_Height := Height ;
    inherited SetBounds( ALeft, ATop, AWidth, AHeight ) ;

    if( not In_SetBounds ) then
    begin
        In_SetBounds := True ;
        if( Application.MainForm <> nil ) then
        begin
            if( MDI and Top_Level_Menu and ( Application.MainForm.FormStyle = fsMDIForm ) ) then
            begin
                if( ( AWidth <> Old_Width ) or ( AHeight <> Old_Height ) ) then
                begin
                    Adjust_MDI_Buttons ;
                end ;
            end ;
        end ;

        if( Align <> Last_Align ) then
        begin
            Adjust_Alignment ;
        end ;

        if( Old_Width <> Width ) then
        begin
            if( Count = 0 ) then
            begin
                for Loop := 0 to ControlCount - 1 do
                begin
                    Controls[ Loop ].Width := ClientWidth ;
                end ;
            end else
            begin
                for Loop := 0 to Count - 1 do
                begin
                    Items[ Loop ].Width := ClientWidth ;
                end ;
            end ;
        end ;
        In_SetBounds := False ;
    end ;
end ; // T2D_Menu.SetBounds


procedure T2D_Menu.SetParent( AParent : TWinControl ) ;

var Style, ExStyle : cardinal ;

begin
    inherited SetParent( AParent ) ;

    if( not ( ( In_Destructor > 0 ) or ( _Items = nil ) ) ) then // Destructor sets parent to nil - don't recreate handle or we blow up
    begin
        if( HandleAllocated ) then
        begin
            if( Parent_Menu = nil ) then
            begin
                DestroyWindowHandle ; // Must do this for top-level window or VCL gets funky 
            end else
            begin
                if( Parent = nil ) then // Not the main menu or an embedded child
                begin
                    Style := WS_POPUPWINDOW and ( not ( WS_BORDER or WS_SYSMENU ) ) ;
                    ExStyle := WS_EX_TOOLWINDOW or WS_EX_TOPMOST ;
                end else
                begin
                    Style := 1174405120 {and ( not WS_CLIPCHILDREN )} ;
                    ExStyle := 65536 ;
                end ;
                SetWindowLong( Handle, GWL_STYLE, Style ) ;
                SetWindowLong( Handle, GWL_EXSTYLE, ExStyle ) ;
                if( ( Parent <> nil ) and ( Parent.HandleAllocated ) ) then
                begin
                    Windows.SetParent( Handle, Parent.Handle ) ;
                end ;
            end ;
        end else
        begin
            Handle ; // Recreate the handle
        end ;
    end ;
end ; // T2D_Menu.SetParent


procedure T2D_Menu.CMVisibleChanged( var Msg : TMessage ) ;

begin
    inherited ;

    Adjust_Size ;
    if( Parent_Menu <> nil ) then
    begin
        Parent_Menu.Adjust_Size ;
        Parent_Menu.Popup_Index( Parent_Menu.Last_Popup_Index ) ;
    end ;
end ;



// MDI message handlers...

procedure T2D_Menu.Handle_MDICreate( Message : TMsg ) ;

begin
    if( MDI and Top_Level_Menu and ( Application.MainForm.FormStyle = fsMDIForm ) ) then
    begin
        Adjust_MDI_Buttons ;
    end ;
end ;


procedure T2D_Menu.Handle_Destroy( Message : TMsg ) ;

begin
    if( MDI and Top_Level_Menu and ( Application.MainForm.FormStyle = fsMDIForm ) ) then
    begin
        Adjust_MDI_Buttons ;
    end ;
end ;


procedure T2D_Menu.Handle_MDIMaximize( Message : TMsg ) ;

begin
    if( MDI and Top_Level_Menu and ( Application.MainForm.FormStyle = fsMDIForm ) ) then
    begin
        Adjust_MDI_Buttons ;
    end ;
end ;


procedure T2D_Menu.Handle_Size( Message : TMsg ) ;

begin
    if( MDI and Top_Level_Menu and ( Application.MainForm.FormStyle = fsMDIForm ) ) then
    begin
        if( IsChild( Application.MainForm.Handle, Message.hwnd ) ) then
        begin
            Adjust_MDI_Buttons ;
        end ;
    end ;
end ;


// API...

procedure T2D_Menu.Activate( Key : word ; Up : boolean ) ;

begin
    if( not ( Enabled or Visible ) ) then
    begin
        exit ;
    end ;
    if( Up ) then // Mouse up...
    begin
        // If we have no children, do the click
        if( Count = 0 ) then
        begin
            if( Key = VK_LBUTTON ) then
            begin
                if( assigned( OnClick ) ) then
                begin
                    if( RadioItem ) then
                    begin
                        Checked := True ;
                    end ;
                    Deactivate_Menus ;
                end ;
                if( Enabled and Check_Decorations ) then
                begin
                    exit ;
                end ;
                Do_Click ;
            end ;
            if( ( Key = VK_RBUTTON ) and assigned( On_Property_Click ) ) then
            begin
                On_Property_Click( self, Menus_Active ) ;
                if( not Menus_Active ) then
                begin
                    Deactivate_Menus ;
                end ;
            end ;
        end else // We have children menus
        if( Highlighted_Menu_Index >= 0 ) then // We have a sub-menu highlighted
        begin
            if( Items[ Highlighted_Menu_Index ].Count = 0 ) then // Highlighted menu has no children
            begin
                Items[ Highlighted_Menu_Index ].Activate( Key, Up ) ;
                Mouse_Move ;
            end else
            begin
                Popup_Index( Highlighted_Menu_Index ) ;
                if( ( Align = alLeft ) or ( Align = alRight ) ) then
                begin
//                    _Selected_Menu_Index := Highlighted_Menu_Index ;
                    Menus_Active := False ;
//                    Invalidate ;
                end else
                begin
                    Menus_Active := True ;
                end ;
            end ;
            if( Enabled ) then
            begin
                Check_Decorations ;
            end ;
        end ;
    end else // Mouse down
    begin
        if(
            Menus_Active // If menus are active
            and
            Top_Level_Menu // If we are the top-level menu
            and
            ( Align <> alLeft ) and ( Align <> alRight ) and ( Align <> alNone ) // Not left/right or popup
          ) then
        begin
            Deactivate_Menus ;
            exit ;
        end ;

        // On Mouse-down, activate menus
        if( Highlighted_Menu_Index < Count ) then
        begin
            Menus_Active := True ;
            Invalidate_Index( Highlighted_Menu_Index ) ;
            if(
                ( Highlighted_Menu_Index >= 0 )
                and
                (
                  ( Items[ Highlighted_Menu_Index ].Count > 0 )
                  or
                  ( Items[ Highlighted_Menu_Index ].ControlCount > 0 )
                )
              ) then
            begin
                Popup_Index( Highlighted_Menu_Index ) ;
            end ;
        end ;
    end ;
end ; // T2D_Menu.Activate


procedure T2D_Menu.Add( M : T2D_Menu ) ;

begin
    // Add the submenu...
    _Items.Add( M ) ;
    Positions.Insert( 0 ) ;
    M._Hide ;
    M.Align := alNone ;
    M.Last_Align := M.Align ;
    M.Parent_Menu := self ;
    Adjust_Size ;
end ;


procedure T2D_Menu.Adjust_Size ;

var Dummy : integer ;

begin
    if( ( Align <> alTop ) and ( Align <> alBottom ) ) then
    begin
        Dummy := Text_Height ;
        if( Height <> Dummy ) then
        begin
            Height := Dummy ;
        end ;
        Dummy := Text_Width ;
        if( Width <> Dummy ) then
        begin
            Width := Dummy ;
        end ;
    end ;
    if( Parent_Menu <> nil ) then
    begin
        Parent_Menu.Adjust_Size ;
    end ;
end ;


function T2D_Menu.Count : integer ;

begin
    if( _Items = nil ) then // We get here during construction so this is true
    begin
        Result := 0 ;
    end else
    begin
        Result := _Items.Count ;
    end ;
end ;


procedure T2D_Menu.DefaultHandler( var Message ) ;

var Rect : TRect ;
    TM : TTextMetric ;

begin
    inherited DefaultHandler( Message ) ;

{$WARNINGS OFF}
    if( TMessage( Message ).Msg = XWM_Get_Hint_Exclusion_Rect ) then
{$WARNINGS ON}
    begin
        if( ( Align <> alTop ) and ( Align <> alBottom ) ) then // Vertical layout
        begin
            if( Highlighted_Menu_Index = -1 ) then
            begin
                exit ;
            end ;

            TMessage( Message ).Result := 1 ; // We handled this
            Rect.Left := 0 ;
            Rect.Right := Width ;
            Rect.Top := Positions.At( Highlighted_Menu_Index ) ;
            GetTextMetrics( Canvas.Handle, TM ) ;
            Rect.Bottom := GetSystemMetrics( SM_CXMENUCHECK ) ;
            if( Rect.Bottom < TM.tmHeight + 4 ) then
            begin
                Rect.Bottom := TM.tmHeight + 4 ;
            end ;
            if( Min_Image_Width > 0 ) then
            begin
                Rect.Bottom := Min_Image_Width ;
            end ;
            Rect.Bottom := Rect.Bottom + BorderWidth ;
            if( Ctl3D ) then
            begin
                Rect.Bottom := Rect.Bottom + 1 ;
            end ;
            Rect.Bottom := Rect.Top + Rect.Bottom ;
            move( Rect, pointer( TMessage( Message ).LParam )^, sizeof( Rect ) ) ;
        end ;
    end ; // if( TMessage( Message ).Msg = XWM_Get_Hint_Exclusion_Rect )
end ; // T2D_Menu.DefaultHandler


procedure T2D_Menu.Paint ;

var B : TBitmap ;
    Handle, Saved_Font, Saved_Pen : THandle ;
    Loop : integer ;
    New_X_Offset : integer ;
    Half_Margin, Margin : integer ;
    Rect, Rects : TRect ;
    Saved_Color : TColor ;
    Special_Width : integer ;
    Temp : integer ;
    This_Bitmap : TBitmap ;
    This_Caption : string ;
    This_Menu : T2D_Menu ;
    TM : TTextMetric ;
    X, Y, Offset : integer ;
    Text_Y_Offset : integer ;
    X_Offset, Y_Offset : integer ;
    W : integer ;

begin
    if( not _Visible ) then // We are currently not visible
    begin
        exit ;
    end ;

    if( ( ControlCount > 0 ) and ( Count = 0 ) ) then
    begin
        inherited Paint ;
        for Loop := 0 to ControlCount - 1 do
        begin
            Controls[ Loop ].Invalidate ;
        end ;
        exit ;
    end ;

    // Setup...
    Handle := Canvas.Handle ;
    GetTextMetrics( Handle, TM ) ;
    Margin := TM.tmAveCharWidth ;
    Positions.DeleteAll ;

    // Do background...
    if( Gradients ) then
    begin
        Bigradient_Fill( Canvas, ClientRect, Gradient_Top, Gradient_Intermediate, Gradient_Bottom ) ;
    end else
    begin
        Canvas.Brush.Color := Color ;
        Canvas.FillRect( ClientRect ) ;
    end ;

    // Do text...
    SetBkMode( Handle, TRANSPARENT ) ;
    SetTextColor( Handle, GetSysColor( COLOR_MENUTEXT ) ) ;
    SetTextAlign( Handle, TA_TOP or TA_LEFT or TA_NOUPDATECP ) ;
    Saved_Font := SelectObject( Handle, Font.Handle ) ;
    Saved_Pen := SelectObject( Handle, GetStockObject( WHITE_PEN ) ) ;
    if( ( Align = alTop ) or ( Align = alBottom ) ) then // Horizontal layout
    begin
        Half_Margin := Margin div 2 ;
        Y := ( Height - Canvas.TextHeight( 'W' ) ) div 2 ;
        X := Margin ; // Note: no images shown on top/bottom alignment menus
        for Loop := 0 to Count - 1 do
        begin
            Positions.Insert( X ) ; // Build list of positions
            This_Menu := Items[ Loop ] ;
            if( This_Menu.Visible ) then
            begin
                This_Caption := This_Menu.Caption ;
                // Note: We don't show accelerators on top/bottom aligned menus
                W := Canvas.TextWidth( This_Caption ) ;
                if( This_Caption = '-' ) then
                begin
                    Canvas.Brush.Color := clDkGray ;
                    Rect.Top := Y ;
                    Rect.Bottom := ClientHeight - Y ;
                    Rect.Left := X + W div 2 ;
                    Rect.Right := Rect.Left + 1 ;
                    Canvas.FillRect( Rect ) ;
                    Canvas.Brush.Color := clWhite ;
                    Rect.Left := Rect.Left + 1 ;
                    Rect.Right := Rect.Right + 1 ;
                    Canvas.FillRect( Rect ) ;
                end else
                begin
                    if( Dialog_Focused or ( not This_Menu.Enabled ) ) then
                    begin
                        SetTextColor( Handle, GetSysColor( COLOR_GRAYTEXT ) ) ;
                    end ;
                    if( This_Menu.ParentFont ) then
                    begin
                        New_X_Offset :=
                            Process_HTML_Text( Handle, Font, This_Caption, True, 0, 0, 32767, Temp, True ) ;
                    end else
                    begin
                        New_X_Offset :=
                            Write_Text( Handle, This_Menu.Font, 0, 0, Width, This_Caption ) ;
                    end ;
                    This_Menu.Process_Decorations( Canvas, New_X_Offset, -1, -1, Y * 2 ) ;
                    SetTextColor( Handle, GetSysColor( COLOR_MENUTEXT ) ) ;
                    if(
                        ( Highlighted_Menu_Index = Loop )
                        and
                        (
                          ( Highlight_Style = THS_Normal )
                          or
                          ( Highlight_Style = THS_Box )
                        )
                      ) then
                    begin
                        // Do border...
                        if( Menus_Active ) then
                        begin
                            SelectObject( Handle, GetStockObject( BLACK_PEN ) ) ;
                        end else
                        begin
                            SelectObject( Handle, GetStockObject( WHITE_PEN ) ) ;
                        end ;
                        Canvas.MoveTo( X - Half_Margin, ClientHeight - 3 ) ;
                        Canvas.LineTo( X - Half_Margin, 2 ) ;
                        Canvas.LineTo( X + W + Half_Margin, 2 ) ;

                        if( Menus_Active ) then
                        begin
                            SelectObject( Handle, GetStockObject( WHITE_PEN ) ) ;
                        end else
                        begin
                            SelectObject( Handle, GetStockObject( BLACK_PEN ) ) ;
                        end ;
                        Canvas.LineTo( X + W + Half_Margin, ClientHeight - 3 ) ;
                        Canvas.LineTo( X - Half_Margin, ClientHeight - 3 ) ;
                    end ;
                end ;
                X := X + W + Margin ;
            end ; // if( This_Menu.Visible )
        end ; // for Loop := 0 to Count - 1
        Positions.Insert( X ) ; // Faked end position
    end else // Vertical layout
    begin
        // Setup for painting...
        X := TM.tmHeight + 4 ; // Height of each menu item
        Special_Width := GetSystemMetrics( SM_CXMENUCHECK ) ;
        if( Special_Width < X ) then
        begin
            Special_Width := X ;
        end ;
        if( Min_Image_Width > 0 ) then
        begin
            Special_Width := Min_Image_Width ;
        end ;
        X_Offset := BorderWidth ;
        if( Ctl3D ) then
        begin
            X_Offset := X_Offset + 1 ;
        end ;
        Y_Offset := X_Offset ;
        X_Offset := X_Offset + Special_Width ;
        Offset := ( ( X - Canvas.TextHeight( 'W' ) ) div 2 ) - 1 ; // Where text starts, relative to rectangle
        Y := Y_Offset ;

        for Loop := 0 to Count - 1 do
        begin
            Positions.Insert( Y ) ;
            This_Menu := Items[ Loop ] ;
            if( This_Menu.Visible ) then
            begin
                // Setup for this loop...
                X := Item_Height( Loop ) ;
                if( X = 0 ) then
                begin
                    continue ;
                end ;
                This_Caption := This_Menu.Caption ;
                if( This_Menu.Shortcut <> 0 ) then
                begin
                    This_Caption := This_Caption + #9 + ShortCutToText( This_Menu.Shortcut ) ;
                end ;

                // Do background...
                Rect.Left := BorderWidth ;
                if( ( This_Menu.Button <> nil ) and This_Menu.Button.Visible ) then
                begin
                    Rect.Left :=    // Why are we adding 2 here?
                        Rect.Left + 2 + This_Menu.Button.Width div This_Menu.Button.NumGlyphs ;
                end ;
                Rect.Right := ClientWidth ;
                Rect.Top := Y ;
                Rect.Bottom := X + Y ;
                SetBkMode( Handle, TRANSPARENT ) ; // Something resets this mode - so we do it again here
                if(
                    ( Highlighted_Menu_Index = Loop )
                    and
                    ( This_Caption <> '-' )
                    and
                    (
                      ( Highlight_Style = THS_Normal )
                      or
                      ( Highlight_Style = THS_Color )
                    )
                  ) then
                begin
                    Canvas.Brush.Color := clHighlight ;
                    Canvas.FillRect( Rect ) ;
                end ; // if

                if( This_Menu.Gradients ) then
                begin
                    Rect.Left := Rect.Left - 2 ; // To compensate for adding 2 above
                    Bigradient_Fill( Canvas, Rect, This_Menu.Gradient_Top,
                        This_Menu.Gradient_Intermediate,
                        This_Menu.Gradient_Bottom ) ;
                    Rect.Left := Rect.Left + 2 ; // Restore rect
                end ;

                // Show child...
                if( This_Caption = '-' ) then
                begin
                    Canvas.Brush.Color := clDkGray ;
                    Rect.Top := Y + X div 4 ;
                    Rect.Bottom := Rect.Top + 1 ;
                    Rect.Left := 4 ;
                    Rect.Right := Width - 4 ;
                    Canvas.FillRect( Rect ) ;
                    SelectObject( Handle, GetStockObject( WHITE_PEN ) ) ;
                    Canvas.MoveTo( 4, Y + X div 4 + 1 ) ;
                    Canvas.LineTo( Width - 4, Y + X div 4 + 1 ) ;
                end else
                begin
                    if( This_Menu.ParentFont ) then
                    begin
                        Saved_Color := Font.Color ;
                        if( not This_Menu.Enabled ) then
                        begin
                            Font.Color := GetSysColor( COLOR_GRAYTEXT ) ;
                        end ;
                        if( This_Menu.ParentFont ) then
                        begin
                            Process_HTML_Text( Handle, Font, This_Caption, False, X_Offset, Y + Offset, 32767, Temp, True ) ;
                            Text_Y_Offset := ( ( X - Temp ) div 2 ) - 1 ;
                            New_X_Offset :=
                                Process_HTML_Text( Handle, Font, This_Caption, True, X_Offset, Y + Offset, 32767, Temp, True ) + X_Offset ;
                        end else
                        begin
                            New_X_Offset :=
                                Write_Text( Handle, Font, X_Offset, Y + Offset, Width, This_Caption ) ;
                        end ;
                        Font.Color := Saved_Color ;
                    end else
                    begin
                        Saved_Color := This_Menu.Font.Color ;
                        if( not This_Menu.Enabled ) then
                        begin
                            This_Menu.Font.Color := GetSysColor( COLOR_GRAYTEXT ) ;
                        end else
                        if(
                            ( Highlighted_Menu_Index = Loop )
                            and
                            (
                              ( Highlight_Style = THS_Normal )
                              or
                              ( Highlight_Style = THS_Color )
                            )
                          ) then
                        begin
                            This_Menu.Font.Color := GetSysColor( COLOR_HIGHLIGHTTEXT ) ;
                        end ;
                        if( Translate_HTML ) then
                        begin
                            Process_HTML_Text( Handle, This_Menu.Font, This_Caption, False, X_Offset, Y, 32767, Temp, True ) ;
                            Text_Y_Offset := ( ( X - Temp ) div 2 ) - 1 ;
                            New_X_Offset :=
                                Process_HTML_Text( Handle, This_Menu.Font, This_Caption, True, X_Offset, Y + Text_Y_Offset, 32767, Temp, True ) + X_Offset ;
                        end else
                        begin
                            New_X_Offset :=
                                Write_Text( Handle, This_Menu.Font, X_Offset, Y + Offset, Width, This_Caption ) ;
                        end ;
                        This_Menu.Font.Color := Saved_Color ;
                    end ; // if( This_Menu.ParentFont )
                    This_Menu.Process_Decorations( Canvas, New_X_Offset, Y + Offset, -1, X ) ;
                    if(
                        ( Align <> alLeft )
                        and
                        ( Align <> alRight )
                        and
                        ( This_Menu.Count > 0 )
                      ) then
                    begin
                        Rect.Left := 0 ;
                        Rect.Top := 0 ;
                        Rect.Right := GetSystemMetrics( SM_CXMENUCHECK ) ; ;
                        Rect.Bottom := Rect.Right ;
                        Arrow_Bitmap.Height := Rect.Right ;
                        Arrow_Bitmap.Width := Arrow_Bitmap.Height ;
                        DrawFrameControl( Arrow_Bitmap.Canvas.Handle, Rect,
                            DFC_MENU, DFCS_MENUARROW ) ;

                        Rect.Top := Y + 1 ;
                        Rect.Bottom := X + Y ;
                        Rect.Right := Width ;
                        Rect.Left := Width - Arrow_Bitmap.Width ;
                        if(
                            ( Highlighted_Menu_Index = Loop )
                            and
                            (
                              ( Highlight_Style = THS_Normal )
                              or
                              ( Highlight_Style = THS_Color )
                            )
                          ) then
                        begin
                            BitBlt( Handle, Rect.Left, Rect.Top, Arrow_Bitmap.Width,
                                Arrow_Bitmap.Width, Arrow_Bitmap.Canvas.Handle,
                                0, 0, MERGEPAINT ) ;
                        end else
                        begin
                            BitBlt( Handle, Rect.Left, Rect.Top, Arrow_Bitmap.Width,
                                Arrow_Bitmap.Width, Arrow_Bitmap.Canvas.Handle,
                                0, 0, SRCAND ) ;
                        end ;
                    end ;
                    if( This_Menu.Button <> nil ) then
                    begin
                        if( This_Menu.Button.Visible ) then
                        begin
                            Rect.Left := BorderWidth + 1 ;
                            Rect.Top := Y + ( X div 2 - This_Menu.Button.Height div 2 ) ;
                            Rect.Right := Rect.Left + This_Menu.Button.Width div This_Menu.Button.NumGlyphs ;
                            Rect.Bottom := Rect.Top + This_Menu.Button.Height ;
                            Rects := This_Menu.Button.ClientRect ;
                            if( This_Menu.Button.NumGlyphs > 1 ) then
                            begin
//ROW
                                Rects.Right := Rects.Right div This_Menu.Button.NumGlyphs ;
                                if( not This_Menu.Enabled ) then
                                begin
                                    OffsetRect(Rects, Rects.Right, 0) ; // Point to disabled glyph
                                end ;
(*
                                if( not This_Menu.Enabled ) then
                                begin
                                    Rects.Left := Rects.Right + 1 ; // Point to disabled glyph
                                end else
                                begin
                                    Rects.Right := Rects.Right div This_Menu.Button.NumGlyphs ;
                                end ;
*)
//ROW-END
                            end ;
                            try
                                B := This_Menu.Button.Glyph ;
                                try
                                    Canvas.CopyRect( Rect, B.Canvas, Rects ) ;
                                finally
                                    B.Free ;
                                end ;
//ROW
(*
                                BitBlt( Handle, Rect.Left, Rect.Top,
                                    This_Menu.Button.Width div This_Menu.Button.NumGlyphs,
                                    This_Menu.Button.Width div This_Menu.Button.NumGlyphs,
                                    This_Menu.Button.Glyph.Canvas.Handle,
                                    0, 0, SRCCOPY ) ;
*)
//ROW-END
                            except // We can get an exception if low on resources
                            end ;
                            if( Highlighted_Menu_Index = Loop ) then
                            begin // Highlight button with box
                                Canvas.Pen.Color := clBtnHighlight ;
                                Canvas.MoveTo( Rect.Left, Rect.Bottom ) ;
                                Canvas.LineTo( Rect.Left, Rect.Top ) ;
                                Canvas.LineTo( Rect.Right - 1, Rect.Top ) ;
                                Canvas.Pen.Color := clBtnShadow ;
                                Canvas.LineTo( Rect.Right - 1, Rect.Bottom - 1 ) ;
                                Canvas.LineTo( Rect.Left, Rect.Bottom - 1 ) ;
                            end ;
                        end ;
                    end else
                    if( This_Menu.Image <> nil ) then
                    begin
                        if( This_Menu.Image.Visible ) then
                        begin
                            Rect.Left := BorderWidth + 1 ;
                            Rect.Top := Y + ( X div 2 - This_Menu.Image.Height div 2 ) ;
                            Rect.Right := Rect.Left + This_Menu.Image.Width ;
                            Rect.Bottom := Rect.Top + This_Menu.Image.Height ;
                            Canvas.CopyRect( Rect, This_Menu.Image.Canvas, This_Menu.Image.ClientRect ) ;
                            if( Highlighted_Menu_Index = Loop ) then
                            begin // Highlight image with box
                                Canvas.Pen.Color := clBtnHighlight ;
                                Canvas.MoveTo( Rect.Left, Rect.Bottom ) ;
                                Canvas.LineTo( Rect.Left, Rect.Top ) ;
                                Canvas.LineTo( Rect.Right - 1, Rect.Top ) ;
                                Canvas.Pen.Color := clBtnShadow ;
                                Canvas.LineTo( Rect.Right - 1, Rect.Bottom - 1 ) ;
                                Canvas.LineTo( Rect.Left, Rect.Bottom - 1 ) ;
                            end ;
                        end ;
                    end else
                    if( This_Menu.Checked ) then
                    begin
                        Rect.Left := 0 ;
                        Rect.Top := 0 ;
                        Rect.Right := GetSystemMetrics( SM_CXMENUCHECK ) ;
                        Rect.Bottom := Rect.Right ;
                        if( This_Menu.RadioItem ) then
                        begin
                            Radio_Button_Bitmap.Height := Rect.Right ;
                            Radio_Button_Bitmap.Width := Rect.Right ;
                            DrawFrameControl( Radio_Button_Bitmap.Canvas.Handle,
                                Rect, DFC_MENU, DFCS_MENUBULLET ) ;
                            This_Bitmap := Radio_Button_Bitmap ;
                        end else
                        begin
                            Check_Bitmap.Height := Rect.Right ;
                            Check_Bitmap.Width := Rect.Right  ;
                            DrawFrameControl( Check_Bitmap.Canvas.Handle, Rect,
                                DFC_MENU, DFCS_MENUCHECK ) ;
                            This_Bitmap := Check_Bitmap ;
                        end ;
                        if(
                            ( Highlighted_Menu_Index = Loop )
                            and
                            (
                              ( Highlight_Style = THS_Normal )
                              or
                              ( Highlight_Style = THS_Color )
                            )
                          ) then
                        begin
                            BitBlt( Handle, BorderWidth + 1, Y + 1, Rect.Right,
                                Rect.Right, This_Bitmap.Canvas.Handle, 0, 0,
                                MERGEPAINT ) ;
                        end else
                        begin
                            BitBlt( Handle, BorderWidth + 1, Y + 1, Rect.Right,
                                Rect.Right, This_Bitmap.Canvas.Handle, 0, 0,
                                SRCAND ) ;
                        end ;
                    end ;
                end ; // if( This_Caption = '-' )

                // Do border...
                if(
                    ( Align = alLeft )
                    or
                    ( Align = alRight )
                    or
                    ( Style = alLeft )
                    or
                    ( Style = alRight )
                  ) then
                begin
                    SelectObject( Handle, GetStockObject( WHITE_PEN ) ) ;
                    Canvas.MoveTo( BorderWidth, Y + X ) ;
                    Canvas.LineTo( BorderWidth, Y ) ;
                    Canvas.LineTo( Width - 1 - BorderWidth, Y ) ;

                    SelectObject( Handle, GetStockObject( BLACK_PEN ) ) ;
                    Canvas.LineTo( Width - 1 - BorderWidth, Y + X - 1 ) ;
                    Canvas.LineTo( BorderWidth, Y + X - 1 ) ;
                end ;

                Y := Y + X ;

                if(
                    ( This_Caption <> '-' )
                    and
                    ( ( Align = alLeft ) or ( Align = alRight ) )
                    and
                    ( _Selected_Menu_Index = Loop )
                    and
                    (
                      ( This_Menu.Count > 0 )
                      or
                      ( This_Menu.ControlCount > 0 )
                    )
                  ) then
                begin
                    Y := Y + This_Menu.Height ;
//                    This_Menu.Paint ;
                end ;
            end ; // if( This_Menu.Visible ) then
        end ; // for Loop := 0 to Count - 1
        Positions.Insert( Y ) ; // Faked end position
    end ; // if (Align...

    // Do border...
    if(
        Ctl3D
        and
        (
          Top_Level_Menu
          or
          ( ( Parent_Menu <> nil ) and ( Parent_Menu.Align <> alLeft ) and ( Parent_Menu.Align <> alRight ) )
        )
      ) then
    begin
        if( BorderStyle = bsSingle ) then
        begin
            SelectObject( Handle, GetStockObject( BLACK_PEN ) ) ;
            Canvas.MoveTo( 0, Height ) ;
            Canvas.LineTo( 0, 0 ) ;
            Canvas.LineTo( Width - 1, 0 ) ;
            Canvas.LineTo( Width - 1, Height - 1 ) ;
            Canvas.LineTo( 0, Height - 1 ) ;
            Offset := 1 ;
        end else
        begin
            Offset := 0 ;
        end ;
        SelectObject( Handle, GetStockObject( WHITE_PEN ) ) ;
        Canvas.MoveTo( Offset, Height - Offset ) ;
        Canvas.LineTo( Offset, Offset ) ;
        Canvas.LineTo( Width - Offset - 1, Offset ) ;

        SelectObject( Handle, GetStockObject( BLACK_PEN ) ) ;
        Canvas.LineTo( Width - Offset - 1, Height - Offset - 1 ) ;
        Canvas.LineTo( Offset, Height - Offset - 1 ) ;
    end ;

    // Restore handles to DC...
    SelectObject( Handle, Saved_Font ) ;
    SelectObject( Handle, Saved_Pen ) ;
end ; // T2D_Menu.Paint


procedure T2D_Menu.Deactivate_Menus ;

begin
    // Change our state, get rid of popups, and make sure we look correct...
    if( Top_Level_Menu ) then
    begin
        Menus_Active := False ;
        Cancel_Popups( nil ) ;
        Invalidate_Index( Highlighted_Menu_Index ) ;
        Mouse_Move ;
    end else
    begin
        Parent_Menu.Deactivate_Menus ;
    end ;
    Menu_Key_Mode := False ;
end ;


function T2D_Menu.Debugger : PDebug_Interface ;

begin
{$IFDEF Debug}
    Debugger := T2D_Menu_Debugger.Create ;
    T2D_Menu_Debugger( Result ).Component := self ;
{$ELSE}
    Debugger := nil ;
{$ENDIF}
end ;


procedure T2D_Menu.Deserialize( const Image : string ) ;

var Dummy, I : integer ;
    S, S1 : string ;

begin
    S := Edit( Image, 4 or 8 or 16 or 32 or 128 or 256 ) ;
    Dummy := pos( '<#INDEX=', S ) ;
    if( Dummy > 0 ) then
    begin
        S1 := copy( S, Dummy + 8, length( S ) ) ;
        S := copy( S, 1, Dummy - 1 ) ;
        Dummy := pos( '>', S1 + '>' ) ;
        S := S + copy( S1, Dummy + 1, length( S1 ) ) ;
        S1 := copy( S1, 1, Dummy - 1 ) ;
        Val( S1, I, Dummy ) ;
        Popup_Index( I ) ;
    end ;
end ;


function T2D_Menu.Display_Height : integer ;

var Loop, X, Y : integer ;
    This_Menu : T2D_Menu ;
    TM : TTextMetric ;

begin
    if( ( Align = alLeft ) or ( Align = alRight ) ) then
    begin
        GetTextMetrics( Canvas.Handle, TM ) ;
        Y := BorderWidth ;
        if( Ctl3D ) then
        begin
            Y := Y + 1 ;
        end ;

        for Loop := 0 to Count - 1 do
        begin
            This_Menu := Items[ Loop ] ;
            if( This_Menu.Visible ) then
            begin
                // Setup for this loop...
                X := Item_Height( Loop ) ;
                if( X > 0 ) then
                begin
                    Y := Y + X ;
                    if(
                        ( This_Menu.Caption <> '-' )
                        and
                        ( _Selected_Menu_Index = Loop )
                        and
                        (
                          ( This_Menu.Count > 0 )
                          or
                          ( This_Menu.ControlCount > 0 )
                        )
                      ) then
                    begin
                        Y := Y + This_Menu.Height ;
                    end ;
                end ;
            end ; // if( This_Menu.Visible ) then
        end ; // for Loop := 0 to Count - 1
        Result := Y ;
    end else
    begin
        Result := Height ;
    end ;
end ; // T2D_Menu.Display_Height


procedure T2D_Menu.Do_Click ;

var Position : TRect ;
    SysMenu : HMENU ;

begin
    if( not Enabled ) then // Ignore clicks on disabled menu
    begin
        exit ;
    end ;

    // Handle system menu uniquely
    if( _System_Menu = self ) then
    begin
        SysMenu := GetSystemMenu( Application.MainForm.ActiveMDIChild.Handle, True ) ;
        DeleteMenu( SysMenu, SC_TASKLIST, MF_BYCOMMAND ) ;
        Position := RectClientToScreen( Parent_Menu.Item_Rectangle( self ) ) ;
        TrackPopupMenu( SysMenu, TPM_LEFTALIGN, Position.Left, Position.Bottom, 0, Handle, nil ) ;
        exit ;
    end ;

    // Handle click
    if( assigned( OnClick ) ) then
    begin
        OnClick( self ) ;
    end ;
end ; // T2D_Menu.Do_Click


procedure T2D_Menu._Hide ;

begin
    Top := Off_Screen - Height ; // Force it out of view even if VCL makes it visible
    if( HandleAllocated ) then
    begin
        ShowWindow( Handle, SW_HIDE ) ;
    end ;
    _Visible := False ;
    if( HandleAllocated and ( ControlCount = 0 ) ) then // Keep handle if we have children - otherwise we start losing data (like captions)
    begin
        DestroyWindowHandle ; // Save on system resources
    end ;
    if( Popup_Form <> nil ) then
    begin
        Popup_Form.Visible := False ;
    end ;
end ;


function T2D_Menu.Index_Of( Menu : T2D_Menu ) : integer ;

var Loop : integer ;

begin
    Result := -1 ;
    for Loop := 0 to _Items.Count - 1 do
    begin
        if( Menu = Items[ Loop ] ) then
        begin
            Result := Loop ;
            exit ;
        end ;
    end ;
end ;


function T2D_Menu.Index_Rectangle( Index : integer ) : TRect ;

var Margin : integer ;
    Rect : TRect ;
    TM : TTextMetric ;

begin
    if( ( Align = alTop ) or ( Align = alBottom ) ) then
    begin
        Rect.Top := 0 ;
        Rect.Bottom := ClientHeight ;
        GetTextMetrics( Canvas.Handle, TM ) ;
        Margin := TM.tmAveCharWidth ;
        Rect.Left := Positions.At( Index ) - Margin ;
        if( Index = Positions.Count - 1 ) then
        begin
            Rect.Right := ClientWidth ;
        end else
        begin
            Rect.Right := Positions.At( Index + 1 ) ;
        end ;
    end else
    begin
        Rect.Top := Positions.At( Index ) ;
        if( Index = Positions.Count - 1 ) then
        begin
            Rect.Bottom := ClientHeight ;
        end else
        begin
            Rect.Bottom := Positions.At( Index + 1 ) ;
        end ;
        Rect.Left := 0 ;
        Rect.Right := ClientWidth ;
    end ;
    Result := Rect ;
end ; // T2D_Menu.Index_Rectangle


procedure T2D_Menu.Invalidate ;

    procedure _Invalidate( C : TWinControl ) ;

    var Control : TControl ;
        Loop : integer ;

    begin
        for Loop := 0 to C.ControlCount - 1 do
        begin
            Control := C.Controls[ Loop ] ;
            Control.Invalidate ;
            if( Control is TWinControl ) then
            begin
                _Invalidate( TWinControl( Control ) ) ;
            end ;
        end ;
    end ;

begin
    inherited Invalidate ;

    _Invalidate( self ) ;
end ;


function T2D_Menu.Item_Rectangle( Index : T2D_Menu ) : TRect ;

begin
    Result := Index_Rectangle( _Items.Indexof( Index ) ) ;
end ;


function T2D_Menu.Popped_Up : boolean ;
// True if we are visible via a call to Popup

begin
    if( Popup_Form = nil ) then
    begin
        Result := False ; // Can't be popped up if no popup form exists
    end else
    if( not Popup_Form.Visible ) then
    begin
        Result := False ; // Can't be popped up if the popup form is invisible
    end else
    begin
        Result := IsChild( Popup_Form.Handle, Handle ) ; // Popped up if a child of the popup form
    end ;
end ;


procedure T2D_Menu.Popup( X, Y : integer ) ;

var Loop : integer ;
    This_Menu : T2D_Menu ;

begin
    if( Align <> alNone ) then
    begin
        Align := alNone ;
    end ;
    BorderStyle := bsSingle ;
    Mouse.Capture := 0 ;
    SendMessage( GetFocus, WM_CANCELMODE, 0, 0 ) ;
    Cancel_All_Popups ;
    if( Parent = nil ) then // Make sure menu is on a form, otherwise we can't focus it
    begin
        Popup_Form := TPopup_Form.Create( Application ) ;
        Popup_Form.Caption := '' ;
        Popup_Form.BorderIcons := [] ;
        Popup_Form.BorderStyle := bsNone ;
        Popup_Form.KeyPreview := True ;
        TPopup_Form( Popup_Form ).On_Message := CB_Popup_Message ;
        Parent := Popup_Form ;
    end else
    begin
        Top := Y ;
        Left := X ;
    end ;
    if( Popup_Form <> nil ) then
    begin
        Popup_Form.Top := Y ;
        Popup_Form.Left := X ;
        Popup_Form.Width := Width ;
        Popup_Form.Height := Height ;
        Popup_Form.Visible := True ;
        SetFocus ;
    end ;
    for Loop := 0 to Count - 1 do
    begin
        This_Menu := T2D_Menu( Items[ Loop ] ) ;
        if( This_Menu.Button <> nil ) then
        begin
            This_Menu.Button.Visible := True ;
        end ;
    end ;
    _Show ;
    Menus_Active := True ;
end ; // T2D_Menu.Popup


procedure T2D_Menu.Popup_Index( Index : integer ) ;

var Loop : integer ;
    Margin : integer ;
    Popup_Position, Position : TRect ;
    This_Menu : T2D_Menu ;
    TM : TTextMetric ;

begin
    // Setup...
    if( Index < 0 ) then // Invalid index
    begin
        exit ;
    end ;
    if( Count = 0 ) then
    begin
        exit ;
    end ;
    if( Index >= Count  ) then
    begin
        exit ;
    end ;
    This_Menu := T2D_Menu( Items[ Index ] ) ;
    if( ( This_Menu.Count = 0 ) and ( This_Menu.ControlCount = 0 ) ) then
    begin
        exit ;
    end ;
    if( not ( This_Menu.Enabled and This_Menu.Visible ) ) then // Do not pop up disabled or invisible menus
    begin
        exit ;
    end ;
    _Last_Popup_Index := Index ;
    if( assigned( This_Menu.On_Popup ) ) then
    begin
        This_Menu.On_Popup( This_Menu ) ;
    end ;

    // Determine top-left corner position of popup menu
    if( ( Align = alLeft ) or ( Align = alRight ) ) then
    begin
        Cancel_All_Popups ;
        for Loop := 0 to Count - 1 do
        begin
            if( Loop <> Index ) then
            begin
                T2D_Menu( Items[ Loop ] )._Hide ;
            end ;
        end ;
        _Selected_Menu_Index := -1 ;
        Paint ; // Do this to recalculate the positions of the items
        Position := Index_Rectangle( Index ) ;
        Popup_Position.Left := Position.Right - This_Menu.Width - 1 ;
        Popup_Position.Top := Position.Bottom ;
        This_Menu.Parent := self ;
        _Selected_Menu_Index := Index ; //Highlighted_Menu_Index ;
        Menus_Active := False ;
    end else
    begin
        Position := RectClientToScreen( Index_Rectangle( Index ) ) ;

        if( ( Align = alTop ) or ( Align = alBottom ) ) then
        begin
            GetTextMetrics( Canvas.Handle, TM ) ;
            Margin := TM.tmAveCharWidth div 2 ;
            Popup_Position.Left := Position.Left + Margin + 1 ;
            Popup_Position.Top := Position.Bottom - 1 ;
        end else
        begin
            Popup_Position.Left := Position.Right - 4 ;
            Popup_Position.Top := Position.Top - BorderWidth - 2 ;
        end ;
        This_Menu.Parent := nil ;
    end ;

    // Now determine lower-right corner of popup...
    Popup_Position.Bottom := This_Menu.Text_Height + Popup_Position.Top ;
    Popup_Position.Right := This_Menu.Width + Popup_Position.Left ;

    if( ( Align <> alLeft ) and ( Align <> alRight ) ) then
    begin
        // See if it fits on the screen and adjust, if necessary...
        if( Popup_Position.Right > Screen.DesktopWidth ) then
        begin
            Popup_Position.Left := Position.Left - This_Menu.Width ;
            if( Popup_Position.Left < 0 ) then
            begin
                Popup_Position.Left := 0 ;
            end ;
        end ;
        Popup_Position.Right := Popup_Position.Left + This_Menu.Width ;
        if( Popup_Position.Bottom > Screen.DesktopHeight ) then
        begin
            Popup_Position.Top := Position.Top - This_Menu.Height ;
            if( Popup_Position.Top < 0 ) then
            begin
                Popup_Position.Top := 0 ;
            end ;
        end ;
        Popup_Position.Bottom := Popup_Position.Top + This_Menu.Text_Height ;
    end ; // if( ( Align <> alLeft ) and ( Align <> alRight ) )

    // Position and show the menu...
    if( This_Menu.ClientHeight <> This_Menu.Height ) then
    begin
        This_Menu.ClientHeight := This_Menu.Height ;
    end ;
    if( This_Menu.ClientWidth <> This_Menu.Width ) then
    begin
        This_Menu.ClientWidth := This_Menu.Width ;
    end ;
    This_Menu.SetBounds( Popup_Position.Left, Popup_Position.Top,
        Popup_Position.Right - Popup_Position.Left,
        Popup_Position.Bottom - Popup_Position.Top ) ;
    This_Menu._Show ;
    Invalidate ;

    if( assigned( This_Menu.On_Popped_Up ) ) then
    begin
        This_Menu.On_Popped_Up( This_Menu ) ;
    end ;
end ; // T2D_Menu.Popup_Index


function T2D_Menu.Process_Accelerator( Key : integer ; Shift : TShiftState ) : boolean ;

var Loop : integer ;

begin
    Result := False ; // Assume no matches
    if( Enabled and Visible ) then
    begin
        if( _Shortcut <> 0 ) then
        begin
            if( _Shortcut = Menus.Shortcut( Key, Shift ) ) then
            begin
                Do_Click ;
                Result := True ;
                exit ;
            end ;
        end ;

        // Search children for matching accelerator
        for Loop := 0 to Count - 1 do
        begin
            Result := Items[ Loop ].Process_Accelerator( Key, Shift ) ;
            if( Result ) then
            begin
                exit ;
            end ;
        end ;
    end ;
end ;


procedure T2D_Menu.Process_Char( Key, Repeats : integer ) ;


    procedure Highlight_Next( Pop : boolean ) ;

    var Last, Original : integer ;

    begin
        Original := Highlighted_Menu_Index ;
        repeat
            { This loop advances up until we get to a visible, non-line menu,
              wrapping if necessary }
            Last := Highlighted_Menu_Index ;
            repeat
                { This loop takes care of trying to highlight a menu that cannot
                  be highlighted by skipping it, and wrapping if necessary }
                if( Last = Count - 1 ) then
                begin
                    Last := 0 ;
                end else
                begin
                    Last := Highlighted_Menu_Index + 1 ;
                end ;
                Highlight_Index( Last, Pop ) ;
                inc( Last ) ;
            until ( Highlighted_Menu_Index >= 0 )
        until (
                ( Original = Highlighted_Menu_Index )
                or
                (
                  ( Items[ Highlighted_Menu_Index ].Caption <> '-' )
                  and
                  ( Items[ Highlighted_Menu_Index ].Visible )
                )
              ) ;
        if( Pop ) then
        begin
            if( Items[ Highlighted_Menu_Index ].Count > 0 ) then
            begin
                Items[ Highlighted_Menu_Index ].Highlight_Index( 0, False ) ;
            end ;
        end ;
    end ; // T2D_Menu.Process_Char.Highlight_Next


    procedure Highlight_Previous( Pop : boolean ) ;

    var Last, Original : integer ;

    begin
        Original := Highlighted_Menu_Index ;
        if( Original = -1 ) then
        begin
            Original := 0 ;
        end ;
        repeat
            { This loop backs up until we get to a visible, non-line menu,
              wrapping if necessary }
            Last := Highlighted_Menu_Index ;
            repeat
                { This loop takes care of trying to highlight a menu that cannot
                  be highlighted by skipping it, and wrapping if necessary }
                if( Last <= 0 ) then
                begin
                    Last := Count - 1 ;
                end else
                begin
                    Last := Last - 1 ;
                end ;
                Highlight_Index( Last, Pop ) ;
                dec( Last ) ;
            until ( Highlighted_Menu_Index >= 0 ) ;
        until (
                ( Original = Highlighted_Menu_Index )
                or
                (
                  ( Items[ Highlighted_Menu_Index ].Caption <> '-' )
                  and
                  ( Items[ Highlighted_Menu_Index ].Visible )
                )
              ) ;
        if( Pop ) then
        begin
            if( Items[ Highlighted_Menu_Index ].Count > 0 ) then
            begin
                Items[ Highlighted_Menu_Index ].Highlight_Index( 0, False ) ;
            end ;
        end ;
    end ; // T2D_Menu.Process_Char.Highlight_Previous

var Dummy : integer ;
    Loop : integer ;
    S : string ;
    This_Menu : T2D_Menu ;

begin // T2D_Menu.Process_Char
    if(
        ( ( Align = alLeft ) or ( Align = alRight ) )
        and
        ( Popped_Menus.Count = 0 )
        and
        ( not In_Process_Key )
      ) then // For top-level vertical menus, switch meanings of arrow keys
    begin
        case Key of
            VK_UP : Key := VK_LEFT ;
            VK_LEFT : Key := VK_UP ;
            VK_DOWN : Key := VK_RIGHT ;
            VK_RIGHT : Key := VK_DOWN ;
        end ;
    end ;

    In_Process_Key := True ;
    if( Key = VK_UP ) then
    begin
        if( Popped_Menus.Count > 0 ) then
        begin
            This_Menu := T2D_Menu( Popped_Menus[ Popped_Menus.Count - 1 ] ) ;
            if( This_Menu = self ) then
            begin
                Highlight_Previous( False ) ;
            end else
            begin
                T2D_Menu( This_Menu ).Process_Char( Key, Repeats ) ;
            end ;
        end else
        begin
            Process_Char( VK_LEFT, Repeats ) ; // If no active menu, treat as left-arrow
        end ;
        In_Process_Key := False ;
        exit ;
    end else
    if( Key = VK_LEFT ) then
    begin
        while( Repeats > 0 ) do
        begin
            dec( Repeats ) ;
            if( Popped_Menus.Count > 1 ) then
            begin
                This_Menu := T2D_Menu( Popped_Menus[ Popped_Menus.Count - 1 ] ) ;
                This_Menu._Hide ;
                Popped_Menus.Count := Popped_Menus.Count - 1 ;
            end else
            begin
                Highlight_Previous( Popped_Menus.Count > 0 ) ;
            end ;
        end ;
        In_Process_Key := False ;
        exit ;
    end else
    if( Key = VK_RIGHT ) then
    begin
        while( Repeats > 0 ) do
        begin
            dec( Repeats ) ;
            if( Popped_Menus.Count > 0 ) then
            begin
                This_Menu := T2D_Menu( Popped_Menus[ Popped_Menus.Count - 1 ] ) ;
                if(
                    ( This_Menu.Highlighted_Menu_Index >= 0 )
                    and
                    ( This_Menu.Items[ This_Menu.Highlighted_Menu_Index ].Count > 0 )
                  ) then // If sitting on a menu with children, pop the child
                begin
                    // Popup the child menu and highlight first index...
                    This_Menu.Popup_Index( This_Menu.Highlighted_Menu_Index ) ;
                    T2D_Menu( Popped_Menus[ Popped_Menus.Count - 1 ] ).Highlight_Index( 0, False ) ;
                    continue ;
                end ;
            end ;
            Highlight_Next( Popped_Menus.Count > 0 ) ;
        end ;
        In_Process_Key := False ;
        exit ;
    end else
    if( Key = VK_DOWN ) then
    begin
        if( Popped_Menus.Count > 0 ) then
        begin
            This_Menu := T2D_Menu( Popped_Menus[ Popped_Menus.Count - 1 ] ) ;
            if( This_Menu = self ) then
            begin
                Highlight_Next( False ) ;
            end else
            begin
                T2D_Menu( This_Menu ).Process_Char( Key, Repeats ) ;
            end ;
        end else
        begin
            if( Items[ Highlighted_Menu_Index ].Count > 0 ) then
            begin
                // Popup the child menu (if no children, ignore key)...
                Menus_Active := True ;
                Invalidate_Index( Highlighted_Menu_Index ) ;
                Popup_Index( Highlighted_Menu_Index ) ;

                // Highlight the first item on the menu
                T2D_Menu( Popped_Menus[ Popped_Menus.Count - 1 ] ).Highlight_Index( 0, False ) ;

                if( Repeats > 1 ) then
                begin
                    Process_Char( Key, Repeats - 1 ) ;
                end ;
            end ;
        end ;
        In_Process_Key := False ;
        exit ;
    end ;

    // No more recursive calls with arrow keys after this point...
    In_Process_Key := False ;
    if( Key = VK_RETURN ) then
    begin
        if( Popped_Menus.Count > 0 ) then
        begin
            This_Menu := T2D_Menu( Popped_Menus[ Popped_Menus.Count - 1 ] ) ;
            if( This_Menu <> self ) then
            begin
                This_Menu.Process_Char( Key, Repeats ) ;
                exit ;
            end ;
        end ;

        // Popup the child menu...
        if( Highlighted_Menu_Index >= 0 ) then
        begin
            if( Items[ Highlighted_Menu_Index ].Count > 0 ) then
            begin
                Menus_Active := True ;
                Invalidate_Index( Highlighted_Menu_Index ) ;
                Popup_Index( Highlighted_Menu_Index ) ;
                T2D_Menu( Popped_Menus[ Popped_Menus.Count - 1 ] ).Highlight_Index( 0, False ) ;
            end else
            begin
                Activate( VK_LBUTTON, True ) ;
            end ;
        end ;
        exit ;
    end ; // if( Key = VK_RETURN )

    // Look for a matching shortcut key
    if( Popped_Menus.Count > 0 ) then
    begin
        This_Menu := T2D_Menu( Popped_Menus[ Popped_Menus.Count - 1 ] ) ;
    end else
    begin
        This_Menu := self ;
    end ;
    for Loop := 0 to This_Menu.Count - 1 do
    begin
        S := This_Menu.Items[ Loop ].Caption ;
        Dummy := pos( '&', S ) ;
        if( Dummy > 0 ) then
        begin
            S := upcase( S[ Dummy + 1 ] ) ;
            if( S = chr( Key ) ) then
            begin
                if( This_Menu.Items[ Loop ].Count > 0 ) then
                begin
                    Menus_Active := True ;
                    This_Menu.Invalidate_Index( Loop ) ;
                    This_Menu.Popup_Index( Loop ) ;
                    T2D_Menu( Popped_Menus[ Popped_Menus.Count - 1 ] ).Highlight_Index( 0, False ) ;
                end else
                begin
                    This_Menu.Items[ Loop ].Activate( VK_LBUTTON, True ) ;
                end ;
                exit ;
            end ;
        end ;
    end ;

    // Keystroke means nothing in this context
    MessageBeep( 0 ) ;
end ; // T2D_Menu.Process_Char


procedure T2D_Menu.Process_Message( Message : TMsg ; var Eat : boolean ) ;

var C : TWinControl ;
    SS : TShiftState ;

begin
    case Message.message of
        WM_MDICREATE : Handle_MDICreate( Message ) ;
        WM_DESTROY : Handle_Destroy( Message ) ;
        WM_MDIMAXIMIZE : Handle_MDIMaximize( Message ) ;
        WM_SIZE : Handle_Size( Message ) ;

        WM_MOUSEMOVE :
            begin
                C := FindControl( Message.hwnd ) ;
                if( ( C = nil ) or ( not ( C is T2D_Menu ) ) ) then // Moved off of a menu
                begin
                    if( ( Align = alTop ) or ( Align = alBottom ) ) then
                    begin
                        if( not ( Menus_Active or Menu_Key_Mode ) ) then
                        begin
                            Highlight_Index( -1, False ) ;
                        end ;
                    end ;
                end ;
            end ;

        WM_LBUTTONDOWN :
            begin
                C := FindControl( Message.hwnd ) ;
                if( ( C = nil ) or ( not ( C is T2D_Menu ) ) ) then
                begin
                    Deactivate_Menus ;
                end ;
            end ;

        WM_CANCELMODE, WM_ACTIVATEAPP, CM_ACTIVATE, CM_DEACTIVATE : Deactivate_Menus ;

        WM_SETFOCUS :
            begin
                Deactivate_Menus ;
                Dialog_Focused := True ;
                Invalidate ;
            end ;

        WM_KEYDOWN, WM_SYSKEYDOWN, WM_CHAR, WM_SYSCHAR :
            begin
                if( Message.WParam = VK_MENU ) then
                begin
                    if( Menus_Active or Menu_Key_Mode ) then
                    begin
                        Deactivate_Menus ;
                    end else
                    begin
                        if( GetKeyState( VK_MENU ) >= 0 ) then // Only on key-up
                        begin
                            Menu_Key_Mode := True ;
                            Highlight_Index( 0, False ) ;
                        end ;
                    end ;
                    Eat := True ;
                    exit ;
                end else
                if( Message.WParam = VK_ESCAPE ) then
                begin
                    if( Menus_Active or Menu_Key_Mode ) then
                    begin
                        Eat := True ;
                    end ;
                    Deactivate_Menus ;
                    exit ;
                end else
                if( ( Menus_Active ) or ( Menu_Key_Mode ) ) then
                begin
                    Eat := True ;
                    Process_Char( Message.WParam, Message.LParam and $FFFF ) ;
                end else
                begin
                    if( ( Message.WParam = VK_SHIFT ) or ( Message.WParam = VK_CONTROL ) ) then
                    begin
                        exit ;
                    end ;
                    SS := [] ;
                    if( GetKeyState( VK_SHIFT ) < 0 ) then
                    begin
                        SS := [ ssShift ]
                    end ;
                    if( GetKeyState( VK_MENU ) < 0 ) then
                    begin
                        SS := SS + [ ssAlt ]
                    end ;
                    if( GetKeyState( VK_Control ) < 0 ) then
                    begin
                        SS := SS + [ ssCtrl ]
                    end ;
                    Eat := Process_Accelerator( Message.WParam, SS ) ;
                end ;
            end ;
    end ; // case Message.message
end ; // T2D_Menu.Process_Message


function T2D_Menu.Highlighted_Index : integer ;

begin
    Result := Highlighted_Menu_Index ;
end ;


procedure T2D_Menu.Highlight_Index( Index : integer ; Pop : boolean ) ;
{ Highlight the specified index.  If Pop is true, popup the highlighted menu's
  children. }

var This_Menu : T2D_Menu ;

begin
    if( Count = 0 ) then
    begin
        exit ;
    end ;
    if( Index <> Highlighted_Menu_Index ) then // Reduce flicker by only doing necessary work
    begin
        if( Highlighted_Menu_Index >= 0 ) then // Something previously selected
        begin
            Invalidate_Index( Highlighted_Menu_Index ) ;
            This_Menu := Items[ Highlighted_Menu_Index ] ;
            if( This_Menu.Count > 0 ) then
            begin
                This_Menu.Highlight_Index( -1, False ) ;
            end ;
        end ;
        if( ( Index >= 0 ) and ( length( Items[ Index ].Caption ) = 0 ) ) then
        begin
            Highlighted_Menu_Index := -1 ;
        end else
        begin
            Highlighted_Menu_Index := Index ;
            Invalidate_Index( Index ) ;
        end ;
        if( Index >= 0 ) then
        begin
            This_Menu := Items[ Index ] ;
            if( assigned( This_Menu.On_Highlight ) ) then
            begin
                This_Menu.On_Highlight( This_Menu ) ;
            end ;
        end ;
        Cancel_Popups( self ) ;
        if( Menus_Active and Pop and ( Align <> alLeft ) and ( Align <> alRight ) ) then
        begin
            Popup_Index( Index ) ;
        end ;
        if( Parent_Menu <> nil ) then
        begin
            Parent_Menu.Child_Highlighted( self ) ;
        end ;
    end ;
end ; // T2D_Menu.Highlight_Index


function T2D_Menu.Serialize( State_Only : boolean ) : string ;

begin
    Result := '<#index=' + inttostr( Last_Popup_Index ) + '>' ;
    if( not State_Only ) then
    begin
        //~~~serialize entire contents
    end ;
end ;


procedure T2D_Menu._Show ;

var Control : TControl ;
    Dummy : integer ;

begin
    ShowWindow( Handle, SW_SHOW ) ;
    _Visible := True ;
    if( ( Align <> alTop ) and ( Align <> alBottom ) ) then
    begin
        if( ( ControlCount = 0 ) or ( Count > 0 ) ) then
        begin
            Dummy := Text_Height ;
            if( Height <> Dummy ) then
            begin
                Height := Dummy ;
            end ;
            Dummy := Text_Width ;
            if( Width <> Dummy ) then
            begin
                Width := Dummy ;
            end ;
        end ;
    end ;
//    BringToFront ;
    for Dummy := 0 to ControlCount - 1 do
    begin
        Control := Controls[ Dummy ] ;
        if( Control is TWinControl ) then
        begin
            TWinControl( Control ).Handle ;
        end ;
//        Control.BringToFront ;
    end ;
    Popped_Menus.Add( self ) ;
end ; // T2D_Menu._Show


function T2D_Menu.Top_Level_Menu : boolean ;

begin
    Result := ( Parent_Menu = nil ) ;
end ;



{ Returns True if we are a top-level menu (no parent). }
initialization
    Popped_Menus := TList.Create ;
    Arrow_Bitmap := TBitmap.Create ;
    Check_Bitmap := TBitmap.Create ;
    Radio_Button_Bitmap := TBitmap.Create ;
    Off_Screen := 32767 - GetSystemMetrics( SM_CYFULLSCREEN ) ;
    { Used to move controls off screen (ie make them invisible to user without
      actually making them invisible).  We don't want anything larger than this
      value or else windows makes the bottom value negative (probably due to
      16-bit integer rounding somewhere). }

finalization
    Popped_Menus.Free ;
    Popped_Menus := nil ;
    Arrow_Bitmap.Free ;
    Check_Bitmap.Free ;
    Radio_Button_Bitmap.Free ;
    Arrow_Bitmap := nil ;
    Check_Bitmap := nil ;
    Radio_Button_Bitmap := nil ;
end.
