{
        Program Name : Operatio
        Package Name : LB
        Purpose      : TOperation_Logger object
        Institution  : Conroy & Conroy Co.
        Date Written : 20-Sep-1996
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 1996 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  The TOperation_Logger object is used to keep track of multi-part
	transactions.
}

unit Operatio ;

interface

type Operation_Handler = procedure( ID : integer ; Data : pointer ) ;

     POperation = ^ROperation ;
     ROperation = record
		      Transaction : longint ; { Transaction number }
		      Typ : integer ; { Transation type }
		      ID : Operation_Handler ; { User-supplied ID }
		      Data : pointer ; { Data to pass to Undo routine }
		      Data_Size : integer ; { Size of Data^ }
		      Next : POperation ; { Pointer to next operation }
		end ;


type POperation_Logger = ^TOperation_Logger ;
     TOperation_Logger = object
			     private
				procedure Check_Extras ;

			     public
				Root : POperation ; { Root of operation list }
				Max_Operations : integer ; { Maximum logged operations }
				Count : integer ; { Count of logged operations }
				Transaction_Number : longint ; { Current transaction number }
				Transaction_Type : integer ; { Current transaction type }
				Original : boolean ; { True if original operation is on the list }

				constructor Init( Max : integer ) ;
				destructor Done ;

				procedure Set_Maximum( Max : integer ) ;
				procedure Zero ;
				procedure New_Transaction( Typ : integer ) ;
				procedure Unwind( Code : integer ) ;
				procedure New_Operation( _ID : Operation_Handler ;
				    Data : pointer ; Size : integer ) ;
				function Query_Type : integer ;
			end ;

implementation

constructor TOperation_Logger.Init( Max : integer ) ;

begin
    Root := nil ; { Root of operation list }
    Count := 0 ; { Count of logged operations }
    Transaction_Number := 0 ;
    Transaction_Type := 0 ;
    Original := TRUE ;
    Set_Maximum( max ) ; { Maximum logged operations }
end ;


destructor TOperation_Logger.Done ;

begin
    Zero ;
end ;


procedure TOperation_Logger.Check_Extras ;
{ Delete any originals which overflow the maximum }

var Temp : POperation ;
    Transaction : longint ;

begin
    if( Root = nil ) then { Nothing logged }
    begin
	exit ;
    end ;

    if( Root^.Transaction = Transaction_Number ) then
    begin
	exit ; { Done if one big transaction }
    end ;

    if( Count > Max_Operations ) then
    begin
        Transaction := 0 ;
	Original := False ;
	while( Count > Max_Operations ) do
	begin
	    Count := Count - 1 ;
	    Temp := Root ;
	    Root := Root^.Next ;
	    Transaction := Temp^.Transaction ;
	    if( Temp^.Data <> nil ) then
	    begin
		freemem( Temp^.Data, Temp^.Data_Size ) ;
	    end ;
	    dispose( Temp ) ;
	end ;
	{ Make sure all operations within a given transaction go away }
	while( Root^.Transaction = Transaction ) do
	begin
	    Temp := Root ;
	    if( Temp^.Data <> nil ) then
	    begin
		freemem( Temp^.Data, Temp^.Data_Size ) ;
	    end ;
	    dispose( Temp ) ;
	    Root := Root^.Next ;
	end ;
    end ;
end ;


procedure TOperation_Logger.Set_Maximum( max : integer ) ;

begin
    { Update maximum operations }
    Max_Operations := max ;

    { Delete any extra operations }
    Check_Extras ;
end ;


procedure TOperation_Logger.Zero ; { Erase all storage }

var Temp : POperation ;

begin
    Original := TRUE ;
    while( Root <> nil ) do
    begin
	Temp := Root ;
	Root := Root^.Next ;
	if( Temp^.Data <> nil ) then
	begin
	    freemem( Temp^.Data, Temp^.Data_Size ) ;
	end ;
	dispose( Temp ) ;
    end ;
    Count := 0 ; { Nothing left }
    Transaction_Number := 0 ;
end ; { TOperation_Logger.Zero }


procedure TOperation_Logger.New_Transaction( Typ : integer ) ;
{ Start a new transaction }

begin
    Transaction_Number := Transaction_Number + 1 ;
    Transaction_Type := Typ ;
end ;


procedure TOperation_Logger.Unwind( Code : integer ) ;
{ Unwind the last transaction on the list }

var Transaction : longint ;
    Temp : POperation ; { Current operation }
    First : POperation ; { First operation of this transaction }
    Previous : POperation ; { Previous operation }
    Last : POperation ; { Last operation of previous transaction }

begin
    { Find beginning of last transaction }
    if( Root = nil ) then
    begin
	exit ; { Nothing to unwind }
    end ;
    Temp := Root ;
    Transaction := Root^.Transaction ;
    Last := nil ;
    First := Root ;
    Previous := nil ;
    while( Temp <> nil ) do { Until EOL }
    begin
	if( Transaction <> Temp^.Transaction ) then { New transaction }
	begin
	    Transaction := Temp^.Transaction ;
	    First := Temp ;
	    Last := Previous ;
	end ;
	Previous := Temp ;
	Temp := Temp^.Next ;
    end ;

    { At this point, Last is the last operation of the previous transaction
      (nil if no previous transaction), First is the first operation of the
      current transaction. }
    if( Last <> nil ) then
    begin
	Last^.Next := nil ; { Terminate the list }
    end else
    begin
        Root := nil ;  { No more items in list }
    end ;
    while( First <> nil ) do
    begin
	{ Find the last transaction on the current operation }
	Temp := First ;
	Previous := Last ;
	while( Temp^.Next <> nil ) do
	begin
	    Previous := Temp ;
	    Temp := Temp^.Next ;
	end ;
	if( Previous <> nil ) then
	begin
	    Previous^.Next := nil ;
	end ;
	Temp^.ID( Code, Temp^.Data ) ;
	if( Temp^.Data <> nil ) then
	begin
	    freemem( Temp^.Data, Temp^.Data_Size ) ;
	end ;
	if( Temp = First ) then { Processed the first operation in the transaction }
	begin
	    First := nil ; { Terminate the loop }
	end ;
	dispose( Temp ) ;
	Count := Count - 1 ; { One less operation }
    end ;
end ; { TOperation_Logger.Unwind }


function TOperation_Logger.Query_Type : integer ;
{ Return type of next undo record }

var Temp : POperation ; { Current operation }

begin
    { Find the last logged record }
    if( Root = nil ) then
    begin
	Query_Type := 0 ; { Nothing to unwind }
	exit ;
    end ;
    Temp := Root ;
    while( Temp^.Next <> nil ) do { Until EOL }
    begin
	Temp := Temp^.Next ;
    end ;
    Query_Type := Temp^.Typ ;
end ;


procedure TOperation_Logger.New_Operation( _ID : Operation_Handler ;
    Data : pointer ; Size : integer ) ;
{ Store new operation on list }

var Temp, Operation : POperation ;

begin
    { Create new operation record }
    new( Operation ) ;
    Operation^.Next := nil ;
    Operation^.Transaction := Transaction_Number ;
    Operation^.Typ := Transaction_Type ;
    Operation^.ID := _ID ;
    Operation^.Data := data ;
    Operation^.Data_Size := size ;

    { Find end of list and insert operation }
    if( Root = nil ) then
    begin
	Root := Operation ;
    end else
    begin
	Temp := Root ;
	while( Temp^.Next <> nil ) do
	begin
	    Temp := Temp^.Next ;
	end ;
	Temp^.Next := Operation ;
    end ;

    { Increment count of operations }
    Count := Count + 1 ;
    Check_Extras ;
end ;



end.
