{$R-}
{$Q-}
{
        Program Name : Randoms
        Package Name : LB
        Purpose      : Generate random numbers
        Institution  : Conroy & Conroy Co.
        Date Written : 30-Dec-1998
        Written By   : Alan Conroy
        Version      : 1.0A

        Copyright (C) 1998 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  1-Nov-2002    EAC         Ignore integer overflows.      *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	This unit contains classes used to generate random numbers.
}

unit Randoms ;

interface

{$I EDEFINES.INC}

type TRNG = object
		private { Instance data... }
		    _X : longint ;

		public { API... }
		    function Generate : longint ;
		    { Generate a new random value }

		    procedure Randomize ; { Randomize X }

		    procedure Set_X( _X_ : longint ) ; { Set X }

		    function X : longint ; { Return X }
	    end ;

type TRNG8 = object
		private { Instance data... }
		    _X : byte ;

		public { API... }
		    function Generate : byte ;
		    { Generate a new random value }

		    procedure Randomize ; { Randomize X }

		    procedure Set_X( _X_ : byte ) ; { Set X }

		    function X : byte ; { Return X }
	    end ;

type TRNG16 = object
		private { Instance data... }
		    _X : word ;

		public { API... }
		    function Generate : word ;
		    { Generate a new random value }

		    procedure Randomize ; { Randomize X }

		    procedure Set_X( _X_ : word ) ; { Set X }

		    function X : word ; { Return X }
	    end ;

{$IFDEF DELPHI}
type TRNG32 = object
		private { Instance data... }
		    _X : cardinal ;

		public { API... }
		    function Generate : cardinal ; stdcall ;
		    { Generate a new random value }

		    procedure Randomize ; stdcall ; { Randomize X }

		    procedure Set_X( _X_ : cardinal ) ; stdcall ; { Set X }

		    function X : cardinal ; stdcall ; { Return X }
	    end ;

type TRNG64 = object
		private { Instance data... }
		    _X : int64 ;

		public { API... }
		    function Generate : int64 ; stdcall ;
		    { Generate a new random value }

		    procedure Randomize ; stdcall ; { Randomize X }

		    procedure Set_X( _X_ : int64 ) ; stdcall ; { Set X }

		    function X : int64 ; stdcall ; { Return X }
	    end ;
{$ENDIF}

implementation

function TRNG.Generate : longint ; { Generate a new random value }

begin
{$WARNINGS OFF}
    _X := ( $BB40E62D * _X + 1 ) ;
{$WARNINGS ON}
    Generate := _X shr 1 ;
end ;


procedure TRNG.Randomize ; { Randomize X }

begin
    System.Randomize ;
    _X := random( 65535 ) ;
    _X := _X shl 16 or random( 65535 ) ;
end ;


procedure TRNG.Set_X( _X_ : longint ) ; { Set X }

begin
    _X := _X_ ;
end ;


function TRNG.X : longint ; { Return X }

begin
    X := _X ;
end ;


function TRNG8.Generate : byte ; { Generate a new random value }

begin
    _X := ( 29 * _X + 1 ) ;
    Generate := _X shr 1 ;
end ;


procedure TRNG8.Randomize ; { Randomize X }

begin
    System.Randomize ;
    _X := random( 265 ) ;
end ;


procedure TRNG8.Set_X( _X_ : byte ) ; { Set X }

begin
    _X := _X_ ;
end ;


function TRNG8.X : byte ; { Return X }

begin
    X := _X ;
end ;


function TRNG16.Generate : word ; { Generate a new random value }

begin
    _X := ( 31413 * _X + 1 ) ;
    Generate := _X shr 1 ;
end ;


procedure TRNG16.Randomize ; { Randomize X }

begin
    System.Randomize ;
    _X := random( 65535 ) ;
end ;


procedure TRNG16.Set_X( _X_ : word ) ; { Set X }

begin
    _X := _X_ ;
end ;


function TRNG16.X : word ; { Return X }

begin
    X := _X ;
end ;


{$IFDEF DELPHI}
function TRNG32.Generate : cardinal ; { Generate a new random value }

begin
    _X := ( 3141592621 * _X + 1 ) ;
    Generate := _X shr 1 ;
end ;


procedure TRNG32.Randomize ; { Randomize X }

begin
    System.Randomize ;
    _X := random( 65535 ) ;
{$WARNINGS OFF}
    _X := _X shl 16 or random( 65535 ) ;
{$WARNINGS ON}
end ;


procedure TRNG32.Set_X( _X_ : cardinal ) ; { Set X }

begin
    _X := _X_ ;
end ;


function TRNG32.X : cardinal ; { Return X }

begin
    X := _X ;
end ;


function TRNG64.Generate : int64 ; { Generate a new random value }

begin
    _X := ( 314159265358973 * _X + 1 ) ;
    Generate := _X shr 1 ;
end ;


procedure TRNG64.Randomize ; { Randomize X }

begin
    System.Randomize ;
    _X := random( 65535 ) ;
    _X := _X shl 16 or random( 65535 ) ;
    _X := _X shl 16 or random( 65535 ) ;
    _X := _X shl 16 or random( 65535 ) ;
end ;


procedure TRNG64.Set_X( _X_ : int64 ) ; { Set X }

begin
    _X := _X_ ;
end ;


function TRNG64.X : int64 ; { Return X }

begin
    X := _X ;
end ;
{$ENDIF}


end.
