{
        Program Name : Sizable_Speed_Button
        Package Name : LB
        Purpose      : Sizable TSpeedButton
        Institution  : Conroy & Conroy Co.
        Date Written : 24-Oct-1998
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 1998 by Alan Conroy.  Released to the Public Domain.
        
        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TSizable_Speed_Button is a TSpeedButton that can resize itself to
        match its parent's height.
}

unit Sizable_Speed_Button ;

interface

uses // Windows...
     Windows,
     Messages,

     // Delphi...
     SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
     Buttons ;

type
  TSizable_Speed_Button = class( TSpeedButton )
      public // Constructors and destructors...
          constructor Init( AOwner : TComponent ; Bitmap : TBitmap ) ;
          destructor Destroy ; override ;

      private // Instance data...
          Height_To_Width_Ratio : real ;
          Original_Bitmap : TBitmap ;

      private // Callbacks...
          procedure CB_Resize( Sender : TObject ) ;

      public
          property Color ;
  end ;

procedure Register ;

implementation

uses // Delphi...
     ExtCtrls ; // TPanel

procedure Register ;

begin
    RegisterComponents( 'LB', [TSizable_Speed_Button] ) ;
end ;


// TSizable_Speed_Button methods...

var Inside_Resize : boolean = False ;

// Constructors and destructors...

constructor TSizable_Speed_Button.Init( AOwner : TComponent ; Bitmap : TBitmap ) ;

var R : real ;

begin
    // Call inherited constructor
    inherited Create( AOwner ) ;

    // Setup...
    OnResize := CB_Resize ;
    Height_To_Width_Ratio := Bitmap.Height ;
    R := Bitmap.Width ;
    Height_To_Width_Ratio := Height_To_Width_Ratio / R ;
    Width := Bitmap.Width ;
    Height := Bitmap.Height ;

    // Save our bitmap
    Original_Bitmap := Bitmap ;

    // Make a copy of the bitmap...
    Glyph := TBitmap.Create ;
    Glyph.Width := Bitmap.Width ;
    Glyph.Height := Bitmap.Height ;
    Glyph.Canvas.CopyRect( Rect( 0, 0, Bitmap.Width, Bitmap.Height ),
        Bitmap.Canvas, Rect( 0, 0, Bitmap.Width, Bitmap.Height ) ) ;
    Glyph.Transparent := Bitmap.Transparent ;
end ;


destructor TSizable_Speed_Button.Destroy ;

begin
    Original_Bitmap.Free ;

    inherited Destroy ;
end ;


// Callbacks...

procedure TSizable_Speed_Button.CB_Resize( Sender : TObject ) ;

var H : integer ; // Parent's height

begin
    // Determine size and shape...
    if( Parent = nil ) then
    begin
        exit ; // Not fully setup yet...
    end ;
    if( Inside_Resize ) then
    begin
        exit ; // No recursive calls
    end ;
    Inside_Resize := True ;
    H := Parent.ClientHeight ;
    if( Parent is TPanel ) then
    begin
        if( TPanel( Parent ).BevelInner <> bvNone ) then
        begin
            H := H - TPanel( Parent ).BevelWidth * 2 ;
        end ;
        if( TPanel( Parent ).BevelOuter <> bvNone ) then
        begin
            H := H - TPanel( Parent ).BevelWidth * 2 ;
        end ;
    end ;
    SetBounds( Left, Top, trunc( H * ( 1.0 / Height_To_Width_Ratio ) ), H ) ;

    // Make a copy of the bitmap...
    Glyph.Width := ClientWidth ;
    Glyph.Height := ClientHeight ;
    Glyph.Canvas.CopyMode := cmSrcCopy ;
    Glyph.Canvas.StretchDraw( ClientRect, Original_Bitmap ) ;
    Glyph.Canvas.Brush.Color := Color ;
    Glyph.Canvas.FloodFill( 0, 0, Glyph.Canvas.Pixels[ 0, 0 ], fsSurface ) ;
    Inside_Resize := False ;
end ; // TSizable_Speed_Button.CB_Resize



end.
