{
        Program Name : SpellChk
        Package Name : Spell
        Purpose      : Spell checker
        Institution  : Conroy & Conroy Co.
        Date Written : 23-Jun-1996
        Written By   : Alan Conroy
        Version      : 1.0A

        Copyright (C) 1996 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
	* 13-Nov-2002   EAC          Fix compilation error.         *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  These functions are used to check the spelling of words.
}

unit SpellChk ;

interface

uses Common, { TBase_Object }
     TypeDefs ; { PString_Array }

const Max_Word_Set = 127 ;

type PSpeller = ^TSpeller ;
     TSpeller = object( TBase_Object )
		    private

			Max_Offsets : array[ 0..Max_Word_Set ] of integer ;
			{ Max offsets in Word_Sets }
			Word_Set_Count : integer ;
			{ How many sets of words defined }
			Word_Sets : array[ 0..Max_Word_Set ] of pSmall_String_Array ;
			{ Defined sets }
			Sorted : array[ 0..Max_Word_Set ] of boolean ;

		    public

			constructor Init ;
			destructor Done ; virtual ;

			procedure Add_Word( S : integer ; X : string ) ;
			function Check( const X : string ) : boolean ;
			procedure Define_Set( const X : string ) ;
			function Is_Word( const X : string ) : boolean ;
		end ;

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses Files,
     _UEHDefs ; { TUEC }

constructor TSpeller.Init ;

begin
    fillchar( Max_Offsets, sizeof( Max_Offsets ), 0 ) ;
    Word_Set_Count := 0 ;
    fillchar( Word_Sets, sizeof( Word_Sets ), 0 ) ;
    fillchar( Sorted, sizeof( Sorted ), 0 ) ;
end ;


destructor TSpeller.Done ;

var Dummy : integer ;

begin
    while( Word_Set_Count > 0 ) do
    begin
	for Dummy := 1 to Max_Offsets[ Word_Set_Count ] do
	begin
	    freemem( Word_Sets[ Word_Set_Count ]^[ Dummy ], sizeof( string ) ) ;
	end ;
	freemem( Word_Sets[ Word_Set_Count ],
	    sizeof( PString ) * Max_Offsets[ Word_Set_Count ] ) ;
	dec( Word_Set_Count ) ;
    end ;
end ; { TSpeller.Done }


procedure TSpeller.Add_Word( S : integer ; X : string ) ;
{ Add word to specified word set. }

var Temp : PSmall_String_Array ; { For reallocation }

begin
    if( ( S < 0 ) or ( S > Max_Word_Set ) ) then
    begin
	exit ;
    end ;
    X := X + #255 ;
    if(
	( Max_Offsets[ S ] = 0 )
	or
	( length( Word_Sets[ S ]^[ Max_Offsets[ S ] ]^ ) + length( X ) > 254 )
      ) then
    begin
	inc( Max_Offsets[ S ] ) ;
	if( Max_Offsets[ S ] > High( Word_Sets[ 0 ]^ ) ) then
	begin
	    exit ; { Too many words to fit }
	end ;

	{ Reallocate word array }
	getmem( Temp, sizeof( PString ) * ( Max_Offsets[ S ] + 1 ) ) ;
	move( Word_Sets[ S ]^, Temp^, sizeof( PString ) * Max_Offsets[ S ] ) ;
	freemem( Word_Sets[ S ], sizeof( PString ) * Max_Offsets[ S ] ) ;
	Word_Sets[ S ] := Temp ;
	Word_Sets[ S ]^[ Max_Offsets[ S ] ] := nil ;
    end ;
    if( Word_Sets[ S ]^[ Max_Offsets[ S ] ] = nil ) then
    begin
	new( Word_Sets[ S ]^[ Max_Offsets[ S ] ] ) ;
	Word_Sets[ S ]^[ Max_Offsets[ S ] ]^ := #255 ;
    end ;
    Word_Sets[ S ]^[ Max_Offsets[ S ] ]^ :=
	Word_Sets[ S ]^[ Max_Offsets[ S ] ]^ + X ;
    Sorted[ S ] := False ;
end ; { TSpeller.Add_Word }


procedure TSpeller.Define_Set( const X : string ) ;
{ Defines a new set of words, loading them from file specified in X.  If X is
  null, an empty word set is created. }

var F : PSFile ;
    Last : string ;
    S : string ;
    Sorted_Set : boolean ;
    UEC : TUEC ;

begin
    if( length( X ) = 0 ) then
    begin
	exit ;
    end ;
    inc( Word_Set_Count ) ;
    if( Word_Set_Count > Max_Word_Set ) then
    begin
	exit ;
    end ;

    getmem( Word_Sets[ Word_Set_Count ], sizeof( PString ) * 2 ) ;
    Word_Sets[ Word_Set_Count ]^[ 0 ] := nil ;
    Word_Sets[ Word_Set_Count ]^[ 1 ] := nil ;
    Sorted_Set := True ; { Assume sorted }
    Last := '' ;
    F := Open( X, FSOM_Input or FSOM_Text, UEC ) ;
    if( UEC.Code = 0 ) then
    begin
	while( F^.IO_Error.Code = 0 ) do
	begin
	    F^.Readln( S ) ;
	    if( F^.IO_Error.Code = 0 ) then
	    begin
		if( Sorted_Set ) then
		begin
		    if( Last > S ) then
		    begin
			Sorted_Set := False ;
		    end else
		    begin
			Last := S ;
		    end ;
		end ;
		Add_Word( Word_Set_Count, S ) ;
	    end ;
	end ;
    end ;
    Sorted[ Word_Set_Count ] := Sorted_Set ;
    dispose( F, Done ) ;
end ; { TSpeller.Define_Set }


function TSpeller.Check( const X : string ) : boolean ;
{ Check X against word sets.  Returns True if found, false otherwise. }

var Dummy, Dummy1, Ending : integer ;
    Mid, Low, High : integer ; { Used to bi-search }

begin
    for Dummy := 1 to Word_Set_Count do { Check each set }
    begin
	Low := 1 ;
	High := Max_Offsets[ Dummy ] ;
	if( Sorted[ Dummy ] ) then
	begin
	    while( Low < High ) do
	    begin
		Mid := ( ( High - Low ) div 2 ) + Low - 1 ; { Mid-point }
		if( Mid < 1 ) then
		begin
		    Mid := 1 ;
		end ;
		if( ( Mid = Low ) or ( Mid = High ) ) then
		begin
		    break ; { Got close enough }
		end ;
		Ending := 2 ;
		while( Word_Sets[ Dummy ]^[ Mid ]^[ Ending ] <> #255 ) do
		begin
		    inc( Ending ) ;
		end ;
		if( X < copy( Word_Sets[ Dummy ]^[ Mid ]^, 2, Ending - 2 ) ) then
		begin
		    High := Mid ;
		end else
		begin
		    Low := Mid ;
		end ;
	    end ; { while( Low < High ) }
	    if( Low > 1 ) then
	    begin
		dec( Low ) ;
	    end ;
	end ; { if( Sorted[ Dummy ] ) }

	{ Check all words in set, in range }
	for Dummy1 := Low to High do
	begin
	    { Check for an early exit }
	    if( Sorted[ Dummy ] ) then
	    begin
		Ending := 2 ;
		while( Word_Sets[ Dummy ]^[ Dummy1 ]^[ Ending ] <> #255 ) do
		begin
		    inc( Ending ) ;
		end ;
		if( X < copy( Word_Sets[ Dummy ]^[ Dummy1 ]^, 2, Ending - 2 ) ) then
		begin
		    break ; { Try next set }
		end ;
	    end ;

	    { See if we can find the word }
	    if( pos( #255 + X + #255, Word_Sets[ Dummy ]^[ Dummy1 ]^ ) > 0 ) then
	    begin
		Check := True ;
		exit ;
	    end ;
	end ;
    end ;
    Check := False ;
end ; { TSpeller.Check }


function TSpeller.Is_Word( const X : string ) : boolean ;
{ Return True if X is a word. }

var Dummy : integer ;

begin
    if( ( length( X ) = 1 ) or ( length( X ) > 253 ) ) then
    begin
	Is_Word := False ;
	exit ;
    end ;
    for Dummy := 1 to length( X ) do
    begin
	if( pos( X[ Dummy ], '-ABCDEFGHIJKLMNOPQRSTUVWXYZ ' + #39 ) = 0 ) then
	begin
	    Is_Word := False ;
	    exit ;
	end ;
    end ;
    Is_Word := True ;
end ; { TSpeller.Is_Word }


{$IFDEF Test}
procedure Test_Unit ;

var Speller : TSpeller ;

begin
    Speller.Init ;
    {TODO: ~~~}
    Speller.Done ;
end ;
{$ENDIF}


end.
