{$N+}
{$A-}
unit StorHeap ;
{
        Program Name : StorHeap
        Package Name : LB
        Purpose      : Store Heap
        Institution  : Conroy & Conroy Co.
        Date Written : 1998
        Written By   : Alan Conroy
        Version      : 1.2E

        Copyright (C) 1998-2002 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  1.0A                                                     *
        *  23-Jan-2002   EAC        Added Separator to TFind_Record.*
        *							    *
        *  1.1                                                      *
        *  4-Mar-2002    EAC        Reorganized all of the code.    *
        *                                                           *
        *  1.2                                                      *
        *  11-Jun-2002   EAC        Added data caching.             *
        *  19-Aug-2002              Removed extra writes during	    *
        *                           extensions.                     *
        *  13-Nov-2002              Support TAllocator rename.	    *
        *  15-Nov-2002              Reduce compiler warnings.	    *
        *                           Fix XSpaceAvail to not return   *
        *                           values < 0.                     *
        *  16-Nov-2002              Remove TAT (now in Allocatr).   *
        *   2-Dec-2002              Handle string initialization.   *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit contains the Store Heap - a TManaged_Store on a TStore.

        Store Heap structure:
                Root (size = resolution, address = 0)
                Header (size = resolution, address = resolution)
                data (size is variable)
                Allocation Table (size is variable)

        Note that in previous versions of the Store Heap, there was no header.
        Thus the current store heap component is incompatible with older
        versions.
}

interface

{$I EDEFINES.INC}

uses
{$IFDEF WINDOWS}
     Objects, Strings,
{$ELSE}
  {$IFDEF DELPHI}
     SysUtils,
  {$ELSE}
     Objects,
  {$ENDIF}
{$ENDIF}
     Allocatr, { TAT }
     _DebugIn, { PSDebug_Interface }
     _UEHDefs, { TUEC }
     _Stores, { PSStore }
     TypeDefs, { pByte_Array }
     MStores ; { TSManaged_Store }

const SHF_NoExtend = 1 ; { Don't extend store }
      SHF_NoClose = 2 ; { Don't close store when done }
      SHF_Fast_Allocate = 4 ; { Don't reallocate deallocated memory }
      SHF_Fast_Deallocate = 8 ; { Don't deallocate memory }
      SHF_Read_Only = 16 ; { Don't allow writes }
      { 32-32768 = reserved }

      { Store Heap error codes... }
      SHErr_Facility = 5 ;
      SHErr_Success = 0 ;
      SHErr_No_Room_On_Device = 1 ;
      SHErr_Corrupted_File_Structure = 2 ;
      SHErr_Invalid_Pointer_Operation = 3 ;
      SHErr_Deallocation_Error = 4 ;
      SHErr_Size_Mismatch = 5 ;
      SHErr_Read_Out_Of_Range = 6 ;
      SHErr_Read_Size_Error = 7 ;
      SHErr_Write_Out_Of_Range = 8 ;
      SHErr_Write_Size_Error = 9 ;
      SHErr_Bad_Minimum = 10 ;

type TStore_Heap_Statistics = record
                          Index_Reads : longint ; { Cached index physical read count }
                          Index_Writes : longint ; { Cached index physical write count }
                          Index_Cache_Hits : longint ; { Number of index reads that were cached }
                      end ;

type PStore_Heap = ^TStore_Heap ;
     TStore_Heap = object( TSManaged_Store )
              private
{$IFDEF Debug}
                  Sanity : string[ 11 ] ;
{$ENDIF}
{$IFDEF Verify}
                  Address_List : TInteger_List ;
                  Size_List : TInteger_List ;
{$ENDIF}
                  Byte_Mapping : longint ; { How much one AT byte maps }

              public
                  AT : TAT ;
                  Store_Heap_Store : PSStore ;
                  Allow_Cache : boolean ; { True if AT is cached }
                  Device : integer ; { File device }
                  Flags : integer ; { Processing flags.  See SHF_* }
                  Memory : PSManaged_Store ; { Where to allocate cached information }
                  Own_Allocator : boolean ; { True if we own Memory }
                  Resolution : longint ; { How much one AT bit maps }
{$IFDEF Verify}
                  Track_FH : boolean ;
{$ENDIF}
                  Serial_Number : longint ;
                  Store_Heap_Statistics : TStore_Heap_Statistics ;
                  Extension_Resolution : longint ; { Minimum amount to extend file by, in percentage of file size }
                  Last_Error : TUEC ;
                  Lowest_Dynamic : TStore_Address ; { Lowest modifiable address }

              public { Constructors and destructors... }
                  constructor Init( Minimum, F : integer ; const Name : string ;
                      var _Result : tUEC ; Mem : PSManaged_Store ) ;
                  constructor Init_On_Store( Minimum, F : integer ; S : PSStore ;
                      var _Result : tUEC ; Mem : PSManaged_Store ) ;
                  destructor Done ; virtual ; { Finish up FH access }

              private { Internal utility routines }
                  procedure Common_Init(  Minimum, F : integer ; S : PSStore ;
                      var _Result : tUEC ; Mem : PSManaged_Store ) ;

                  procedure Log_Allocation( Address : TStore_Address ; _Size : TStore_Address ) ;
                  procedure Log_Deallocation( Address : TStore_Address ; _Size : TStore_Address ) ;
                  procedure Log_Read( Address : TStore_Address ; _Size : TStore_Address ) ;
                  procedure Log_Write( Address : TStore_Address ; _Size : TStore_Address ) ;
                  procedure Log_Error( C, S : integer ) ;
                  procedure Update_Header ;

              public { Overrides... }
                  function Extended_Size : TStore_Address ; virtual ;

              public { API... }
                  procedure Determine_AT_Metrics( var UEC : TUEC ) ; virtual ;
                  procedure New( _Size : TStore_Address ) ; virtual ;
                  { Create new heap, if empty }

                  function Allocate( _Size : TStore_Address ) : TStore_Address ;
                      virtual ;
                  { Allocate space }

                  function Allocate_At( _Position, _Size : TStore_Address ) : boolean ;
                      virtual ;
                  { Allocate space at a specific location }

                  procedure Deallocate( PTR : TStore_Address ; _Size : TStore_Address ) ;
                      virtual ;
                  { Deallocate space in Heap }

                  procedure Read_Heap( PTR : TStore_Address ; _Size : TStore_Address ;
                      var Buffer ) ; virtual ; { Read data from Heap }
                  function Reallocate( PTR : TStore_Address ;
                      Old, _New : TStore_Address ) : longint ; virtual ;
                      { Reallocate existing data }

                  procedure Write_Heap( PTR : TStore_Address ; _Size : TStore_Address ;
                      var Buffer ) ; virtual ; { Write data to Heap }
                  procedure Set_Root( var Buffer ) ; virtual ;
                  { Write data to root block of Heap }

                  procedure Get_Root( var Buffer ) ; virtual ;
                  { Read data from root block of Heap }

                  function Path : string ; virtual ;

                  function SpaceAvail : TStore_Address ; virtual ;
                  function XSpaceAvail : TStore_Address ; virtual ;
                  function MaxSpace : TStore_Address ; virtual ;
                  function Debugger : PSDebug_Interface ; virtual ;

                  procedure Set_Cache_Flag( Flag : boolean ) ;
                  function Valid_Data_Address( Address : TStore_Address ) : boolean ;
                  procedure Write_Cache ;
	   end ; { TStore_Heap }

implementation

uses
{$IFDEF DELPHI}
     { Borland... }
     Windows, { SECURITY_ATTRIBUTES }
     Forms, { Application }
  {$IFDEF TEST}
     Dialogs, { ShowMessage }
  {$ENDIF}
{$ELSE}
     { Borland... }
  {$IFNDEF WINDOWS}
     Strings, { strpas }
  {$ENDIF}
{$ENDIF}

     { Conroy & Conroy Co. }
     Cache, { PCache }
     Collect,
     CommonUt, { Boolean_To_String }
     Compatib, { ReallocMem }
     DebugInt, { TText_Debugger }
     ErrMastr, { Error_Code }
     Files, { Must be AFTER TypeDefs due to PFile }
     Num1s, { Num1 }
     O_S, { OS }
     Standard ; { Bit_Values }

type TStore_Heap_Header = record { 16-bytes }
                     ID : smallint ;
                     ID1 : byte ;
                     Version : byte ; { FH format version 10 = V1.0 }
                     AT_Offset : longint ; { Current AT location }
                     AT_Size : longint ; { Current AT size }
                     AT_Offset_High : longint ; { Upper 32 bits for 64-bit file systems }
                 end ;

{$IFDEF Verify}
var Global_Address_List : TInteger_List ;
    Global_Size_List : TInteger_List ;
{$ENDIF}

{$IFDEF DELPHI}
type tStore_Heap_EI = class( tError_Interface )
                            public
                                function Get_Facility : longint ;
                                    override ; stdcall ;
                                function Get_Name : PChar ; override ; stdcall ;
                                function Translate_Error( Code : longint ) : PChar ;
                                    override ; stdcall ;
                                procedure Done ; override ; stdcall ;
                                function Facility_Version : longint ; override ;
                                function Version : longint ; override ;
                                function Severity( Code : longint ) : longint ;
                                    override ;
                                function Translate_Error_Ex(  Code : longint ;
                                    var Size, Typ : longint ) : PChar ;
                                    override ; stdcall ;
                  end ;
     pStore_Heap_EI = tStore_Heap_EI ;
{$ELSE}
type tStore_Heap_EI = object( tError_Interface )
                            public
                                function Get_Facility : longint ; virtual ;
                                function Get_Name : PChar ; virtual ;
                                function Translate_Error( Code : longint ) : PChar ;
                                    virtual ;
                                function Facility_Version : longint ; virtual ;
                                function Version : longint ; virtual ;
                                function Severity( Code : longint ) : longint ;
                                    virtual ;
                  end ;
     pStore_Heap_EI = ^tStore_Heap_EI ;
{$ENDIF}


{ tStore_Heap_EI methods... }

function tStore_Heap_EI.Get_Facility : longint ;

begin
    Get_Facility := SHErr_Facility ;
end ;


const Store_Heap_Facility_Name : PChar = 'FH' ;

function tStore_Heap_EI.Get_Name : PChar ;

begin
    Get_Name := Store_Heap_Facility_Name ;
end ;


var Store_Heap_Error_Text : array[ 0..127 ] of char ;

function tStore_Heap_EI.Translate_Error( Code : longint ) : PChar ;

var Err : string ;

begin
    case Code of
        SHErr_Success : Err := 'Success' ;
        SHErr_No_Room_On_Device : Err := 'No Room On Device' ;
        SHErr_Corrupted_File_Structure : Err := 'Corrupted File Structure' ;
        SHErr_Invalid_Pointer_Operation : Err := 'Invalid Pointer Operation' ;
        SHErr_Deallocation_Error : Err := 'Deallocation Error' ;
        SHErr_Size_Mismatch : Err := 'Size Mismatch' ;
        SHErr_Read_Out_Of_Range : Err := 'Read Out Of Range' ;
        SHErr_Read_Size_Error : Err := 'Read Size Error' ;
        SHErr_Write_Out_Of_Range : Err := 'Write Out Of Range' ;
        SHErr_Write_Size_Error : Err := 'Write Size Error' ;
        SHErr_Bad_Minimum : Err := 'Bad minimum' ;
        else Err := 'Unknown error code ' + num1( Code ) ;
    end ;
    Err := Err + #0 ;
    move( Err[ 1 ], Store_Heap_Error_Text[ 0 ], length( Err ) ) ;
    Translate_Error := @Store_Heap_Error_Text ;
end ; { tStore_Heap_EI.Translate_Error }


function tStore_Heap_EI.Severity( Code : longint ) : longint ;

begin
    Severity := 0 ;
end ;


function tStore_Heap_EI.Translate_Error_Ex(  Code : longint ;
    var Size, Typ : longint ) : PChar ;

begin
    Translate_Error_Ex := Translate_Error( Code ) ;
end ;


{$IFDEF DELPHI}
procedure tStore_Heap_EI.Done ;

begin
    { Done by unit }
end ;
{$ENDIF}


function tStore_Heap_EI.Facility_Version : longint ;

begin
    Facility_Version := 0 ;
end ;


function tStore_Heap_EI.Version : longint ;

begin
    Version := 10 ; { Compliant with UEH V1.0 specification }
end ;



{ THeap_Statistics_Debugger... }

{$IFDEF DELPHI}
type THeap_Statistics_Debugger = object( TSText_Debugger )
                                     private
                                         _Instance : PStore_Heap ;
                                         S : string ;

                                     public
                                         constructor Init( I : PStore_Heap ) ;

                                         destructor Done ; virtual ;

                                         function Count : longint ; virtual ;

                                         function Child( Ordinal : longint ) : PSDebug_Interface ;
                                             virtual ;
                                 end ;
     PHeap_Statistics_Debugger = ^THeap_Statistics_Debugger ;

constructor THeap_Statistics_Debugger.Init( I : PStore_Heap ) ;

begin
    inherited Init ;

    _Instance := I ;
    Initialize( S ) ;
end ;


destructor THeap_Statistics_Debugger.Done ;

begin
    Finalize( S ) ;

    inherited Done ;
end ;


{$ELSE}
type PHeap_Statistics_Debugger = ^THeap_Statistics_Debugger ;
     THeap_Statistics_Debugger = object( TText_Debugger )
                          private
                              _Instance : PStore_Heap ;
                              S : string[ 255 ] ;

                          public
                              constructor Init( I : PStore_Heap ) ;

                              function Count : longint ; virtual ;

                              function Child( Ordinal : longint ) : PDebug_Interface ;
                                  virtual ;
                      end ;

constructor THeap_Statistics_Debugger.Init( I : PStore_Heap ) ;

begin
    inherited Init ;

    _Instance := I ;
end ;
{$ENDIF}


function THeap_Statistics_Debugger.Count : longint ;

begin
    Count := 21 ;
end ;


function THeap_Statistics_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

var R : PSText_Debugger ;

begin
{$IFDEF DELPHI}
    R := new( PSText_Debugger, Init ) ;
    case Ordinal of
        0 : S := 'Allocations = ' + Num1( _Instance.Statistics.Allocations ) ;
        1 : S := 'Bytes_Allocated = ' + Num1( _Instance.Statistics. Bytes_Allocated) ;
        2 : S := 'Bytes_Copied = ' + Num1( _Instance.Statistics.Bytes_Copied ) ;
        3 : S := 'Bytes_Deallocated = ' + Num1( _Instance.Statistics.Bytes_Deallocated ) ;
        4 : S := 'Bytes_Filled = ' + Num1( _Instance.Statistics.Bytes_Filled ) ;
        5 : S := 'Bytes_Read = ' + Num1( _Instance.Store_Statistics.Bytes_Read ) ;
        6 : S := 'Bytes_Written = ' + Num1( _Instance.Store_Statistics.Bytes_Written ) ;
        7 : S := 'Copies = ' + Num1( _Instance.Statistics.Copies ) ;
        8 : S := 'Deallocations = ' + Num1( _Instance.Statistics.Deallocations ) ;
        9 : S := 'Error_Count = ' + Num1( _Instance.Store_Statistics.Error_Count ) ;
        10 : S := 'Extensions = ' + Num1( _Instance.Statistics.Extensions ) ;
        11 : S := 'Fills = ' + Num1( _Instance.Statistics.Fills ) ;
        12 : S := 'Index_Cache_Hits = ' + Num1( _Instance.Store_Heap_Statistics.Index_Cache_Hits ) ;
        13 : S := 'Index_Reads = ' + Num1( _Instance.Store_Heap_Statistics.Index_Reads ) ;
        14 : S := 'Index_Writes = ' + Num1( _Instance.Store_Heap_Statistics.Index_Writes ) ;
        15 : S := 'Reads = ' + Num1( _Instance.Store_Statistics.Reads ) ;
        16 : S := 'Reallocations = ' + Num1( _Instance.Statistics.Reallocations ) ;
        17 : S := 'Reallocations_In_Place = ' + Num1( _Instance.Statistics.Reallocations_In_Place ) ;
        18 : S := 'Root_Reads = ' + Num1( _Instance.Statistics.Root_Reads ) ;
        19 : S := 'Root_Writes = ' + Num1( _Instance.Statistics.Root_Writes ) ;
        20 : S := 'Writes = ' + Num1( _Instance.Store_Statistics.Writes ) ;
    end ;
    R.Title := PChar( S ) ;
{$ELSE}
    new( R, Init ) ;
    R^.Set_Title( @S[ 1 ] ) ;
{$ENDIF}
    Child := PSDebug_Interface( R ) ;
end ;



{ Internal utility functions... }
procedure Add_To_Store_Heap_List( X : pointer ) ; forward ;
procedure Remove_From_Store_Heap_List( X : pointer ) ; forward ;
procedure Flush_FH( T : pointer ; const X : string ) ; forward ;


{ Store heap variables.  These variables simply make life easier. }

var SH_Buffer : array[ 0..127 ] of byte ; { 128 bytes of zero (0). }


{$IFDEF Debug}
procedure Check_FH( O : TStore_Heap ) ;

begin
    if( O.Sanity <> 'TStore_Heap' ) then
    begin
        raise Exception.Create( 'Invalid type-cast' ) ;
        halt ;
    end ;
end ;
{$ENDIF}

{ TStore_Heap_Debugger... }

{$IFDEF DELPHI}
type TStore_Heap_Debugger = object( TSText_Debugger )
                                private
                                    _Instance : PStore_Heap ;
                                    S : string ;

                                public
                                    constructor Init( I : PStore_Heap ) ;

                                    destructor Done ; virtual ;

                                    function Count : longint ; virtual ;

                                    function Child( Ordinal : longint ) : PSDebug_Interface ;
                                        virtual ;
                            end ;
     PStore_Heap_Debugger = ^TStore_Heap_Debugger ;

constructor TStore_Heap_Debugger.Init( I : PStore_Heap ) ;

begin
    inherited Init ;

    _Instance := I ;
    Initialize( S ) ;
end ;


destructor TStore_Heap_Debugger.Done ;

begin
    Finalize( S ) ;

    inherited Done ;
end ;


{$ELSE}
type PStore_Heap_Debugger = ^TStore_Heap_Debugger ;
     TStore_Heap_Debugger = object( TSText_Debugger )
                          private
                              _Instance : PStore_Heap ;
                              S : string[ 255 ] ;

                          public
                              constructor Init( I : PStore_Heap ) ;

                              function Count : longint ; virtual ;

                              function Child( Ordinal : longint ) : PDebug_Interface ;
                                  virtual ;
                      end ;

constructor TStore_Heap_Debugger.Init( I : PStore_Heap ) ;

begin
    inherited Init ;

    _Instance := I ;
end ;
{$ENDIF}


function TStore_Heap_Debugger.Count : longint ;

begin
{$IFDEF Verify}
    Count := 14 ;
{$ELSE}
    Count := 13 ;
{$ENDIF}
end ;


function TStore_Heap_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

var R : PSText_Debugger ;

begin
{$IFDEF DELPHI}
    R := new( PSText_Debugger, Init ) ;
    case Ordinal of
        0 : S := 'Allow_Cache = ' + Boolean_To_String( _Instance.Allow_Cache ) ;
        1 : begin
                dispose( R, Done ) ;
                R := PSText_Debugger( _Instance.AT.Debugger ) ;
                if( R = nil ) then
                begin
                    R := new( PSText_Debugger, Init ) ;
                end ;
                S := 'AT' ;
            end ;
        2 : S := 'Byte_Mapping = ' + Num1( _Instance.Byte_Mapping ) ;
        3 : S := 'Device = ' + Num1( _Instance.Device ) ;
        4 : S := 'Extension_Resolution = ' + Num1( _Instance.Extension_Resolution ) ;
        5 : begin
                dispose( R, Done ) ;
                if( _Instance.Store_Heap_Store = nil ) then
                begin
                    R := nil ;
                end else
                begin
                    R := PSText_Debugger( _Instance.Store_Heap_Store.Debugger ) ;
                end ;
                if( R = nil ) then
                begin
                    R := new( PSText_Debugger, Init ) ;
                end ;
                S := 'Store_Heap_Store = ' + Num1( longint( _Instance.Store_Heap_Store ) ) ;
            end ;
        6 : S := 'Flags = ' + Num1( _Instance.Flags ) ;
        7 : S := 'Lowest_Dynamic = ' + Num1( _Instance.Lowest_Dynamic ) ;
        8 : begin
                 dispose( R, Done ) ;
                 if( _Instance.Memory = nil ) then
                 begin
                     R := nil ;
                 end else
                 begin
                     R := PSText_Debugger( _Instance.Memory.Debugger ) ;
                 end ;
                 if( R = nil ) then
                 begin
                     R := new( PSText_Debugger, Init ) ;
                 end ;
                 S := 'Memory = ' + Num1( longint( _Instance.Memory ) ) ;
             end ;
        9 : S := 'Own_Allocator = ' + Boolean_To_String( _Instance.Own_Allocator ) ;
        10 : S := 'Resolution = ' + Num1( _Instance.Resolution ) ;
        11 : S := 'Serial_Number = ' + Num1( _Instance.Serial_Number ) ;
        12 : begin
                R.Free ;
                R := new( PHeap_Statistics_Debugger, Init( _Instance ) ) ;
                S := 'Statistics' ;
            end ;
  {$IFDEF Verify}
        13 : S := 'Track_FH = ' + Boolean_To_String( _Instance.Track_FH ) ;
  {$ENDIF}
    end ;
    R.Title := PChar( S ) ;
{$ELSE}
    new( R, Init ) ;
    case Ordinal of
        0 : S := 'Allow_Cache = ' + Boolean_To_String( _Instance^.Allow_Cache ) ;
        1 : begin
                R.Free ;
                R := PText_Debugger( _Instance^.AT.Debugger ) ;
                if( R = nil ) then
                begin
                    new( R, Init ) ;
                end ;
                S := 'AT' ;
            end ;
        2 : S := 'Byte_Mapping = ' + Num1( _Instance^.Byte_Mapping ) ;
        3 : S := 'Device = ' + Num1( _Instance^.Device ) ;
        4 : S := 'Extension_Resolution = ' + Num1( _Instance^.Extension_Resolution ) ;
        5 : begin
                dispose( R, done ) ;
                if( _Instance^.Store_Heap_Store = nil ) then
                begin
                    R := nil ;
                end else
                begin
                    R := PText_Debugger( _Instance^.Store_Heap_Store^.Debugger ) ;
                end ;
                if( R = nil ) then
                begin
                    new( R, Init ) ;
                end ;
                S := 'Store_Heap_Store = ' + Num1( longint( _Instance^.Store_Heap_Store ) ) ;
            end ;
        6 : S := 'Flags = ' + Num1( _Instance^.Flags ) ;
        7 : S := 'Lowest_Dynamic = ' + Num1( _Instance^.Lowest_Dynamic ) ;
        8 : begin
                dispose( R, done ) ;
                if( _Instance^.Memory = nil ) then
                begin
                    R := nil ;
                end else
                begin
                    R := PText_Debugger( _Instance^.Memory^.Debugger ) ;
                end ;
                if( R = nil ) then
                begin
                    new( R, Init ) ;
                end ;
                S := 'Memory = ' + Num1( longint( _Instance^.Memory ) ) ;
            end ;
        9 : S := 'Own_Allocator = ' + Boolean_To_String( _Instance^.Own_Allocator ) ;
        10 : S := 'Resolution = ' + Num1( _Instance^.Resolution ) ;
        11 : S := 'Serial_Number = ' + Num1( _Instance^.Serial_Number ) ;
        12 : begin
                dispose( R, done ) ;
                new( R, Init ) ;
                S := 'Statistics' ;
            end ;
  {$IFDEF Verify}
        13 : S := 'Track_FH = ' + Boolean_To_String( _Instance^.Track_FH ) ;
  {$ENDIF}
    end ;
    R^.Set_Title( @S[ 1 ] ) ;
{$ENDIF}
    Child := PSDebug_Interface( R ) ;
end ; { TStore_Heap_Debugger.Child }



{ TStore_Heap (store heap) methods... }

{ Internal utility routines }

procedure TStore_Heap.Log_Error( C, S : integer ) ;

var
{$IFDEF DELPHI}
    R : tUEC ;
{$ELSE}
    R : double ;
{$ENDIF}
    UEC : tUEC ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Last_Error.Facility := SHErr_Facility ;
    Last_Error.Code := C ;
    R := Error_Code( SHErr_Facility, C ) ;
    UEC := tUEC( R ) ;
    ErrMastr.Log_Error( UEC, '', S ) ;
    inc( Store_Statistics.Error_Count ) ;
end ;


{ API... }

procedure TStore_Heap.Determine_AT_Metrics( var UEC : TUEC ) ;

var Store_Size{, Temp} : TStore_Size ;
    {Dumm : byte ;}
{$IFNDEF DELPHI}
    Dummy : TRecord_Size ;
{$ENDIF}
    Header : TStore_Heap_Header ;
{    Size : TRecord_Size ;}

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    fillchar( UEC, sizeof( UEC ), 0 ) ;
    Store_Size := Store_Heap_Store^.Max_Storage ;
    if( ( Flags and SHF_NoExtend ) <> 0 ) then { Static store }
    begin
        AT.Set_Offset( Store_Size - ( Store_Size div ( Byte_Mapping + 1 ) ) ) ;
{$IFDEF DELPHI}
        AT.Set_Size( Store_Size - AT.Get_Offset ) ; { Size of allocation table }
{$ELSE}
        AT.Set_Size( trunc( Store_Size - AT.Get_Offset ) ) ; { Size of allocation table }
{$ENDIF}
        AT.Set_Modified( False ) ; { We are matching the in-memory AT to the on-store AT, so nothing is modified }
{$IFNDEF DELPHI}
        Dummy :=
{$ENDIF}
            Store_Heap_Store^.Read_Data( Header, Resolution, Sizeof( Header ), UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            exit ;
        end ;
        if(
            ( Header.AT_Size = 0 ) { Invalid AT size }
            or
            ( Header.AT_Size >= Store_Size ) { Ditto }
            or
            ( Header.AT_Offset < Resolution * 2 ) { Invalid AT offset }
            or
            ( ( Header.AT_Offset_High <> 0 ) and ( sizeof( Store_Size ) = 4 ) )
            or
            ( Header.ID <> -1 ) { Not a valid header }
            or
            ( Header.ID1 <> 1 ) { Ditto }
          ) then
        begin
            AT.Set_Size( Store_Size div Byte_Mapping ) ;
            AT.Set_Offset( Store_Size div 2 ) ; { Center the AT in the fixed store }
{$IFDEF DELPHI}
            AT.Set_Offset( ( ( AT.Get_Offset + Resolution - 1 ) div Resolution ) * Resolution ) ; { Align on resolution }
{            Temp := AT.Get_Offset ;}
{$ELSE}
            AT.Set_Offset( ( trunc( AT.Get_Offset + Resolution - 1 ) div Resolution ) * Resolution ) ; { Align on resolution }
{            Temp := trunc( AT.Get_Offset ) ;}
{$ENDIF}
(*
 EAC, 10-Aug-2002 - Shouldn't write to store yet - we may just be seeing if this IS a valid file system
            Size := Resolution ;
            if( Size > sizeof( SH_Buffer ) ) then
            begin
                Size := sizeof( SH_Buffer ) ;
            end ;
            for Dummy := 0 to AT.Get_Size div Size do
            begin
                Store_Heap_Store^.Write_Data( SH_Buffer, Temp, Size, UEC ) ;
                Temp := Temp + Size ;
            end ;
            Dumm := 3 ;
{$IFDEF DELPHI}
            Dummy := Store_Heap_Store^.Write_Data( Dumm, AT.Get_Offset, 1, UEC ) ; { Root and header are preallocated }
{$ELSE}
            Dummy := Store_Heap_Store^.Write_Data( Dumm, trunc( AT.Get_Offset ), 1, UEC ) ; { Root and header are preallocated }
{$ENDIF}
            Update_Header ;
*)
        end else
        begin
            AT.Set_Offset( Header.AT_Offset ) ;
            AT.Set_Size( Header.AT_Size ) ;
        end ;
    end else
    begin
        { Dynamic store... }
        AT.Set_Offset( Store_Size - ( Store_Size div ( Byte_Mapping + 1 ) ) ) ;
{$IFDEF DELPHI}
        AT.Set_Size( Store_Size - AT.Get_Offset ) ; { Size of allocation table }
{$ELSE}
        AT.Set_Size( trunc( Store_Size - AT.Get_Offset ) ) ; { Size of allocation table }
{$ENDIF}
        if( AT.Get_Offset > 0 ) then { Heap was previously written to }
        begin
{$IFNDEF DELPHI}
            Dummy :=
{$ENDIF}
                Store_Heap_Store^.Read_Data( Header, Resolution, Sizeof( Header ), UEC ) ;
            AT.Set_Offset( Header.AT_Offset ) ;
            AT.Set_Size( Header.AT_Size ) ;
            if( Store_Size > AT.Get_Offset + AT.Get_Size ) then
            begin
{$IFDEF DELPHI}
                Store_Heap_Store^.Set_Max_Storage( AT.Get_Offset + AT.Get_Size, UEC ) ;
{$ELSE}
                Store_Heap_Store^.Set_Max_Storage( trunc( AT.Get_Offset + AT.Get_Size ), UEC ) ;
{$ENDIF}
            end ;
        end ;
    end ;
end ; { TStore_Heap.Determine_AT_Metrics }


constructor TStore_Heap.Init( Minimum, F : integer ; const Name : string ;
    var _Result : tUEC ; Mem : PSManaged_Store ) ;
{ Initialize for FH access.  Note that the requested file MUST exist.  _Result
  is result of initialization.  if the file is non-zero-length or a device is
  being used, the file/device must have already been initialized with FH format.
  F is the initialization flags.  Name is the file name.  Minimum is the
  allocation resolution.  F is flags. }

begin
    inherited Init ;
    AT.Init( 16 ) ;

    { Open file/device }
    if( ( length( Name ) = 2 ) and ( Name[ 2 ] = ':' ) ) then { Device }
    begin
        Store_Heap_Store := System.New( PBlock_Device, Init ) ;
        PBlock_Device( Store_Heap_Store )^.Reset( Name ) ;
        if( PBlock_Device( Store_Heap_Store )^.IO_Error.Code <> 0 ) then
        begin
            _Result := PBlock_Device( Store_Heap_Store )^.IO_Error ;
            Last_Error := _Result ;
            exit ;
        end ;
        Minimum := Store_Heap_Store^.Min_Storage ;
        Device := -1 ; { Flag this as a device rather than a file }
    end else
    begin
        Store_Heap_Store := System.New( PSFile, Init( nil, false, 1 ) ) ;
        if( ( F and SHF_Read_Only ) <> 0 ) then
        begin
            PSFile( Store_Heap_Store )^.File_Mode := FSOM_Read_Only ;
        end ;
        PSFile( Store_Heap_Store )^.Reset( uppercase( Explicit( Name ) ) ) ;
        if( PSFile( Store_Heap_Store )^.IO_Error.Code  <> 0 ) then
        begin
            _Result := PSFile( Store_Heap_Store )^.IO_Error ;
            Last_Error := _Result ;
            exit ;
        end ;
        if( ( F and SHF_NoExtend ) <> 0 ) then
        begin
            PSFile( Store_Heap_Store )^.Extendable := False ;
        end ;
        Add_To_Store_Heap_List( @self ) ;
    end ;
    Device := 0 ;
    if( length( Name ) > 1 ) then
    begin
        if( Name[ 2 ] = ':' ) then
        begin
            Device := ord( Name[ 1 ] ) - 64 ;
        end ;
    end ;

    Common_Init( Minimum, F, Store_Heap_Store, _Result, Mem ) ;
end ; { TStore_Heap.Init }


constructor TStore_Heap.Init_On_Store( Minimum, F : integer ; S : PSStore ;
    var _Result : tUEC ; Mem : PSManaged_Store ) ;

begin
    inherited Init ;

    AT.Init( 16 ) ;
    Device := 0 ;
    Store_Heap_Store := nil ;
    Common_Init( Minimum, F, S, _Result, Mem ) ;
end ;


procedure TStore_Heap.Common_Init( Minimum, F : integer ; S : PSStore ;
    var _Result : tUEC ; Mem : PSManaged_Store ) ;

var Dummy : integer ;
    Found : boolean ;

begin
{$IFDEF Debug}
    Sanity := 'TStore_Heap' ;
{$ENDIF}

{$IFDEF Verify}
    Address_List := TInteger_List.Create ;
    Size_List := TInteger_List.Create ;
{$ENDIF}
    Memory := Mem ;
    Own_Allocator := False ;
    fillchar( Statistics, sizeof( Statistics ), 0 ) ;
    Allow_Cache := True ;
    Extension_Resolution := 0 ; { Extend only by amount necessary }
    Lowest_Dynamic := 0 ; { Allow write access to entire heap }

    Flags := F ;
    _Result.Facility := SHErr_Facility ;
    _Result.Code := SHErr_Bad_Minimum ;
    Last_Error := _Result ;
    Found := false ;
    Store_Heap_Store := nil ;
    if( Minimum = 0 ) then
    begin
        Minimum := 16 ; { Default resolution }
    end ;
    for Dummy := 0 to 15 do
    begin
        if( Minimum = Bit_Values[ Dummy ] ) then
        begin
            Found := true ;
        end ;
    end ;
    { See if a power of 2 from 1 to 32768. }

    if( not Found ) then
    begin
        exit ;
    end ;

    Store_Heap_Store := S ;

    if( Store_Heap_Store^.Extend( 0 ) = -1 ) then
    begin
        Flags := Flags or SHF_NoExtend ; { Store is not expandable }
    end ;
    if( Store_Heap_Store^.Read_Only ) then
    begin
        Flags := Flags or SHF_Read_Only ; { Store is read-only }
    end ;

    { Now set up various variables and read in AT segment 0 }

    Resolution := Minimum ; { Number of bytes allocated per AT bit }
    Byte_Mapping := 8 * Resolution ; { How much one AT byte maps }

    { Allocate AT memory }
    AT.Init( Resolution ) ;
    AT.Allow_Cache := Allow_Cache ;
    AT.Set_Store( Store_Heap_Store ) ;

    Determine_AT_Metrics( _Result ) ;
    if( _Result.Code <> 0 ) then
    begin
        exit ;
    end ;
    At.Load_Table ;

    _Result.Facility := SHErr_Facility ;
    _Result.Code := 0 ; { Setup was OK }
    Last_Error := _Result ;
    Serial_Number := Get_Serial_Number ;

    Initialize ;
end ; { TStore_Heap.Common_Init }


destructor TStore_Heap.Done ; { Finish up FH access }

begin
{$IFDEF Verify}
    Address_List.Free ;
    Size_List.Free ;
{$ENDIF}
    AT.Done ;
    if( ( ( Flags and SHF_NoClose ) = 0 ) and ( Store_Heap_Store <> nil ) ) then { Not prevented from closing the store heap }
    begin
        dispose( Store_Heap_Store, Done ) ;
        Store_Heap_Store := nil ;
    end ;

    { Release memory }
    if( Own_Allocator ) then
    begin
        dispose( Memory, Done ) ;
        Memory := nil ;
    end ;

    { Notify FH list }
    if( Device > 0 ) then
    begin
        Remove_From_Store_Heap_List( @self ) ;
    end ;

{$IFDEF Debug}
    Sanity := ' ' ;
{$ENDIF}

    { Inherited destructor }
    inherited Done ;
end ; { TStore_Heap.Done }


procedure TStore_Heap.Update_Header ;

var Header : TStore_Heap_Header ;
{$IFNDEF DELPHI}
    Dummy : TRecord_Size ;
{$ENDIF}
    UEC : TUEC ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    { Now write FH header... }
    fillchar( Header, sizeof( Header ), 0 ) ;
    Header.ID := -1 ;
    Header.ID1 := 1 ;
{$IFDEF DELPHI}
    Header.AT_Offset := AT.Get_Offset ;
{$ELSE}
    Header.AT_Offset := trunc( AT.Get_Offset ) ;
{$ENDIF}
    Header.AT_Size := AT.Get_Size ;
{$IFNDEF DELPHI}
    Dummy :=
{$ENDIF}
        Store_Heap_Store^.Write_Data( Header, Resolution, sizeof( Header ), UEC ) ;
end ;


function TStore_Heap.Path : string ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Path := Store_Heap_Store^.Get_Name ;
end ;


procedure TStore_Heap.New( _Size : longint ) ; { Create new FH, if empty }

var Counter : longint ;
    {Dummy : TRecord_Size ;}
    M : TStore_Address ;
    P : TStore_Address ;
    {UEC : TUEC ;}

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    { Determine initial file _Size and _Size of FAT }
    AT.Set_Offset( ( _Size + Byte_Mapping + Resolution * 2 ) and
        ( not ( Byte_Mapping - 1 ) ) ) ;
    { Round up to multiple of FH byte map (plus root and header) }
{$IFDEF DELPHI}
    AT.Set_Size( AT.Get_Offset div Byte_Mapping ) ; { Allocation table _Size }
{$ELSE}
    AT.Set_Size( trunc( AT.Get_Offset ) div Byte_Mapping ) ; { Allocation table _Size }
{$ENDIF}

    { Extend file up to, and including, start of AT... }
    Store_Heap_Store^.Extend( AT.Get_Offset + 1 ) ;

    { Write out allocation table... }
    AT.Allocate_At( 0, 2 * Resolution ) ;

    P := AT.Get_Offset + 1 ;
    Counter := AT.Get_Size - 1 ;
    M := Store_Heap_Store^.Max_Storage ;
    if( P + Counter > M ) then
    begin
        Store_Heap_Store^.Extend( ( P + Counter ) - M ) ; { Extend the store to include entire AT }
    end ;
(*
    while( Counter > 0 ) do
    begin
        if( Counter > sizeof( SH_Buffer ) ) then
        begin
            Dummy := Store_Heap_Store^.Write_Data( SH_Buffer, P, sizeof( SH_Buffer ), UEC ) ;
        end else
        begin
            Dummy := Store_Heap_Store^.Write_Data( SH_Buffer, P, Counter, UEC ) ;
        end ;
        P := P + Dummy ;
        Counter := Counter - sizeof( SH_Buffer ) ;
    end ;
*)
    inc( Statistics.Extensions ) ;
    AT.Flush ;
    Update_Header ;
end ; { TStore_Heap.New }


function TStore_Heap.Allocate( _Size : TStore_Address ) : TStore_Address ;
{ Allocate _Size bytes in store heap and return pointer.  Return 0 if error. }

var Actual_Size : longint ;
    {Dumm : TRecord_Size ;}
    Dummy, Dummy1{, Dummy2} : longint ; { Dummy variables }
    File_Size : longint ;
    Offset : TStore_Address ;
    {P : TStore_Address ;}
    {UEC : TUEC ;}
    {Write_Count : longint ; { Count of 128 byte sections to write at EOF }

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Allocate := 0 ; { Return zero (0) if error }
    if( _Size < 1 ) then { Cannot allocate nothing }
    begin
        exit ;
    end ;
    _Size := ( _Size + Resolution - 1 ) and ( not ( Resolution - 1 ) ) ;
    { Allocate minimum multiple bytes at a time }

    { Determine _Size for file extensions... }
    Actual_Size := _Size ;
    if( Extension_Resolution > 0 ) then
    begin
        File_Size := Store_Heap_Store^.Max_Storage ;
        Actual_Size := Extension_Resolution * File_Size div 100 ;
        if( Actual_Size <= 0 ) then { Handle overflow for big files }
        begin
            Actual_Size := Extension_Resolution * ( File_Size div 100 ) ;
        end ;
        if( Actual_Size < _Size ) then
        begin
            Actual_Size := _Size ;
        end ;
        if( Actual_Size > Max_Memory - File_Size ) then
        begin
            Actual_Size := Max_Memory - File_Size ;
        end ;
    end ; { if( Extension_Resolution > 0 ) }

    if( AT.Get_Offset = 0 ) then { File is empty }
    begin
        { Setup... }
        if( ( Flags and SHF_NoExtend ) <> 0 ) then { Cannot extend file }
        begin
            Log_Error( SHErr_No_Room_On_Device, 2 ) ;
            exit ;
        end ;
        New( Actual_Size ) ;
    end ; { if AT_Offset = 0 }

    Offset := AT.Allocate( _Size ) ;
    if( Offset = 0 ) then { No room in AT }
    begin
        { Need to expand file... }
        if( ( Flags and SHF_NoExtend ) <> 0 ) then { Cannot extend file }
        begin
            exit ;
        end ;

        { We always extend by at least the AT _Size so that we don't
          overwrite the original AT when we copy it (so we can recover
          if something goes wrong during the extend operation). }
        if( Actual_Size < AT.Get_Size ) then
        begin
            Actual_Size := AT.Get_Size ;
        end ;

        Dummy := ( Actual_Size + Byte_Mapping - 1 ) div Byte_Mapping ;
        { How big next AT segment will need to be (minimum) to provide space for
          current allocation request (and any additional extension amount).  It
          must be rounded up to a full byte. }

        { Physically extend the FH file... }
        Dummy1 := Dummy + Dummy * Byte_Mapping ; { Additional space needed in the store }
        if( Store_Heap_Store^.Extend( Dummy1 ) <> Dummy1 ) then { Failed to extend by required amount }
        begin
            Log_Error( SHErr_No_Room_On_Device, 2 ) ;
            exit ;
        end ;
(*
        P := Store_Heap_Store^.Max_Storage ; { Go to EOS }
        Dummy1 := Dummy ;
        while( Dummy1 > 0 ) do { Extend file, as needed }
        begin
            dec( Dummy1 ) ;
            { One less AT byte worth of space to extend }
            Store_Heap_Store^.Extend( 1 ) ;
            Dumm := Store_Heap_Store^.Write_Data( SH_Buffer, P, 1, UEC ) ;
            { Need a byte for AT }
            P := P + Dumm ;

            if( Byte_Mapping > sizeof( SH_Buffer ) ) then
            { Can only write out SH_Buffer zeroes at a time }
            begin
                Dummy2 := sizeof( SH_Buffer ) ;
            end else
            begin
                Dummy2 := Byte_Mapping ;
            end ;
            Write_Count := Byte_Mapping div Dummy2 ;
            while( Write_Count > 0 ) do
            begin
                Store_Heap_Store^.Extend( Dummy2 ) ;
                Dumm := Store_Heap_Store^.Write_Data( SH_Buffer, P, Dummy2, UEC ) ;
                { Need space for what AT byte can map }
                P := P + Dumm ;
                dec( Write_Count ) ;
            end ;
        end ; { while( Dummy1 > 0 ) }
*)

        AT.Set_Size( AT.Get_Size + Dummy ) ; { New AT Size }
	AT.Set_Offset( Store_Heap_Store^.Max_Storage - AT.Get_Size ) ; { New AT position }

        inc( Statistics.Extensions ) ;
        Update_Header ;
        Offset := AT.Allocate( _Size ) ;
    end ; { if( Offset = 0 ) }

    if( Offset < 2 * Resolution ) then
    begin
        { The only way we should get here is if the allocation table didn't have
          the first 2 Resolution bytes allocated, which indicates corruption of
          the AT. }
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( SHErr_Corrupted_File_Structure, 3 ) ;
        halt ;
    end ;
    Allocate := Offset ; { Return offset of allocation }
    AT.Flush ;
    Log_Allocation( Offset, _Size ) ;
end ; { TStore_Heap.Allocate }


{ Allocate space at a specific location }
function TStore_Heap.Allocate_At( _Position, _Size : TStore_Address ) : boolean ;

var Actual_Size : longint ;
    {Dumm : TRecord_Size ;}
    Dummy, Dummy1{, Dummy2} : longint ; { Dummy variables }
    File_Size : longint ;
    {P : TStore_Address ;}
    {UEC : TUEC ;}
    {Write_Count : longint ; { Count of 128 byte sections to write at EOF }

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Allocate_At := False ; { Return false if error }
    if( _Size < 1 ) then { Cannot allocate nothing }
    begin
        exit ;
    end ;
    _Size := ( _Size + Resolution - 1 ) and ( not ( Resolution - 1 ) ) ;
    { Allocate minimum multiple bytes at a time }

    { Determine _Size for file extensions... }
    Actual_Size := _Size ;
    if( Extension_Resolution > 0 ) then
    begin
        File_Size := Store_Heap_Store^.Max_Storage ;
        Actual_Size := Extension_Resolution * File_Size div 100 ;
        if( Actual_Size <= 0 ) then { Handle overflow for big files }
        begin
            Actual_Size := Extension_Resolution * ( File_Size div 100 ) ;
        end ;
        if( Actual_Size < _Size ) then
        begin
            Actual_Size := _Size ;
        end ;
        if( Actual_Size > Max_Memory - File_Size ) then
        begin
            Actual_Size := Max_Memory - File_Size ;
        end ;
    end ; { if( Extension_Resolution > 0 ) }

    if( AT.Get_Offset = 0 ) then { File is empty }
    begin
        { Setup... }
        if( ( Flags and SHF_NoExtend ) <> 0 ) then { Cannot extend file }
        begin
            Log_Error( SHErr_No_Room_On_Device, 2 ) ;
            exit ;
        end ;
        New( Actual_Size ) ;
    end ; { if AT_Offset = 0 }

    if( not AT.Allocate_At( _Position, _Size ) ) then { No room in AT }
    begin
        { Need to expand file... }
        if( ( Flags and SHF_NoExtend ) <> 0 ) then { Cannot extend file }
        begin
            exit ;
        end ;
        if( _Position + _Size < AT.Get_Offset ) then { Not due to store being too small }
        begin
            exit ; { Exit, otherwise we needlessly expand the store }
        end ;

        { We always extend by at least the AT _Size so that we don't
          overwrite the original AT when we copy it (so we can recover
          if something goes wrong during the extend operation). }
        if( Actual_Size < AT.Get_Size ) then
        begin
            Actual_Size := AT.Get_Size ;
        end ;

        Dummy := ( Actual_Size + Byte_Mapping - 1 ) div Byte_Mapping ;
        { How big next AT segment will need to be (minimum) to provide space for
          current allocation request (and any additional extension amount).  It
          must be rounded up to a full byte. }

        { Physically extend the FH file... }
        Dummy1 := Dummy + Dummy * Byte_Mapping ; { Additional space needed in the store }
        if( Store_Heap_Store^.Extend( Dummy1 ) <> Dummy1 ) then { Failed to extend by required amount }
        begin
            exit ;
        end ;
(*
        P := Store_Heap_Store^.Max_Storage ; { Go to EOS }
        Dummy1 := Dummy ;
        while( Dummy1 > 0 ) do { Extend file, as needed }
        begin
            dec( Dummy1 ) ;
            { One less AT byte worth of space to extend }
            Store_Heap_Store^.Extend( 1 ) ;
            Dumm := Store_Heap_Store^.Write_Data( SH_Buffer, P, 1, UEC ) ;
            { Need a byte for AT }
            P := P + Dumm ;

            if( Byte_Mapping > sizeof( SH_Buffer ) ) then
            { Can only write out SH_Buffer zeroes at a time }
            begin
                Dummy2 := sizeof( SH_Buffer ) ;
            end else
            begin
                Dummy2 := Byte_Mapping ;
            end ;
            Write_Count := Byte_Mapping div Dummy2 ;
            while( Write_Count > 0 ) do
            begin
                Store_Heap_Store^.Extend( Dummy2 ) ;
                Dumm := Store_Heap_Store^.Write_Data( SH_Buffer, P, Dummy2, UEC ) ;
                { Need space for what AT byte can map }
                P := P + Dumm ;
                dec( Write_Count ) ;
            end ;
        end ; { while( Dummy1 > 0 ) }
*)

        AT.Set_Size( AT.Get_Size + Dummy ) ; { New AT Size }
	AT.Set_Offset( Store_Heap_Store^.Max_Storage - AT.Get_Size ) ; { New AT position }

        inc( Statistics.Extensions ) ;
        Update_Header ;
        if( not AT.Allocate_At( _Position, _Size ) ) then
        begin
            exit ;
        end ;
    end ; { if( not AT.Allocate_At( _Position, _Size ) }

    Allocate_At := True ; { Return offset of allocation }
    AT.Flush ;
    Log_Allocation( _Position, _Size ) ;
end ; { TStore_Heap.Allocate_At }


procedure TStore_Heap.Deallocate( PTR, _Size : longint ) ; { Deallocate space in FH }

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    if( ( Flags and SHF_Fast_Deallocate ) <> 0 ) then
    begin
        _Size := 0 ; { Don't deallocate }
    end ;
    if( _Size = 0 ) then
    begin
        inc( Statistics.Deallocations ) ;
        exit ;
    end ;
    if(
        ( PTR < Resolution )
        or
        ( PTR < Lowest_Dynamic )
        or
        ( PTR + _Size > AT.Get_Offset )
        or
        ( _Size < 1 )
      ) then
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( SHErr_Invalid_Pointer_Operation, 3 ) ;
        halt ;
    end ;
    Log_Deallocation( PTR, _Size ) ;
    AT.Deallocate( PTR, _Size ) ;
end ; { FH.Deallocate }


procedure TStore_Heap.Read_Heap( PTR, _Size : longint ; var Buffer ) ; { Read data from FH }

var Dummy : TRecord_Size ;
    Len : longint ;
    UEC : TUEC ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    if( _Size = 0 ) then
    begin
        inc( Store_Statistics.Reads ) ;
        exit ; { Do nothing }
    end ;
    if(
        ( PTR < Resolution )
        or
        ( PTR + _Size > AT.Get_Offset )
        or
        ( _Size < 1 )
      ) then
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( SHErr_Invalid_Pointer_Operation, 3 ) ;
        halt ;
    end ;
    Log_Read( PTR, _Size ) ;
    if( _Cache <> nil ) then
    begin
        Len := _Cache^.Read_Cache( Buffer, PTR, _Size ) ;
        if( Len = _Size ) then { Data was found in cache }
        begin
            Store_Statistics.Bytes_Read := Store_Statistics.Bytes_Read + Len ;
            exit ;
        end ;
    end ;
    Dummy := Store_Heap_Store^.Read_Data( Buffer, PTR, _Size, UEC ) ;
    Store_Statistics.Bytes_Read := Store_Statistics.Bytes_Read + Dummy ;
end ; { TStore_Heap.Read_Heap }


function TStore_Heap.Reallocate( PTR, Old, _New : longint ) : longint ;
{ Reallocate a pointer }

var Dummy : longint ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Dummy := not ( Resolution - 1 ) ;
    if( ( ( Old - 1 ) and Dummy ) = ( ( _New - 1 ) and Dummy ) ) then
    { No actual reallocation is needed: new allocation fits within the old one
      due to minimum resolution. }
    begin
        Reallocate := PTR ;
        inc( Statistics.Reallocations_In_Place ) ;
    end else
    begin
        Reallocate := inherited Reallocate( PTR, Old, _New ) ;
    end ;
end ;


procedure TStore_Heap.Write_Heap( PTR, _Size : longint ; var Buffer ) ;
{ Write data to FH }

var Dummy : TRecord_Size ;
    UEC : TUEC ;
    
begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    if( _Size = 0 ) then
    begin
        inc( Store_Statistics.Writes ) ;
        exit ;
    end ;
    if(
        ( PTR < Resolution )
        or
        ( PTR < Lowest_Dynamic )
        or
        ( PTR + _Size > AT.Get_Offset )
        or
        ( _Size < 1 )
      ) then
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( SHErr_Invalid_Pointer_Operation, 3 ) ;
        halt ;
    end ;
    Log_Write( PTR, _Size ) ;
    Dummy := Store_Heap_Store^.Write_Data( Buffer, PTR, _Size, UEC ) ;
    Store_Statistics.Bytes_Written := Store_Statistics.Bytes_Written + Dummy ;

    if( _Cache <> nil ) then
    begin
        _Cache^.Write_Cache( Buffer, PTR, _Size ) ;
    end ;
end ; { TStore_Heap.Write_Heap }


procedure TStore_Heap.Set_Root( var Buffer ) ; { Write data to root block of FH }

var
{$IFNDEF DELPHI}
    Dumm : TRecord_Size ;
{$ENDIF}
    UEC : TUEC ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    if( AT.Get_Offset = 0 ) then { File is empty }
    begin
	New( Resolution ) ;
    end ;
    if( Device = -1 ) then { Reserve first block of device for bootstrap }
    begin
{$IFNDEF DELPHI}
        Dumm :=
{$ENDIF}
            Store_Heap_Store^.Write_Data( Buffer, Resolution, Resolution, UEC ) ;
    end else
    begin
{$IFNDEF DELPHI}
        Dumm :=
{$ENDIF}
            Store_Heap_Store^.Write_Data( Buffer, 0, Resolution, UEC ) ;
    end ;
    inc( Statistics.Root_Writes ) ;
end ;


procedure TStore_Heap.Get_Root( var Buffer ) ; {Read data from root block of FH}

var
{$IFNDEF DELPHI}
    Dummy : TRecord_Size ;
{$ENDIF}
    UEC : TUEC ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    if( AT.Get_Offset = 0 ) then { File is empty }
    begin
	fillchar( Buffer, Resolution, 0 ) ;
    end else
    begin
	if( Device = -1 ) then
	{ First block of device is reserved for bootstrap }
	begin
{$IFNDEF DELPHI}
	    Dummy :=
{$ENDIF}
                Store_Heap_Store^.Read_Data( Buffer, Resolution, Resolution, UEC ) ;
	end else
	begin
{$IFNDEF DELPHI}
	    Dummy :=
{$ENDIF}
                Store_Heap_Store^.Read_Data( Buffer, 0, Resolution, UEC ) ;
	end ;
	inc( Statistics.Root_Reads ) ;
    end ;
end ;


function TStore_Heap.XSpaceAvail : longint ;
{ Return available space TStore_Address store heap including file expansion }

var R, R1 : TStore_Address ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    R := SpaceAvail ; { Space available within store heap }
    if( Store_Heap_Store^.Extend( 0 ) = 0 ) then { Can extend }
    begin
        if( Diskfree( Device ) + R > OS.Max_File_Size ) then
        begin
            R1 := OS.Max_File_Size - Store_Heap_Store^.Max_Storage ; { Amount file can expand }
            R1 := R1 div ( ( Resolution * 8 ) + 1 ) * Resolution * 8 ; { Subtract overhead }
            R := R1 + R ;
        end else
        begin
            R := R + Diskfree( Device ) ;
        end ;
    end ;
{$WARNINGS OFF}
    if( ( R > $7FFFFFFF ) or ( R < 0 ) ) then { Larger value than fits in longint function result }
{$WARNINGS ON}
    begin
        R := $7FFFFFFF ;
    end ;
    XSpaceAvail := R ;
end ;


function TStore_Heap.SpaceAvail : longint ;
{ Return available space within store heap without expanding file }

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    SpaceAvail := AT.SpaceAvail ;
end ;


function TStore_Heap.MaxSpace : longint ;
{ Return largest contiguous space without extending file. }

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    MaxSpace := AT.MaxSpace ;
end ; { TStore_Heap.MaxSpace }


function TStore_Heap.Debugger : PSDebug_Interface ;

var R : PStore_Heap_Debugger ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    system.new( R, Init( @Self ) ) ;
    Debugger := R ;
end ;


procedure TStore_Heap.Set_Cache_Flag( Flag : boolean ) ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    if( Flag <> Allow_Cache ) then
    begin
	Allow_Cache := Flag ;
	if( not Flag ) then
	begin
	    if( AT.Get_Modified ) then
	    begin
		AT.Flush ;
	    end ;
	end ;
        AT.Allow_Cache := Flag ;
    end ;
end ;


function TStore_Heap.Extended_Size : TStore_Address ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Extended_Size := XSpaceAvail ;
end ;


function TStore_Heap.Valid_Data_Address( Address : TStore_Address ) : boolean ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Valid_Data_Address := ( Address >= Resolution ) and ( Address < AT.Get_Offset ) ;
end ;


procedure TStore_Heap.Write_Cache ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    AT.Flush ;
    inc( Store_Heap_Statistics.Index_Writes ) ;
end ;


procedure TStore_Heap.Log_Allocation( Address, _Size : longint ) ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    inc( Statistics.Allocations ) ;
    Statistics.Bytes_Allocated := Statistics.Bytes_Allocated + _Size ;
{$IFDEF Verify}
    if( Track_FH ) then
    begin
        Global_Address_List.Add( Address ) ;
        Global_Size_List.Add( _Size ) ;
    end ;
{$ENDIF}
end ;


procedure TStore_Heap.Log_Deallocation( Address, _Size : longint ) ;

{$IFDEF Verify}
var Dummy : integer ;
{$ENDIF}

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    inc( Statistics.Deallocations ) ;
    Statistics.Bytes_Deallocated := Statistics.Bytes_Deallocated + _Size ;
{$IFDEF Verify}
    if( Track_FH ) then
    begin
        Dummy := Global_Address_List.Indexof( Address ) ;
        if( Dummy = -1 ) then
        begin
            Log_Error( SHErr_Deallocation_Error, 2 ) ;
        end ;
        if( _Size <> Size_List[ Dummy ] ) then
        begin
            Log_Error( SHErr_Size_Mismatch, 2 ) ;
        end ;
    end ;
{$ENDIF}
end ; { TStore_Heap.Log_Deallocation }


procedure TStore_Heap.Log_Read( Address, _Size : longint ) ;

{$IFDEF Verify}
var Dummy : integer ;
{$ENDIF}

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    inc( Store_Statistics.Reads ) ;
{$IFDEF Verify}
    if( Track_FH ) then
    begin
        Dummy := Global_Address_List.Indexof( Address ) ;
        if( Dummy = -1 ) then
        begin
            Log_Error( SHErr_Read_Out_Of_Range, 2 ) ;
        end ;
        if( _Size <> Size_List[ Dummy ] ) then
        begin
            Log_Error( SHErr_Read_Size_Error, 2 ) ;
        end ;
    end ;
{$ENDIF}
end ;


procedure TStore_Heap.Log_Write( Address, _Size : longint ) ;

{$IFDEF Verify}
var Dummy : integer ;
{$ENDIF}

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    inc( Store_Statistics.Writes ) ;
{$IFDEF Verify}
    if( Track_FH ) then
    begin
        Dummy := Global_Address_List.Indexof( Address ) ;
        if( Dummy = -1 ) then
        begin
            Log_Error( SHErr_Write_Out_Of_Range, 2 ) ;
        end ;
        if( _Size <> Size_List[ Dummy ] ) then
        begin
            Log_Error( SHErr_Write_Size_Error, 2 ) ;
        end ;
    end ;
{$ENDIF}
end ;



var SH_List : TCollection ;

procedure Add_To_Store_Heap_List( X : pointer ) ;

begin
    SH_List.Insert( X ) ;
end ;


procedure Remove_From_Store_Heap_List( X : pointer ) ;

begin
    SH_List.Delete( X ) ;
end ;


procedure Flush_FH( T : pointer ; const X : string ) ;

var Dummy : integer ;
    P : PStore_Heap ;

begin
    for Dummy := 0 to SH_List.Count - 1 do
    begin
	P := PStore_Heap( SH_List.At( Dummy ) ) ;
	if( P <> T ) then { Don't recurse }
	begin
	    if( P^.AT.Get_Modified ) then
	    begin
		if( P^.Store_Heap_Store^.Get_Name = X ) then
		begin
		    P^.Write_Cache ;
		end ;
	    end ;
	end ;
    end ;
end ;


{ Unit initialization and finalization... }

{$IFDEF DELPHI}
var FHEI : pStore_Heap_EI = nil ;
{$ELSE}
var FHEI : pStore_Heap_EI ;
{$ENDIF}

procedure Unit_Initialization ;

begin
    { Set up SH_Buffer with zeroes.  This is so we can quickly blockwrite
      zeroes to the store heap. }
    fillchar( SH_Buffer, sizeof( SH_Buffer ), 0 ) ;

{$IFDEF Verify}
    Global_Address_List := TInteger_List.Create ;
    Global_Size_List := TInteger_List.Create ;
{$ENDIF}

    { Create store heap list. }
    SH_List.Init( 4, 4 ) ;

    { Create error interfaces... }
{$IFDEF DELPHI}
    FHEI := tStore_Heap_EI.Create ;
    Register_Error_Interface( PSError_Interface( FHEI ) ) ;
{$ELSE}
    new( FHEI, Init ) ;
    Register_Error_Interface( @FHEI ) ;
{$ENDIF}
end ; { Unit_Initialization }


procedure Unit_Deallocation ;

begin
    SH_List.Done ;
{$IFDEF Verify}
    Global_Address_List.Free ;
    Global_Size_List.Free ;
{$ENDIF}
{$IFDEF DELPHI}
    Unregister_Error_Interface( FHEI.Get_Facility ) ;
    FHEI.Free ;
    FHEI := nil ;
{$ELSE}
    Unregister_Error_Interface( FHEI^.Get_Facility ) ;
    dispose( FHEI ) ;
    FHEI := nil ;
{$ENDIF}
end ; { Unit_Deallocation }


{$IFDEF DELPHI}
initialization
{$ELSE}
begin
{$ENDIF}
    Unit_Initialization ;
{$IFDEF DELPHI}

finalization
    Unit_Deallocation ;
{$ENDIF}
end.
