{$N+}
{$E+}
unit Stores ;
{
        Program Name : Stores
        Package Name : TBase_Store
        Purpose      : Base Storage objects
        Institution  : Conroy & Conroy Co.
        Date Written : 1999
        Written By   : Alan Conroy
        Version      : 1.1B

        Copyright (C) 1999-2005 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  1-Nov-2002    EAC         Use TObject from Common.       *
        * 13-Nov-2002    EAC         Rename Tallocator to           *
        *                            TManaged_Store.                *
        * 16-Nov-2005    EAC         Changed const parameters to var*
        *                            to compile with BP.            *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TStore is the base class for all storage-related classes.  It
        implements the generic store functionality upon which descendant stores
        are built.
}

interface

{$I EDEFINES.INC}

uses { Layer 1... }
     _Cache, { PSCache }
     _DebugIn, { PSDebug_Interface }
     _Stores, { TSStore }
     _UEHDefs ; { TUEC }

const StoreErr_Facility = 48 ;

      StoreErr_Success = 0 ;
      StoreErr_Invalid_Operation = 1 ; { The operation is not valid }
      StoreErr_Abstract = 2 ;

type TSBase_Store = object( TSStore )
                        public { Instance data... } {protected}
                            _Cache : PSCache ;
                            _Read_Only : boolean ;
                            _Write_Only : boolean ;

                        public { API... }
                            function Read_Data( var Data ;
                                Address, _Size : TStore_Address ;
                                var UEC : TUEC ) : TStore_Address ; virtual ;
                            { Read data from specified Address in store to Data.
                              Read _Size bytes.  Returns success of operation in UEC.
                              Function returns the number of bytes actually read. }

                            function Write_Data( var Data ; Address,
                                _Size : TStore_Address ;
                                var UEC : TUEC ) : TStore_Address ; virtual ;
                            { Write data from Data to specified Address in store.
                              Write _Size bytes.  Returns success of operation in UEC.
                              Function returns the number of bytes actually written. }

                            function Max_Storage : TStore_Size ; virtual ;
                            { Returns the maximum (current) size of the store, in
                              bytes.  In other words, this is the highest valid
                              address (plus 1). }

                            function Min_Storage : TStore_Address ; virtual ;
                            { Returns the minimum atomic size of the store.  For
                              instance, this would be the sector size for a disk. }

                            function Extend( Amount : TStore_Address ) : TStore_Address ;
                                virtual ;
                            { Requests extension of store by the amount specified.
                              Returns the amount actually added to store.  If Amount
                              is 0, function returns 0 if store is extendable, and -1
                              if it is not. }

                            function Read_Only : boolean ; virtual ;
                            { Indicates that the store can only be read from. }

                            function Write_Only : boolean ; virtual ;
                            { Indicates that the store can only be written to. }

                            procedure Initialize ; virtual ;
                            { Initializes the TStore object. }

                            procedure Terminate ; virtual ;
                            { Terminates the TStore object. }

                            procedure Format ; virtual ;
                            { Does a low-level format of the store. }

                            function Get_Name : string ; virtual ;
                            { Returns name of store. }

                            function Debugger : PSDebug_Interface ; virtual ;
                            { Returns a debugger object. }

                            function Get_Cache : PSCache ; virtual ;
                            { Returns current cache object. }

                            procedure Set_Cache( Value : PSCache ) ; virtual ;
                            { Sets the current cache object. }

                            function Contiguous_Store : boolean ; virtual ;
                            { Returns True if the store's addresses are contiguous
                              (non-sparse). }

                            procedure Set_Max_Storage( Value : TStore_Address ;
                                var Res : TUEC ) ; virtual ;
                            { Sets the maximum (current) size of the store, in bytes.
                              In other words, this is the new highest valid address
                              (plus 1).  If the store is not resizable, the call
                              fails and an error code is returned in Res.  This should
                              only be used to shrink the size of the store, not to
                              extend it.  To extend, use the Extend method. }

                            function Extended_Size : TStore_Address ; virtual ;
                            { Returns the maximum theoretical size that the store can
                              be extended to.  If the store is not extendable, this
                              returns the same value as Max_Storage. }

                            procedure Set_Read_Only( Value : boolean ) ; virtual ;
                            { Sets the store read-only value. }

                            procedure Set_Write_Only( Value : boolean ) ; virtual ;
                            { Sets the store write-only value. }
                    end ; { TSBase_Store }
     PSBase_Store = ^TSBase_Store ;

{$IFDEF DELPHI}
type TSBase_Store64 = object( TSStore64 )
                          protected { Instance data... }
                              _Cache : PSCache ;
                              _Read_Only : boolean ;
                              _Write_Only : boolean ;

                          public { API... }
                              function Read_Data( var Data ;
                                  Address, _Size : TStore_Address64 ;
                                  var UEC : TUEC ) : TStore_Address64 ; virtual ;
                              { Read data from specified Address in store to Data.
                                Read _Size bytes.  Returns success of operation in UEC.
                                Function returns the number of bytes actually read. }

                              function Write_Data( var Data ; Address,
                                  _Size : TStore_Address64 ;
                                  var UEC : TUEC ) : TStore_Address64 ; virtual ;
                              { Write data from Data to specified Address in store.
                                Write _Size bytes.  Returns success of operation in UEC.
                                Function returns the number of bytes actually written. }

                              function Max_Storage : TStore_Size64 ; virtual ;
                              { Returns the maximum (current) size of the store, in
                                bytes.  In other words, this is the highest valid
                                address (plus 1). }

                              function Min_Storage : TStore_Address64 ; virtual ;
                              { Returns the minimum atomic size of the store.  For
                                instance, this would be the sector size for a disk. }

                              function Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
                                  virtual ;
                              { Requests extension of store by the amount specified.
                                Returns the amount actually added to store.  If Amount
                                is 0, function returns 0 if store is extendable, and -1
                                if it is not. }

                              function Read_Only : boolean ; virtual ;
                              { Indicates that the store can only be read from. }

                              function Write_Only : boolean ; virtual ;
                              { Indicates that the store can only be written to. }

                              procedure Initialize ; virtual ;
                              { Initializes the TStore object. }

                              procedure Terminate ; virtual ;
                              { Terminates the TStore object. }

                              procedure Format ; virtual ;
                              { Does a low-level format of the store. }

                              function Get_Name : string ; virtual ;
                              { Returns name of store. }

                              function Debugger : PSDebug_Interface ; virtual ;
                              { Returns a debugger object. }

                              function Get_Cache : PSCache ; virtual ;
                              { Returns current cache object. }

                              procedure Set_Cache( Value : PSCache ) ; virtual ;
                              { Sets the current cache object. }

                              function Contiguous_Store : boolean ; virtual ;
                              { Returns True if the store's addresses are contiguous
                                (non-sparse). }

                              procedure Set_Max_Storage( Value : TStore_Address64 ;
                                  var Res : TUEC ) ; virtual ;
                              { Sets the maximum (current) size of the store, in bytes.
                                In other words, this is the new highest valid address
                                (plus 1).  If the store is not resizable, the call
                                fails and an error code is returned in Res.  This should
                                only be used to shrink the size of the store, not to
                                extend it.  To extend, use the Extend method. }

                              function Extended_Size : TStore_Address64 ; virtual ;
                              { Returns the maximum theoretical size that the store can
                                be extended to.  If the store is not extendable, this
                                returns the same value as Max_Storage. }

                              procedure Set_Read_Only( Value : boolean ) ; virtual ;
                              { Sets the store read-only value. }

                              procedure Set_Write_Only( Value : boolean ) ; virtual ;
                              { Sets the store write-only value. }
                      end ; { TSBase_Store64 }
     PSBase_Store64 = ^TSBase_Store64 ;

type TBase_Store = class( TStore )
                      protected { Instance data... }
                          Store_Statistics : TStore_Statistics ;
                          _Cache : TCache ;
                          _Read_Only : boolean ;
                          _Write_Only : boolean ;

                      public { API... }
                          function Read_Data( var Data ; Address, _Size : TStore_Address ;
                              var UEC : TUEC ) : TStore_Address ; override ;
                          { Read data from specified Address in store to Data.
                            Read _Size bytes.  Returns success of operation in UEC.
                            Function returns the number of bytes actually read. }

                          function Write_Data( var Data ; Address, _Size : TStore_Address ;
                              var UEC : TUEC ) : TStore_Address ; override ;
                          { Write data from Data to specified Address in store.
                            Write _Size bytes.  Returns success of operation in UEC.
                            Function returns the number of bytes actually written. }

                          function Max_Storage : TStore_Size ; override ;
                          { Returns the maximum (current) size of the store, in
                            bytes.  In other words, this is the highest valid
                            address (plus 1). }

                          function Min_Storage : TStore_Address ;
                              override ;
                          { Returns the minimum atomic size of the store.  For
                            instance, this would be the sector size for a disk. }

                          function Extend( Amount : TStore_Address ) : TStore_Address ;
                              override ;
                          { Requests extension of store by the amount specified.
                            Returns the amount actually added to store.  If Amount
                            is 0, function returns 0 if store is extendable, and -1
                            if it is not. }

                          function Read_Only : boolean ; override ;
                          { Indicates that the store can only be read from. }

                          function Write_Only : boolean ; override ;
                          { Indicates that the store can only be written to. }

                          procedure Initialize ; override ;
                          { Initializes the TStore object. }

                          procedure Terminate ; override ;
                          { Terminates the TStore object. }

                          procedure Format ; override ;
                          { Does a low-level format of the store. }

                          function Get_Name : string ; override ;
                          { Returns name of store. }

                          function Get_Cache : TCache ; override ;
                          { Returns current cache object. }

                          procedure Set_Cache( Value : TCache ) ;
                              override ;
                          { Sets the current cache object. }

                          function Contiguous_Store : boolean ;
                              override ;
                          { Returns True if the store's addresses are contiguous
                            (non-sparse). }

                          procedure Set_Max_Storage( Value : TStore_Address ;
                              var Res : TUEC ) ; override ;
                          { Sets the maximum (current) size of the store, in bytes.
                            In other words, this is the new highest valid address
                            (plus 1).  If the store is not resizable, the call
                            fails and an error code is returned in Res.  This should
                            only be used to shrink the size of the store, not to
                            extend it.  To extend, use the Extend method. }

                          function Extended_Size : TStore_Address ;
                              override ;
                          { Returns the maximum theoretical size that the store can
                            be extended to.  If the store is not extendable, this
                            returns the same value as Max_Storage. }

                          procedure Set_Read_Only( Value : boolean ) ; override ;
                          { Sets the store read-only value. }

                          procedure Set_Write_Only( Value : boolean ) ; override ;
                          { Sets the store write-only value. }
                   end ;

type TBase_Store64 = class( TStore64 )
                         protected { Instance data... }
                             Store_Statistics : TStore_Statistics ;
                             _Cache : TCache ;
                             _Read_Only : boolean ;
                             _Write_Only : boolean ;

                         public { API... }
                             function Read_Data( var Data ; Address, _Size : TStore_Address64 ;
                                 var UEC : TUEC ) : TStore_Address64 ;
                                 override ;
                             { Read data from specified Address in store to Data.
                               Read _Size bytes.  Returns success of operation in UEC.
                               Function returns the number of bytes actually read. }

                             function Write_Data( var Data ; Address, _Size : TStore_Address64 ;
                                 var UEC : TUEC ) : TStore_Address64 ;
                                 override ;
                             { Write data from Data to specified Address in store.
                               Write _Size bytes.  Returns success of operation in UEC.
                               Function returns the number of bytes actually written. }

                             function Max_Storage : TStore_Size64 ;
                                 override ;
                             { Returns the maximum (current) size of the store, in
                               bytes.  In other words, this is the highest valid
                               address (plus 1). }

                             function Min_Storage : TStore_Address64 ;
                                 override ;
                             { Returns the minimum atomic size of the store.  For
                               instance, this would be the sector size for a disk. }

                             function Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
                                 override ;
                             { Requests extension of store by the amount specified.
                               Returns the amount actually added to store.  If Amount
                               is 0, function returns 0 if store is extendable, and -1
                               if it is not. }

                             function Read_Only : boolean ; override ;
                             { Indicates that the store can only be read from. }

                             function Write_Only : boolean ; override ;
                             { Indicates that the store can only be written to. }

                             procedure Initialize ; override ;
                             { Initializes the TStore object. }

                             procedure Terminate ; override ;
                             { Terminates the TStore object. }

                             procedure Format ; override ;
                             { Does a low-level format of the store. }

                             function Get_Name : string ; override ;
                             { Returns name of store. }

                             function Get_Cache : TCache ; override ;
                             { Returns current cache object. }

                             procedure Set_Cache( Value : TCache ) ;
                                 override ;
                             { Sets the current cache object. }

                             function Contiguous_Store : boolean ;
                                 override ;
                             { Returns True if the store's addresses are contiguous
                               (non-sparse). }

                             procedure Set_Max_Storage( Value : TStore_Address64 ;
                                 var Res : TUEC ) ; override ;
                             { Sets the maximum (current) size of the store, in bytes.
                               In other words, this is the new highest valid address
                               (plus 1).  If the store is not resizable, the call
                               fails and an error code is returned in Res.  This should
                               only be used to shrink the size of the store, not to
                               extend it.  To extend, use the Extend method. }

                             function Extended_Size : TStore_Address64 ;
                                 override ;
                             { Returns the maximum theoretical size that the store can
                               be extended to.  If the store is not extendable, this
                               returns the same value as Max_Storage. }

                            procedure Set_Read_Only( Value : boolean ) ;
                                override ;
                            { Sets the store read-only value. }

                            procedure Set_Write_Only( Value : boolean ) ;
                                override ;
                            { Sets the store write-only value. }
                     end ; { TBase_Store64 }

type TStandard_COM_Store = class( TCOM_Store )
                               private { Instance data... }
                                   Store_Statistics : TStore_Statistics ;
                                   _Cache : TCOM_Cache ;
                                   _Read_Only : boolean ;
                                   _Write_Only : boolean ;

                               public { API... }
                                   function Read_Data( var Data ;
                                       Address, _Size : TStore_Address ;
                                       var UEC : TUEC ) : TStore_Address ;
                                       override ;
                                   { Read data from specified Address in store to Data.
                                     Read _Size bytes.  Returns success of operation in UEC.
                                     Function returns the number of bytes actually read. }

                                   function Write_Data( var Data ;
                                       Address, _Size : TStore_Address ;
                                       var UEC : TUEC ) : TStore_Address ;
                                           override ;
                                   { Write data from Data to specified Address in store.
                                     Write _Size bytes.  Returns success of operation in UEC.
                                     Function returns the number of bytes actually written. }

                                   function Max_Storage : TStore_Size ;
                                       override ;
                                   { Returns the maximum (current) size of the store, in
                                     bytes.  In other words, this is the highest valid
                                     address (plus 1). }

                                   function Min_Storage : TStore_Address ;
                                       override ;
                                   { Returns the minimum atomic size of the store.  For
                                     instance, this would be the sector size for a disk. }

                                   function Extend( Amount : TStore_Address ) : TStore_Address ;
                                      override ;
                                   { Requests extension of store by the amount specified.
                                     Returns the amount actually added to store.  If Amount
                                     is 0, function returns 0 if store is extendable, and -1
                                     if it is not. }

                                   function Read_Only : boolean ;
                                       override ;
                                   { Indicates that the store can only be read from. }

                                   function Write_Only : boolean ;
                                       override ;
                                   { Indicates that the store can only be written to. }

                                   procedure Format ; override ;
                                   { Does a low-level format of the store. }

                                   function Get_Name : PChar ; override ;
                                   { Returns name of store. }

                                   function Get_Cache : TCOM_Cache ;
                                       override ;
                                   { Returns current cache object. }

                                   procedure Set_Cache( Value : TCOM_Cache ) ;
                                       override ;
                                   { Sets the current cache object. }

                                   function Contiguous_Store : boolean ;
                                       override ;
                                   { Returns True if the store's addresses are contiguous
                                     (non-sparse). }

                                   procedure Set_Max_Storage( Value : TStore_Address ;
                                       var Res : TUEC ) ; override ;
                                   { Sets the maximum (current) size of the store, in bytes.
                                     In other words, this is the new highest valid address
                                     (plus 1).  If the store is not resizable, the call
                                     fails and an error code is returned in Res.  This should
                                     only be used to shrink the size of the store, not to
                                     extend it.  To extend, use the Extend method. }

                                   function Extended_Size : TStore_Address ;
                                       override ;
                                   { Returns the maximum theoretical size that the store can
                                     be extended to.  If the store is not extendable, this
                                     returns the same value as Max_Storage. }

                                   function Get_Bytes_Read : longint ;
                                       override ; { Data bytes read }
                                   function Get_Bytes_Written : longint ;
                                       override ; { Data bytes written }
                                   function Get_Reads : longint ;
                                       override ; { Data read count }
                                   function Get_Writes : longint ;
                                       override ; { Data write count }
                                   function Get_Error_Count : longint ;
                                       override ; { Error count }
                                   procedure Set_Bytes_Read( Value : longint ) ;
                                       override ; { Data bytes read }
                                   procedure Set_Bytes_Written( Value : longint ) ;
                                       override ; { Data bytes written }
                                   procedure Set_Reads( Value : longint ) ;
                                       override ; { Data read count }
                                   procedure Set_Writes( Value : longint ) ;
                                       override ; { Data write count }
                                   procedure Set_Error_Count( Value : longint ) ;
                                       override ; { Error count }
                                   procedure Set_Read_Only( Value : boolean ) ;
                                       override ;
                                   { Sets the store read-only value. }
                                   procedure Set_Write_Only( Value : boolean ) ;
                                       override ;
                                   { Sets the store write-only value. }
                           end ;

type TStandard_COM_Store64 = class( TCOM_Store64 )
                                 private { Instance data... }
                                     Store_Statistics : TStore_Statistics ;
                                     _Cache : TCOM_Cache ;
                                     _Read_Only : boolean ;
                                     _Write_Only : boolean ;

                                 public { API... }
                                     function Read_Data( var Data ; Address, _Size : TStore_Address64 ;
                                         var UEC : TUEC ) : TStore_Address64 ;
                                         override ;
                                     { Read data from specified Address in store to Data.
                                       Read _Size bytes.  Returns success of operation in UEC.
                                       Function returns the number of bytes actually read. }

                                     function Write_Data( var Data ; Address, _Size : TStore_Address64 ;
                                         var UEC : TUEC ) : TStore_Address64 ;
                                         override ;
                                     { Write data from Data to specified Address in store.
                                       Write _Size bytes.  Returns success of operation in UEC.
                                       Function returns the number of bytes actually written. }

                                     function Max_Storage : TStore_Size64 ;
                                         override ;
                                     { Returns the maximum (current) size of the store, in
                                       bytes.  In other words, this is the highest valid
                                       address (plus 1). }

                                     function Min_Storage : TStore_Address64 ;
                                         override ;
                                     { Returns the minimum atomic size of the store.  For
                                       instance, this would be the sector size for a disk. }

                                     function Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
                                         override ;
                                     { Requests extension of store by the amount specified.
                                       Returns the amount actually added to store.  If Amount
                                       is 0, function returns 0 if store is extendable, and -1
                                       if it is not. }

                                     function Read_Only : boolean ;
                                         override ;
                                     { Indicates that the store can only be read from. }

                                     function Write_Only : boolean ;
                                         override ;
                                     { Indicates that the store can only be written to. }

                                     procedure Format ; override ;
                                     { Does a low-level format of the store. }

                                     function Get_Name : PChar ; override ;
                                     { Returns name of store. }

                                     function Get_Cache : TCOM_Cache ;
                                         override ;
                                     { Returns current cache object. }

                                     procedure Set_Cache( Value : TCOM_Cache ) ;
                                         override ;
                                     { Sets the current cache object. }

                                     function Contiguous_Store : boolean ;
                                         override ;
                                     { Returns True if the store's addresses are contiguous
                                       (non-sparse). }

                                     procedure Set_Max_Storage( Value : TStore_Address64 ;
                                         var Res : TUEC ) ;
                                         override ;
                                     { Sets the maximum (current) size of the store, in bytes.
                                       In other words, this is the new highest valid address
                                       (plus 1).  If the store is not resizable, the call
                                       fails and an error code is returned in Res.  This should
                                       only be used to shrink the size of the store, not to
                                       extend it.  To extend, use the Extend method. }

                                     function Extended_Size : TStore_Address64 ;
                                         override ;
                                     { Returns the maximum theoretical size that the store can
                                       be extended to.  If the store is not extendable, this
                                       returns the same value as Max_Storage. }

                                   function Get_Bytes_Read : longint ;
                                       override ; { Data bytes read }
                                   function Get_Bytes_Written : longint ;
                                       override ; { Data bytes written }
                                   function Get_Reads : longint ;
                                       override ; { Data read count }
                                   function Get_Writes : longint ;
                                       override ; { Data write count }
                                   function Get_Error_Count : longint ;
                                       override ; { Error count }
                                   procedure Set_Bytes_Read( Value : longint ) ;
                                       override ; { Data bytes read }
                                   procedure Set_Bytes_Written( Value : longint ) ;
                                       override ; { Data bytes written }
                                   procedure Set_Reads( Value : longint ) ;
                                       override ; { Data read count }
                                   procedure Set_Writes( Value : longint ) ;
                                       override ; { Data write count }
                                   procedure Set_Error_Count( Value : longint ) ;
                                       override ; { Error count }
                                   procedure Set_Read_Only( Value : boolean ) ;
                                       override ;
                                   { Sets the store read-only value. }
                                   procedure Set_Write_Only( Value : boolean ) ;
                                       override ;
                                   { Sets the store write-only value. }
                             end ; { TStandard_COM_Store64 }
{$ENDIF}

implementation

{ TSBase_Store methods... }

{ API... }

function TSBase_Store.Read_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : TUEC ) : TStore_Address ;

begin
    Read_Data := 0 ;
    UEC.Facility := StoreErr_Facility ;
    UEC.Code := StoreErr_Abstract ;
end ;


function TSBase_Store.Write_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : TUEC ) : TStore_Address ;

begin
    Write_Data := 0 ;
    UEC.Facility := StoreErr_Facility ;
    UEC.Code := StoreErr_Abstract ;
end ;


function TSBase_Store.Max_Storage : TStore_Size ;

begin
    Max_Storage := 0 ;
end ;


function TSBase_Store.Min_Storage : TStore_Address ;

begin
    Min_Storage := 1 ;
end ;


function TSBase_Store.Extend( Amount : TStore_Address ) : TStore_Address ;

begin
    if( Amount = 0 ) then
    begin
        Extend := -1 ; { Cannot extend }
    end else
    begin
        Extend := 0 ; { Nothing added because we cannot extend }
    end ;
end ;


function TSBase_Store.Read_Only : boolean ;

begin
    Read_Only := _Read_Only ;
end ;


function TSBase_Store.Write_Only : boolean ;

begin
    Write_Only := _Write_Only ;
end ;


procedure TSBase_Store.Initialize ;

begin
    _Cache := nil ;
    _Read_Only := False ;
    _Write_Only := False ;
    fillchar( Store_Statistics, sizeof( Store_Statistics ), 0 ) ;
end ;


procedure TSBase_Store.Terminate ;

begin
    { This routine left intentionally blank. }
end ;


procedure TSBase_Store.Format ;

begin
    { This routine left intentionally blank. }
end ;


function TSBase_Store.Get_Name : string ;

begin
    Get_Name := '' ;
end ;


function TSBase_Store.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ;
end ;


function TSBase_Store.Get_Cache : PSCache ;

begin
    Get_Cache := _Cache ;
end ;


procedure TSBase_Store.Set_Cache( Value : PSCache ) ;

begin
    if( Value <> _Cache ) then
    begin
        if( Value <> nil ) then
        begin
            Value^.Attach ;
        end ;
        if( _Cache <> nil ) then
        begin
            _Cache^.Terminate ;
            _Cache^.Initialize ;
            _Cache^.Detach ;
        end ;
        _Cache := Value ;
    end ;
end ;


function TSBase_Store.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ; { Assume not }
end ;


procedure TSBase_Store.Set_Max_Storage( Value : TStore_Address ; var Res : TUEC ) ;

begin
    Res.Facility := StoreErr_Facility ;
    Res.Code := StoreErr_Invalid_Operation ;
end ;


function TSBase_Store.Extended_Size : TStore_Address ;

begin
    Extended_Size := Max_Storage ;
end ;


procedure TSBase_Store.Set_Read_Only( Value : boolean ) ;

begin
    _Read_Only := Value ;
end ;


procedure TSBase_Store.Set_Write_Only( Value : boolean ) ;

begin
    _Write_Only := Value ;
end ;


{$IFDEF DELPHI}
{ TSBase_Store64 methods... }

{ API... }

function TSBase_Store64.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUEC ) : TStore_Address64 ;

begin
    Read_Data := 0 ;
    UEC.Facility := StoreErr_Facility ;
    UEC.Code := StoreErr_Abstract ;
end ;


function TSBase_Store64.Write_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUEC ) : TStore_Address64 ;

begin
    Write_Data := 0 ;
    UEC.Facility := StoreErr_Facility ;
    UEC.Code := StoreErr_Abstract ;
end ;


function TSBase_Store64.Max_Storage : TStore_Size64 ;

begin
    Max_Storage := 0 ;
end ;


function TSBase_Store64.Min_Storage : TStore_Address64 ;

begin
    Min_Storage := 1 ;
end ;


function TSBase_Store64.Extend( Amount : TStore_Address64 ) : TStore_Address64 ;

begin
    if( Amount = 0 ) then
    begin
        Extend := -1 ; { Cannot extend }
    end else
    begin
        Extend := 0 ; { Nothing added because we cannot extend }
    end ;
end ;


function TSBase_Store64.Read_Only : boolean ;

begin
    Read_Only := _Read_Only ;
end ;


function TSBase_Store64.Write_Only : boolean ;

begin
    Write_Only := _Write_Only ;
end ;


procedure TSBase_Store64.Initialize ;

begin
    _Cache := nil ;
    _Read_Only := False ;
    _Write_Only := False ;
    fillchar( Store_Statistics, sizeof( Store_Statistics ), 0 ) ;
end ;


procedure TSBase_Store64.Terminate ;

begin
    { This routine left intentionally blank. }
end ;


procedure TSBase_Store64.Format ;

begin
    { This routine left intentionally blank. }
end ;


function TSBase_Store64.Get_Name : string ;

begin
    Get_Name := '' ;
end ;


function TSBase_Store64.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ;
end ;


function TSBase_Store64.Get_Cache : PSCache ;

begin
    Get_Cache := _Cache ;
end ;


procedure TSBase_Store64.Set_Cache( Value : PSCache ) ;

begin
    if( Value <> _Cache ) then
    begin
        if( Value <> nil ) then
        begin
            Value^.Attach ;
        end ;
        if( _Cache <> nil ) then
        begin
            _Cache^.Terminate ;
            _Cache^.Initialize ;
            _Cache^.Detach ;
        end ;
        _Cache := Value ;
    end ;
end ;


function TSBase_Store64.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ; { Assume not }
end ;


procedure TSBase_Store64.Set_Max_Storage( Value : TStore_Address64 ; var Res : TUEC ) ;

begin
    Res.Facility := StoreErr_Facility ;
    Res.Code := StoreErr_Invalid_Operation ;
end ;


function TSBase_Store64.Extended_Size : TStore_Address64 ;

begin
    Extended_Size := Max_Storage ;
end ;


procedure TSBase_Store64.Set_Read_Only( Value : boolean ) ;

begin
    _Read_Only := Value ;
end ;


procedure TSBase_Store64.Set_Write_Only( Value : boolean ) ;

begin
    _Write_Only := Value ;
end ;



{ TBase_Store methods... }

{ API... }

function TBase_Store.Read_Data( var Data ; Address, _Size : TStore_Address ;
  var UEC : TUEC ) : TStore_Address ;

begin
    Read_Data := 0 ;
    UEC.Facility := StoreErr_Facility ;
    UEC.Code := StoreErr_Abstract ;
end ;


function TBase_Store.Write_Data( var Data ; Address, _Size : TStore_Address ;
  var UEC : TUEC ) : TStore_Address ;

begin
    Write_Data := 0 ;
    UEC.Facility := StoreErr_Facility ;
    UEC.Code := StoreErr_Abstract ;
end ;


function TBase_Store.Max_Storage : TStore_Size ;

begin
    Max_Storage := 0 ;
end ;


function TBase_Store.Min_Storage : TStore_Address ;

begin
    Min_Storage := 1 ;
end ;


function TBase_Store.Extend( Amount : TStore_Address ) : TStore_Address ;

begin
    if( Amount = 0 ) then
    begin
        Extend := -1 ; { Cannot extend }
    end else
    begin
        Extend := 0 ; { Nothing added because we cannot extend }
    end ;
end ;


function TBase_Store.Read_Only : boolean ;

begin
    Read_Only := False ;
end ;


function TBase_Store.Write_Only : boolean ;

begin
    Write_Only := False ;
end ;


procedure TBase_Store.Initialize ;

begin
    { This routine left intentionally blank. }
end ;


procedure TBase_Store.Terminate ;

begin
    Free ;
end ;


procedure TBase_Store.Format ;

begin
    { This routine left intentionally blank. }
end ;


function TBase_Store.Get_Name : string ;

begin
    Get_Name := '' ;
end ;


function TBase_Store.Get_Cache : TCache ;

begin
    Get_Cache := _Cache ;
end ;


procedure TBase_Store.Set_Cache( Value : TCache ) ;

begin
    if( Value <> _Cache ) then
    begin
        if( Value <> nil ) then
        begin
            Value.Attach ;
        end ;
        if( _Cache <> nil ) then
        begin
            _Cache.Terminate ;
            _Cache.Initialize ;
            _Cache.Detach ;
        end ;
        _Cache := Value ;
    end ;
end ;


function TBase_Store.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ; { Assume not }
end ;


procedure TBase_Store.Set_Max_Storage( Value : TStore_Address ;
    var Res : TUEC ) ;

begin
    Res.Facility := StoreErr_Facility ;
    Res.Code := StoreErr_Invalid_Operation ;
end ;


function TBase_Store.Extended_Size : TStore_Address ;

begin
    Extended_Size := Max_Storage ;
end ;


procedure TBase_Store.Set_Read_Only( Value : boolean ) ;

begin
    _Read_Only := Value ;
end ;


procedure TBase_Store.Set_Write_Only( Value : boolean ) ;

begin
    _Write_Only := Value ;
end ;



{ TBase_Store64 methods... }

{ API... }

function TBase_Store64.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
   var UEC : TUEC ) : TStore_Address64 ;

begin
    Read_Data := 0 ;
    UEC.Facility := StoreErr_Facility ;
    UEC.Code := StoreErr_Abstract ;
end ;


function TBase_Store64.Write_Data( var Data ; Address, _Size : TStore_Address64 ;
   var UEC : TUEC ) : TStore_Address64 ;

begin
    Write_Data := 0 ;
    UEC.Facility := StoreErr_Facility ;
    UEC.Code := StoreErr_Abstract ;
end ;


function TBase_Store64.Max_Storage : TStore_Size64 ;

begin
    Max_Storage := 0 ;
end ;


function TBase_Store64.Min_Storage : TStore_Address64 ;

begin
    Min_Storage := 1 ;
end ;


function TBase_Store64.Extend( Amount : TStore_Address64 ) : TStore_Address64 ;

begin
    if( Amount = 0 ) then
    begin
        Extend := -1 ; { Cannot extend }
    end else
    begin
        Extend := 0 ; { Nothing added because we cannot extend }
    end ;
end ;


function TBase_Store64.Read_Only : boolean ;

begin
    Read_Only := False ;
end ;


function TBase_Store64.Write_Only : boolean ;

begin
    Write_Only := False ;
end ;


procedure TBase_Store64.Initialize ;

begin
    { This routine left intentionally blank. }
end ;


procedure TBase_Store64.Terminate ;

begin
    Free ;
end ;


procedure TBase_Store64.Format ;

begin
    { This routine left intentionally blank. }
end ;


function TBase_Store64.Get_Name : string ;

begin
    Get_Name := '' ;
end ;


function TBase_Store64.Get_Cache : TCache ;

begin
    Get_Cache := _Cache ;
end ;


procedure TBase_Store64.Set_Cache( Value : TCache ) ;

begin
    if( Value <> _Cache ) then
    begin
        if( Value <> nil ) then
        begin
            Value.Attach ;
        end ;
        if( _Cache <> nil ) then
        begin
            _Cache.Terminate ;
            _Cache.Initialize ;
            _Cache.Detach ;
        end ;
        _Cache := Value ;
    end ;
end ;


function TBase_Store64.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ; { Assume not }
end ;


procedure TBase_Store64.Set_Max_Storage( Value : TStore_Address64 ;
    var Res : TUEC ) ;

begin
    Res.Facility := StoreErr_Facility ;
    Res.Code := StoreErr_Invalid_Operation ;
end ;


function TBase_Store64.Extended_Size : TStore_Address64 ;

begin
    Extended_Size := Max_Storage ;
end ;


procedure TBase_Store64.Set_Read_Only( Value : boolean ) ;

begin
    _Read_Only := Value ;
end ;


procedure TBase_Store64.Set_Write_Only( Value : boolean ) ;

begin
    _Write_Only := Value ;
end ;



{ TStandard_COM_Store methods... }

{ API... }

function TStandard_COM_Store.Read_Data( var Data ;
    Address, _Size : TStore_Address ; var UEC : TUEC ) : TStore_Address ;

begin
    Read_Data := 0 ;
    UEC.Facility := StoreErr_Facility ;
    UEC.Code := StoreErr_Abstract ;
end ;


function TStandard_COM_Store.Write_Data( var Data ;
    Address, _Size : TStore_Address ; var UEC : TUEC ) : TStore_Address ;

begin
    Write_Data := 0 ;
    UEC.Facility := StoreErr_Facility ;
    UEC.Code := StoreErr_Abstract ;
end ;


function TStandard_COM_Store.Max_Storage : TStore_Size ;

begin
    Max_Storage := 0 ;
end ;


function TStandard_COM_Store.Min_Storage : TStore_Address ;

begin
    Min_Storage := 1 ;
end ;


function TStandard_COM_Store.Extend( Amount : TStore_Address ) : TStore_Address ;

begin
    if( Amount = 0 ) then
    begin
        Extend := -1 ; { Cannot extend }
    end else
    begin
        Extend := 0 ; { Nothing added because we cannot extend }
    end ;
end ;


function TStandard_COM_Store.Read_Only : boolean ;

begin
    Read_Only := False ;
end ;


function TStandard_COM_Store.Write_Only : boolean ;

begin
    Write_Only := False ;
end ;


procedure TStandard_COM_Store.Format ;

begin
    { This routine left intentionally blank. }
end ;


function TStandard_COM_Store.Get_Name : PChar ;

begin
    Get_Name := nil ;
end ;


function TStandard_COM_Store.Get_Cache : TCOM_Cache ;

begin
    Get_Cache := _Cache ;
end ;


procedure TStandard_COM_Store.Set_Cache( Value : TCOM_Cache ) ;

begin
    if( Value <> _Cache ) then
    begin
        if( Value <> nil ) then
        begin
            Value.Attach ;
        end ;
        if( _Cache <> nil ) then
        begin
            _Cache.Terminate ;
            _Cache.Initialize ;
            _Cache.Detach ;
        end ;
        _Cache := Value ;
    end ;
end ;


function TStandard_COM_Store.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ; { Assume not }
end ;


procedure TStandard_COM_Store.Set_Max_Storage( Value : TStore_Address ;
    var Res : TUEC ) ;

begin
    Res.Facility := StoreErr_Facility ;
    Res.Code := StoreErr_Invalid_Operation ;
end ;


function TStandard_COM_Store.Extended_Size : TStore_Address ;

begin
    Extended_Size := Max_Storage ;
end ;


function TStandard_COM_Store.Get_Bytes_Read : longint ;

begin
    Result := Store_Statistics.Bytes_Read ;
end ;


function TStandard_COM_Store.Get_Bytes_Written : longint ;

begin
    Result := Store_Statistics.Bytes_Written ;
end ;


function TStandard_COM_Store.Get_Reads : longint ;

begin
    Result := Store_Statistics.Reads ;
end ;


function TStandard_COM_Store.Get_Writes : longint ;

begin
    Result := Store_Statistics.Writes ;
end ;


function TStandard_COM_Store.Get_Error_Count : longint ;

begin
    Result := Store_Statistics.Error_Count ;
end ;


procedure TStandard_COM_Store.Set_Bytes_Read( Value : longint ) ;

begin
    Store_Statistics.Bytes_Read := Value ;
end ;


procedure TStandard_COM_Store.Set_Bytes_Written( Value : longint ) ;

begin
    Store_Statistics.Bytes_Written := Value ;
end ;


procedure TStandard_COM_Store.Set_Reads( Value : longint ) ;

begin
    Store_Statistics.Reads := Value ;
end ;


procedure TStandard_COM_Store.Set_Writes( Value : longint ) ;

begin
    Store_Statistics.Writes := Value ;
end ;


procedure TStandard_COM_Store.Set_Error_Count( Value : longint ) ;

begin
    Store_Statistics.Error_Count := Value ;
end ;


procedure TStandard_COM_Store.Set_Read_Only( Value : boolean ) ;

begin
    _Read_Only := Value ;
end ;


procedure TStandard_COM_Store.Set_Write_Only( Value : boolean ) ;

begin
    _Write_Only := Value ;
end ;



{ TStandard_COM_Store64 methods... }

{ API... }

function TStandard_COM_Store64.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUEC ) : TStore_Address64 ;

begin
    Read_Data := 0 ;
    UEC.Facility := StoreErr_Facility ;
    UEC.Code := StoreErr_Abstract ;
end ;


function TStandard_COM_Store64.Write_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUEC ) : TStore_Address64 ;

begin
    Write_Data := 0 ;
    UEC.Facility := StoreErr_Facility ;
    UEC.Code := StoreErr_Abstract ;
end ;


function TStandard_COM_Store64.Max_Storage : TStore_Size64 ;

begin
    Max_Storage := 0 ;
end ;


function TStandard_COM_Store64.Min_Storage : TStore_Address64 ;

begin
    Min_Storage := 1 ;
end ;


function TStandard_COM_Store64.Extend( Amount : TStore_Address64 ) : TStore_Address64 ;

begin
    if( Amount = 0 ) then
    begin
        Extend := -1 ; { Cannot extend }
    end else
    begin
        Extend := 0 ; { Nothing added because we cannot extend }
    end ;
end ;


function TStandard_COM_Store64.Read_Only : boolean ;

begin
    Read_Only := False ;
end ;


function TStandard_COM_Store64.Write_Only : boolean ;

begin
    Write_Only := False ;
end ;


procedure TStandard_COM_Store64.Format ;

begin
    { This routine left intentionally blank. }
end ;


function TStandard_COM_Store64.Get_Name : PChar ;

begin
    Get_Name := nil ;
end ;


function TStandard_COM_Store64.Get_Cache : TCOM_Cache ;

begin
    Get_Cache := _Cache ;
end ;


procedure TStandard_COM_Store64.Set_Cache( Value : TCOM_Cache ) ;

begin
    if( Value <> _Cache ) then
    begin
        if( Value <> nil ) then
        begin
            Value.Attach ;
        end ;
        if( _Cache <> nil ) then
        begin
            _Cache.Terminate ;
            _Cache.Initialize ;
            _Cache.Detach ;
        end ;
        _Cache := Value ;
    end ;
end ;


function TStandard_COM_Store64.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ; { Assume not }
end ;


procedure TStandard_COM_Store64.Set_Max_Storage( Value : TStore_Address64 ;
     var Res : TUEC ) ;

begin
    Res.Facility := StoreErr_Facility ;
    Res.Code := StoreErr_Invalid_Operation ;
end ;


function TStandard_COM_Store64.Extended_Size : TStore_Address64 ;

begin
    Extended_Size := Max_Storage ;
end ;


function TStandard_COM_Store64.Get_Bytes_Read : longint ;

begin
    Result := Store_Statistics.Bytes_Read ;
end ;


function TStandard_COM_Store64.Get_Bytes_Written : longint ;

begin
    Result := Store_Statistics.Bytes_Written ;
end ;


function TStandard_COM_Store64.Get_Reads : longint ;

begin
    Result := Store_Statistics.Reads ;
end ;


function TStandard_COM_Store64.Get_Writes : longint ;

begin
    Result := Store_Statistics.Writes ;
end ;


function TStandard_COM_Store64.Get_Error_Count : longint ;

begin
    Result := Store_Statistics.Error_Count ;
end ;


procedure TStandard_COM_Store64.Set_Bytes_Read( Value : longint ) ;

begin
    Store_Statistics.Bytes_Read := Value ;
end ;


procedure TStandard_COM_Store64.Set_Bytes_Written( Value : longint ) ;

begin
    Store_Statistics.Bytes_Written := Value ;
end ;


procedure TStandard_COM_Store64.Set_Reads( Value : longint ) ;

begin
    Store_Statistics.Reads := Value ;
end ;


procedure TStandard_COM_Store64.Set_Writes( Value : longint ) ;

begin
    Store_Statistics.Writes := Value ;
end ;


procedure TStandard_COM_Store64.Set_Error_Count( Value : longint ) ;

begin
    Store_Statistics.Error_Count := Value ;
end ;
{$ENDIF}


procedure TStandard_COM_Store64.Set_Read_Only( Value : boolean ) ;

begin
    _Read_Only := Value ;
end ;


procedure TStandard_COM_Store64.Set_Write_Only( Value : boolean ) ;

begin
    _Write_Only := Value ;
end ;


end.
