{$N+}
{
        Program Name : Tables
        Package Name : LB
        Purpose      : Manipulate tables
        Institution  : Conroy & Conroy Co.
        Date Written : 14-Jan-1994
        Written By   : Alan Conroy
        Version      : 1.3

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  31-May-95	 EAC         Added ability for table lists. *
        *                                                           *
        *  01-Jun-95     EAC         Added multiplier ability.      *
        *                                                           *
        *  20-Jul-95	 EAC	     Allow roll equations.          *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit Tables ;

interface

{$I EDEFINES.INC}

uses Common, { TBase_Object }
{$IFNDEF DELPHI}
     Objects, { TObject }
{$ENDIF}
     Files ;

type PEntry_Record = ^TEntry_Record ;
     TEntry_Record = record
                         Low, High : Integer ;
                         Next : PEntry_Record ;
                         Txt : string ;
                     end ;
     PTable_Record = ^TTable_Record ;
     TTable_Record = record
                         Table_ID : Longint ;
                         Roll : PChar ;
			 Indirection : Integer ;
                         First_Entry : PEntry_Record ;
                         Next : PTable_Record ;
                     end ;

     PTable = ^TTable ;
     TTable = object( TBase_Object )

                  Minimum_Heap : Longint ;
                  Root : PTable_Record ;

                  constructor Init( MinFree : Longint ) ;
		  procedure Find_Table( Table : Longint ;
			Inp : String ; var Current : PTable_Record ) ; virtual ;
                  function Get_Entry( Table, Ent : Longint;
                      Inp : String ) : String ; virtual ;
                  function Generate_Entry( Table : Longint;
                      Inp : String ) : String ; virtual ;
                  procedure Free_Memory ; virtual ;
                  destructor Done ; virtual ;

              end ;

implementation

uses Cache,
{$IFDEF DELPHI}
     SysUtils, { StrPCopy }
{$ELSE}
     Strings, { StrPCopy }
{$ENDIF}
     CommonUt, { Edit }
     Express,
     NUM1S,
     Standard, { MaxAvail }
     _UEHDefs ; { TUEC }

procedure Append_String( var Result, Temp : string ) ;
{Append Temp to Result, using ellipses for overflow.  Ignore if Temp is null}

begin
    if( length( edit( Temp, -1 ) ) = 0 ) then {Nothing to append}
    begin
	exit ;
    end ;
    if( length( Result ) + length( Temp ) > 255 ) then
    {Result is too large}
    begin
	if( length( Result ) < 253 ) then
	begin
	    Result := Result + '...' ; {Indicate overflow}
	end ;
    end else
    begin
	if( length( Result ) > 0 ) then
	begin
	    Result := Result + ', ' ;
	end ;
	Result := Result + Temp ;
    end ;
end ;


function Die( Count, Size : integer ) : integer ;
{Generate a random number}

var Res : integer ;

begin
    Res := 0 ;
    while Count > 0 do
    begin
	Res := Res + random( Size ) + 1 ;
	Count := Count - 1 ;
    end ;
    Die := Res ;
end ;


function Do_Die( X : string ) : integer ;

var Dummy, Dummy0, Dummy1, Dummy2 : integer ;

begin
    Do_Die := -1 ;
    Fedit( X, X, -1 ) ;
    Dummy := pos( 'D', X ) ;
    val( copy( X, 1, Dummy - 1 ), Dummy1, Dummy0 ) ;
    if( Dummy0 <> 0 ) then
    begin
	exit ;
    end ;
    val( copy( X, Dummy + 1, 255 ), Dummy2, Dummy0 ) ;
    if( Dummy0 <> 0 ) then
    begin
	exit ;
    end ;
    Do_Die := Die( Dummy1, Dummy2 ) ;
end ;



{TTable methods}

constructor TTable.Init( MinFree : Longint ) ;

begin
    Minimum_Heap := MinFree ;
    Root := nil ;
end ;


procedure TTable.Find_Table( Table : Longint ; Inp : string ;
    var Current : PTable_Record ) ;
{ Return record for specified table number and file }

var Dummy : Longint ;
    Dummy1 : Integer ;
    Current_Entry : PEntry_Record ;
    F : PSFile ;
    R : TEntry_Record ;
    S : string ;
    T : TTable_Record ;
    UEC : TUEC ;

begin
    Current := Root ;
    while( Current <> nil ) do
    begin
        if( Table = Current^.Table_ID ) then { Found the entry in memory }
        begin
	    exit ;
        end ;
        Current := Current^.Next ; { Follow link }
    end ;

    { If we get here, the entry is not in memory, so find it in file }
    Current_Entry := nil ;
    S := NUM1( Table ) + chr( 9 ) ;
    F := Open( Inp, FSOM_Text, UEC ) ;
    while( not F^.EOF ) do { Until end of file is reached }
    begin
        F^.Readln( R.Txt ) ;
        if( copy( R.Txt, 1, length( S ) ) = S ) then { Found table }
        begin
            Current := Root ;
            if( Current <> nil ) then
	    begin
                while( Current^.Next <> nil ) do
		begin
                    Current := Current^.Next ; {Find last item in list}
		end ;
	    end ;
            T.Table_ID := Table ;
            T.Next := nil ;
            F^.Readln( R.TXT ) ;
            fedit( R.TXT, R.TXT, -1 ) ;
	    T.Indirection := 0 ;
	    T.Roll := nil ;
            if R.TXT[ 1 ] = '@' then {Table redirection}
            begin
                T.First_Entry := nil ;
                Val( copy( R.TXT, 2, 254 ), T.Indirection, Dummy1 ) ;
                if( Root = nil ) then
                begin
                    New( Root ) ;
                    Root^ := T ;
		    Current := Root ;
                end else
                begin
		    if( MaxAvail < sizeof( T ) ) then
		    begin
			Free_Cache( sizeof( T ), $4FFFFFFF, nil ) ;
		    end ;
                    while( ( Root <> nil ) and ( MaxAvail < sizeof( T ) ) ) do
		    begin
			Free_Memory ;
		    end ;
                    New( Current^.Next ) ;
                    Current^.Next^ := T ;
                end ;
                dispose( F, Done ) ;
                exit ;
            end ;
	    getmem( T.Roll, length( R.TXT ) + 1 ) ;
	    strPCopy( T.Roll, R.TXT ) ;
            T.First_Entry := nil ;
	    if( MaxAvail < sizeof( T ) ) then
	    begin
		Free_Cache( sizeof( T ), $4FFFFFFF, nil ) ;
	    end ;
            while( ( Root <> nil ) and ( MaxAvail < sizeof( T ) ) ) do
	    begin
		Free_Memory ;
	    end ;
            if( Root = nil ) then { First table in memory }
            begin
                New( Root ) ;
                Root^ := T ;
                Current := Root ;
            end else
            begin
                New( Current^.Next ) ; { Add new table to end of list }
                Current^.Next^ := T ;
                Current := Current^.Next ;
            end ;
            { At this point, Current points to the table header }

            while( length( R.TXT ) <> 0 ) do { Until end of table }
            begin
                if( F^.EOF ) then { End of File is also end of table }
                begin
                    R.TXT := '' ;
                end else
                begin
                    F^.Readln( R.TXT ) ;
                    if( length( R.TXT ) > 0 ) then { Not end of table }
                    begin
                        Dummy := pos( chr( 9 ), R.TXT ) ;
                        S := copy( R.TXT, 1, Dummy - 1 ) ;
                        R.TXT := Copy( R.TXT, Dummy + 1, 254 ) ;
                        Dummy := pos( '-', S ) ;
                        Val( copy( S, Dummy + 1, 254 ), R.High, Dummy1 ) ;
                        if( Dummy = 0 )
                            then
                                R.Low := R.High
                            else
                                Val( copy( S, 1, Dummy - 1 ), R.Low, Dummy1 ) ;
                        if( R.High = 0 ) then
                        begin
                            Val( '1' + copy( S, Dummy + 1, 254 ), R.High,
                                Dummy1 ) ;
                        end ;
                        R.Next := nil ;
                        if( Current^.First_Entry = nil ) then { First entry }
                        begin
                            GetMem( Current^.First_Entry, 9 +
                                length( R.TXT ) ) ;
                            Current^.First_Entry^.Next := nil ;
                            Current^.First_Entry^.Low := R.Low ;
                            Current^.First_Entry^.High := R.High ;
{$IFDEF DELPHI}
                            Initialize( Current^.First_Entry^.Txt ) ;
{$ENDIF}
                            Current^.First_Entry^.Txt := R.Txt ;
                            Current_Entry := Current^.First_Entry ;
                        end else
                        begin
                            GetMem( Current_Entry^.Next, 9 +
                                length( R.TXT ) ) ;
                            Current_Entry := Current_Entry^.Next ;
                            Current_Entry^.Next := nil ;
                            Current_Entry^.Low := R.Low ;
                            Current_Entry^.High := R.High ;
{$IFDEF DELPHI}
                            Initialize( Current^.First_Entry^.Txt ) ;
{$ENDIF}
                            Current_Entry^.Txt := R.Txt ;
                        end ; { if Current^.First_Entry = nil }
                    end ; { if length( R.TXT ) > 0 }
                end ; { if F.EOF }
            end ; { while length( R.TXT ) <> 0 }
            dispose( F, Done ) ;
            exit ;
        end ; {if copy( R.Txt, 1, length( S ) ) = S}
	{Skip down to end of this table, since it is not a match}
        while ( not F^.EOF ) and ( length( R.Txt ) > 0 ) do F^.Readln( R.Txt ) ;
    end ; {while not F.EOF do}
    Current := nil ; {Table not found}
    dispose( F, Done ) ;
end ; { TTable.Find_Table }


function TTable.Get_Entry( Table, Ent : Longint ; Inp : String ) : String ;
{Get a specific entry from a table}

var Current : PTable_Record ;
    Dummy : Longint ;
    Dummy1, Dummy2 : Integer ;
    Entry : PEntry_Record ;
    Res, Temp, Work : string ;

begin
    Get_Entry := ' ' ;
    Find_Table( Table, Inp, Current ) ;
    if( Current <> nil ) then
    begin
        Entry := Current^.First_Entry ;
        if( Entry = nil ) then {Table redirection}
        begin
	    if Current^.Indirection = 0
		then
		    Get_Entry := 'use Standard Encounter Tables'
		else
	            Get_Entry := Generate_Entry( Current^.Indirection, Inp ) ;
            exit ;
        end ;
        while( Entry <> nil ) do
        begin
            if ( Entry^.Low <= Ent ) and ( Entry^.High >= Ent ) then
            begin
		if Entry^.Txt = '@0' then {Special case}
		begin
		    Get_Entry := 'use Standard Encounter Tables' ;
		    exit ;
		end ;
		Res := '' ;
		Work := Entry^.TXT ;
                while( ( length( Work ) > 0 ) and ( Work[ 1 ] = '@' ) ) do
                begin
		    Dummy2 := pos( ',', Work + ',' ) ;
                    Val( copy( Work, 2, Dummy2 - 2 ), Dummy, Dummy1 ) ;
		    if( Dummy1 > 1 ) then
		    begin
			if( upcase( Work[ Dummy1 + 1 ] ) = 'X' ) then
			begin
			    Temp := copy( Work, Dummy1 + 2, Dummy2 - Dummy1 - 2 ) ;
	                    Val( copy( Work, 2, Dummy1 - 1 ), Dummy, Dummy1 ) ;
			    Dummy1 := Do_Die( Temp ) ;
			    while( Dummy1 > 0 ) do
			    begin
				Temp := Generate_Entry( Dummy, Inp ) ;
				Append_String( Res, Temp ) ;
				Dummy1 := Dummy1 - 1 ;
			    end ;
			end ;
		    end ;
                    if Dummy1 <> 0 {Invalid table indirection}
                        then
                            Temp := ''
                        else
                            Temp := Generate_Entry( Dummy, Inp ) ;
		    Work := copy( Work, Dummy2 + 1, 254 ) ;
		    Append_String( Res, Temp ) ;
		end ;
		if( length( Work ) > 0 ) then {Something leftover}
		begin
		    Append_String( Res, Work ) ;
		end ;
		if( length( Res ) = 0 ) then
		begin
                    Get_Entry := Entry^.Txt ;
		end else
		begin
		    Get_Entry := Res ;
		end ;
                exit ;
            end ;
            Entry := Entry^.Next ; {Follow link to next entry}
        end ;
        Get_Entry := ' ' ; { Entry not found }
        exit ;
    end ;
end ; { TTable.Get_Entry }


Function TTable.Generate_Entry( Table : Longint; Inp : String ) : String ;
{Get a random entry from a table}

var Current : PTable_Record ;
    Dummy, Dummy1 : extended ;
    Dummy2, Dummy3 : integer ;

begin
    Find_Table( Table, Inp, Current ) ;
    if( Current = nil ) then
    begin
	Generate_Entry := '' ; { Not found }
    end else
    begin
	Eval( strpas( Current^.Roll ), Dummy, Dummy1, Dummy2, Dummy3 ) ;
	Generate_Entry := Get_Entry( Table, trunc( Dummy ), Inp ) ;
    end ;
end ;


procedure TTable.Free_Memory ; { Free first table in list }

var Current, Temp : PEntry_Record ;

begin
    if( Root <> nil ) then
    begin
	strdispose( Root^.Roll ) ;
	Current := Root^.First_Entry ;
	while( Current <> nil ) do
	begin
	    Temp := Current ;
	    Current := Current^.Next ;
	    dispose( Temp ) ; { TODO: Need Free due to morphable size (see getmem usage) }
	end ;
	Root := Root^.Next ;
    end ;
end ;


destructor TTable.Done ;

begin
    while( Root <> nil ) do
    begin
	Free_Memory ; { Free memory until done }
    end ;
end ;



end.
