{
        Program Name : _Streams
        Package Name : Subroutine Library - Layer 0
        Purpose      : Stream object definitions
        Institution  : Conroy & Conroy Co.
        Date Written : 2002
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2002-2005 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit _Streams ;

interface

{$I EDEFINES.INC}

uses ComInter, { TCommon_COM_Object }
     _Common, { TObject }
     _UEHDefs ; { TUEC }

{$IFDEF DELPHI}
type TCOM_Stream = class( TCommon_COM_Interface )
                       public { API... }
                          { Returns True if there is no more data in the stream to
                            read. }
                          function At_End : boolean ;
                              virtual ; stdcall ; abstract ;

                          { Returns last error code, and resets the error code. }
                          procedure Last_Error( var UEC : TUEC ) ;
                              virtual ; stdcall ; abstract ;

                          { Reads the specified number of bytes from the stream.
                            Size is modified to be the actual bytes transferred. }
                          procedure Read( var Buffer ; var _Size : longint ) ;
                              virtual ; stdcall ; abstract ;

                          { Reads one line of input, up to the size of the buffer.
                            The line is assumed to end at an ASCII 13 code.  The
                            ASCII 13 code is not included in the returned data.
                            Size is modified to be the actual bytes transferred. }
                          procedure Read_Line( var Buffer ; var _Size : longint ) ;
                              virtual ; stdcall ; abstract ;

                          { Position to specified byte within streamed data. }
                          procedure Seek( Position : longint ) ;
                              virtual ; stdcall ; abstract ;

                          { Returns size of stream data, in bytes.  -1 indicate that
                            the size is unknown or larger than 2^31. }
                          function Size : longint ; virtual ;
                              stdcall ; abstract ;

                          { Writes the specified buffer, of the specified size in
                            bytes, to the stream. }
                          procedure Write( var Buffer ; _Size : longint ) ;
                              virtual ; stdcall ; abstract ;

                          { Writes the specified null-terminated text to the stream.
                            An ASCII code 13 is appended to the text on output. }
                          procedure Write_Line( Buffer : PChar ) ;
                              virtual ; stdcall ; abstract ;
                   end ;

type TStream = class( TDelphi_Object )
                   public { API... }
                      { Returns True if there is no more data in the stream to
                        read. }
                      function At_End : boolean ; virtual ; abstract ;

                      { Returns last error code, and resets the error code. }
                      procedure Last_Error( var UEC : TUEC ) ;
                          virtual ; abstract ;

                      { Reads the specified number of bytes from the stream.
                        Size is modified to be the actual bytes transferred. }
                      procedure Read( var Buffer ; var _Size : longint ) ;
                          virtual ; abstract ;

                      { Reads one line of input, up to the size of the buffer.
                        The line is assumed to end at an ASCII 13 code.  The
                        ASCII 13 code is not included in the returned data.
                        Size is modified to be the actual bytes transferred. }
                      procedure Read_Line( var Buffer ; var _Size : longint ) ;
                          virtual ; abstract ;

                      { Position to specified byte within streamed data. }
                      procedure Seek( Position : longint ) ;
                          virtual ; abstract ;

                      { Returns size of stream data, in bytes.  -1 indicate that
                        the size is unknown or larger than 2^31. }
                      function Size : longint ; virtual ; abstract ;

                      { Writes the specified buffer, of the specified size in
                        bytes, to the stream. }
                      procedure Write( const Buffer ; _Size : longint ) ;
                          virtual ; abstract ;

                      { Writes the specified null-terminated text to the stream.
                        An ASCII code 13 is appended to the text on output. }
                      procedure Write_Line( Buffer : PChar ) ;
                          virtual ; abstract ;
               end ;
{$ENDIF}

type TSStream = object( TObject )
                   public { API... }
                      { Returns True if there is no more data in the stream to
                        read. }
                      function At_End : boolean ; virtual ;

                      { Returns last error code, and resets the error code. }
                      procedure Last_Error( var UEC : TUEC ) ; virtual ;

                      { Reads the specified number of bytes from the stream.
                        Size is modified to be the actual bytes transferred. }
                      procedure Read( var Buffer ; var _Size : longint ) ;
                          virtual ;

                      { Reads one line of input, up to the size of the buffer.
                        The line is assumed to end at an ASCII 13 code.  The
                        ASCII 13 code is not included in the returned data.
                        Size is modified to be the actual bytes transferred. }
                      procedure Read_Line( var Buffer ; var _Size : longint ) ;
                          virtual ;

                      { Reads one line of input.  The line is assumed to end at
                        an ASCII 13 code.  The ASCII 13 code is not included in
                        the returned data. }
                      function Read_String : string ; virtual ;

                      { Position to specified byte within streamed data. }
                      procedure Seek( Position : longint ) ; virtual ;

                      { Writes the specified buffer, of the specified size in
                        bytes, to the stream. }
                      procedure Write( var Buffer ; _Size : longint ) ;
                          virtual ;

                      { Writes the specified string to the stream. }
                      procedure Write_String( const Buffer : string ) ;

                      { Writes the specified null-terminated text to the stream.
                        An ASCII code 13 is appended to the text on output. }
                      procedure Write_Line( Buffer : PChar ) ; virtual ;

                      { Writes the specified string to the stream.
                        An ASCII code 13 is appended to the text on output. }
                      procedure Write_Line_String( const Buffer : string ) ;
                          virtual ;
               end ; { TSStream }
     PSStream = ^TSStream ;

implementation

{ TSStream methods... }

function TSStream.At_End : boolean ;

begin
    At_End := True ;
    Abstract_Error ;
end ;


procedure TSStream.Last_Error( var UEC : TUEC ) ;

begin
    fillchar( UEC, sizeof( UEC ), 0 ) ;
    Abstract_Error ;
end ;


procedure TSStream.Read( var Buffer ; var _Size : longint ) ;

begin
    _Size := 0 ;
    Abstract_Error ;
end ;


procedure TSStream.Read_Line( var Buffer ; var _Size : longint ) ;

begin
    _Size := 0 ;
    Abstract_Error ;
end ;


function TSStream.Read_String : string ;

begin
    Abstract_Error ;
end ;


procedure TSStream.Seek( Position : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSStream.Write( var Buffer ; _Size : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSStream.Write_String( const Buffer : string ) ;

begin
    Abstract_Error ;
end ;


procedure TSStream.Write_Line( Buffer : PChar ) ;

begin
    Abstract_Error ;
end ;


procedure TSStream.Write_Line_String( const Buffer : string ) ;

begin
    Abstract_Error ;
end ;



end.
