.RADIX  10
.8086

;              Program Name : EDIT
;              Package Name : LB
;              Purpose      : Edit function
;              Institution  : Conroy & Conroy Company
;              Date Written : 01-Nov-86
;              Written By   : Alan Conroy
;              Version      : 9.2-10
;
;	 Copyright (C) 1986 by Alan Conroy.  Released to the public domain.
;
;	 TO THE GLORY OF GOD THROUGH JESUS CHRIST
;
;        *********************************************************
;        *                                                       *
;        *        M O D I F I C A T I O N   H I S T O R Y        *
;        *                                                       *
;        *     DATE      BY             REASON                   *
;        *                                                       *
;        *********************************************************
;
;        *********************************************************
;        *                                                       *
;        *           P R O G R A M   P U R P O S E               *
;        *                                                       *
;        *********************************************************
;
; This routine edits strings.  Flags are in CX.  Input string is the input
; buffer and the output string is the work area (io.asm).  Flags:
;
; AND     1 = Reserved
; AND     2 = Remove all spaces and tabs
; AND     4 = Remove all nulls, linefeeds, formfeeds, carriage returns, deletes
; AND     8 = Remove leading spaces
; AND    16 = Reduce spaces and tabs to one space
; AND    32 = Convert lower to upper case
; AND    64 = Convert '(' to '[' and ')' to ']'
; AND   128 = Remove all trailing spaces
; AND   256 = Leave quoted characters unmodified
; AND   512-16384 = Reserved
;

EDIT:   PUSH  AX ; Save registers
        PUSH  BX
        PUSH  CX
        PUSH  DX
        PUSH  SI
        PUSH  DI
        MOV   DI, 0    ; Start at first byte in target string
        MOV   SI, DI   ; Start at first byte in source string
        MOV   BX, OFFSET WorkBLen ; Get length of destination string
        MOV   AL, 0
        MOV   [BX], AL ; Zero length of destination
        MOV   BX, OFFSET InBLen ; Get length of source string
        MOV   DL, [BX] ; Save in working register
        MOV   BX, OFFSET Flags
        MOV   AX, CX ; Get flags
        MOV   [BX], AX ; Save flags
        MOV   BX, OFFSET Last
        MOV   BYTE PTR [BX], 0  ; Last=0
        MOV   BX, OFFSET Leading
        MOV   BYTE PTR [BX], 1 ; Leading_Space=true
        MOV   BX, OFFSET Space
        MOV   BYTE PTR [BX], 0 ; Space=False
        MOV   BX, OFFSET Quotes
        MOV   BYTE PTR [BX], 0 ; Quotes=False

Edit0:  ; Beginning of edit loop
        INC   SI       ; Increment source offset
        MOV   BX, OFFSET InBuff
        MOV   AL, [BX+SI] ; Get next character
        AND   AL, 127  ; Strip off high bit
        MOV   BX, OFFSET Quotes
        CMP   BYTE PTR [BX], 0  ; Are quotes false
        JE    Edit0a   ; They are false
        JMP   EditOK   ; They are true

Edit0a: ; See if end of leading spaces
        CMP   AL, 32   ; Space?
        JE    Edit1    ; Yup
        CMP   AL, 9    ; Tab?
        JE    Edit1    ; Yup
        MOV   BX, OFFSET Leading
        MOV   BYTE PTR [BX], 0  ; Leading_Space=False
        JMP   Edit2    ; Since not a tab or space, ignore the next code

Edit1:  MOV   BX, OFFSET Flags
        TEST  WORD PTR [BX], 2  ; Ignore all tabs and spaces?
        JZ    Edit2    ; No
        JMP   EditLoop ; Yes

Edit2:  MOV   BX, OFFSET Flags
        TEST  WORD PTR [BX], 4  ; Discard LF, VT, FF, CR, ESC
        JZ    Edit3    ; Nope
        CMP   AL, 10   ; LF?
        JNE   Edit2a   ; No
        JMP   EditLoop ; Yes
Edit2a: CMP   AL, 11   ; VT?
        JNE   Edit2b   ; No
        JMP   EditLoop ; Yes
Edit2b: CMP   AL, 12   ; FF?
        JNE   Edit2c   ; No
        JMP   EditLoop ; Yes
Edit2c: CMP   AL, 13   ; CR?
        JNE   Edit2d   ; No
        JMP   EditLoop ; Yes
Edit2d: CMP   AL, 27   ; ESC?
        JNE   Edit3    ; No
        JMP   EditLoop ; Yes

Edit3:  MOV   BX, OFFSET Flags
        TEST  WORD PTR [BX], 8  ; Discard leading spaces and tabs?
        JZ    Edit4    ; Nope
        MOV   BX, OFFSET Leading
        TEST  BYTE PTR [BX], 1  ; Nothing but spaces so far?
        JNZ   EditLoop ; Yes, so ignore

Edit4:  MOV   BX, OFFSET Flags
        TEST  WORD PTR [BX], 16 ; Change tabs to spaces?
        JZ    Edit5    ; Nope
        CMP   AL, 9    ; Is it a tab?
        JNE   Edit5    ; Nope
        MOV   AL, 32   ; Convert to a space
        MOV   BX, OFFSET Space
        MOV   BYTE PTR [BX], 1 ; Space=True

Edit5:  MOV   BX, OFFSET Flags
        TEST  WORD PTR [BX], 32 ; Convert to uppercase?
        JZ    Edit6    ; Nope
        CMP   AL, 97   ; Compare with 'a'
        JL    Edit6    ; Less than
        CMP   AL, 122  ; Compare with 'z'
        JG    Edit6    ; Greater than
        SUB   AL, 32   ; Convert to uppercase

Edit6:  MOV   BX, OFFSET Flags
        TEST  WORD PTR [BX], 64 ; Convert () to []?
        JZ    EditOK   ; Nope
        CMP   AL, 40   ; See if (
        JNE   Edit7    ; Nope
        MOV   AL, 91   ; Change to [
        JMP   EditOK   ; Done with this character

Edit7:  CMP   AL, 41   ; See if )
        JNE   EditOK   ; Nope
        MOV   AL, 93   ; Change to ]

EditOK: ; Character is OK, put into destination
        INC   DI       ; Increment destination offset
        MOV   BX, OFFSET WorkBuff
        MOV   [BX+DI], AL ; Save character
        CMP   AL, 34   ; Is character a quote?
        JNE   Edit8    ; No
        MOV   BX, OFFSET Flags
        TEST  WORD PTR [BX], 256 ; Leave quoted characters alone?
        JZ    Edit8    ; No
        MOV   BX, OFFSET Quotes
        NOT   BYTE PTR [BX] ; Complement flag

Edit8:  MOV   BX, OFFSET Space
        CMP   AL, 32   ; See if character is a space
        JE    Edit9    ; yes
        CMP   AL, 9    ; See if character is a tab
        JE    Edit9    ; yes
        MOV   BYTE PTR [BX], 0 ; Space=False
        MOV   BX, OFFSET Last
        PUSH  AX       ; Save AX
        MOV   AX, DI
        MOV   [BX], AL ; Last non-space position
        POP   AX       ; Restore AX
        JMP   EditLoop

Edit9:  MOV   BYTE PTR [BX], 1 ; Space=True

EditLoop: ; End of edit loop - see if we are done
        DEC   DL    ; One less byte to process
        PUSH  AX    ; Save AX
        MOV   AL,DL ; Get bytes left
        CMP   AL,0  ; Any left?
        POP   AX    ; Get AX back
        JE    Edit9a ; No
        JMP   Edit0 ; Go back for next
Edit9a: MOV   BX, OFFSET Flags
        TEST  WORD PTR [BX], 128 ; Ignore trailing spaces?
        JZ    Edit10 ; Nope
        MOV   BX, OFFSET Last
        MOV   AL, [BX] ; Get last valid position in target string
        MOV   AH, 0
        MOV   DI, AX

Edit10: MOV   BX, OFFSET WorkBLen
        MOV   AX, DI
        MOV   [BX], AL ; Set length of target string
        POP   DI    ; Restore registers
        POP   SI
        POP   DX
        POP   CX
        POP   BX
        POP   AX
        RET         ; All done now
