{
        Program Name : ExtStore
        Package Name : LB
        Purpose      : Extent Store
        Institution  : Conroy & Conroy Co.
        Date Written : 12-Apr-2003
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2003-2006 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          Extent stores are extendable stores that exist on a managed store.

          LEVEL 2 CERTIFICATION: 1-June-2006 by EAC
}

unit ExtStore ;

interface

{$I EDEFINES.INC}

uses TypeDefs, { Max_Memory }

     { C&C Subroutine Library - Layer 0 }
     _Stores, { TStore_Address }

     { C&C Subroutine Library - Layer 1 }
     MStores, { TManaged_Store }
     _UEHDefs ; { TUEC }

const { Extent store error codes }
      ExtentStoreErr_Facility = 66 ;
      ExtentStoreErr_Success = 0 ;
      ExtentStoreErr_No_Store = 1 ; { No store has been set }
      ExtentStoreErr_Invalid_Length = 2 ; { Invalid read/write length }
      ExtentStoreErr_Corruption = 3 ; { Data corruption }
      ExtentStoreErr_Cannot_Extend = 4 ; { Could not extend store }
      ExtentStoreErr_No_Room = 5 ; { No room }
      ExtentStoreErr_Read_Only = 6 ; { Cannot write to read-only store }
      ExtentStoreErr_No_Memory = 7 ; { No scratch memory is available }
      ExtentStoreErr_End_Of_Data = 8 ; { Read/write past end of data }

const { Extent_store64 error codes }
      ExtentStore64Err_Facility = 116 ;
      ExtentStore64Err_Success = 0 ;
      ExtentStore64Err_No_Store = 1 ; { No store has been set }
      ExtentStore64Err_Invalid_Length = 2 ; { Invalid read/write length }
      ExtentStore64Err_Corruption = 3 ; { Data corruption }
      ExtentStore64Err_Cannot_Extend = 4 ; { Could not extend store }
      ExtentStore64Err_No_Room = 5 ; { No room }
      ExtentStore64Err_Read_Only = 6 ; { Cannot write to read-only store }
      ExtentStore64Err_No_Memory = 7 ; { No scratch memory is available }
      ExtentStore64Err_End_Of_Data = 8 ; { Read/write past end of data }

type TAllocation_Record = array[ 0..Max_Memory div sizeof( TStore_Address ) ] of TStore_Address ;
     PAllocation_Record = ^TAllocation_Record ;
     { And allocation record is two or more TStore_Addresses.  The last address
       in the record is the pointer to the next allocation record in the list. }

     PSExtent_Store = ^TSExtent_Store ;
     TSExtent_Store = object( TSStore )
                        private { Instance data... }
                            _Header : record
                                          Facility : longint ; { facility code for Extent store header }
                                          Record_Type : longint ; { 0 = Header }
                                          Version : longint ; { Header format version }
                                          Flags : longint ; { HF_* flags }

                                          Size : TStore_Address ; { Size of store, in bytes }
                                          Clustersize : TStore_Address ; { Minimum allocation size, in bytes }
                                          Allocation_Info : TStore_Address ; { Location of first set of allocation pointers.  If contiguous, this is the starting location of the data. }
                                          { Remainder of header is currently reserved }
                                      end ;
                            _Root : TStore_Address ; { Location of Header on _Store }
                            _Store : PSManaged_Store ;
                            _Allocation_Record : PAllocation_Record ;
                            _Allocation_Record_Size : longint ; { Size of allocation record blocks, in bytes }
                            Clusters_In_Block : longint ; { Number of cluster pointers in allocation record }
                            Next_Record_Offset : longint ;

                        private { Internal utility routines... }
                            function Allocate_Entry : TStore_Address ;

                            function Allocation_Record : PAllocation_Record ;

                            procedure Copy_To_Store( Store : PSExtent_Store ) ;
                            { Copy ourselves to the passed store. }

                            function Get_Absolute( Offset : longint ) : TStore_Address ;
                            { Translate a store-relative offset into a _Store
                              address.  Offset = Cluster offset within store. }

                            procedure Set_Error( Value : integer ) ;

                            procedure Update_Header ;

                        public { Overrides... }
                            function Read_Data( var Data ; Address, Len : TStore_Address ;
                                var UEC : TUEC ) : TStore_Address ; virtual ;

                            function Write_Data( var Data ;
                                Address, Len : TStore_Address ;
                                var UEC : TUEC ) : TStore_Address ; virtual ;

                            function Max_Storage : TStore_Size ; virtual ;

                            function Min_Storage : TStore_Address ; virtual ;

                            function Extend( Amount : TStore_Address ) : TStore_Address ;
                                virtual ;

                            function Read_Only : boolean ; virtual ;

                            function Write_Only : boolean ; virtual ;

                            procedure Initialize ; virtual ;

                            procedure Terminate ; virtual ;

                            function Contiguous_Store : boolean ; virtual ;

                            procedure Set_Max_Storage( Value : TStore_Address ;
                                var Res : TUEC ) ; virtual ;

                            function Extended_Size : TStore_Address ; virtual ;

                        public { API... }
                            Last_Error : TUEC ;

                            function Get_Clustersize : TStore_Address ; virtual ;
                            function Get_Contiguous_Flag : boolean ; virtual ;
                            function Get_Placed_Flag : boolean ; virtual ;
                            function Get_Read_Only_Flag : boolean ; virtual ;
                            function Get_Write_Only_Flag : boolean ; virtual ;
                            function Get_Root : TStore_Address ; virtual ;
                            function Get_Store : PSStore ; virtual ;
                            procedure Set_Clustersize( Value : TStore_Address ) ; virtual ;
                            procedure Set_Contiguous_Flag( Value : boolean ) ; virtual ;
                            procedure Set_Placed_Flag( Value : boolean ) ; virtual ;
                            procedure Set_Read_Only_Flag( Value : boolean ) ; virtual ;
                            procedure Set_Root( Value : integer ) ; virtual ;
                            procedure Set_Store( S : PSManaged_Store ) ; virtual ;
                            procedure Set_Read_Only( Value : boolean ) ; virtual ;
                            procedure Set_Write_Only( Value : boolean ) ; virtual ;

                            procedure Delete ; virtual ;
                            procedure Load_From_Store( S : PSManaged_Store ;
                                R : TStore_Address ) ; virtual ;
                     end ; { TSExtent_Store }

type TAllocation_Record64 = array[ 0..Max_Memory div sizeof( TStore_Address64 ) ] of TStore_Address64 ;
     PAllocation_Record64 = ^TAllocation_Record64 ;
     { And allocation record is two or more TStore_Addresse64s.  The last address
       in the record is the pointer to the next allocation record in the list. }

     PSExtent_Store64 = ^TSExtent_Store64 ;
     TSExtent_Store64 = object( TSStore64 )
                        private { Instance data... }
                            _Header : record
                                          Facility : longint ; { facility code for Extent store header }
                                          Record_Type : longint ; { 0 = Header }
                                          Version : longint ; { Header format version }
                                          Flags : longint ; { HF_* flags }

                                          Size : TStore_Address64 ; { Size of store, in bytes }
                                          Clustersize : TStore_Address64 ; { Minimum allocation size, in bytes }
                                          Allocation_Info : TStore_Address64 ; { Location of first set of allocation pointers.  If contiguous, this is the starting location of the data. }
                                          { Remainder of header is currently reserved }
                                      end ;
                            _Root : TStore_Address64 ; { Location of Header on _Store }
                            _Store : PSManaged_Store64 ;
                            _Allocation_Record : PAllocation_Record64 ;
                            _Allocation_Record_Size : longint ; { Size of allocation record blocks, in bytes }
                            Clusters_In_Block : longint ; { Number of cluster pointers in allocation record }
                            Next_Record_Offset : longint ;

                        private { Internal utility routines... }
                            function Allocate_Entry : TStore_Address64 ;

                            function Allocation_Record : PAllocation_Record64 ;

                            procedure Copy_To_Store( Store : PSExtent_Store64 ) ;
                            { Copy ourselves to the passed store. }

                            function Get_Absolute( Offset : longint ) : TStore_Address64 ;
                            { Translate a store-relative offset into a _Store
                              address.  Offset = Cluster offset within store. }

                            procedure Set_Error( Value : integer ) ;

                            procedure Update_Header ;

                        public { Overrides... }
                            function Read_Data( var Data ; Address, Len : TStore_Address64 ;
                                var UEC : TUEC ) : TStore_Address64 ; virtual ;

                            function Write_Data( var Data ;
                                Address, Len : TStore_Address64 ;
                                var UEC : TUEC ) : TStore_Address64 ; virtual ;

                            function Max_Storage : TStore_Size64 ; virtual ;

                            function Min_Storage : TStore_Address64 ; virtual ;

                            function Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
                                virtual ;

                            function Read_Only : boolean ; virtual ;

                            function Write_Only : boolean ; virtual ;

                            procedure Initialize ; virtual ;

                            procedure Terminate ; virtual ;

                            function Contiguous_Store : boolean ; virtual ;

                            procedure Set_Max_Storage( Value : TStore_Address64 ;
                                var Res : TUEC ) ; virtual ;

                            function Extended_Size : TStore_Address64 ;
                                virtual ;

                        public { API... }
                            Last_Error : TUEC ;

                            function Get_Clustersize : TStore_Address64 ;
                                virtual ;
                            function Get_Contiguous_Flag : boolean ; virtual ;
                            function Get_Placed_Flag : boolean ; virtual ;
                            function Get_Read_Only_Flag : boolean ; virtual ;
                            function Get_Write_Only_Flag : boolean ; virtual ;
                            function Get_Root : TStore_Address64 ; virtual ;
                            function Get_Store : PSStore64 ; virtual ;
                            procedure Set_Clustersize( Value : TStore_Address64 ) ;
                                virtual ;
                            procedure Set_Contiguous_Flag( Value : boolean ) ;
                                virtual ;
                            procedure Set_Placed_Flag( Value : boolean ) ;
                                virtual ;
                            procedure Set_Read_Only_Flag( Value : boolean ) ;
                                virtual ;
                            procedure Set_Root( Value : integer ) ; virtual ;
                            procedure Set_Store( S : PSManaged_Store64 ) ;
                                virtual ;
                            procedure Set_Read_Only( Value : boolean ) ;
                                virtual ;
                            procedure Set_Write_Only( Value : boolean ) ;
                                virtual ;

                            procedure Delete ; virtual ;
                            procedure Load_From_Store( S : PSManaged_Store64 ;
                                R : TStore_Address64 ) ; virtual ;
                     end ; { TSExtent_Store64 }

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses { C&C Subroutine Library - Layer 1 }
     BigFH, { TSBig_FH }
     FSFH, { TSFH }
     Heaps ; { TSHeap } { TSFH }

const { Extent store header flags... }
      HF_Read_Only = 1 ; { Store is read-only }
      HF_Contiguous = 2 ; { Store's data is contiguous }
      HF_Placed = 4 ; { Store's data is placed at a specific location (only has meaning for contiguous stores) }
      HF_Write_Only = 8 ; { Store is write-only }
      { Other flags reserved for future use }

{ TSExtent_Store methods... }

{ Internal utility routines... }

{ Allocate a block of allocation pointers }
function TSExtent_Store.Allocate_Entry : TStore_Address ;

var
{$IFNDEF DELPHI}
    B : PAllocation_Record ;
{$ENDIF}
    _Result : TStore_Address ;

begin
{$IFNDEF DELPHI}
    B :=
{$ENDIF}
      Allocation_Record ; { Make sure metrics are determined }
    _Result := _Store^.Allocate( _Allocation_Record_Size ) ;
    if( _Result <> 0 ) then { Successfully allocated room on the store } { UNTESTED: Else }
    begin
        _Store^.Fill( _Result, _Allocation_Record_Size, 0 ) ;
    end ;
    Allocate_Entry := _Result ;
end ;


function TSExtent_Store.Allocation_Record : PAllocation_Record ;

begin
    if( _Allocation_Record = nil ) then
    begin
        if( _Store <> nil ) then { This should always be true } { Untested: Else }
        begin
            _Allocation_Record_Size := _Store^.Min_Storage ;
            if( _Allocation_Record_Size < 2 * sizeof( TStore_Address ) ) then
            begin
                _Allocation_Record_Size := 2 * sizeof( TStore_Address ) ;
            end ;
            getmem( _Allocation_Record, _Allocation_Record_Size ) ;
            Next_Record_Offset :=
                _Allocation_Record_Size div sizeof( TStore_Address ) - 1 ;
            Clusters_In_Block := Next_Record_Offset ; { Number of clusters in block }
        end ;
    end ;
    Allocation_Record := _Allocation_Record ;
end ;


procedure TSExtent_Store.Copy_To_Store( Store : PSExtent_Store ) ;

var Buffer : pointer ;
    Count, _Pos, Size : integer ;

begin
    Set_Error( 0 ) ;
    Size := _Header.Clustersize ;
    if( Size < 1024 ) then
    begin
        Size := 1024 ;
    end ;
    getmem( Buffer, Size ) ;
    if( Buffer = nil ) then { Clustersize must be too large }
    begin
        { Untested }
        Size := 1024 ;
        getmem( Buffer, Size ) ;
        if( Buffer = nil ) then
        begin
            Set_Error( ExtentStoreErr_No_Memory ) ;
            exit ;
        end ;
    end ;
    Count := _Header.Size ;
    _Pos := 0 ;
    while( Count > Size ) do
    begin
        Read_Data( Buffer^, _Pos, Size, Last_Error ) ;
        if( Last_Error.Code <> 0 ) then
        begin
            freemem( Buffer, Size ) ; { UNTESTED }
            exit ;
        end ;
        Store^.Write_Data( Buffer^, _Pos, Size, Last_Error ) ;
        if( Last_Error.Code <> 0 ) then
        begin
            freemem( Buffer, Size ) ; { UNTESTED }
            exit ;
        end ;
        Count := Count - Size ;
        _Pos := _Pos + Size ;
    end ; { while( Count > Size ) }
    if( Count > 0 ) then
    begin
        Read_Data( Buffer^, _Pos, Count, Last_Error ) ;
        if( Last_Error.Code <> 0 ) then
        begin
            freemem( Buffer, Size ) ; { Untested }
            exit ;
        end ;
        Store^.Write_Data( Buffer^, _Pos, Count, Last_Error ) ;
    end ;
    freemem( Buffer, Size ) ;
end ; { TSExtent_Store.Copy_To_Store }


function TSExtent_Store.Get_Absolute( Offset : longint ) : TStore_Address ;
{ Translate an extent-relative offset into an address on _Store.
  Offset = Cluster offset within extent. }

var Allocation : TStore_Address ;
    _Record : PAllocation_Record ;
    UEC : TUEC ;

begin
    if( Get_Contiguous_Flag ) then
    begin
        Get_Absolute := _Header.Allocation_Info + ( Offset * _Header.Clustersize ) ; { We should never get here } { UNTESTED }
        exit ;
    end ;

    _Record := Allocation_Record ; { Allocate a buffer }

    { Get the first allocation record }
    _Store^.Read_Data( _Record^, _Header.Allocation_Info, _Allocation_Record_Size, UEC ) ;

    while( Offset >= Clusters_In_Block ) do
    begin
	Offset := Offset - Clusters_In_Block ;
	Allocation := _Record^[ Next_Record_Offset ] ;
	_Store^.Read_Data( _Record^, Allocation, _Allocation_Record_Size, UEC ) ;
	if( Allocation = 0 ) then
	begin
	    Get_Absolute := 0 ; { Beyond end of data } { Note: We should never get here } { UNTESTED }
	    exit ;
	end ;
    end ;
    Get_Absolute := _Record^[ Offset ] ;
    {TODO: Keep track of where we are for performance}
end ; { TSExtent_Store.Get_Absolute }


procedure TSExtent_Store.Set_Error( Value : integer ) ;

begin
    Last_Error.Facility := ExtentStoreErr_Facility ;
    Last_Error.Code := Value ;
end ;


procedure TSExtent_Store.Update_Header ; { Write header to store }

var UEC : TUEC ;

begin
    if( _Store <> nil ) then
    begin
        if( _Root = 0 ) then
        begin
            _Root := _Store.Allocate( sizeof( _Header ) ) ;
        end ;
        _Store^.Write_Data( _Header, _Root, sizeof( _Header ), UEC )
    end ;
end ;


{ Overrides... }

function TSExtent_Store.Read_Data( var Data ; Address, Len : TStore_Address ;
    var UEC : TUEC ) : TStore_Address ;

var _Buffer : PByte_Array ;
    Cluster : longint ; { Current cluster being read }
    Count : TStore_Address ;
    Dummy, Dummy1 : longint ;
    Offset : integer ; { Source buffer offset }
    This_Count : word ; { Count of bytes to move for current cluster }

begin
    Set_Error( 0 ) ;
    Read_Data := 0 ;
    if( _Store = nil ) then
    begin
        Set_Error( ExtentStoreErr_No_Store ) ;
        UEC := Last_Error ;
        exit ;
    end ;
    if( Len < 0 ) then
    begin
        Set_Error( ExtentStoreErr_Invalid_Length ) ;
        UEC := Last_Error ;
        exit ;
    end ;
    if( Len = 0 ) then
    begin
        exit ;
    end ;

    { Setup... }
    Set_Error( 0 ) ;
    UEC := Last_Error ;
    if( Address + Len > _Header.Size ) then { Need to truncate passed length }
    begin
	Len := _Header.Size - Address ; { Only read what is there }
        if( Len < 0 ) then
        begin
            Set_Error( ExtentStoreErr_End_Of_Data ) ;
            UEC := Last_Error ;
            exit ;
        end ;
    end ;
    if( Len = 0 ) then
    begin
        exit ;
    end ;

    if( Get_Contiguous_Flag ) then
    begin
        Read_Data := _Store^.Read_Data( Data, _Header.Allocation_Info + Address, Len, UEC ) ;
        Last_Error := UEC ;
        exit ;
    end ;

    { Now read from store... }
    getmem( _Buffer, _Header.Clustersize ) ; { Allocate a cluster-sized buffer }
    Cluster := Address div _Header.Clustersize ; { Cluster to start reading from }
    Dummy1 := Address - ( Cluster * _Header.Clustersize ) ; { Offset }
    Offset := 0 ; { Offset in source buffer }
    Count := 0 ;
    while( Len > 0 ) do { Until done reading from store }
    begin
	Dummy := Get_Absolute( Cluster ) ;
        if( Dummy <= 0 ) then { This should not happen }
        begin
            Set_Error( ExtentStoreErr_Corruption ) ; { UNTESTED }
            exit ;
        end ;
	Count := Count + _Store^.Read_Data( _Buffer^, Dummy, _Header.Clustersize, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Last_Error := UEC ; { UNTESTED }
            exit ;
        end ;
	This_Count := _Header.Clustersize - Dummy1 ; { Amount remaining in cluster }
	if( This_Count > Len ) then
	begin
	    This_Count := Len ; { Move no more than asked to }
	end ;
	move( _Buffer^[ Dummy1 ], TByte_Array( Data )[ Offset ], This_Count ) ;
	Offset := Offset + This_Count ; { New source position }
	Len := Len - This_Count ; { Bytes left to read }
	Cluster := Cluster + 1 ; { Next cluster next time }
	Dummy1 := 0 ; { Next source starts at offset 0 }
    end ;
    freemem( _Buffer, _Header.Clustersize ) ; { Deallocate it }
    Read_Data := Count ;
end ; { TSExtent_Store.Read_Data }


function TSExtent_Store.Write_Data( var Data ; Address, Len : TStore_Address ;
    var UEC : TUEC ) : TStore_Address ;
{ Write data from Data to specified Address in store.
  Write Size bytes.  Returns success of operation in UEC.
  Function returns the number of bytes actually written. }

var _Buffer : PByte_Array ;
    Cluster : longint ; { Current cluster being allocated }
    Count : TStore_Address ;
    Dummy, Dummy1 : longint ;
    Offset : integer ; { Source buffer offset }
    This_Count : word ; { Count of bytes to move for current cluster }

begin
    Write_Data := 0 ;
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentStoreErr_Read_Only ) ;
        UEC := Last_Error ;
        exit ;
    end ;
    if( _Store = nil ) then
    begin
        Set_Error( ExtentStoreErr_No_Store ) ;
        UEC := Last_Error ;
        exit ;
    end ;

    { Setup... }
    Set_Error( 0 ) ;
    UEC := Last_Error ;

    if( Address + Len > _Header.Size ) then { Need to extend }
    begin
	Extend( ( Address + Len ) - _Header.Size ) ;
        if( Last_Error.Code <> 0 ) then { Could not extend the store }
        begin
            UEC := Last_Error ; { UNTESTED }
            exit ;
        end ;
        if( Address + Len > _Header.Size ) then { Did not extend enough }
        begin
            Set_Error( ExtentStoreErr_Cannot_Extend ) ; { UNTESTED }
            UEC := Last_Error ;
            exit ;
        end ;
    end ;
    if( Len = 0 ) then
    begin
        exit ;
    end ;

    if( Get_Contiguous_Flag ) then
    begin
        Write_Data := _Store^.Write_Data( Data, Address + _Header.Allocation_Info, Len, UEC ) ;
        Last_Error := UEC ;
        exit ;
    end ;

    { Now write data to store... }
    getmem( _Buffer, _Header.Clustersize ) ; { Allocate a cluster-sized buffer }
    Cluster := Address div _Header.Clustersize ; { Cluster to start writing to }
    Dummy1 := Address - ( Cluster * _Header.Clustersize ) ; { Offset }
    Offset := 0 ; { Offset in source buffer }
    Count := 0 ;
    while( Len > 0 ) do { Until done writing to store }
    begin
	Dummy := Get_Absolute( Cluster ) ;
        if( Dummy <= 0 ) then { This should not happen }
        begin
            Set_Error( ExtentStoreErr_Corruption ) ; { UNTESTED }
            exit ;
        end ;
	_Store^.Read_Data( _Buffer^, Dummy, _Header.Clustersize, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            exit ; { UNTESTED }
        end ;
	This_Count := _Header.Clustersize - Dummy1 ;
	if( This_Count > Len ) then
	begin
	    This_Count := Len ; { Move no more than asked to }
	end ;
	move( TByte_Array( Data )[ Offset ], _Buffer^[ Dummy1 ], This_Count ) ;
	Count := Count + _Store^.Write_Data( _Buffer^, Dummy, _Header.Clustersize, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            exit ; { UNTESTED }
        end ;
	Offset := Offset + This_Count ; { New source position }
	Len := Len - This_Count ; { Bytes left to write }
	Cluster := Cluster + 1 ; { Next cluster next time }
	Dummy1 := 0 ; { Next destination starts at offset 0 }
    end ;
    Freemem( _Buffer, _Header.Clustersize ) ;
    Write_Data := Count ;
end ; { TSExtent_Store.Write_Data }


function TSExtent_Store.Max_Storage : TStore_Size ;
{ Returns the maximum (current) size of the store, in bytes.  In other words,
  this is the highest valid address (plus 1). }

begin
    Max_Storage := _Header.Size ;
end ;


function TSExtent_Store.Min_Storage : TStore_Address ;

var L : TStore_Address ;

begin
    if( _Store <> nil ) then
    begin
        L := _Store^.Min_Storage ;
        if( L < _Header.Clustersize ) then
        begin
            L := _Header.Clustersize ;
        end ;
        Min_Storage := L ;
    end else
    begin
        Min_Storage := _Header.Clustersize ;
    end ;
end ;


function TSExtent_Store.Extend( Amount : TStore_Address ) : TStore_Address ;
{ Requests extension of store by the amount specified.  Returns the amount
  actually added to store.  If Amount is 0, function returns 0 if store is
  extendable, and -1 if it is not. }

var Cluster : TStore_Address ; { Current cluster being allocated }
    Current : TStore_Address ; { Current allocation record pointer }
    Dummy, Dummy1 : TStore_Address ;
    New_Size : TStore_Address ;
    _Record : PAllocation_Record ;

begin
    Extend := 0 ;
    if( Amount = 0 ) then
    begin
        if( ( _Store <> nil ) and ( _Store^.Read_Only ) ) then
        begin
            Extend := -1 ;
        end else
        if( Get_Read_Only_Flag ) then
        begin
            Extend := -1 ;
        end ;
        exit ;
    end ;
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentStoreErr_Read_Only ) ;
        exit ;
    end ;

    { Determine new size (rounded to clustersize...) }
    New_Size := _Header.Size + Amount ;
    if( New_Size <> _Header.Clustersize ) then
    begin
        New_Size := ( New_Size + _Header.Clustersize - 1 ) ;
        New_Size := New_Size div _Header.Clustersize ;
        New_Size := New_Size * _Header.Clustersize ;
    end ;
    if( _Store = nil ) then
    begin
        _Header.Size := New_Size ;
        exit ;
    end ;

    if( _Store^.XSpaceAvail < New_Size - _Header.Size ) then { No room }
    begin
	Set_Error( ExtentStoreErr_Cannot_Extend ) ; { UNTESTED }
	exit ;
    end ;

    if( _Header.Size <> 0 ) then
    begin
        Set_Contiguous_Flag( False ) ; { Extend removes contiguous flag on non-zero-length stores }
    end ;
    if( Get_Contiguous_Flag ) then
    begin
        _Header.Allocation_Info := _Store^.Allocate( New_Size ) ;
	if( _Header.Allocation_Info = 0 ) then
	begin
	    Set_Error( ExtentStoreErr_No_Room ) ; { UNTESTED }
	    exit ;
	end ;
        _Header.Size := New_Size ;
        Update_Header ;
        exit ;
    end ;

    _Record := Allocation_Record ; { Allocate a buffer }

    { Now allocate additional space... }
    if( _Header.Allocation_Info = 0 ) then { Store is empty }
    begin
 	_Header.Allocation_Info := Allocate_Entry ;
	if( _Header.Allocation_Info = 0 ) then
	begin
	    Set_Error( ExtentStoreErr_No_Room ) ; { UNTESTED }
	    exit ;
	end ;
        Update_Header ;
    end ;
    _Store^.Read_Data( _Record^, _Header.Allocation_Info, _Allocation_Record_Size, Last_Error ) ;
    if( Last_Error.Code <> 0 ) then
    begin
        exit ; { UNTESTED }
    end ;

    { TODO: For performance, start at current end of allocation pointers... }
    Current := _Header.Allocation_Info ;
    Dummy := New_Size div _Header.Clustersize ; { Maximum used clusters }
    Dummy1 := 1 ;
    while( Dummy1 <= Dummy ) do
    begin
	for Cluster := 0 to Clusters_In_Block - 1 do
        begin
            if( Dummy1 <= Dummy ) then { Still more clusters to allocate }
            begin
                if( _Record^[ Cluster ] = 0 ) then { Cluster not yet allocated }
                begin
                    _Record^[ Cluster ] := _Store^.Allocate( _Header.Clustersize ) ;
                    if( _Record^[ Cluster ] = 0 ) then
                    begin
                        Set_Error( ExtentStoreErr_No_Room ) ; { TEST: Untested }
                        exit ;
                    end ;
                end ; { if( _Record^[ Cluster ] = 0 ) }
                Dummy1 := Dummy1 + 1 ;
            end else
            begin
                break ; { Once we are done, skip of out loop }
            end ; { if( Dummy1 <= Dummy ) }
        end ; { for Cluster := 0 to Clusters_In_Block - 1 }
	_Store^.Write_Data( _Record^, Current, _Allocation_Record_Size, Last_Error ) ;
        if( Last_Error.Code <> 0 ) then
        begin
            exit ; { TEST: Untested }
        end ;
	if( Dummy1 <= Dummy ) then { More allocation needed }
	begin
	    if( _Record^[ Next_Record_Offset ] = 0 ) then
	    begin
		_Record^[ Next_Record_Offset ] := Allocate_Entry ;
		if( _Record^[ Next_Record_Offset ] = 0 ) then
		begin
		    Set_Error( ExtentStoreErr_No_Room ) ; { TEST: Untested }
		    exit ;
		end ;
		_Store^.Write_Data( _Record^, Current, _Allocation_Record_Size, Last_Error ) ;
                if( Last_Error.Code <> 0 ) then
                begin
                    exit ; { TEST: Untested }
                end ;
		Current := _Record^[ Next_Record_Offset ] ;
		fillchar( _Record^, _Allocation_Record_Size, 0 ) ;
	    end else
	    begin
		_Store^.Write_Data( _Record^, Current, _Allocation_Record_Size, Last_Error ) ;
                if( Last_Error.Code <> 0 ) then
                begin
                    exit ; { TEST: Untested }
                end ;
		Current := _Record^[ Next_Record_Offset ] ;
		_Store^.Read_Data( _Record^, Current, _Allocation_Record_Size, Last_Error ) ;
                if( Last_Error.Code <> 0 ) then
                begin
                    exit ; { TEST: Untested }
                end ;
	    end ; { if( _Record^[ Next_Record_Offset ] = 0 ) }
	end ; { if( Dummy1 <= Dummy ) }
    end ; { while( Dummy1 <= Dummy ) }
    _Store^.Write_Data( _Record^, Current, _Allocation_Record_Size, Last_Error ) ;
    if( Last_Error.Code <> 0 ) then
    begin
        exit ; { TEST: Untested }
    end ;

    _Header.Size := New_Size ;
    Update_Header ;
end ; { TSExtent_Store.Extend }


function TSExtent_Store.Read_Only : boolean ;

begin
    if( ( _Store = nil ) or ( _Store^.Read_Only ) ) then
    begin
        Read_Only := True ;
    end else
    begin
        Read_Only := Get_Read_Only_Flag ;
    end ;
end ;


function TSExtent_Store.Write_Only : boolean ;

begin
    if( _Store <> nil ) then
    begin
        Write_Only := _Store^.Write_Only ;
    end else
    begin
        Write_Only := False ;
    end ;
end ;


procedure TSExtent_Store.Initialize ;

begin
    _Store := nil ;
    _Root := 0 ;
    fillchar( _Header, sizeof( _Header ), 0 ) ;
    _Header.Facility := ExtentStoreErr_Facility ;
    _Header.Record_Type := 0 ;
    _Header.Version := 0 ;
    _Header.Size := 0 ;
    _Header.Flags := 0 ;
    _Header.Clustersize := 1 ;
    _Header.Allocation_Info := 0 ;
    _Allocation_Record := nil ;
    fillchar( Last_Error, sizeof( Last_Error ), 0 ) ;
end ;


procedure TSExtent_Store.Terminate ;

begin
    Set_Store( nil ) ;
end ;


function TSExtent_Store.Contiguous_Store : boolean ;

begin
    Contiguous_Store := True ;
end ;


procedure TSExtent_Store.Set_Max_Storage( Value : TStore_Address ;
    var Res : TUEC ) ;
{ Sets the maximum (current) size of the store, in bytes.  In other words, this
  is the new highest valid address (plus 1).  If the store is not resizable, the
  call fails and an error code is returned in Res.  This should only be used to
  shrink the size of the store, not to extend it.  To extend, use the Extend
  method. }

var Current : longint ;
    _Record : pAllocation_Record ;
    Next, Offset : TStore_Address ;
    UEC : TUEC ;

begin
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentStoreErr_Read_Only ) ;
        exit ;
    end ;
    if( _Store = nil ) then
    begin
        _Header.Size := Value ;
        exit ;
    end ;

    { Setup... }
    if( Value < 0 ) then
    begin
        Value := 0 ;
    end ;
    Set_Error( 0 ) ;
    Res := Last_Error ;
    if( Value <> _Header.Clustersize ) then
    begin
        Value := ( Value + _Header.Clustersize - 1 ) ;
        Value := Value div _Header.Clustersize ;
        Value := Value * _Header.Clustersize ;
    end ;

    { Resize... }
    if( Value > _Header.Size ) then { Expand }
    begin
        Extend( Value - _Header.Size ) ;
    end else
    if( Value < _Header.Size ) then { Truncate }
    begin
        if( Get_Contiguous_Flag ) then
        begin
            _Store.Reallocate( _Header.Allocation_Info, _Header.Size, Value ) ;
            _Header.Size := Value ; { New size }
            Update_Header ;
            exit ;
        end ;

        _Header.Size := Value ; { New size }
        Update_Header ;
        _Record := Allocation_Record ; { Allocate a buffer }

        { Get the first allocation record }
        _Store^.Read_Data( _Record^, _Header.Allocation_Info, _Allocation_Record_Size, UEC ) ;

        { Free unused (previously allocated) clusters in last allocation block... }
        Offset := Value div _Header.Clustersize ;
        Current := _Header.Allocation_Info ;
        while( Offset >= Clusters_In_Block ) do { Move to allocation block where truncation begins }
        begin
            Offset := Offset - Clusters_In_Block ;
            Current := _Record^[ Next_Record_Offset ] ;
            if( Current = 0 ) then 
            begin
                exit ; { Beyond end of data.  This should never happen } { Test: UNTESTED }
            end ;
            _Store^.Read_Data( _Record^, Current, _Allocation_Record_Size, UEC ) ;
        end ;
        while( Offset < Next_Record_Offset ) do
        begin
            if( _Record^[ Offset ] <> 0 ) then
            begin
                _Store.Deallocate( _Record^[ Offset ], _Header.Clustersize ) ;
                _Record^[ Offset ] := 0 ;
            end ;
            inc( Offset ) ;
        end ;
        Next := _Record^[ Next_Record_Offset ] ;
        _Record[ Next_Record_Offset ] := 0 ;
        _Store^.Write_Data( _Record^, Current, _Allocation_Record_Size, UEC ) ;

        { Delete all remaining allocation blocks... }
        Current := Next ;
        while( Current <> 0 ) do
        begin
            _Store^.Read_Data( _Record^, Next, _Allocation_Record_Size, UEC ) ;
            for Offset := 0 to Clusters_In_Block - 1 do
            begin
                if( _Record^[ Offset ] <> 0 ) then
                begin
                    _Store.Deallocate( _Record^[ Offset ], _Header.Clustersize ) ;
                end ;
            end ;
            Next := _Record^[ Next_Record_Offset ] ;
            _Store^.Deallocate( Current, _Allocation_Record_Size ) ;
            Current := Next ;
        end ;
        if( _Header.Size = 0 ) then { Get rid of first allocation record }
        begin
            _Store^.Deallocate( _Header.Allocation_Info, _Allocation_Record_Size ) ;
            _Header.Allocation_Info := 0 ;
            Update_Header ;
        end ;
    end ; { if( Value < _Header.Size ) }
end ; { TSExtent_Store.Set_Max_Storage }


function TSExtent_Store.Extended_Size : TStore_Address ;
{ Returns the maximum theoretical size that the store can be extended to.  If
  the store is not extendable, this returns the same value as Max_Storage. }

var _Result : TStore_Address ;

begin
    Set_Error( 0 ) ;
    if( _Store = nil ) then
    begin
        _Result := Max_Storage ;
    end else
    begin
        _Result := _Header.Size + _Store^.XSpaceAvail ;
        if( _Result < 0 ) then { UNTESTED: Else - spaceavail on heap is always $7FFFFFFF, plus _Header.Size is < 0 on 32-bit systems }
        begin
            _Result := $7FFFFFFF ;
        end ;
    end ;
    Extended_Size := _Result ;
end ;


function TSExtent_Store.Get_Clustersize : TStore_Address ;

begin
    Set_Error( 0 ) ;
    Get_Clustersize := _Header.Clustersize ;
end ;


function TSExtent_Store.Get_Contiguous_Flag : boolean ;

begin
    Set_Error( 0 ) ;
    Get_Contiguous_Flag := ( ( _Header.Flags and HF_Contiguous ) <> 0 ) ;
end ;


function TSExtent_Store.Get_Placed_Flag : boolean ;

begin
    Set_Error( 0 ) ;
    Get_Placed_Flag := ( ( _Header.Flags and HF_Placed ) <> 0 ) ;
end ;


function TSExtent_Store.Get_Read_Only_Flag : boolean ;

begin
    Set_Error( 0 ) ;
    Get_Read_Only_Flag := ( ( _Header.Flags and HF_Read_Only ) <> 0 ) ;
end ;


function TSExtent_Store.Get_Write_Only_Flag : boolean ;

begin
    Set_Error( 0 ) ;
    Get_Write_Only_Flag := ( ( _Header.Flags and HF_Write_Only ) <> 0 ) ;
end ;


function TSExtent_Store.Get_Root : TStore_Address ;

begin
    Set_Error( 0 ) ;
    Get_Root := _Root ;
end ;


function TSExtent_Store.Get_Store : PSStore ;

begin
    Set_Error( 0 ) ;
    Get_Store := _Store ;
end ;


procedure TSExtent_Store.Set_Clustersize( Value : TStore_Address ) ;

var Saved : TStore_Address ;
    Temp : TSExtent_Store ;

begin
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentStoreErr_Read_Only ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    if( Value < 0 ) then
    begin
        Value := 1 ;
    end ;
    if( _Store <> nil ) then
    begin
        if( Value < _Store.Min_Storage ) then
        begin
            Value := _Store.Min_Storage ; { Cannot be smaller clustersize than _Store }
        end ;
    end ;
    if( Value <> _Header.Clustersize ) then
    begin
        if( _Header.Size > 0 ) then
        begin
            { Copy data to new list with new clustersize... }
            Temp.Init ;
            Temp.Initialize ;
            Temp.Set_Clustersize( Value ) ;
            Temp.Set_Store( _Store ) ;
            Copy_To_Store( @Temp ) ;
            if( Last_Error.Code <> 0 ) then
            begin
                { UNTESTED }
                Temp.Delete ;
                Temp.Done ;
                exit ;
            end ;

            { Delete existing data... }
            Saved := _Header.Size ;
            Set_Max_Storage( 0, Last_Error ) ;
            _Header.Size := Saved ;
        end ;

        { Update header... }
        _Header.Clustersize := Value ;
        if( _Header.Size > 0 ) then
        begin
            _Header.Allocation_Info := Temp._Header.Allocation_Info ;
            Temp.Set_Store( nil ) ;
            Temp.Done ;
        end ;
        Update_Header ;
    end ;
end ; { TSExtent_Store.Set_Clustersize }


procedure TSExtent_Store.Set_Contiguous_Flag( Value : boolean ) ;

var Current, Last, Original, _Size : TStore_Address ;
    Index : longint ;
    _Record : PAllocation_Record ;
    Temp : TSExtent_Store ;

begin
    Set_Error( 0 ) ;
    if( Value <> Get_Contiguous_Flag ) then
    begin
        if( Value ) then { Converting to contiguous }
        begin
            if( _Header.Size > 0 ) then
            begin
                Temp.Init ;
                Temp.Initialize ;
                Temp.Set_Clustersize( _Header.Clustersize ) ;
                Temp.Set_Contiguous_Flag( Value ) ;
                Temp.Set_Max_Storage( _Header.Size, Last_Error ) ;
                Temp.Set_Store( _Store ) ;
                Copy_To_Store( @Temp ) ;
                if( Last_Error.Code <> 0 ) then
                begin
                    Temp.Delete ; { UNTESTED }
                    Temp.Done ;
                    exit ;
                end ;
            end ;
            _Header.Flags := _Header.Flags or HF_Contiguous ;
            Set_Max_Storage( _Header.Size, Last_Error ) ;
            if( _Header.Size > 0 ) then
            begin
                _Header.Allocation_Info := Temp._Header.Allocation_Info ;
                Temp.Set_Store( nil ) ;
                Temp.Done ;
            end ;
        end else
        begin
            { Converting from contiguous to non-contiguous is easy: we simply
              create an allocation chain that points to the existing data.  This
              keeps contiguous data pysically contiguous, even if not logically
              so. }
            Current := _Header.Allocation_Info ; { Start of contiguous data }
            _Size := _Header.Size ;
            Last := Allocate_Entry ; { Create initial allocation record }
            if( Last = 0 ) then
            begin
                Set_Error( ExtentStoreErr_No_Room ) ; { UNTESTED }
                exit ;
            end ;
            Original := _Header.Allocation_Info ;
            _Header.Allocation_Info := Last ;
            _Record := Allocation_Record ; { Allocate a buffer }
            Last := _Header.Allocation_Info ;
            _Store^.Read_Data( _Record^, Last, _Allocation_Record_Size, Last_Error ) ;
            Index := 0 ;
            while( _Size > 0 ) do
            begin
                _Record[ Index ] := Current ;
                Current := Current + _Header.Clustersize ;
                _Size := _Size - _Header.Clustersize ;
                inc( Index ) ;
                if( ( Index >= Clusters_In_Block - 1 ) and ( _Size > 0 ) ) then { End of allocation block but not of data }
                begin
                    _Record[ Clusters_In_Block - 1 ] := Allocate_Entry ; { Create next allocation record in chain }
                    if( _Record[ Clusters_In_Block - 1 ] = 0 ) then
                    begin
                        Set_Error( ExtentStoreErr_No_Room ) ; { UNTESTED }
                        _Header.Allocation_Info := Original ;
                        exit ;
                    end ;
                    _Store^.Write_Data( _Record^, Last, _Allocation_Record_Size, Last_Error ) ;
                    Last := _Record[ Clusters_In_Block - 1 ] ;
                    _Store^.Read_Data( _Record^, Last, _Allocation_Record_Size, Last_Error ) ;
                    Index := 0 ;
                end ; { if }
            end ; { while( _Size > 0 ) }
            _Store^.Write_Data( _Record^, Last, _Allocation_Record_Size, Last_Error ) ;
            _Header.Flags := _Header.Flags and ( not HF_Contiguous ) ;
        end ; { if( Value ) }
        Update_Header ;
    end ; { if( Value <> Get_Contiguous_Flag ) }
end ; { TSExtent_Store.Set_Contiguous_Flag }


procedure TSExtent_Store.Set_Placed_Flag( Value : boolean ) ;

begin
    Set_Error( 0 ) ;
    if( Value <> Get_Placed_Flag ) then
    begin
        if( Value ) then
        begin
            _Header.Flags := _Header.Flags or HF_Placed ;
        end else
        begin
            _Header.Flags := _Header.Flags and ( not HF_Placed ) ;
        end ;
        Update_Header ;
    end ;
end ;


procedure TSExtent_Store.Set_Read_Only_Flag( Value : boolean ) ;

begin
    Set_Error( 0 ) ;
    if( Value <> Get_Read_Only_Flag ) then
    begin
        if( Value ) then
        begin
            _Header.Flags := _Header.Flags or HF_Read_Only ;
        end else
        begin
            _Header.Flags := _Header.Flags and ( not HF_Read_Only ) ;
        end ;
        Update_Header ;
    end ;
end ;


procedure TSExtent_Store.Set_Root( Value : integer ) ;

begin
    Set_Error( 0 ) ;
    _Root := Value ;
end ;


procedure TSExtent_Store.Set_Store( S : PSManaged_Store ) ;

var Saved : TStore_Address ;

begin
    if( _Store <> S ) then
    begin
        Saved := _Header.Size ;
        _Store := S ;
        _Root := 0 ;
        _Header.Allocation_Info := 0 ;
        if( _Store <> nil ) then
        begin
            Update_Header ;
        end ;
        if( _Allocation_Record <> nil ) then
        begin
            freemem( _Allocation_Record, _Allocation_Record_Size ) ;
            _Allocation_Record := nil ;
        end ;
        if( S <> nil ) then
        begin
            _Header.Size := 0 ; { So following will actually extend the store }
            Set_Max_Storage( Saved, Last_Error ) ;
        end ;
    end ;
end ;


procedure TSExtent_Store.Delete ;

begin
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentStoreErr_Read_Only ) ;
        exit ;
    end ;
    Set_Max_Storage( 0, Last_Error ) ;
    if( _Store <> nil ) then
    begin
        _Store^.Deallocate( _Root, sizeof( _Header ) ) ;
        _Root := 0 ;
    end ;
end ;


procedure TSExtent_Store.Load_From_Store( S : PSManaged_Store ;
    R : TStore_Address ) ; { UNTESTED }

var UEC : TUEC ;

begin
    _Store := S ;
    _Root := R ;
    if( _Root = 0 ) then
    begin
        _Store^.Read_Data( _Header, _Root, sizeof( _Header ), UEC )
    end ;
end ;


procedure TSExtent_Store.Set_Read_Only( Value : boolean ) ;

begin
    Set_Read_Only_Flag( Value ) ;
end ;


procedure TSExtent_Store.Set_Write_Only( Value : boolean ) ;

begin
    Set_Error( 0 ) ;
    if( Value <> Get_Write_Only_Flag ) then
    begin
        if( Value ) then
        begin
            _Header.Flags := _Header.Flags or HF_Write_Only ;
        end else
        begin
            _Header.Flags := _Header.Flags and ( not HF_Write_Only ) ;
        end ;
        Update_Header ;
    end ;
end ;



{ TSExtent_Store64 methods... }

{ Internal utility routines... }

{ Allocate a block of allocation pointers }
function TSExtent_Store64.Allocate_Entry : TStore_Address64 ;

var
{$IFNDEF DELPHI}
    B : PAllocation_Record64 ;
{$ENDIF}
    _Result : TStore_Address64 ;

begin
{$IFNDEF DELPHI}
    B :=
{$ENDIF}
      Allocation_Record ; { Make sure metrics are determined }
    _Result := _Store^.Allocate( _Allocation_Record_Size ) ;
    if( _Result <> 0 ) then { Successfully allocated room on the store } { UNTESTED: Else }
    begin
        _Store^.Fill( _Result, _Allocation_Record_Size, 0 ) ;
    end ;
    Allocate_Entry := _Result ;
end ;


function TSExtent_Store64.Allocation_Record : PAllocation_Record64 ;

begin
    if( _Allocation_Record = nil ) then
    begin
        if( _Store <> nil ) then { This should always be true } { Untested: Else }
        begin
            _Allocation_Record_Size := _Store^.Min_Storage ;
            if( _Allocation_Record_Size < 2 * sizeof( TStore_Address64 ) ) then
            begin
                _Allocation_Record_Size := 2 * sizeof( TStore_Address64 ) ;
            end ;
            getmem( _Allocation_Record, _Allocation_Record_Size ) ;
            Next_Record_Offset :=
                _Allocation_Record_Size div sizeof( TStore_Address64 ) - 1 ;
            Clusters_In_Block := Next_Record_Offset ; { Number of clusters in block }
        end ;
    end ;
    Allocation_Record := _Allocation_Record ;
end ;


procedure TSExtent_Store64.Copy_To_Store( Store : PSExtent_Store64 ) ;

var Buffer : pointer ;
    Count, _Pos, Size : integer ;

begin
    Set_Error( 0 ) ;
    Size := _Header.Clustersize ;
    if( Size < 1024 ) then
    begin
        Size := 1024 ;
    end ;
    getmem( Buffer, Size ) ;
    if( Buffer = nil ) then { Clustersize must be too large }
    begin
        { Untested }
        Size := 1024 ;
        getmem( Buffer, Size ) ;
        if( Buffer = nil ) then
        begin
            Set_Error( ExtentStore64Err_No_Memory ) ;
            exit ;
        end ;
    end ;
    Count := _Header.Size ;
    _Pos := 0 ;
    while( Count > Size ) do
    begin
        Read_Data( Buffer^, _Pos, Size, Last_Error ) ;
        if( Last_Error.Code <> 0 ) then
        begin
            freemem( Buffer, Size ) ; { UNTESTED }
            exit ;
        end ;
        Store^.Write_Data( Buffer^, _Pos, Size, Last_Error ) ;
        if( Last_Error.Code <> 0 ) then
        begin
            freemem( Buffer, Size ) ; { UNTESTED }
            exit ;
        end ;
        Count := Count - Size ;
        _Pos := _Pos + Size ;
    end ; { while( Count > Size ) }
    if( Count > 0 ) then
    begin
        Read_Data( Buffer^, _Pos, Count, Last_Error ) ;
        if( Last_Error.Code <> 0 ) then
        begin
            freemem( Buffer, Size ) ; { Untested }
            exit ;
        end ;
        Store^.Write_Data( Buffer^, _Pos, Count, Last_Error ) ;
    end ;
    freemem( Buffer, Size ) ;
end ; { TSExtent_Store64.Copy_To_Store }


function TSExtent_Store64.Get_Absolute( Offset : longint ) : TStore_Address64 ;
{ Translate an extent-relative offset into an address on _Store.
  Offset = Cluster offset within extent. }

var Allocation : TStore_Address64 ;
    _Record : PAllocation_Record64 ;
    UEC : TUEC ;

begin
    if( Get_Contiguous_Flag ) then
    begin
        Get_Absolute := _Header.Allocation_Info + ( Offset * _Header.Clustersize ) ; { We should never get here } { UNTESTED }
        exit ;
    end ;

    _Record := Allocation_Record ; { Allocate a buffer }

    { Get the first allocation record }
    _Store^.Read_Data( _Record^, _Header.Allocation_Info, _Allocation_Record_Size, UEC ) ;

    while( Offset >= Clusters_In_Block ) do
    begin
	Offset := Offset - Clusters_In_Block ;
	Allocation := _Record^[ Next_Record_Offset ] ;
	_Store^.Read_Data( _Record^, Allocation, _Allocation_Record_Size, UEC ) ;
	if( Allocation = 0 ) then
	begin
	    Get_Absolute := 0 ; { Beyond end of data } { Note: We should never get here } { UNTESTED }
	    exit ;
	end ;
    end ;
    Get_Absolute := _Record^[ Offset ] ;
    {TODO: Keep track of where we are for performance}
end ; { TSExtent_Store64.Get_Absolute }


procedure TSExtent_Store64.Set_Error( Value : integer ) ;

begin
    Last_Error.Facility := ExtentStore64Err_Facility ;
    Last_Error.Code := Value ;
end ;


procedure TSExtent_Store64.Update_Header ; { Write header to store }

var UEC : TUEC ;

begin
    if( _Store <> nil ) then
    begin
        if( _Root = 0 ) then
        begin
            _Root := _Store.Allocate( sizeof( _Header ) ) ;
        end ;
        _Store^.Write_Data( _Header, _Root, sizeof( _Header ), UEC ) ;
    end ;
end ;


{ Overrides... }

function TSExtent_Store64.Read_Data( var Data ; Address, Len : TStore_Address64 ;
    var UEC : TUEC ) : TStore_Address64 ;

var _Buffer : PByte_Array ;
    Cluster : longint ; { Current cluster being read }
    Count : TStore_Address64 ;
    Dummy, Dummy1 : longint ;
    Offset : integer ; { Source buffer offset }
    This_Count : word ; { Count of bytes to move for current cluster }

begin
    Set_Error( 0 ) ;
    Read_Data := 0 ;
    if( _Store = nil ) then
    begin
        Set_Error( ExtentStore64Err_No_Store ) ;
        UEC := Last_Error ;
        exit ;
    end ;
    if( Len < 0 ) then
    begin
        Set_Error( ExtentStore64Err_Invalid_Length ) ;
        UEC := Last_Error ;
        exit ;
    end ;
    if( Len = 0 ) then
    begin
        exit ;
    end ;

    { Setup... }
    Set_Error( 0 ) ;
    UEC := Last_Error ;
    if( Address + Len > _Header.Size ) then { Need to truncate passed length }
    begin
	Len := _Header.Size - Address ; { Only read what is there }
        if( Len < 0 ) then
        begin
            Set_Error( ExtentStore64Err_End_Of_Data ) ;
            UEC := Last_Error ;
            exit ;
        end ;
    end ;
    if( Len = 0 ) then
    begin
        exit ;
    end ;

    if( Get_Contiguous_Flag ) then
    begin
        Read_Data := _Store^.Read_Data( Data, _Header.Allocation_Info + Address, Len, UEC ) ;
        Last_Error := UEC ;
        exit ;
    end ;

    { Now read from store... }
    getmem( _Buffer, _Header.Clustersize ) ; { Allocate a cluster-sized buffer }
    Cluster := Address div _Header.Clustersize ; { Cluster to start reading from }
    Dummy1 := Address - ( Cluster * _Header.Clustersize ) ; { Offset }
    Offset := 0 ; { Offset in source buffer }
    Count := 0 ;
    while( Len > 0 ) do { Until done reading from store }
    begin
	Dummy := Get_Absolute( Cluster ) ;
        if( Dummy <= 0 ) then { This should not happen }
        begin
            Set_Error( ExtentStore64Err_Corruption ) ; { UNTESTED }
            exit ;
        end ;
	Count := Count + _Store^.Read_Data( _Buffer^, Dummy, _Header.Clustersize, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Last_Error := UEC ; { UNTESTED }
            exit ;
        end ;
	This_Count := _Header.Clustersize - Dummy1 ; { Amount remaining in cluster }
	if( This_Count > Len ) then
	begin
	    This_Count := Len ; { Move no more than asked to }
	end ;
	move( _Buffer^[ Dummy1 ], TByte_Array( Data )[ Offset ], This_Count ) ;
	Offset := Offset + This_Count ; { New source position }
	Len := Len - This_Count ; { Bytes left to read }
	Cluster := Cluster + 1 ; { Next cluster next time }
	Dummy1 := 0 ; { Next source starts at offset 0 }
    end ;
    freemem( _Buffer, _Header.Clustersize ) ; { Deallocate it }
    Read_Data := Count ;
end ; { TSExtent_Store64.Read_Data }


function TSExtent_Store64.Write_Data( var Data ; Address, Len : TStore_Address64 ;
    var UEC : TUEC ) : TStore_Address64 ;
{ Write data from Data to specified Address in store.
  Write Size bytes.  Returns success of operation in UEC.
  Function returns the number of bytes actually written. }

var _Buffer : PByte_Array ;
    Cluster : longint ; { Current cluster being allocated }
    Count : TStore_Address64 ;
    Dummy, Dummy1 : longint ;
    Offset : integer ; { Source buffer offset }
    This_Count : word ; { Count of bytes to move for current cluster }

begin
    Write_Data := 0 ;
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentStore64Err_Read_Only ) ;
        UEC := Last_Error ;
        exit ;
    end ;
    if( _Store = nil ) then
    begin
        Set_Error( ExtentStore64Err_No_Store ) ;
        UEC := Last_Error ;
        exit ;
    end ;

    { Setup... }
    Set_Error( 0 ) ;
    UEC := Last_Error ;

    if( Address + Len > _Header.Size ) then { Need to extend }
    begin
	Extend( ( Address + Len ) - _Header.Size ) ;
        if( Last_Error.Code <> 0 ) then { Could not extend the store }
        begin
            UEC := Last_Error ; { UNTESTED }
            exit ;
        end ;
        if( Address + Len > _Header.Size ) then { Did not extend enough }
        begin
            Set_Error( ExtentStore64Err_Cannot_Extend ) ; { UNTESTED }
            UEC := Last_Error ;
            exit ;
        end ;
    end ;
    if( Len = 0 ) then
    begin
        exit ;
    end ;

    if( Get_Contiguous_Flag ) then
    begin
        Write_Data := _Store^.Write_Data( Data, Address + _Header.Allocation_Info, Len, UEC ) ;
        Last_Error := UEC ;
        exit ;
    end ;

    { Now write data to store... }
    getmem( _Buffer, _Header.Clustersize ) ; { Allocate a cluster-sized buffer }
    Cluster := Address div _Header.Clustersize ; { Cluster to start writing to }
    Dummy1 := Address - ( Cluster * _Header.Clustersize ) ; { Offset }
    Offset := 0 ; { Offset in source buffer }
    Count := 0 ;
    while( Len > 0 ) do { Until done writing to store }
    begin
	Dummy := Get_Absolute( Cluster ) ;
        if( Dummy <= 0 ) then { This should not happen }
        begin
            Set_Error( ExtentStore64Err_Corruption ) ; { UNTESTED }
            exit ;
        end ;
	_Store^.Read_Data( _Buffer^, Dummy, _Header.Clustersize, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            exit ; { UNTESTED }
        end ;
	This_Count := _Header.Clustersize - Dummy1 ;
	if( This_Count > Len ) then
	begin
	    This_Count := Len ; { Move no more than asked to }
	end ;
	move( TByte_Array( Data )[ Offset ], _Buffer^[ Dummy1 ], This_Count ) ;
	Count := Count + _Store^.Write_Data( _Buffer^, Dummy, _Header.Clustersize, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            exit ; { UNTESTED }
        end ;
	Offset := Offset + This_Count ; { New source position }
	Len := Len - This_Count ; { Bytes left to write }
	Cluster := Cluster + 1 ; { Next cluster next time }
	Dummy1 := 0 ; { Next destination starts at offset 0 }
    end ;
    Freemem( _Buffer, _Header.Clustersize ) ;
    Write_Data := Count ;
end ; { TSExtent_Store64.Write_Data }


function TSExtent_Store64.Max_Storage : TStore_Size64 ;
{ Returns the maximum (current) size of the store, in bytes.  In other words,
  this is the highest valid address (plus 1). }

begin
    Max_Storage := _Header.Size ;
end ;


function TSExtent_Store64.Min_Storage : TStore_Address64 ;

var L : TStore_Address64 ;

begin
    if( _Store <> nil ) then
    begin
        L := _Store^.Min_Storage ;
        if( L < _Header.Clustersize ) then
        begin
            L := _Header.Clustersize ;
        end ;
        Min_Storage := L ;
    end else
    begin
        Min_Storage := _Header.Clustersize ;
    end ;
end ;


function TSExtent_Store64.Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
{ Requests extension of store by the amount specified.  Returns the amount
  actually added to store.  If Amount is 0, function returns 0 if store is
  extendable, and -1 if it is not. }

var Cluster : TStore_Address ; { Current cluster being allocated }
    Current : TStore_Address64 ; { Current allocation record pointer }
    Dummy, Dummy1 : TStore_Address64 ;
    New_Size : TStore_Address64 ;
    _Record : PAllocation_Record64 ;

begin
    Extend := 0 ;
    if( Amount = 0 ) then
    begin
        if( ( _Store <> nil ) and ( _Store^.Read_Only ) ) then
        begin
            Extend := -1 ;
        end else
        if( Get_Read_Only_Flag ) then
        begin
            Extend := -1 ;
        end ;
        exit ;
    end ;
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentStore64Err_Read_Only ) ;
        exit ;
    end ;

    { Determine new size (rounded to clustersize...) }
    New_Size := _Header.Size + Amount ;
    if( New_Size <> _Header.Clustersize ) then
    begin
        New_Size := ( New_Size + _Header.Clustersize - 1 ) ;
        New_Size := New_Size div _Header.Clustersize ;
        New_Size := New_Size * _Header.Clustersize ;
    end ;
    if( _Store = nil ) then
    begin
        _Header.Size := New_Size ;
        exit ;
    end ;

    if( _Store^.XSpaceAvail < New_Size - _Header.Size ) then { No room }
    begin
	Set_Error( ExtentStore64Err_Cannot_Extend ) ; { UNTESTED }
	exit ;
    end ;

    if( _Header.Size <> 0 ) then
    begin
        Set_Contiguous_Flag( False ) ; { Extend removes contiguous flag on non-zero-length stores }
    end ;
    if( Get_Contiguous_Flag ) then
    begin
        _Header.Allocation_Info := _Store^.Allocate( New_Size ) ;
	if( _Header.Allocation_Info = 0 ) then
	begin
	    Set_Error( ExtentStore64Err_No_Room ) ; { UNTESTED }
	    exit ;
	end ;
        _Header.Size := New_Size ;
        Update_Header ;
        exit ;
    end ;

    _Record := Allocation_Record ; { Allocate a buffer }

    { Now allocate additional space... }
    if( _Header.Allocation_Info = 0 ) then { Store is empty }
    begin
 	_Header.Allocation_Info := Allocate_Entry ;
	if( _Header.Allocation_Info = 0 ) then
	begin
	    Set_Error( ExtentStore64Err_No_Room ) ; { UNTESTED }
	    exit ;
	end ;
        Update_Header ;
    end ;
    _Store^.Read_Data( _Record^, _Header.Allocation_Info, _Allocation_Record_Size, Last_Error ) ;
    if( Last_Error.Code <> 0 ) then
    begin
        exit ; { UNTESTED }
    end ;

    { TODO: For performance, start at current end of allocation pointers... }
    Current := _Header.Allocation_Info ;
    Dummy := New_Size div _Header.Clustersize ; { Maximum used clusters }
    Dummy1 := 1 ;
    while( Dummy1 <= Dummy ) do
    begin
	for Cluster := 0 to Clusters_In_Block - 1 do
        begin
            if( Dummy1 <= Dummy ) then { Still more clusters to allocate }
            begin
                if( _Record^[ Cluster ] = 0 ) then { Cluster not yet allocated }
                begin
                    _Record^[ Cluster ] := _Store^.Allocate( _Header.Clustersize ) ;
                    if( _Record^[ Cluster ] = 0 ) then
                    begin
                        Set_Error( ExtentStore64Err_No_Room ) ; { TEST: Untested }
                        exit ;
                    end ;
                end ; { if( _Record^[ Cluster ] = 0 ) }
                Dummy1 := Dummy1 + 1 ;
            end else
            begin
                break ; { Once we are done, skip of out loop }
            end ; { if( Dummy1 <= Dummy ) }
        end ; { for Cluster := 0 to Clusters_In_Block - 1 }
	_Store^.Write_Data( _Record^, Current, _Allocation_Record_Size, Last_Error ) ;
        if( Last_Error.Code <> 0 ) then
        begin
            exit ; { TEST: Untested }
        end ;
	if( Dummy1 <= Dummy ) then { More allocation needed }
	begin
	    if( _Record^[ Next_Record_Offset ] = 0 ) then
	    begin
		_Record^[ Next_Record_Offset ] := Allocate_Entry ;
		if( _Record^[ Next_Record_Offset ] = 0 ) then
		begin
		    Set_Error( ExtentStore64Err_No_Room ) ; { TEST: Untested }
		    exit ;
		end ;
		_Store^.Write_Data( _Record^, Current, _Allocation_Record_Size, Last_Error ) ;
                if( Last_Error.Code <> 0 ) then
                begin
                    exit ; { TEST: Untested }
                end ;
		Current := _Record^[ Next_Record_Offset ] ;
		fillchar( _Record^, _Allocation_Record_Size, 0 ) ;
	    end else
	    begin
		_Store^.Write_Data( _Record^, Current, _Allocation_Record_Size, Last_Error ) ;
                if( Last_Error.Code <> 0 ) then
                begin
                    exit ; { TEST: Untested }
                end ;
		Current := _Record^[ Next_Record_Offset ] ;
		_Store^.Read_Data( _Record^, Current, _Allocation_Record_Size, Last_Error ) ;
                if( Last_Error.Code <> 0 ) then
                begin
                    exit ; { TEST: Untested }
                end ;
	    end ; { if( _Record^[ Next_Record_Offset ] = 0 ) }
	end ; { if( Dummy1 <= Dummy ) }
    end ; { while( Dummy1 <= Dummy ) }
    _Store^.Write_Data( _Record^, Current, _Allocation_Record_Size, Last_Error ) ;
    if( Last_Error.Code <> 0 ) then
    begin
        exit ; { TEST: Untested }
    end ;

    _Header.Size := New_Size ;
    Update_Header ;
end ; { TSExtent_Store64.Extend }


function TSExtent_Store64.Read_Only : boolean ;

begin
    if( ( _Store = nil ) or ( _Store^.Read_Only ) ) then
    begin
        Read_Only := True ;
    end else
    begin
        Read_Only := Get_Read_Only_Flag ;
    end ;
end ;


function TSExtent_Store64.Write_Only : boolean ;

begin
    if( _Store <> nil ) then
    begin
        Write_Only := _Store^.Write_Only ;
    end else
    begin
        Write_Only := False ;
    end ;
end ;


procedure TSExtent_Store64.Initialize ;

begin
    _Store := nil ;
    _Root := 0 ;
    fillchar( _Header, sizeof( _Header ), 0 ) ;
    _Header.Facility := ExtentStore64Err_Facility ;
    _Header.Record_Type := 0 ;
    _Header.Version := 0 ;
    _Header.Size := 0 ;
    _Header.Flags := 0 ;
    _Header.Clustersize := 1 ;
    _Header.Allocation_Info := 0 ;
    _Allocation_Record := nil ;
    fillchar( Last_Error, sizeof( Last_Error ), 0 ) ;
end ;


procedure TSExtent_Store64.Terminate ;

begin
    Set_Store( nil ) ;
end ;


function TSExtent_Store64.Contiguous_Store : boolean ;

begin
    Contiguous_Store := True ;
end ;


procedure TSExtent_Store64.Set_Max_Storage( Value : TStore_Address64 ;
    var Res : TUEC ) ;
{ Sets the maximum (current) size of the store, in bytes.  In other words, this
  is the new highest valid address (plus 1).  If the store is not resizable, the
  call fails and an error code is returned in Res.  This should only be used to
  shrink the size of the store, not to extend it.  To extend, use the Extend
  method. }

var Current : longint ;
    _Record : PAllocation_Record64 ;
    Next : TStore_Address64 ;
    Offset : longint ;
    UEC : TUEC ;

begin
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentStore64Err_Read_Only ) ;
        exit ;
    end ;
    if( _Store = nil ) then
    begin
        _Header.Size := Value ;
        exit ;
    end ;

    { Setup... }
    if( Value < 0 ) then
    begin
        Value := 0 ;
    end ;
    Set_Error( 0 ) ;
    Res := Last_Error ;
    if( Value <> _Header.Clustersize ) then
    begin
        Value := ( Value + _Header.Clustersize - 1 ) ;
        Value := Value div _Header.Clustersize ;
        Value := Value * _Header.Clustersize ;
    end ;

    { Resize... }
    if( Value > _Header.Size ) then { Expand }
    begin
        Extend( Value - _Header.Size ) ;
    end else
    if( Value < _Header.Size ) then { Truncate }
    begin
        if( Get_Contiguous_Flag ) then
        begin
            _Store.Reallocate( _Header.Allocation_Info, _Header.Size, Value ) ;
            _Header.Size := Value ; { New size }
            Update_Header ;
            exit ;
        end ;

        _Header.Size := Value ; { New size }
        Update_Header ;
        _Record := Allocation_Record ; { Allocate a buffer }

        { Get the first allocation record }
        _Store^.Read_Data( _Record^, _Header.Allocation_Info, _Allocation_Record_Size, UEC ) ;

        { Free unused (previously allocated) clusters in last allocation block... }
        Offset := Value div _Header.Clustersize ;
        Current := _Header.Allocation_Info ;
        while( Offset >= Clusters_In_Block ) do { Move to allocation block where truncation begins }
        begin
            Offset := Offset - Clusters_In_Block ;
            Current := _Record^[ Next_Record_Offset ] ;
            if( Current = 0 ) then 
            begin
                exit ; { Beyond end of data.  This should never happen } { Test: UNTESTED }
            end ;
            _Store^.Read_Data( _Record^, Current, _Allocation_Record_Size, UEC ) ;
        end ;
        while( Offset < Next_Record_Offset ) do
        begin
            if( _Record^[ Offset ] <> 0 ) then
            begin
                _Store.Deallocate( _Record^[ Offset ], _Header.Clustersize ) ;
                _Record^[ Offset ] := 0 ;
            end ;
            inc( Offset ) ;
        end ;
        Next := _Record^[ Next_Record_Offset ] ;
        _Record[ Next_Record_Offset ] := 0 ;
        _Store^.Write_Data( _Record^, Current, _Allocation_Record_Size, UEC ) ;

        { Delete all remaining allocation blocks... }
        Current := Next ;
        while( Current <> 0 ) do
        begin
            _Store^.Read_Data( _Record^, Next, _Allocation_Record_Size, UEC ) ;
            for Offset := 0 to Clusters_In_Block - 1 do
            begin
                if( _Record^[ Offset ] <> 0 ) then
                begin
                    _Store.Deallocate( _Record^[ Offset ], _Header.Clustersize ) ;
                end ;
            end ;
            Next := _Record^[ Next_Record_Offset ] ;
            _Store^.Deallocate( Current, _Allocation_Record_Size ) ;
            Current := Next ;
        end ;
        if( _Header.Size = 0 ) then { Get rid of first allocation record }
        begin
            _Store^.Deallocate( _Header.Allocation_Info, _Allocation_Record_Size ) ;
            _Header.Allocation_Info := 0 ;
            Update_Header ;
        end ;
    end ; { if( Value < _Header.Size ) }
end ; { TSExtent_Store64.Set_Max_Storage }


function TSExtent_Store64.Extended_Size : TStore_Address64 ;
{ Returns the maximum theoretical size that the store can be extended to.  If
  the store is not extendable, this returns the same value as Max_Storage. }

var _Result : TStore_Address64 ;

begin
    if( _Store = nil ) then
    begin
        _Result := Max_Storage ;
    end else
    begin
        _Result := _Header.Size + _Store^.XSpaceAvail ;
        if( _Result < 0 ) then { UNTESTED: Else - spaceavail on heap is always $7FFFFFFF, plus _Header.Size is < 0 on 32-bit systems }
        begin
            _Result := $7FFFFFFFFFFFFFFF ;
        end ;
    end ;
    Extended_Size := _Result ;
end ;


function TSExtent_Store64.Get_Clustersize : TStore_Address64 ;

begin
    Get_Clustersize := _Header.Clustersize ;
end ;


function TSExtent_Store64.Get_Contiguous_Flag : boolean ;

begin
    Get_Contiguous_Flag := ( ( _Header.Flags and HF_Contiguous ) <> 0 ) ;
end ;


function TSExtent_Store64.Get_Placed_Flag : boolean ;

begin
    Get_Placed_Flag := ( ( _Header.Flags and HF_Placed ) <> 0 ) ;
end ;


function TSExtent_Store64.Get_Read_Only_Flag : boolean ;

begin
    Get_Read_Only_Flag := ( ( _Header.Flags and HF_Read_Only ) <> 0 ) ;
end ;


function TSExtent_Store64.Get_Write_Only_Flag : boolean ;

begin
    Get_Write_Only_Flag := ( ( _Header.Flags and HF_Write_Only ) <> 0 ) ;
end ;


function TSExtent_Store64.Get_Root : TStore_Address64 ;

begin
    Get_Root := _Root ;
end ;


function TSExtent_Store64.Get_Store : PSStore64 ;

begin
    Get_Store := _Store ;
end ;


procedure TSExtent_Store64.Set_Clustersize( Value : TStore_Address64 ) ;

var Saved : TStore_Address64 ;
    Temp : TSExtent_Store64 ;

begin
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentStore64Err_Read_Only ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    if( Value < 0 ) then
    begin
        Value := 1 ;
    end ;
    if( _Store <> nil ) then
    begin
        if( Value < _Store.Min_Storage ) then
        begin
            Value := _Store.Min_Storage ; { Cannot be smaller clustersize than _Store }
        end ;
    end ;
    if( Value <> _Header.Clustersize ) then
    begin
        if( _Header.Size > 0 ) then
        begin
            { Copy data to new list with new clustersize... }
            Temp.Init ;
            Temp.Initialize ;
            Temp.Set_Clustersize( Value ) ;
            Temp.Set_Store( _Store ) ;
            Copy_To_Store( @Temp ) ;
            if( Last_Error.Code <> 0 ) then
            begin
                { UNTESTED }
                Temp.Delete ;
                Temp.Done ;
                exit ;
            end ;

            { Delete existing data... }
            Saved := _Header.Size ;
            Set_Max_Storage( 0, Last_Error ) ;
            _Header.Size := Saved ;
        end ;

        { Update header... }
        _Header.Clustersize := Value ;
        if( _Header.Size > 0 ) then
        begin
            _Header.Allocation_Info := Temp._Header.Allocation_Info ;
            Temp.Set_Store( nil ) ;
            Temp.Done ;
        end ;
        Update_Header ;
    end ;
end ; { TSExtent_Store64.Set_Clustersize }


procedure TSExtent_Store64.Set_Contiguous_Flag( Value : boolean ) ;

var Current, Last, Original, _Size : TStore_Address64 ;
    Index : longint ;
    _Record : PAllocation_Record64 ;
    Temp : TSExtent_Store64 ;

begin
    Set_Error( 0 ) ;
    if( Value <> Get_Contiguous_Flag ) then
    begin
        if( Value ) then { Converting to contiguous }
        begin
            if( _Header.Size > 0 ) then
            begin
                Temp.Init ;
                Temp.Initialize ;
                Temp.Set_Clustersize( _Header.Clustersize ) ;
                Temp.Set_Contiguous_Flag( Value ) ;
                Temp.Set_Max_Storage( _Header.Size, Last_Error ) ;
                Temp.Set_Store( _Store ) ;
                Copy_To_Store( @Temp ) ;
                if( Last_Error.Code <> 0 ) then
                begin
                    Temp.Delete ; { UNTESTED }
                    Temp.Done ;
                    exit ;
                end ;
            end ;
            _Header.Flags := _Header.Flags or HF_Contiguous ;
            Set_Max_Storage( _Header.Size, Last_Error ) ;
            if( _Header.Size > 0 ) then
            begin
                _Header.Allocation_Info := Temp._Header.Allocation_Info ;
                Temp.Set_Store( nil ) ;
                Temp.Done ;
            end ;
        end else
        begin
            { Converting from contiguous to non-contiguous is easy: we simply
              create an allocation chain that points to the existing data.  This
              keeps contiguous data pysically contiguous, even if not logically
              so. }
            Current := _Header.Allocation_Info ; { Start of contiguous data }
            _Size := _Header.Size ;
            Last := Allocate_Entry ; { Create initial allocation record }
            if( Last = 0 ) then
            begin
                Set_Error( ExtentStoreErr_No_Room ) ; { UNTESTED }
                exit ;
            end ;
            Original := _Header.Allocation_Info ;
            _Header.Allocation_Info := Last ;
            _Record := Allocation_Record ; { Allocate a buffer }
            Last := _Header.Allocation_Info ;
            _Store^.Read_Data( _Record^, Last, _Allocation_Record_Size, Last_Error ) ;
            Index := 0 ;
            while( _Size > 0 ) do
            begin
                _Record[ Index ] := Current ;
                Current := Current + _Header.Clustersize ;
                _Size := _Size - _Header.Clustersize ;
                inc( Index ) ;
                if( ( Index >= Clusters_In_Block - 1 ) and ( _Size > 0 ) ) then { End of allocation block but not of data }
                begin
                    _Record[ Clusters_In_Block - 1 ] := Allocate_Entry ; { Create next allocation record in chain }
                    if( _Record[ Clusters_In_Block - 1 ] = 0 ) then
                    begin
                        Set_Error( ExtentStoreErr_No_Room ) ; { UNTESTED }
                        _Header.Allocation_Info := Original ;
                        exit ;
                    end ;
                    _Store^.Write_Data( _Record^, Last, _Allocation_Record_Size, Last_Error ) ;
                    Last := _Record[ Clusters_In_Block - 1 ] ;
                    _Store^.Read_Data( _Record^, Last, _Allocation_Record_Size, Last_Error ) ;
                    Index := 0 ;
                end ; { if }
            end ; { while( _Size > 0 ) }
            _Store^.Write_Data( _Record^, Last, _Allocation_Record_Size, Last_Error ) ;
            _Header.Flags := _Header.Flags and ( not HF_Contiguous ) ;
        end ; { if( Value ) }
        Update_Header ;
    end ; { if( Value <> Get_Contiguous_Flag ) }
end ; { TSExtent_Store64.Set_Contiguous_Flag }


procedure TSExtent_Store64.Set_Placed_Flag( Value : boolean ) ;

begin
    Set_Error( 0 ) ;
    if( Value <> Get_Placed_Flag ) then
    begin
        if( Value ) then
        begin
            _Header.Flags := _Header.Flags or HF_Placed ;
        end else
        begin
            _Header.Flags := _Header.Flags and ( not HF_Placed ) ;
        end ;
        Update_Header ;
    end ;
end ;


procedure TSExtent_Store64.Set_Read_Only_Flag( Value : boolean ) ;

begin
    Set_Error( 0 ) ;
    if( Value <> Get_Read_Only_Flag ) then
    begin
        if( Value ) then
        begin
            _Header.Flags := _Header.Flags or HF_Read_Only ;
        end else
        begin
            _Header.Flags := _Header.Flags and ( not HF_Read_Only ) ;
        end ;
        Update_Header ;
    end ;
end ;


procedure TSExtent_Store64.Set_Root( Value : integer ) ;

begin
    Set_Error( 0 ) ;
    _Root := Value ;
end ;


procedure TSExtent_Store64.Set_Store( S : PSManaged_Store64 ) ;

var Saved : TStore_Address64 ;

begin
    if( _Store <> S ) then
    begin
        Saved := _Header.Size ;
        _Store := S ;
        _Root := 0 ;
        _Header.Allocation_Info := 0 ;
        if( _Store <> nil ) then
        begin
            Update_Header ;
        end ;
        if( _Allocation_Record <> nil ) then
        begin
            freemem( _Allocation_Record, _Allocation_Record_Size ) ;
            _Allocation_Record := nil ;
        end ;
        if( S <> nil ) then
        begin
            _Header.Size := 0 ; { So following will actually extend the store }
            Set_Max_Storage( Saved, Last_Error ) ;
        end ;
    end ;
end ;


procedure TSExtent_Store64.Load_From_Store( S : PSManaged_Store64 ;
    R : TStore_Address64 ) ; { UNTESTED }

var UEC : TUEC ;

begin
    _Store := S ;
    _Root := R ;
    if( _Root = 0 ) then
    begin
        _Store^.Read_Data( _Header, _Root, sizeof( _Header ), UEC )
    end ;
end ;


procedure TSExtent_Store64.Delete ;

begin
    if( Get_Read_Only_Flag ) then
    begin
        Set_Error( ExtentStore64Err_Read_Only ) ;
        exit ;
    end ;
    Set_Max_Storage( 0, Last_Error ) ;
    if( _Store <> nil ) then
    begin
        _Store^.Deallocate( _Root, sizeof( _Header ) ) ;
        _Root := 0 ;
    end ;
end ;


procedure TSExtent_Store64.Set_Read_Only( Value : boolean ) ;

begin
    Set_Read_Only_Flag( Value ) ;
end ;


procedure TSExtent_Store64.Set_Write_Only( Value : boolean ) ;

begin
    Set_Error( 0 ) ;
    if( Value <> Get_Write_Only_Flag ) then
    begin
        if( Value ) then
        begin
            _Header.Flags := _Header.Flags or HF_Write_Only ;
        end else
        begin
            _Header.Flags := _Header.Flags and ( not HF_Write_Only ) ;
        end ;
        Update_Header ;
    end ;
end ;


{$IFDEF Test}
procedure Test_Unit ;

    procedure Test_Extent_Store ;

        procedure Error ;

        begin
            writeln( 'Error' ) ;
            halt ;
        end ;


        procedure Check_UEC( UEC : TUEC ; F, C : integer ) ;

        begin
            if( ( UEC.Facility <> F ) or ( UEC.Code <> C ) ) then
            begin
                Error ;
            end ;
        end ;

    var S : string[ 255 ] ;
        Test_Store : TSExtent_Store ;

    procedure Check_Contents( Size : integer ) ;

    var S1 : string[ 255 ] ;
        UEC : TUEC ;

    begin
        S1 := '                                ' ;
        S1[ 0 ] := chr( Size ) ;
        Test_Store.Read_Data( S1[ 1 ], 0, Size, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( S1 <> copy( S, 1, Size ) ) then
        begin
            Error ;
        end ;
    end ;


    var Dummy : integer ;
        F : file ;
        Managed_Store : TSHeap ;
        FH_Store : TFH ;
        S1 : string[ 255 ] ;
        UEC : TUEC ;

    begin
        Managed_Store.Init ;
        Managed_Store.Initialize ;
        Test_Store.Init ;
        Test_Store.Initialize ;
        assign( F, 'c:\lb\qc\testextstore.dat' ) ;
        rewrite( F ) ;
        close( F ) ;
        FH_Store.Init( 1, 0, 'c:\lb\qc\testextstore.dat', UEC, @Managed_Store ) ;
        FH_Store.Initialize ;

        { Tests with no store attached... }
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Extended_Size <> Test_Store.Max_Storage ) then
        begin
            Error ;
        end ;

        if( Test_Store.Max_Storage <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Min_Storage <> 1 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Get_Clustersize <> 1 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        Test_Store.Delete ;
        if( Test_Store.Max_Storage <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Get_Clustersize <> 1 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Get_Contiguous_Flag ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Get_Read_Only_Flag ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Get_Root <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Root( 1 ) ;
        if( Test_Store.Get_Root <> 1 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Get_Store <> nil ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( not Test_Store.Contiguous_Store ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( not Test_Store.Read_Only ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Write_Only ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Get_Placed_Flag ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Placed_Flag( True ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( not Test_Store.Get_Placed_Flag ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Placed_Flag( False ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Get_Placed_Flag ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Placed_Flag( False ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Get_Placed_Flag ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        Test_Store.Extend( 12 ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Max_Storage <> 12 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Extend( 0 ) = -1 ) then
        begin
            Error ;
        end ;

        Test_Store.Set_Read_Only_Flag( True ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Extend( 0 ) = 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Extend( 1 ) ;
        if( Test_Store.Last_Error.Code = 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Write_Data( Dummy, 0, 0, UEC ) ;
        Check_UEC( UEC, 66, 6 ) ;
        Check_UEC( Test_Store.Last_Error, 66, 6 ) ;
        Test_Store.Set_Clustersize( 0 ) ;
        Check_UEC( Test_Store.Last_Error, 66, 6 ) ;
        Test_Store.Delete ;
        Check_UEC( Test_Store.Last_Error, 66, 6 ) ;
        Test_Store.Set_Max_Storage( 12, UEC ) ;
        Check_UEC( UEC, 66, 6 ) ;
        Check_UEC( Test_Store.Last_Error, 66, 6 ) ;

        Test_Store.Set_Read_Only_Flag( False ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Read_Only_Flag( False ) ; { Repeat operation }
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        Test_Store.Read_Data( Dummy, 0, 0, UEC ) ;
        Check_UEC( UEC, 66, 1 ) ;
        Check_UEC( Test_Store.Last_Error, 66, 1 ) ;
        Test_Store.Write_Data( Dummy, 0, 0, UEC ) ;
        Check_UEC( UEC, 66, 1 ) ;
        Check_UEC( Test_Store.Last_Error, 66, 1 ) ;

        Test_Store.Set_Store( @Managed_Store ) ;
        if( Test_Store.Min_Storage <> 16 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Read_Only ) then
        begin
            Error ;
        end ;
        Managed_Store.Set_Read_Only( True ) ;
        if( Test_Store.Extend( 0 ) <> -1 ) then
        begin
            Error ;
        end ;
        Managed_Store.Set_Read_Only( False ) ;

        S := '1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZ' ;
        Test_Store.Write_Data( S[ 1 ], 0, 15, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Max_Storage <> 15 ) then
        begin
            Error ;
        end ;
        setlength( S1, 15 ) ;
        Test_Store.Read_Data( S1[ 1 ], 0, 15, UEC ) ; { Read what we just wrote }
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( copy( S, 1, 15 ) <> S1 ) then
        begin
            Error ;
        end ;
        Test_Store.Read_Data( S1[ 1 ], 0, 31, UEC ) ; { Try to read more than exists }
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( copy( S, 1, 15 ) <> S1 ) then
        begin
            Error ;
        end ;
        Test_Store.Read_Data( S1[ 1 ], 0, 0, UEC ) ; { Try to read 0 bytes }
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Write_Data( S[ 1 ], 0, 16, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Max_Storage <> 16 ) then
        begin
            Error ;
        end ;
        setlength( S1, 16 ) ;
        Test_Store.Read_Data( S1[ 1 ], 0, 17, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( copy( S, 1, 16 ) <> S1 ) then
        begin
            Error ;
        end ;
        Test_Store.Write_Data( S[ 1 ], 0, 17, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Max_Storage <> 17 ) then
        begin
            Error ;
        end ;
        setlength( S1, 17 ) ;
        Test_Store.Read_Data( S1[ 1 ], 0, 18, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( copy( S, 1, 17 ) <> S1 ) then
        begin
            Error ;
        end ;
        Test_Store.Write_Data( S[ 1 ], 0, 31, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Max_Storage <> 31 ) then
        begin
            Error ;
        end ;
        setlength( S1, 31 ) ;
        Test_Store.Read_Data( S1[ 1 ], 0, 32, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( copy( S, 1, 31 ) <> S1 ) then
        begin
            Error ;
        end ;
        Test_Store.Write_Data( S[ 1 ], 0, 0, UEC ) ; { Write 0 bytes }
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Write_Data( S[ 1 ], 0, 1, UEC ) ; { Overwrite existing data without extending }
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Read_Data( S1[ 1 ], 0, -2, UEC ) ; { Read negative bytes }
        if( UEC.Code = 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Read_Data( S1[ 1 ], 1024, 32, UEC ) ; { Read past end of data }
        if( UEC.Code = 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Extended_Size <= Test_Store.Max_Storage ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Max_Storage( -1, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Write_Only ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Max_Storage( Test_Store.Max_Storage, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;

        { Convert empty file to contiguous and back }
        Test_Store.Set_Clustersize( 32 ) ;
        Test_Store.Set_Contiguous_Flag( True ) ;

        { Convert non-empty file to contiguous and back }
        Test_Store.Set_Contiguous_Flag( False ) ;
        Test_Store.Write_Data( S[ 1 ], 0, 32, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ; { Read what we just wrote }
        Test_Store.Set_Contiguous_Flag( True ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ; 
        Test_Store.Set_Contiguous_Flag( False ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        if( Test_Store.Get_Write_Only_Flag ) then
        begin
            Error ;
        end ;

        Test_Store.Set_Clustersize( 2048 ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Clustersize( 0 ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Read_Only_Flag( True ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Clustersize( 1 ) ;
        if( Test_Store.Last_Error.Code = 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Read_Only_Flag( False ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Clustersize( -1 ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Read_Only_Flag( True ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Clustersize( 2 ) ;
        if( Test_Store.Last_Error.Code = 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Contiguous_Flag( False ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Max_Storage( 16, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ; { Read what we just wrote }

        Test_Store.Set_Read_Only_Flag( False ) ;
        Test_Store.Delete ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        Test_Store.Write_Data( S[ 1 ], 0, 32, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Contiguous_Flag( True ) ;
        Test_Store.Set_Max_Storage( 8, UEC ) ; { Truncate a contiguous file }
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Contiguous_Flag( False ) ;
        Test_Store.Set_Max_Storage( 256, UEC ) ;
        Test_Store.Set_Max_Storage( 32, UEC ) ;
        Test_Store.Set_Max_Storage( 256, UEC ) ;
        Test_Store.Set_Max_Storage( 64, UEC ) ;

        Test_Store.Set_Store( @FH_Store ) ;
        Test_Store.Write_Data( S[ 1 ], 0, 32, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Clustersize( 64 ) ;
        if( Test_Store.Min_Storage <> 64 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Write_Only( True ) ;
        Test_Store.Set_Write_Only( False ) ;
        Test_Store.Set_Write_Only( False ) ;
        Test_Store.Set_Read_OnlY( False ) ;

        Test_Store.Set_Max_Storage( 256, UEC ) ;
        Test_Store.Set_Contiguous_Flag( True ) ;
        Test_Store.Set_Contiguous_Flag( False ) ;

        Test_Store.Set_Store( nil ) ;
        Test_Store.Terminate ;
        Test_Store.Done ;
        Managed_Store.Done ;
        FH_Store.Terminate ;
        FH_Store.Done ;
    end ;


    procedure Test_Extent_Store64 ;

        procedure Error ;

        begin
            writeln( 'Error' ) ;
            halt ;
        end ;


        procedure Check_UEC( UEC : TUEC ; F, C : integer ) ;

        begin
            if( ( UEC.Facility <> F ) or ( UEC.Code <> C ) ) then
            begin
                Error ;
            end ;
        end ;

    var S : string[ 255 ] ;
        Test_Store : TSExtent_Store64 ;

    procedure Check_Contents( Size : integer ) ;

    var S1 : string[ 255 ] ;
        UEC : TUEC ;

    begin
        S1 := '                                ' ;
        S1[ 0 ] := chr( Size ) ;
        Test_Store.Read_Data( S1[ 1 ], 0, Size, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( S1 <> copy( S, 1, Size ) ) then
        begin
            Error ;
        end ;
    end ;


    var Dummy : integer ;
        F : file ;
        Managed_Store : TSHeap64 ;
        FH_Store : TSBig_FH ;
        S1 : string[ 255 ] ;
        UEC : TUEC ;

    begin
        Managed_Store.Init ;
        Managed_Store.Initialize ;
        Test_Store.Init ;
        Test_Store.Initialize ;
        assign( F, 'c:\lb\qc\testextstore.dat' ) ;
        rewrite( F ) ;
        close( F ) ;
        FH_Store.Init( 1, 0, 'c:\lb\qc\testextstore.dat', UEC, @Managed_Store ) ;
        FH_Store.Initialize ;

        { Tests with no store attached... }
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Extended_Size <> Test_Store.Max_Storage ) then
        begin
            Error ;
        end ;

        if( Test_Store.Max_Storage <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Min_Storage <> 1 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Get_Clustersize <> 1 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        Test_Store.Delete ;
        if( Test_Store.Max_Storage <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Get_Clustersize <> 1 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Get_Contiguous_Flag ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Get_Read_Only_Flag ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Get_Root <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Root( 1 ) ;
        if( Test_Store.Get_Root <> 1 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Get_Store <> nil ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( not Test_Store.Contiguous_Store ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( not Test_Store.Read_Only ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Write_Only ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        if( Test_Store.Get_Placed_Flag ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Placed_Flag( True ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( not Test_Store.Get_Placed_Flag ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Placed_Flag( False ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Get_Placed_Flag ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Placed_Flag( False ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Get_Placed_Flag ) then
        begin
            Error ;
        end ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        Test_Store.Extend( 12 ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Max_Storage <> 12 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Extend( 0 ) = -1 ) then
        begin
            Error ;
        end ;

        Test_Store.Set_Read_Only_Flag( True ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Extend( 0 ) = 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Extend( 1 ) ;
        if( Test_Store.Last_Error.Code = 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Write_Data( Dummy, 0, 0, UEC ) ;
        Check_UEC( UEC, 66, 6 ) ;
        Check_UEC( Test_Store.Last_Error, 66, 6 ) ;
        Test_Store.Set_Clustersize( 0 ) ;
        Check_UEC( Test_Store.Last_Error, 66, 6 ) ;
        Test_Store.Delete ;
        Check_UEC( Test_Store.Last_Error, 66, 6 ) ;
        Test_Store.Set_Max_Storage( 12, UEC ) ;
        Check_UEC( UEC, 66, 6 ) ;
        Check_UEC( Test_Store.Last_Error, 66, 6 ) ;

        Test_Store.Set_Read_Only_Flag( False ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Read_Only_Flag( False ) ; { Repeat operation }
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        Test_Store.Read_Data( Dummy, 0, 0, UEC ) ;
        Check_UEC( UEC, 66, 1 ) ;
        Check_UEC( Test_Store.Last_Error, 66, 1 ) ;
        Test_Store.Write_Data( Dummy, 0, 0, UEC ) ;
        Check_UEC( UEC, 66, 1 ) ;
        Check_UEC( Test_Store.Last_Error, 66, 1 ) ;

        Test_Store.Set_Store( @Managed_Store ) ;
        if( Test_Store.Min_Storage <> 16 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Read_Only ) then
        begin
            Error ;
        end ;
        Managed_Store.Set_Read_Only( True ) ;
        if( Test_Store.Extend( 0 ) <> -1 ) then
        begin
            Error ;
        end ;
        Managed_Store.Set_Read_Only( False ) ;

        S := '1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZ' ;
        Test_Store.Write_Data( S[ 1 ], 0, 15, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Max_Storage <> 15 ) then
        begin
            Error ;
        end ;
        setlength( S1, 15 ) ;
        Test_Store.Read_Data( S1[ 1 ], 0, 15, UEC ) ; { Read what we just wrote }
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( copy( S, 1, 15 ) <> S1 ) then
        begin
            Error ;
        end ;
        Test_Store.Read_Data( S1[ 1 ], 0, 31, UEC ) ; { Try to read more than exists }
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( copy( S, 1, 15 ) <> S1 ) then
        begin
            Error ;
        end ;
        Test_Store.Read_Data( S1[ 1 ], 0, 0, UEC ) ; { Try to read 0 bytes }
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Write_Data( S[ 1 ], 0, 16, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Max_Storage <> 16 ) then
        begin
            Error ;
        end ;
        setlength( S1, 16 ) ;
        Test_Store.Read_Data( S1[ 1 ], 0, 17, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( copy( S, 1, 16 ) <> S1 ) then
        begin
            Error ;
        end ;
        Test_Store.Write_Data( S[ 1 ], 0, 17, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Max_Storage <> 17 ) then
        begin
            Error ;
        end ;
        setlength( S1, 17 ) ;
        Test_Store.Read_Data( S1[ 1 ], 0, 18, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( copy( S, 1, 17 ) <> S1 ) then
        begin
            Error ;
        end ;
        Test_Store.Write_Data( S[ 1 ], 0, 31, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Max_Storage <> 31 ) then
        begin
            Error ;
        end ;
        setlength( S1, 31 ) ;
        Test_Store.Read_Data( S1[ 1 ], 0, 32, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( copy( S, 1, 31 ) <> S1 ) then
        begin
            Error ;
        end ;
        Test_Store.Write_Data( S[ 1 ], 0, 0, UEC ) ; { Write 0 bytes }
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Write_Data( S[ 1 ], 0, 1, UEC ) ; { Overwrite existing data without extending }
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Read_Data( S1[ 1 ], 0, -2, UEC ) ; { Read negative bytes }
        if( UEC.Code = 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Read_Data( S1[ 1 ], 1024, 32, UEC ) ; { Read past end of data }
        if( UEC.Code = 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Extended_Size <= Test_Store.Max_Storage ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Max_Storage( -1, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        if( Test_Store.Write_Only ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Max_Storage( Test_Store.Max_Storage, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;

        { Convert empty file to contiguous and back }
        Test_Store.Set_Clustersize( 32 ) ;
        Test_Store.Set_Contiguous_Flag( True ) ;

        { Convert non-empty file to contiguous and back }
        Test_Store.Set_Contiguous_Flag( False ) ;
        Test_Store.Write_Data( S[ 1 ], 0, 32, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ; { Read what we just wrote }
        Test_Store.Set_Contiguous_Flag( True ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ; 
        Test_Store.Set_Contiguous_Flag( False ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        if( Test_Store.Get_Write_Only_Flag ) then
        begin
            Error ;
        end ;

        Test_Store.Set_Clustersize( 2048 ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Clustersize( 0 ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Read_Only_Flag( True ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Clustersize( 1 ) ;
        if( Test_Store.Last_Error.Code = 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Read_Only_Flag( False ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Clustersize( -1 ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Read_Only_Flag( True ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Clustersize( 2 ) ;
        if( Test_Store.Last_Error.Code = 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Contiguous_Flag( False ) ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ;
        Test_Store.Set_Max_Storage( 16, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Check_Contents( 32 ) ; { Read what we just wrote }

        Test_Store.Set_Read_Only_Flag( False ) ;
        Test_Store.Delete ;
        if( Test_Store.Last_Error.Code <> 0 ) then
        begin
            Error ;
        end ;

        Test_Store.Write_Data( S[ 1 ], 0, 32, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Contiguous_Flag( True ) ;
        Test_Store.Set_Max_Storage( 8, UEC ) ; { Truncate a contiguous file }
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Contiguous_Flag( False ) ;
        Test_Store.Set_Max_Storage( 256, UEC ) ;
        Test_Store.Set_Max_Storage( 32, UEC ) ;
        Test_Store.Set_Max_Storage( 256, UEC ) ;
        Test_Store.Set_Max_Storage( 64, UEC ) ;

        Test_Store.Set_Store( @FH_Store ) ;
        Test_Store.Write_Data( S[ 1 ], 0, 32, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Clustersize( 64 ) ;
        if( Test_Store.Min_Storage <> 64 ) then
        begin
            Error ;
        end ;
        Test_Store.Set_Write_Only( True ) ;
        Test_Store.Set_Write_Only( False ) ;
        Test_Store.Set_Write_Only( False ) ;
        Test_Store.Set_Read_OnlY( False ) ;

        Test_Store.Set_Max_Storage( 256, UEC ) ;
        Test_Store.Set_Contiguous_Flag( True ) ;
        Test_Store.Set_Contiguous_Flag( False ) ;

        Test_Store.Set_Store( nil ) ;
        Test_Store.Terminate ;
        Test_Store.Done ;
        Managed_Store.Done ;
        FH_Store.Terminate ;
        FH_Store.Done ;
    end ;

begin
    Test_Extent_Store ;
    Test_Extent_Store64 ;
end ;
{$ENDIF}


end.

