/* Font Manager routines */

#include <stdio.h>
#include "fontmana.h"

static struct RFonts *Font_List = NULL ; /* Font list root */


int Create_Font( LOGFONT LogFont )
/* Create a font and return an index.  0 indicates a failure */
{
    int Index ;
    LOGFONT Log_Font ;
    TFont *Temp ;
    RFonts *Current, *Previous ;

    /* Create font and get actual LogFont for it */
    Temp = new TFont( &LogFont ) ;
    Temp->GetObject( Log_Font ) ;

    if( Font_List == NULL ) /* First one */
    {
	Font_List = ( RFonts* ) malloc( sizeof( RFonts ) ) ;
	Font_List->Next = NULL ;
	Font_List->Usage_Count = 1 ;
	Font_List->LogFont = Log_Font ;
	Font_List->Font = Temp ;
	return 1 ;
    } ;

    /* See if font is already in use */
    Index = 1 ;
    Current = Font_List ;
    while( Current != NULL )
    {
	Previous = Current ; /* Save this one */
	/* We must do the following comparison in two parts because the
	   font name is a ASCIIZ string, and the garbage following the Null
	   doesn't have to match in order for the fonts to be the same. */
	if( memcmp( &Current->LogFont, &Log_Font, sizeof( LOGFONT ) - LF_FACESIZE ) == 0 )
	/* Found a match on all but the font name */
	{
	    if( strcmp( Current->LogFont.lfFaceName, Log_Font.lfFaceName ) == 0 )
	    /* Found a match on the font name */
	    {
		Current->Usage_Count++ ;
		delete Temp ;
		return Index ;
	    } ;
	} ;
	Current = Current->Next ; /* Try next one */
	Index = Index + 1 ;
    } ;

    /* Create new font */
    Previous->Next = ( RFonts* ) malloc( sizeof( RFonts ) ) ;
    Previous->Next->Next = NULL ;
    Previous->Next->Usage_Count = 1 ;
    Previous->Next->LogFont = Log_Font ;
    Previous->Next->Font = Temp ;
    return Index ;
} ;

void Delete_Font( int Index ) /* Delete a font */
{
    RFonts *Current ;

    if( Index < 1 ) /* Ignore invalid font */
    {
	return ;
    } ;

    /* Find specified font */
    Current = Font_List ;
    while( Current != NULL )
    {
	Index = Index - 1 ;
	if( Index == 0 ) /* Found match */
	{
	    if( Current->Usage_Count > 0 )
	    {
		Current->Usage_Count = Current->Usage_Count - 1 ;
	    } ;
	    return ;
	} ;
	Current = Current->Next ;
    } ;
} ;

int Duplicate_Font( int Index ) /* Duplicate a font */
{
    RFonts *Current ;
    int Result ;

    if( Index < 1 ) /* Ignore invalid font */
    {
	return 0 ;
    } ;
    Result = Index ;

    /* Find specified font */
    Current = Font_List ;
    while( Current != NULL )
    {
	Index = Index - 1 ;
	if( Index == 0 ) /* Found match */
	{
	    Current->Usage_Count = Current->Usage_Count + 1 ;
	    return Result ;
	} ;
	Current = Current->Next ;
    } ;
    return 0 ; /* Not found */
} ;

RFonts* Get_Font( int Index ) /* Return specified font record */
{
    RFonts *Current ;

    if( Index < 1 ) /* Ignore invalid font */
    {
	return NULL ;
    } ;

    /* Find specified font */
    Current = Font_List ;
    while( Current != NULL )
    {
	Index = Index - 1 ;
	if( Index == 0 ) /* Found match */
	{
	    return Current ;
	} ;
	Current = Current->Next ;
    } ;
    return NULL ; /* Not found */
} ;

void Free_Fonts() /* Delete all fonts and free memory */
{
    RFonts *Current ;

    while( Font_List != NULL )
    {
	Current = Font_List->Next ;
	delete Font_List->Font ;
	free( Font_List ) ;
	Font_List = Current ;
    } ;
} ;

int Translate_Font( int Index )
/* Return index, pretending that unused fonts are not there */
{
    RFonts *Current ;
    int Actual ;

    if( Index < 1 ) /* Ignore invalid font */
    {
	return 0 ;
    } ;

    /* Find specified font */
    Actual = 1 ;
    Current = Font_List ;
    while( Current != NULL )
    {
	Index = Index - 1 ;
	if( Index == 0 ) /* Found match */
	{
	    return Actual ;
	} ;
	Current = Current->Next ;
	if( Current->Usage_Count > 0 )
	{
	    Actual = Actual + 1 ;
	} ;
    } ;
    return 0 ; /* Not found */
} ;


void Store_Font( FILE *File, BOOL All )
{
    RFonts *Current ;
    RFonts EOL ;

    Current = Font_List ;
    while( Current != NULL )
    {
	if( ( Current->Usage_Count > 0 ) || All )
	{
	    fwrite( Current, sizeof( LOGFONT ) + sizeof( int ), 1, File ) ;
	} ;
	Current = Current->Next ;
    } ;
    EOL.Usage_Count = -1 ;
    fwrite( &EOL, sizeof( LOGFONT ) + sizeof( int ), 1, File ) ;
} ;

void Load_Font( FILE *File )
/* Load a set of fonts from a file */
{
    RFonts Temp ;
    int Dummy ;

    fread( &Temp, sizeof( LOGFONT ) + sizeof( int ), 1, File ) ;
    /* Get first font */

    while( Temp.Usage_Count != -1 ) /* Until end of fonts on stream */
    {
	Dummy = Create_Font( Temp.LogFont ) ; /* Create a font */
	Delete_Font( Dummy ) ; /* But with no usage */

	while( Temp.Usage_Count > 0 ) /* Increase usage count, if needed */
	{
	    Duplicate_Font( Dummy ) ;
	    Temp.Usage_Count = Temp.Usage_Count - 1 ;
	} ;
	fread( &Temp, sizeof( LOGFONT ) + sizeof( int ), 1, File ) ;
	/* Get next font */
    } ;
} ;

