{
        Program Name : GenNum_Dialog
        Package Name : LB
        Purpose      : Generic number entry dialog
        Institution  : Conroy & Conroy Co.
        Date Written : 19-April-2005
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2005 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This dialog is used to get a number from the user.  It is a general-
        purpose dialog with several advanced features.  Currently, this is
        limited to numeric input from -2^63 to 2^63 - 1.
}

unit Gennum_Dialog ;

interface

uses // Borland...
     Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
     Dialogs, StdCtrls, Buttons, ExtCtrls, ComCtrls, Spin,

     // C&C...
     TDialogs ; // TDialog

type TMode = ( M_INTEGER, M_ASCII, M_EBCDIC, M_RADIX50 ) ;

type
  TGenNum_Form = class(TForm)
    Button_Panel: TPanel;
    OK_Button: TBitBtn;
    Cancel_Button: TBitBtn;
    Panel1: TPanel;
    Value_Edit: TEdit;
    Instruction_Label: TLabel;
    UpDown: TUpDown;
    Base_Panel: TPanel;
    Binary_RB: TRadioButton;
    Octal_RB: TRadioButton;
    Decimal_RB: TRadioButton;
    Hexadecimal_RB: TRadioButton;
    Other_RB: TRadioButton;
    Base_Spin: TSpinEdit;
    ASCII_RB: TRadioButton;
    EBCDIC_RB: TRadioButton;
    Radix50_RB: TRadioButton;
    procedure FormShow(Sender: TObject);
    procedure UpDownChangingEx(Sender: TObject; var AllowChange: Boolean;
      NewValue: Smallint; Direction: TUpDownDirection);
    procedure Value_EditKeyPress(Sender: TObject; var Key: Char);
    procedure Value_EditChange(Sender: TObject);
    procedure Base_SpinChange(Sender: TObject);
    procedure Binary_RBClick(Sender: TObject);
    procedure Octal_RBClick(Sender: TObject);
    procedure Decimal_RBClick(Sender: TObject);
    procedure Hexadecimal_RBClick(Sender: TObject);
    procedure ASCII_RBClick(Sender: TObject);
    procedure EBCDIC_RBClick(Sender: TObject);
    procedure Radix50_RBClick(Sender: TObject);

  public // Constructors and destructors...
      constructor Create( AOwner : TComponent ) ; override ;

  private // Internal data...
      _Base : word ;
      _Maximum : int64 ;
      _Minimum : int64 ;
      _Mode : TMode ;
      _Max_Size : word ;

  protected // Property handlers...
      function Get_Base : word ;
      function Get_Value : int64 ;
      procedure Set_Base( Value : word ) ;
      procedure Set_Maximum( Value : int64 ) ;
      procedure Set_Minimum( Value : int64 ) ;
      procedure Set_Mode( Value : TMode ) ;
      procedure Set_Value( Value : int64 ) ;

  public // API...
      property Mode : TMode
          read _Mode
          write Set_Mode ;
      function Valid : boolean ;

  published // API...
      property Base : word
          read Get_Base
          write Set_Base ;
      property Max_Size : word
          read _Max_Size
          write _Max_Size ;
      property Maximum : int64
          read _Maximum
          write Set_Maximum ;
      property Minimum : int64
          read _Minimum
          write Set_Minimum ;
      property Value : int64
          read Get_Value
          write Set_Value ;
  end ;

implementation

{$R *.dfm}

uses // C&C...
     ASCIIDef, // NUL
     CVT, // CvtB
     EBCDICs,
     Num1s, // Num1
     Radix50s ;

// TGenNum_Form...

procedure TGenNum_Form.FormShow(Sender: TObject);

begin
    UpDown.Height := Value_Edit.Height ;
    UpDown.Width := UpDown.Height ;
    UpDown.Left := Value_Edit.Left + Value_Edit.Width ;
end ;



procedure TGenNum_Form.UpDownChangingEx(Sender: TObject;
    var AllowChange: Boolean; NewValue: Smallint;
    Direction: TUpDownDirection ) ;

begin
    if( Valid ) then
    begin
        case Direction of
            updNone : exit ;
            updUp : Value := Value + 1 ;
            updDown : Value := Value - 1 ;
        end ;
    end ;
    UpDown.Position := 50 ;
end ;


// Constructors and destructors...

constructor TGenNum_Form.Create( AOwner : TComponent ) ;

begin
    inherited Create( AOwner ) ;

    _Base := 10 ;
    _Maximum := -1 ;
end ;


// Property handlers...

function TGenNum_Form.Get_Base : word ;

begin
    Result := _Base ;
end ;


function TGenNum_Form.Get_Value : int64 ;

var S : integer ;
    Work : string ;

begin
    Result := 0 ;
    S := sizeof( Result ) ;
    if( length( Value_Edit.Text ) < S ) then
    begin
        S := length( Value_Edit.Text ) ;
    end ;
    move( Value_Edit.Text[ 1 ], Result, S ) ;
    try
        case Mode of
            M_Integer : Result := strtoint64( CvtB( Base, 10, Value_Edit.Text ) ) ;
            M_ASCII : ; // Already did this above
            M_EBCDIC :
                begin
                    Work := To_EBCDIC( Value_Edit.Text ) ;
                    move( Work[ 1 ], Result, S ) ;
                end ;
            M_Radix50 : Result := Rad50( Value_Edit.Text ) ;
        end ;
    except
    end ;
end ;


procedure TGenNum_Form.Set_Base( Value : word ) ;

var Old : word ;

begin
    if( Value <> _Base ) then
    begin
        case Mode of
            M_Integer :
		begin
                    if( ( Value > 1 ) and ( Value < 50 ) ) then
                    begin
                        case Value of
                            2 : Binary_RB.Checked := True ;
                            8 : Octal_RB.Checked := True ;
                            10 : Decimal_RB.Checked := True ;
                            16 : Hexadecimal_RB.Checked := True ;
                            else Other_RB.Checked := True ;
                        end ;
                        Old := Base ;
                        _Base := Value ;
                        if( Valid_Base( Value_Edit.Text, Old ) ) then
                        begin
                            Value_Edit.Text := CvtB( Old, Value, Value_Edit.Text ) ;
                        end ;
                        Base_Spin.Value := Value ;
                    end ; // if
                end ; // M_Integer
        end ; // case Mode of
    end ;
end ;


procedure TGenNum_Form.Set_Maximum( Value : int64 ) ;

begin
    _Maximum := Value ;
    Value_EditChange( Value_Edit ) ;
end ;


procedure TGenNum_Form.Set_Minimum( Value : int64 ) ;

begin
    _Minimum := Value ;
    Value_EditChange( Value_Edit ) ;
end ;


procedure TGenNum_Form.Set_Mode( Value : TMode ) ;

var Saved : int64 ;

begin
    if( _Mode <> Value ) then
    begin
        if( Valid ) then
        begin
            Saved := Get_Value ;
            _Mode := Value ;
            Set_Value( Saved ) ;
        end else
        begin
            _Mode := Value ;
        end ;
    end ;
end ;


procedure TGenNum_Form.Set_Value( Value : int64 ) ;

var S : string ;

begin
    setlength( S, sizeof( Value ) ) ;
    move( Value, S[ 1 ], length( S ) ) ;
    case Mode of
        M_Integer : Value_Edit.Text := CvtB( 10, Base, num1( Value ) ) ;
        M_ASCII : Value_Edit.Text := S ;
        M_EBCDIC : Value_Edit.Text := From_EBCDIC( S ) ;
        M_Radix50 : Value_Edit.Text := Rad( Value ) ;
    end ;
end ;



procedure TGenNum_Form.Value_EditKeyPress(Sender: TObject; var Key: Char);

begin
    if( Key = BS ) then // Delete is valid
    begin
       exit ;
    end else
    case Mode of
        M_Integer :
            begin
                Key := upcase( Key ) ;
                if( not Valid_Base( Key, Base ) ) then
                begin
                    Key := NUL ;
                    MessageBeep( 0 ) ;
                end ;
            end ;
        M_EBCDIC :
            if( not Valid_EBCDIC( Key ) ) then
            begin
                Key := NUL ;
                MessageBeep( 0 ) ;
            end ;
        M_Radix50 :
            if( not Valid_Radix50( Key ) ) then
            begin
                Key := NUL ;
                MessageBeep( 0 ) ;
            end ;
    end ;
end ;


procedure TGenNum_Form.Value_EditChange(Sender: TObject);

begin
    OK_Button.Enabled := Valid ;
end ;


procedure TGenNum_Form.Base_SpinChange(Sender: TObject);

begin
    Mode := M_Integer ;
    Base := Base_Spin.Value ;
end ;


procedure TGenNum_Form.Binary_RBClick(Sender: TObject);

begin
    Mode := M_Integer ;
    Base := 2 ;
end ;

procedure TGenNum_Form.Octal_RBClick(Sender: TObject);

begin
    Mode := M_Integer ;
    Base := 8 ;
end ;


procedure TGenNum_Form.Decimal_RBClick(Sender: TObject);

begin
    Mode := M_Integer ;
    Base := 10 ;
end ;


procedure TGenNum_Form.Hexadecimal_RBClick(Sender: TObject);

begin
    Mode := M_Integer ;
    Base := 16 ;
end ;


procedure TGenNum_Form.ASCII_RBClick(Sender: TObject);

begin
    Mode := M_ASCII ;
    Set_Base( 0 ) ;
end ;


function TGenNum_Form.Valid : boolean ;

begin
    Result := True ;
    case Mode of
        M_Integer : Result := Valid_Base( Value_Edit.Text, Base ) ;
        M_EBCDIC : Result := ( Valid_EBCDIC( Value_Edit.Text ) and ( length( Value_Edit.Text ) < Max_Size ) ) ;
        M_Radix50 : Result := ( Valid_Radix50( Value_Edit.Text ) and ( length( Value_Edit.Text ) * 2 div 3 < Max_Size ) ) ;
        M_ASCII : Result := ( length( Value_Edit.Text ) < Max_Size ) ;
    end ;
    if( Result ) then // Validate range
    begin
        Result := (
                    ( ( Value >= Minimum ) and ( Value <= Maximum ) )
                    or
                    ( Minimum > Maximum )
                  ) ;
    end ;
end ;


procedure TGenNum_Form.EBCDIC_RBClick(Sender: TObject);

begin
    Mode := M_EBCDIC ;
    Set_Base( 0 ) ;
end ;


procedure TGenNum_Form.Radix50_RBClick(Sender: TObject);

begin
    Mode := M_Radix50 ;
    Set_Base( 0 ) ;
end ;


end.
