/* 
              Program Name : LIB
              Package Name : LB
              Purpose      : Librarian routines
              Institution  : Conroy & Conroy Co.
              Date Written : 15-Dec-1988
              Written By   : Alan Conroy
              Version      : 1.0

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *            F I L E   D E S C R I P T I O N S          *
        *                                                       *
        *********************************************************

 */

#include <dos.h>
#include <stdio.h>
#include <string.h>
#include <io.h>

struct FILE *Object_File, *Output, *Libr ;
int hObject_File, hLibr ;
char Lib_Buffer[ 256 ] ;
int Lib_Hour, Lib_Minute ;
int Lib_Day, Lib_Month, Lib_Year ;

    /* Library file format:
     Block    Description
     -----    -----------
       1      header block
       2+     name and data blocks

     Header block format:
         Byte(s)    Description
         -------    -----------
         0          =0 : Good, <>0 : Bad
         1          Version (10 = V1.0)
         2          0 (Quest), 1 (Sirius), 2 (Graphics images)
         3-5        First name block
         6-8        Number of deleted blocks
         9          File type (2=Library)
         10         Day of creation
         11         Month of creation
         12         Year of creation
         13         Hour of creation
         14         Minute of creation
         15         Day of last update
         16         Month of last update
         17         Year of last update
         18         Hour of last update
         19         Minute of last update
         20         Library type (0 = OLB, 1 = ILB)

     Name block format:
         Byte(s)    Description
         -------    -----------
           0-15     Module name
          16-18     Back Link
          19-21     Forward Link
          22-24     Module start block
          25-27     Module size (in blocks)
          28        Module insertion day
          29        Module insertion month
          30        Module insertion year
          31        Module insertion hour
          32        Module insertion minute
          33        Flags:
                        AND 1 = No vector table
                        AND 2-128 = reserved
          34-255    reserved */

static int A ;
static long int L, L1, Link, Back_Link, Forward_Link ; /* File link */
static union REGS Reg ;

static long int Cvtbl( unsigned short int X ) /* Convert byte to long integer */
{

long int Y ;

    Y = X & 255 ;
    return Y ;
} ;

static void Stuff( int Offset, long int Value )
/* Stuff a 3-byte value into the buffer */
{
    Lib_Buffer[ Offset + 2 ] = Value / 65536 ;
    Value = Value - Cvtbl( Lib_Buffer[ Offset + 2 ] ) * 65536 ;
    Lib_Buffer[ Offset + 1 ] = Value / 256 ;
    Value = Value - Cvtbl( Lib_Buffer[ Offset + 1 ] ) * 256 ;
    Lib_Buffer[ Offset ] = Value ;
} ;

long int Get_Value( int Offset ) /* Get and unpack value from buffer */
{
    return Cvtbl( Lib_Buffer[ Offset ] ) +
	Cvtbl( Lib_Buffer[ Offset + 1 ] ) * 256 +
        Cvtbl( Lib_Buffer[ Offset + 2 ] ) * 65536 ;
} ;

void Get_Date()
{
    Reg.w.ax = 0x2A00 ;
    intdos( &Reg, &Reg ) ; /* Call MSDOS for date */
    Lib_Year = Reg.w.cx ;
    Lib_Month = Reg.h.dh ;
    Lib_Day = Reg.h.dl ;
    Lib_Year = Lib_Year - 1900 ;
    while( Lib_Year > 99 )
    {
	Lib_Year = Lib_Year - 100 ;
    } ;
} ;

void Get_Time()
{
    Reg.w.ax = 0x2C00 ;
    intdos( &Reg, &Reg ) ; /* Get time from MSDOS */
    Lib_Hour = Reg.h.ch ;
    Lib_Minute = Reg.h.cl ;
} ;

void Update() /* Update library header with current date/time for change */
{
    fseek( Libr, 0, SEEK_SET ) ;
    fread( Lib_Buffer, 256, 1, Libr ) ; /* Read header */
    Get_Date() ;
    Get_Time() ;
    Lib_Buffer[ 15 ] = Lib_Day ; /* Day of last change */
    Lib_Buffer[ 16 ] = Lib_Month ; /* Month of last change */
    Lib_Buffer[ 17 ] = Lib_Year ; /* Year of last change */
    Lib_Buffer[ 18 ] = Lib_Hour ; /* Hour of last change */
    Lib_Buffer[ 19 ] = Lib_Minute ; /* Minute of last change */
    fseek( Libr, 0, SEEK_SET ) ; /* Reposition */
    fwrite( Lib_Buffer, 256, 1, Libr ) ; /* Write updated header */
} ;

void Insertion()
/* Change links so as to insert this module (name).  L is the name record # */
{
    fseek( Libr, Back_Link * 256, SEEK_SET ) ; /* Get previous logical record */
    fread( Lib_Buffer, 256, 1, Libr ) ;
    Stuff( 19, L ) ; /* Set forward link */
    fseek( Libr, Back_Link * 256, SEEK_SET ) ;
    fwrite( Lib_Buffer, 256, 1, Libr ) ; /* Update record */
    fseek( Libr, Forward_Link * 256, SEEK_SET ) ; /* Get next logical record */
    fread( Lib_Buffer, 256, 1, Libr ) ;
    Stuff( 16, L ) ; /* Set back link */
    fseek( Libr, Forward_Link * 256, SEEK_SET ) ; /* Update record */
    fwrite( Lib_Buffer, 256, 1, Libr ) ;
} ;

int Delete( char *Command ) /* Delete module specified in Command */
{

int A ;
char Inp[ 256 ] ;
long int Back_Link, Forward_Link, L, L1, Link ;

    fseek( Libr, 0, SEEK_SET ) ;
    fread( Lib_Buffer, 256, 1, Libr ) ; /* Get library header */
    Link = Get_Value( 3 ) ; /* Get first name link */
    while( Link != 0 )
    {
        fseek( Libr, Link * 256, SEEK_SET ) ;
        fread( Lib_Buffer, 256, 1, Libr ) ;
        for( A = 1; A <= 15; A++ )
	{
	    Inp[ A - 1 ] = Lib_Buffer[ A ] ;
	} ;
	Inp[ Lib_Buffer[ 0 ] ] = 0 ;
        Forward_Link = Get_Value( 19 ) ;
        if( strcmp( Inp, Command ) == 0 ) /* Found module */
        {
            Back_Link = Get_Value( 16 ) ;
            L1 = Get_Value( 25 ) ; /* Get module size */
            if( Back_Link > 0 ) /* A previous record */
            {
                fseek( Libr, Back_Link * 256, SEEK_SET ) ;
		/* Get previous record */
                fread( Lib_Buffer, 256, 1, Libr ) ;
                Stuff( 19, Forward_Link ) ;
                fseek( Libr, Back_Link * 256, SEEK_SET ) ;
                fwrite( Lib_Buffer, 256, 1, Libr ) ; /* Patch forward link */
            } ;
            if( Forward_Link > 0 ) /* If a following record */
            {
                fseek( Libr, Forward_Link * 256, SEEK_SET ) ;
		/* Get next record */
                fread( Lib_Buffer, 256, 1, Libr ) ;
                Stuff( 16, Back_Link ) ;
                fseek( Libr, Forward_Link * 256, SEEK_SET ) ;
                fwrite( Lib_Buffer, 256, 1, Libr ) ; /* Patch back link */
            } ;
            fseek( Libr, 0, SEEK_SET ) ;
            fread( Lib_Buffer, 256, 1, Libr ) ; /* Get header record */
            L = Get_Value( 6 ) + L1 + 10 ; /* Add up deleted block count */
            Stuff( 6, L ) ;
            if( Back_Link == 0 ) /* If deleted module was first module */
	    {
		Stuff( 3, Forward_Link ) ;
	    } ;
            fseek( Libr, 0, SEEK_SET ) ;
            fwrite( Lib_Buffer, 256, 1, Libr ) ; /* Update header */
            Update() ;
            return 0 ;
        } ;
        if( strcmp( Inp, Command ) > 0 )
	{
	    Link = 0 ;
	}
	else
	{
	    Link = Forward_Link ;
	} ;
    } ;
    return 1 ; /* Module not found */
} ;

int Insert( char *Command, int Vector, int Start_Block )
/* This function inserts a module with the name in Command, from the Object
   file.  If Vector is false, we don't include vector table. */
{

int A, Dummy ;
char Inp[ 256 ] ;
long int Back_Link, Forward_Link, L, L1, Last_Link, Link ;

    fseek( Libr, 0, SEEK_SET ) ;
    fread( Lib_Buffer, 256, 1, Libr ) ; /* Read library header */
    Link = Get_Value( 3 ) ; /* Get first name link */
    L = filelength( hLibr ) / 256 ; /* Get current library size */
    if( Link == 0 ) /* First module */
    {
        Back_Link = 0 ;
        Forward_Link = 0 ;
        Stuff( 3, L ) ;
        fseek( Libr, 0, SEEK_SET ) ;
        fwrite( Lib_Buffer, 256, 1, Libr ) ; /* Update library header */
    }
    else
    {
	while( Link != 0 ) /* Until end of list */
	{
            fseek( Libr, Link * 256, SEEK_SET ) ;
	    fread( Lib_Buffer, 256, 1, Libr ) ;
	    for( A = 1; A <= 15; A++ )
	    {
		Inp[ A - 1 ] = Lib_Buffer[ A ] ;
	    } ;
	    Inp[ Lib_Buffer[ 0 ] ] = 0 ;
	    if( strcmp( Command, Inp ) == 0 )
	    {
		return 2 ;
	    } ;
	    Back_Link = Link ;
	    Forward_Link = Link ;
	    if( strcmp( Inp, Command ) > 0 )
	    /* Insert new module just before this one */
	    {
		if( Lib_Buffer[ 16 ] == 0 && Lib_Buffer[ 17 ] == 0 &&
			Lib_Buffer[ 18 ] == 0 ) /* No back link */
		{
                    Stuff( 16, L ) ;
                    fseek( Libr, Link * 256, SEEK_SET ) ;
                    fwrite( Lib_Buffer, 256, 1, Libr ) ;
                    Forward_Link = Link ;
                    Back_Link = 0 ;
                    Link = 0 ;
                    fseek( Libr, 0, SEEK_SET ) ;
                    fread( Lib_Buffer, 256, 1, Libr ) ;
                    fseek( Libr, 0, SEEK_SET ) ;
                    Stuff( 3, L ) ;
                    fwrite( Lib_Buffer, 256, 1, Libr ) ;
                }
		else /* Must insert this module between two others */
                {
                    Last_Link = Link ; /* Save current record number */
                    L1 = Get_Value( 16 ) ; /* Get back link */
                    Stuff( 16, L ) ; /* Set back link to new record */
                    fseek( Libr, Link * 256, SEEK_SET ) ;
                    fwrite( Lib_Buffer, 256, 1, Libr ) ; /* Update record */
                    fseek( Libr, L1 * 256, SEEK_SET ) ;
		    /* Get previous module */
                    fread( Lib_Buffer, 256, 1, Libr ) ;
                    Stuff( 19, L ) ; /* Set forward link */
                    fseek( Libr, L1 * 256, SEEK_SET ) ;
                    fwrite( Lib_Buffer, 256, 1, Libr ) ;
		    /* Update previous record */
                    Forward_Link = Link ;
                    Back_Link = L1 ; /* Set up links for new record */
                    Link = 0 ; /* Terminate loop */
                } ;
                /* Follow back link */
            }
	    else
	    { /* Inp < Command */
                if( Lib_Buffer[ 19 ] == 0 && Lib_Buffer[ 20 ] == 0 &&
                    Lib_Buffer[ 21 ] == 0 )
                /* No forward link */
                {
                    Stuff( 19, L ) ;
                    fseek( Libr, Link * 256, SEEK_SET ) ;
                    fwrite( Lib_Buffer, 256, 1, Libr ) ;
                    Back_Link = Link ;
                    Forward_Link = 0 ;
                    Link = 0 ;
                }
		else
                {
                    Last_Link = Link ; /* Save current record number */
                    L1 = Get_Value( 19 ) ; /* Get forward link */
                    if( L1 == Last_Link )
                    {
                        Forward_Link = Last_Link ;
                        Back_Link = Link ;
                        Insertion() ;
                        Link = 0 ;
                    }
		    else
		    {
			Link = L1 ;
		    } ;
		} ;
		/* Follow forward link */
            } ;
        } ;
    } ;
    fseek( Libr, L * 256, SEEK_SET ) ;
    for( A = 1; A <= 15; A++ )
    {
	Lib_Buffer[ A ] = Command[ A - 1 ] ;
    } ;
    Lib_Buffer[ 0 ] = strlen( Command ) ;
    Stuff( 16, Back_Link ) ;
    Stuff( 19, Forward_Link ) ;
    Stuff( 22, L + 1 ) ;
    L1 = filelength( hObject_File ) / 256 ;
    Stuff( 25, L1 ) ;
    Get_Date() ;
    Get_Time() ;
    Lib_Buffer[ 28 ] = Lib_Day ; /* Day of last change */
    Lib_Buffer[ 29 ] = Lib_Month ; /* Month of last change */
    Lib_Buffer[ 30 ] = Lib_Year ; /* Year of last change */
    Lib_Buffer[ 31 ] = Lib_Hour ; /* Hour of last change */
    Lib_Buffer[ 32 ] = Lib_Minute ; /* Minute of last change */
    Lib_Buffer[ 33 ] = 0 ;
    if( !Vector )
    {
	Lib_Buffer[ 33 ] = Lib_Buffer[ 33 ] | 1 ;
    } ;
    fwrite( Lib_Buffer, 256, 1, Libr ) ; /* Write out name block */
    rewind( Object_File ) ;
    Dummy = 0 ;
    while( fread( Lib_Buffer, 256, 1, Object_File ) != 0 )
    {
        if( Vector || Dummy == 0 || Dummy >= Start_Block )
	{
            fwrite( Lib_Buffer, 256, 1, Libr ) ;
	    /* Write unless vector and /NV */
	} ;
        Dummy = Dummy + 1 ;
    } ;
    return 0 ;
} ;

int Create_Library( char *Command, int Language_Id, int Typ )
{

int A ;

    Libr = fopen( Command, "w+b" ) ;
    if( Libr == NULL )
    {
	return errno ;
    } ;
    Lib_Buffer[ 0 ] = 0 ; /* Anything but 0 indicates a bad file */
    Lib_Buffer[ 1 ] = 10 ; /* Version 1.0 */
    Lib_Buffer[ 2 ] = Language_Id ;
    Lib_Buffer[ 3 ] = 0 ; /* First name block */
    Lib_Buffer[ 4 ] = 0 ;
    Lib_Buffer[ 5 ] = 0 ;
    Lib_Buffer[ 6 ] = 0 ; /* Unused blocks count */
    Lib_Buffer[ 7 ] = 0 ;
    Lib_Buffer[ 8 ] = 0 ;
    Lib_Buffer[ 9 ] = 2 ; /* Library file */
    Get_Date() ;
    Get_Time() ;
    Lib_Buffer[ 10 ] = Lib_Day ; /* Day of creation */
    Lib_Buffer[ 11 ] = Lib_Month ; /* Month of creation */
    Lib_Buffer[ 12 ] = Lib_Year ; /* Year of creation */
    Lib_Buffer[ 13 ] = Lib_Hour ; /* Hour of creation */
    Lib_Buffer[ 14 ] = Lib_Minute ; /* Minute of creation */
    Lib_Buffer[ 15 ] = Lib_Day ; /* Day of last change */
    Lib_Buffer[ 16 ] = Lib_Month ; /* Month of last change */
    Lib_Buffer[ 17 ] = Lib_Year ; /* Year of last change */
    Lib_Buffer[ 18 ] = Lib_Hour ; /* Hour of last change */
    Lib_Buffer[ 19 ] = Lib_Minute ; /* Minute of last change */
    Lib_Buffer[ 20 ] = Typ ; /* Library file type */
    if( fwrite( Lib_Buffer, 256, 1, Libr ) != 1 )
    {
	return errno ;
    } ;
    fclose( Libr ) ;

    return 0 ;
} ;
