{
              Program Name : CEF_Assembler
              Package Name : CEF
              Purpose      : Default master assembler
              Institution  : Conroy & Conroy
              Date Written : 19-Mar-87
              Written By   : Alan Conroy
              Version      : 1.1

              Copyright (C) 1987 by Alan Conroy.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit defines the master assembler for CEF Z80.
}

unit CEF_Assembler ;

interface

uses // Borland...
     Classes, // TStringList
     Forms, // Application

     { CEF... }
     CEF, // TMaster_Assembler

     { Other... }
     Files, // TSFile
     Parse, // TParser
     _Streams, // TCOM_Stream
     Standard, // TInteger64_List
     Symbols, // TSymbol_Table
     _UEHDefs ; { TUEC }

const CEFAssemblerErr_Facility = 54 ;
      CEFAssemblerErr_Success = 0 ;
      CEFAssemblerErr_Invalid_Digits = 1 ; // Invalid digits for radix
      CEFAssemblerErr_Illegal_Expression = 2 ; // Illegal expression
      CEFAssemblerErr_Unterminated_String_Literal = 3 ; // Unterminated string literal
      CEFAssemblerErr_Undefined_Symbol = 4 ; // Undefined symbol
      CEFAssemblerErr_Illegal_Instruction = 5 ; // Illegal instruction
      CEFAssemblerErr_Missing_Value = 6 ; // Missing value
      CEFAssemblerErr_Multiply_Defined = 7 ; // Multiply defined symbol

type TAssembler_Context = class
                              public // Constructors and destructors...
                                  constructor Create ;
                                  destructor Destroy ; override ;

                              private // Instance data...
                                  Filenames : TStringList ;
                                  Addresses : Tinteger64_List ;
                                  Mappings : TList ;

                              private // Internal utility routines...
                                  function Add_Mapping( Filename : string ;
                                      Address : int64 ;
                                      Line : integer ) : integer ;
                                  procedure Delete_Mapping( Index : integer ) ;

                              public // API...
                                  Symbol_Table : TSymbol_Table ;

                                  { Returns filename and line number
                                     corresponding to the passed address.  If
                                     there is no mapping, it returns -1. }
                                  function Mapping( Address : int64 ;
                                      var Filename : string ) : integer ;

                                  { Get value associated with passed symbol.  If
                                    symbol is unknown, it returns False. }
                                  function Symbol_Value( const Name : string ;
                                      var Value : int64 ) : boolean ;

                                  { Get size associated with passed symbol.  If
                                    symbol is unknown, it returns False. }
                                  function Symbol_Size( const Name : string ;
                                      var Value : integer ) : boolean ;
                          end ;

     TStringList_Parser = class( TParser )
                              private // Instance data...
                                  _List : TStringList ;
                                  _Index : integer ;
                                  Parser : TString_Parser ;

                              public { Constructors and destructors... }
                                  constructor Create( List : TStringList ) ;
                                  destructor Destroy ; override ;

                              public { Property handlers... }
                                  function Get_Current_Line : longint ;
                                  function Get_Token_Line : longint ;
                                  function Get_Non_Terminals : string ;
                                  procedure Set_Non_Terminals( S : string ) ;

                              public { API... }
                                  function Grab_Line : string ; override ;
                                  procedure Put_Token( const S : string ) ;
                                      override ;
                                  procedure Reset ;
                                  function Token : string ; override ;
                                  function Token_EOL : boolean ; override ;

                                  property Non_Terminals : string
                                      read Get_Non_Terminals
                                      write Set_Non_Terminals ;
                          end ; // TStringList_Parser

     TState = class
                  public // Constructors and destructors...
                      destructor Destroy ; override ;

                  public // API...
                      State : integer ;
                      Context : string ;
                      List : TStringList ;
                      Tokens : TParser ;
              end ;

     TCEF_Assembler = class( TMaster_Assembler )
                          public // Instance data...
                              Altered_Address : int64 ; // If Altered_PC is true, this is the new PC
                              Altered_PC : boolean ; // True if an origin change
                              _Assembler_Context : TAssembler_Context ;
                              Backpatch_Table : TStringList ;
                              Base : integer ;
                              Err : boolean ; { True if an error }
                              _CPU : TComponent ;
                              PC : int64 ; { Current target address }
                              Segment : integer ; { Current segment }
                              Tokens : TFile_Parser ;
                              Terminated : boolean ; { True if we processed a .END directive }
                              In_Assembly_Status : TAssembler_Status ;
                              Forwards_Names : TStringList ;
                              Forwards_Values : TStringList ;
                              Initial_Symbols : string ;
                              Valid_Symbols : string ;
                              Temp_Expand : string ; // Used by .Expand
                              Last_Mapping_Index : integer ;
                              States : TList ; // Current set of states
                              Start_Address : int64 ; // Starting execution address
                              Have_Start_Address : boolean ; // True if a start address specified
                              _UI : TUI_Interface ;
                              Macros, Macro_Arguments : TStringList ;
                              Current_Created_Local_Label : integer ;

                          public // Constructors and destructors...
                              constructor Create( CPU : TComponent ;
                                  UI : TUI_Interface ) ;

                              destructor Destroy ; override ;

                          private // Internal utility routines...
                              function Forward_Declare( Value : string ;
                                  const Input_Buffer : string ) : boolean ;

                              function Grab_Current_Line : string ;

                              function Handle_Directives( var Value : string ;
                                  var Any_Code : boolean ;
                                  Status : TAssembler_Status ) : boolean ;

                              function Get_Assembler_Context : TAssembler_Context ;

                              procedure Enter_State( State : integer ;
                                  const Context : string ; List : TStringList ;
                                  Stream : TCom_Stream ) ;

                              procedure Exit_State ;

                              function _Get_Token : string ;

                              function _Grab_Line( Current : boolean ) : string ;

                              procedure _Put_Token( const Token : string ) ;

                              function _Token_EOL : boolean ;

                              function Process_If( Condition, Argument : string ;
                                  Status : TAssembler_Status ) : boolean ;

                              procedure Process_IRP( Symbol, Value : string ;
                                  Source, Target : TStringList ) ;

                          public // API...
                              procedure Define_Macro( Name : string ;
                                  const Arguments : string ;
                                  SL : TStringList ) ;

                              procedure Delete_Macro( Name : string ) ;

                              // True if error.
                              function Process_Macro( const Name : string ;
                                  Status : TAssembler_Status ) : boolean ;

                          public // API (overrides)...
                              property Assembler_Context : TAssembler_Context
                                  read Get_Assembler_Context
                                  write _Assembler_Context ;

                              function Assemble( Input, Output, Listing : TCOM_Stream ;
                                  Status : TAssembler_Status ) : TUEC ;
                                  override ; stdcall ;

                              procedure Add_Reference( Name : PChar ;
                                  Size : longint ; Address : int64 ) ;
                                  override ; stdcall ;

                              procedure Backpatch( Status : TAssembler_Status ;
                                  Stream : TCOM_Stream ) ;
                                  override ; stdcall ;

                              function Expand( Source : PChar ;
                                  var Res : PChar ;
                                  var Res_Length : longint ;
                                   Status : TAssembler_Status ) : TUEC ;
                                  override ; stdcall ;

                              function Evaluate( Value : PChar ;
                                  var _Result : int64 ) : TUEC ;
                                  override ; stdcall ;

                              function Facility_Code : longint ;
                                  override ; stdcall ;

                              procedure In_Line( Input : TCOM_Stream ) ;
                                  override ; stdcall ;

                              procedure Log_Error( Text : PChar ;
                                  Severity : longint ) ; override ; stdcall ;

                              function Get_Token : PChar ; override ; stdcall ;

                              procedure Put_Token( Token : PChar ) ;
                                  override ; stdcall ;

                              function Peek_Token( Same_Line : boolean ) : PChar ;
                                  override ; stdcall ;

                              procedure Push_Scope ; override ; stdcall ;

                              procedure Pop_Scope ; override ; stdcall ;

                              function Get_Symbol( Name : PChar ) : PSymbol_Record ;
                                  override ; stdcall ;

                              function Add_Symbol( Name : PChar ;
                                  P : pSymbol_Record ) : TUEC ;
                                  override ; stdcall ;

                              procedure Set_Case_Sensitive( Value : boolean ) ;
                                  override ; stdcall ;

                              function Get_Case_Sensitive : boolean ;
                                  override ; stdcall ;

                              function Grab_Line( Current : boolean ) : PChar ;
                                  override ; stdcall ;

                              procedure Map( Address : int64 ) ;
                                  override ; stdcall ;

                              procedure UnMap ; override ; stdcall ;
                      end ; // TCEF_Assembler

implementation

uses // Borland...
     Dialogs, // ShowMessage
     SysUtils, // uppercase

     // C&C...
     CommonUt, // FEdit
     CVT,
     Express,
     _Stores, // TStore_Address
     Instrs, // Instr
     Maths,
     NUM1S,
     Radix50s, // Valid_Radix50
     SStreams, // TCOM_String_Stream
     TypeDefs, // TRecord_Size
     UStrings ; // String_To_BCD

// Nested states...
const State_Repeat = 1 ;
      State_Include = 2 ;
      State_Macro = 3 ;
      State_IRP = 4 ;
      State_If = 5 ;
      State_Inline = 6 ;

// TStringList_Parser methods...

// Constructors and destructors...

constructor TStringList_Parser.Create( List : TStringList ) ;

begin
    inherited Create ;

    _List := List ;
    _Index := 0 ;
    Parser := TString_Parser.Create ;
    Reset ;
end ;


destructor TStringList_Parser.Destroy ;

begin
    Parser.Free ;
    Parser := nil ;

    inherited Destroy ;
end ;



// Property handlers...

function TStringList_Parser.Get_Current_Line : longint ;

begin
    Result := _Index ;
end ;


function TStringList_Parser.Get_Token_Line : longint ;

begin
    Result := _Index ;
end ;


function TStringList_Parser.Get_Non_Terminals : string ;

begin
    Result := Parser.Get_Non_Terminals ;
end ;


procedure TStringList_Parser.Set_Non_Terminals( S : string ) ;

begin
    Parser.Set_Non_Terminals( S ) ;
end ;


// API...

function TStringList_Parser.Grab_Line : string ;

label Try_Again ;

begin
    Result := '' ;
Try_Again:
    if( Parser.Token_EOL ) then
    begin
        inc( _Index ) ;
        if( _Index < _List.Count ) then
        begin
            Parser.Set_Source( _List[ _Index ] ) ;
            goto Try_Again ;
        end ;
        exit ;
    end ;
    Result := Parser.Grab_Line ;

    // Now prepare the next line...
    inc( _Index ) ;
    if( _Index < _List.Count ) then
    begin
        Parser.Set_Source( _List[ _Index ] ) ;
    end ;
end ;


procedure TStringList_Parser.Put_Token( const S : string ) ;

begin
   Parser.Put_Token( S ) ;
end ;


procedure TStringList_Parser.Reset ;

begin
    _Index := 0 ;
    if( _List.Count > 0 ) then
    begin
        Parser.Set_Source( _List[ 0 ] ) ;
    end ;
end ;


function TStringList_Parser.Token : string ;

label Try_Again ;

begin
Try_Again:
    Result := Parser.Token ;
    if( length( Result ) = 0 ) then
    begin
        inc( _Index ) ;
        if( _Index < _List.Count ) then
        begin
            Parser.Set_Source( _List[ _Index ] ) ;
            goto Try_Again ;
        end ;
    end ;
end ;


function TStringList_Parser.Token_EOL : boolean ;

begin
    Result := Parser.Token_EOL ;
end ;



type TStreamer_File = object( TSFile )
                          public
                              Stream : TCOM_Stream ;

                              function Extend( Amount : TStore_Address ) : TStore_Address ;
                                  virtual ;

                              function Get_Name : string ; virtual ;

                              function Max_Storage : TStore_Address ; virtual ;

                              function Min_Storage : TStore_Address ; virtual ;

                              function Contiguous_Store : boolean ; virtual ;

                              procedure Set_Max_Storage( Value : TStore_Address ;
                                  var Res : TUEC ) ; virtual ;

                              procedure Reset( const Name : String ) ; virtual ;
                              procedure Rewrite( const Name : String ) ;
                                  virtual ;
                              procedure Append( const Name : String ) ;
                                  virtual ;
                              procedure Close ; virtual ;
                              destructor Done ; virtual ;
                              procedure Seek( Pos : TStore_Address ) ; virtual ;

                              procedure Blockread( var Buf ;
                                  Count : TRecord_Size ;
                                  var Res : TRecord_Size ) ; virtual ;
                              procedure Blockwrite( var Buf ; Count : TRecord_Size ;
                                  var Res : TRecord_Size ) ; virtual ;
                              function EOF : boolean ; virtual ;
                              function FilePos : TStore_Address ; virtual ;
                              function Get_Size : TStore_Address ; virtual ;
                              procedure Set_Size( New_Size : TStore_Address ) ;
                                  virtual ;
                              procedure Flush ; virtual ;
                              procedure GetFAttr( var Attr : Word ) ;
                                  virtual ;
                              procedure Erase( const Name : String ) ; virtual ;
                              function FileName : string ; virtual ;
                              procedure SetFAttr( Attr : Word ) ; virtual ;
                              procedure GetFTime( var _Time : Longint ) ;
                                  virtual ;
                              procedure SetFTime( _Time : longint ) ; virtual ;
                              procedure SetTextBuf( var buf ; size : Word ) ;
                                  virtual ;
                              procedure Set_Extend( State : boolean ) ;
                                  virtual ;
                              procedure Truncate( Size : TStore_Address ) ;
                                  virtual ;
                              procedure Initialize ; virtual ;
                              function Get_Attributes : longint ; virtual ;
                              procedure Set_Attributes( Value : longint ) ;
                                  virtual ;
                              function Facility : longint ; virtual ;

                              function Version : longint ; virtual ;

                              function Facility_Name : PChar ; virtual ;
                              function Opened : boolean ; virtual ;
                      end ; // TStreamer_File
     PStreamer_File = ^TStreamer_File ;

function TStreamer_File.Extend( Amount : TStore_Address ) : TStore_Address ;

begin
    // Has no meaning for us
    Extend := -1 ;
end ;


function TStreamer_File.Get_Name : string ;

begin
    Result := '' ;
end ;


function TStreamer_File.Max_Storage : TStore_Address ;

begin
    Result := 0 ;
end ;


function TStreamer_File.Min_Storage : TStore_Address ;

begin
    Result := 0 ;
end ;


function TStreamer_File.Contiguous_Store : boolean ;

begin
    Result := True ;
end ;


procedure TStreamer_File.Set_Max_Storage( Value : TStore_Address ;
    var Res : TUEC ) ;

begin
    // Has no meaning for us
end ;


procedure TStreamer_File.Reset( const Name : String ) ;

begin
    Stream.Seek( 0 ) ;
end ;


procedure TStreamer_File.Rewrite( const Name : String ) ;

begin
    Stream.Seek( 0 ) ;
    Truncate( FilePos ) ;
end ;


procedure TStreamer_File.Append( const Name : String ) ;

var C : char ;
    S : integer ;

begin
    while( not Stream.At_End ) do
    begin
        S := 1 ;
        Stream.Read( C, S ) ;
    end ;
end ;


procedure TStreamer_File.Close ;

begin
    // Has no meaning for us
end ;


destructor TStreamer_File.Done ;

begin
end ;


procedure TStreamer_File.Seek( Pos : TStore_Address ) ;

begin
    Stream.Seek( Pos ) ;
end ;


procedure TStreamer_File.Blockread( var Buf ; Count : TRecord_Size ;
    var Res : TRecord_Size ) ;

begin
    Res := Count ;
    Stream.Read( Buf, Res ) ;
end ;


procedure TStreamer_File.Blockwrite( var Buf ; Count : TRecord_Size ;
    var Res : TRecord_Size ) ;

begin
    Res := Count ;
    Stream.Write( Buf, Res ) ;
end ;


function TStreamer_File.EOF : boolean ;

begin
    Result := Stream.At_End ;
end ;


function TStreamer_File.FilePos : TStore_Address ;

begin
    FilePos := 0 ;
end ;


function TStreamer_File.Get_Size : TStore_Address ;

begin
    Get_Size := 0 ;
end ;


procedure TStreamer_File.Set_Size( New_Size : TStore_Address ) ;

begin
    // Has no meaning for us
end ;


procedure TStreamer_File.Flush ;

begin
    // Has no meaning for us
end ;


procedure TStreamer_File.GetFAttr( var Attr : Word ) ;

begin
    Attr := 0 ;
end ;


procedure TStreamer_File.Erase( const Name : String ) ;

begin
    // Has no meaning for us
end ;


function TStreamer_File.FileName : string ;

begin
    Result := '' ;
end ;


procedure TStreamer_File.SetFAttr( Attr : Word ) ;

begin
    // Has no meaning for us
end ;


procedure TStreamer_File.GetFTime( var _Time : Longint ) ;

begin
    Time := 0 ;
end ;


procedure TStreamer_File.SetFTime( _Time : longint ) ;

begin
    // Has no meaning for us
end ;


procedure TStreamer_File.SetTextBuf( var buf ; size : Word ) ;

begin
    // Has no meaning for us
end ;


procedure TStreamer_File.Set_Extend( State : boolean ) ;

begin
    // Has no meaning for us
end ;


procedure TStreamer_File.Truncate( Size : TStore_Address ) ;

begin
    // Has no meaning for us
end ;


procedure TStreamer_File.Initialize ;

begin
end ;


function TStreamer_File.Get_Attributes : longint ;

begin
    Result := 0 ;
end ;


procedure TStreamer_File.Set_Attributes( Value : longint ) ;

begin
    // Has no meaning for us
end ;


function TStreamer_File.Facility : longint ;

begin
    Result := -1 ;
end ;


function TStreamer_File.Version : longint ;

begin
    Result := 10 ;
end ;


function TStreamer_File.Facility_Name : PChar ;

begin
    Result := nil ;
end ;


function TStreamer_File.Opened : boolean ;

begin
    Result := True ;
end ;


type TMapping_Info = class
                         public
                             File_Index : integer ;
                             Line : integer ;
                     end ;

// TAssembler_Context

// Constructors and destructors...

constructor TAssembler_Context.Create ;

begin
    inherited Create ;

    Symbol_Table.Init( nil ) ;
    Filenames := TStringList.Create ;
    Addresses := Tinteger64_List.Create ;
    Mappings := TList.Create ;
end ;


destructor TAssembler_Context.Destroy ;

var Loop : integer ;

begin
    Symbol_Table.Done ;
    Filenames.Free ;
    Filenames := nil ;
    Addresses.Free ;
    Addresses := nil ;
    for Loop := 0 to Mappings.Count - 1 do
    begin
        TMapping_Info( Mappings[ Loop ] ).Free ;
        Mappings[ Loop ] := nil ;
    end ;
    Mappings.Free ;
    Mappings := nil ;

    inherited Destroy ;
end ;


// Internal utility routines...

function TAssembler_Context.Add_Mapping( Filename : string ; Address : int64 ;
    Line : integer ) : integer ;

var Address_Index, Index : integer ;
    Info : TMapping_Info ;

begin
    Filename := uppercase( Filename ) ;
    Index := Filenames.IndexOf( Filename ) ;
    if( Index = -1 ) then
    begin
        Index := Filenames.Add( Filename ) ;
    end ;

    Info := TMapping_Info.Create ;
    Info.File_Index := Index ;
    Info.Line := Line ;
    Address_Index := Addresses.Indexof( Address ) ;
    if( Address_Index = -1 ) then
    begin
        Address_Index := Addresses.Add( Address ) ;
        Mappings.Add( Info ) ;
    end else
    begin
        TMapping_Info( Mappings[ Address_Index ] ).Free ;
        Mappings[ Address_Index ] := Info ;
    end ;
    Result := Address_Index ;
end ;


procedure TAssembler_Context.Delete_Mapping( Index : integer ) ;

begin
    TMapping_Info( Mappings[ Index ] ).Free ;
    Mappings[ Index ] := nil ;
    Mappings.Delete( Index ) ;
    Addresses.Delete( Index ) ;
end ;


// API...

function TAssembler_Context.Mapping( Address : int64 ;
    var Filename : string ) : integer ;

var Index : integer ;
    Info : TMapping_Info ;

begin
    Index := Addresses.IndexOf( Address ) ;
    if( Index = -1 ) then
    begin
        Result := -1 ;
        exit ;
    end ;
    Info := TMapping_Info( Mappings[ Index ] ) ;
    Filename := Filenames[ Info.File_Index ] ;
    Result := Info.Line ;
end ;


function TAssembler_Context.Symbol_Value( const Name : string ;
    var Value : int64 ) : boolean ;

var Err, Symbol_D_T, Symbol_Siz, Symbol_Flg : integer ;
    Symbol_Addr : int64 ;
    Symbol_Data : pointer ;

begin
    Err := Symbol_Table.Find( Name, Symbol_Addr,
        Symbol_Flg, Symbol_D_T, Symbol_Siz, Symbol_Data ) ;
    if( Err <> 0 ) then
    begin
        Result := False ;
        exit ;
    end ;
    Value := pSymbol_Record( Symbol_Data )^.Address ;
    Result := True ;
end ;


function TAssembler_Context.Symbol_Size( const Name : string ;
    var Value : integer ) : boolean ;

var Err, Symbol_D_T, Symbol_Siz, Symbol_Flg : integer ;
    Symbol_Addr : int64 ;
    Symbol_Data : pointer ;

begin
    Err := Symbol_Table.Find( Name, Symbol_Addr,
        Symbol_Flg, Symbol_D_T, Symbol_Siz, Symbol_Data ) ;
    if( Err <> 0 ) then
    begin
        Result := False ;
        exit ;
    end ;
    Value := Symbol_Siz ;
    Result := True ;
end ;



// TCEF_Assembler methods...

// Constructors and destructors...

constructor TCEF_Assembler.Create( CPU : TComponent ; UI : TUI_Interface ) ;

begin
    inherited Create ;

    _CPU := CPU ;
    _UI := UI ;
    if( CPU = nil ) then
    begin
        Base := 16 ;
    end else
    begin
        Base := CPU.CPU.Default_Base ;
    end ;
    PC := 0 ;
    Tokens := TFile_Parser.Create ;
    Tokens.Non_Terminals := Tokens.Non_Terminals + '.' ;
    States := TList.Create ;
    Macros := TStringList.Create ;
    Macro_Arguments := TStringList.Create ;
end ;


destructor TCEF_Assembler.Destroy ;

var Loop : integer ;

begin
    Tokens.Free ;
    Tokens := nil ;
    Backpatch_Table.Free ;
    Backpatch_Table := nil ;
    Forwards_Names.Free ;
    Forwards_Names := nil ;
    Forwards_Values.Free ;
    Forwards_Values := nil ;
    _Assembler_Context.Free ;
    _Assembler_Context := nil ;
    States.Free ;
    States := nil ;

    for Loop := 0 to Macros.Count - 1 do
    begin
        TStringList( Macros.Objects[ Loop ] ).Free ;
        Macros.Objects[ Loop ] := nil ;
    end ;
    Macros.Free ;
    Macros := nil ;
    Macro_Arguments.Free ;
    Macro_Arguments := nil ;

    inherited Destroy ;
end ;


// Internal utility routines...

function TCEF_Assembler.Evaluate( Value : PChar ; var _Result : int64 ) : TUEC ;

    function Va( X : string ; var ER : boolean ) : int64 ; { Return value of X }

    var A : String ;
        B : Real ;
        C, D, E : Integer ;

        procedure Strip( Y : Integer ) ; { Strip specifier and set E }

        begin
            E := Y ;
            X := Copy( X, 1, Length( X ) - 1 ) ;
        end ;

    begin
        ER := False ;
        Va := 0 ;
        if ( X[ 1 ] = Chr( 39 ) ) or ( X[ 1 ] = '"' ) then { ASCII literal }
        begin
            if X[ Length( X ) ] <> X[ 1 ] then { No closing quote of matching type }
            begin
                Er := True ;
                Exit ;
            end ;
            X := Copy( X, 2, Length( X ) - 2 ) ; { Strip quotes }
            if ( Length( X ) > 2 ) or ( Length( X ) = 0 ) then
            begin
                Er := True ;
                Exit ;
            end ;
            B := Ord( X[ 1 ] ) ;
            if( Length( X ) = 2 ) then
            begin
                B := B + Ord( X[ 2 ] ) * 256 ;
            end ;
        end else
        begin
            if ( Pos( '.', X ) = 0 ) and ( Pos( '-', X ) = 0 ) then
            begin
                D := Length( X ) ;
                E := Base ;
                if( ( X[ D ] = 'H' ) and ( Base < 18 ) ) then Strip( 16 ) else
                   if ( X[ D ] = 'D' ) and ( Base < 14 ) then Strip( 10 ) else
                      if ( X[ D ] = 'B' ) and ( Base < 11 ) then Strip( 2 ) else
                         if ( X[ D ] = 'O' ) or ( X[ D ] = 'Q' ) then Strip( 8 ) ;
                for C := 1 to Length( X ) do
                  if Pos( X[ C ], Copy( '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ', 1, E + 1 ) ) = 0 then
                begin
                    Er := True ;
                    Exit
                end ;
                A := Cvtb( E, 10, X ) ;
            end else
            begin
                A := X ;
            end ;
            Val( A, B, C ) ;
            if( ( C <> 0 ) or ( B < low( Result ) ) or ( B > high( Result ) ) ) then
            begin
                Er := True ;
                Exit ;
            end ;
        end ;
        if( B > high( Result ) ) then
        begin
            B := B - high( Result ) - high( Result ) - 2 ;
        end ;
        Va := Trunc( B ) ;
    end ; // .Va.Strip


    procedure Invalid( S, V : string ) ;

    var Dummy : integer ;

    begin
	for Dummy := 1 to length( S ) do
	begin
	    if( pos( S[ Dummy ], V ) = 0 ) then
	    begin
                Result.Code := CEFAssemblerErr_Invalid_Digits ;
		exit ;
	    end ;
	end ;
    end ;

var Err : integer ;
    Dummy : integer ;
    ER : boolean ; { True if Va() returned an error }
    Expression : string ; { Expression to evaluate }
    I, R : Extended ;
    Work : string ;
    Symbol_Addr : int64 ;
    Symbol_Flg, Symbol_D_T, Symbol_Siz : integer ;
    Symbol_Data : pointer ;
    X : string ;

begin // TCEF_Assembler.Evaluate
    { Setup... }
    X := Value ;
    Expression := '' ;
    Result.Facility := CEFAssemblerErr_Facility ;
    Result.Code := 0 ;

    { Parse each term }
    while( length( X ) > 0 ) do
    begin
        X := Edit( X, 8 or 128 ) ;
	if( pos( copy( X, 1, 1 ), '0123456789' ) > 0 ) then { Numeric literal }
	begin
	    Dummy := 1 ;
	    while(
		   ( Dummy <= length( X ) )
		   and
		   ( pos( copy( X, Dummy, 1 ), '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ.' ) > 0 )
		 ) do
	    begin
		inc( Dummy ) ;
	    end ;
	    Work := copy( X, 1, Dummy - 1 ) ;
	    Expression := Expression + ' ' + num1( Va( Work, ER ) ) ;
	    if( ER ) then
	    begin
		Result.Code := CEFAssemblerErr_Illegal_Expression ;
		exit ;
	    end ;

	    X := copy( X, Dummy, length( X ) ) ;
	end else
	if( ( copy( X, 1, 1 ) = #39 ) or ( copy( X, 1, 1 ) = '"' ) ) then // String literal
	begin
	    Dummy := instr( 2, X, copy( X, 1, 1 ) ) ;
	    if( Dummy = 0 ) then
	    begin
		Result.Code := CEFAssemblerErr_Unterminated_String_Literal ;
		exit ;
	    end ;
	    Expression := Expression + ' ' + num1( Va( copy( X, 1, Dummy ), ER ) ) ;
	    if( ER ) then
	    begin
		Result.Code := CEFAssemblerErr_Illegal_Expression ;
		exit ;
	    end ;
	    X := copy( X, Dummy + 1, length( X ) ) ;
	end else
	if( ( copy( X, 1, 1 ) = '.' ) or ( copy( X, 1, 1 ) = '$' ) ) then // PC
	begin
	    Expression := Expression + ' ' + num1( PC ) ;
	    X := copy( X, 2, length( X ) ) ;
	end else { if length( Expression ) = 0 }
	if( pos( X[ 1 ], Initial_Symbols ) > 0 ) then { Symbol }
	begin
	    Dummy := 1 ;
	    while(
		   ( Dummy <= length( X ) )
		   and
		   ( pos( copy( X, Dummy, 1 ), Valid_Symbols ) > 0 ) // What about $ and _?
		 ) do
	    begin
		inc( Dummy ) ;
	    end ;
            Work := copy( X, 1, Dummy - 1 ) ;
            if(
                ( Work = 'SHR' )
                or
                ( Work = 'SHL' )
                or
                ( Work = 'AND' )
                or
                ( Work = 'OR' )
                or
                ( Work = 'XOR' )
              ) then
            begin
                Expression := Expression + ' ' + Work + ' ' ;
            end else
            begin
                Err := Assembler_Context.Symbol_Table.Find( Work, Symbol_Addr,
                    Symbol_Flg, Symbol_D_T, Symbol_Siz, Symbol_Data ) ;
                if( Err <> 0 ) then // Not found
                begin
                    if( Valid_Base( Work, Base ) ) then
                    begin
                        Expression := Expression + num1( Va( Work, ER ) ) ;
                        X := copy( X, Dummy, length( X ) ) ;
                        continue ;
                    end ;
                    Result.Code := CEFAssemblerErr_Undefined_Symbol ;
                    exit ;
                end ; // if( Err <> 0 )
                Expression := Expression + num1( pSymbol_Record( Symbol_Data )^.Address ) ;
	    end ; // if
	    X := copy( X, Dummy, length( X ) ) ;
	end else
        begin
            if( X[ 1 ] <> ' ' ) then
            begin
                Expression := Expression + copy( X, 1, 1 ) ;
            end ;
            X := copy( X, 2, length( X ) ) ;
        end ;
    end ; // while( length( X ) > 0 )

    Eval( Expression, R, I, Err, Dummy ) ;
    if( Err <> 0 ) then
    begin
	Result.Code := CEFAssemblerErr_Illegal_Expression ;
    end else
    begin
	_Result := trunc( int( R ) ) ;
    end ;
end ; // TCEF_Assembler.Evaluate


function ET( Code : integer ) : string ;

begin
    case Code of
        CEFAssemblerErr_Invalid_Digits : Result := 'Invalid digits' ;
        CEFAssemblerErr_Illegal_Expression : Result := 'Illegal expression' ;
        CEFAssemblerErr_Unterminated_String_Literal : Result := 'Unterminated string literal' ;
        CEFAssemblerErr_Undefined_Symbol : Result := 'Undefined symbol' ;
        CEFAssemblerErr_Illegal_Instruction : Result := 'Illegal Instruction' ;
    end ;
end ;


function TCEF_Assembler.Forward_Declare( Value : string ;
    const Input_Buffer : string ) : boolean ;

begin
    Value := uppercase( Value ) ;
    if( Forwards_Names = nil ) then
    begin
        Forwards_Names := TStringList.Create ;
        Forwards_Values := TStringList.Create ;
    end ;
    Result := ( Forwards_Names.Indexof( Value ) <> -1 ) ;
    Forwards_Names.Add( Value ) ;
    Forwards_Values.Add( Input_Buffer ) ;
end ;


function TCEF_Assembler.Grab_Current_Line : string ;

var Temp : string ;

begin
    Result := _Grab_Line( True ) ;
    Temp := Parse_Parameter( ';', Result ) ; // Trim comments
    Result := Temp ;
end ;


function Valid_Condition( Value : string ) : boolean ;

begin
    Result := False ;
    Value := Edit( Value, 8 or 32 or 128 ) ;
    if(
        ( Value = 'EQUAL' )
        or
        ( Value = 'EQ' )
        or
        ( Value = 'NOT_EQUAL' )
        or
        ( Value = 'NE' )
        or
        ( Value = 'GREATER' )
        or
        ( Value = 'G' )
        or
        ( Value = 'LESS_EQUAL' )
        or
        ( Value = 'LE' )
        or
        ( Value = 'LESS_THAN' )
        or
        ( Value = 'LT' )
        or
        ( Value = 'GREATER_EQUAL' )
        or
        ( Value = 'GE' )
        or
        ( Value = 'DEFINED' )
        or
        ( Value = 'DF' )
        or
        ( Value = 'NOT_DEFINED' )
        or
        ( Value = 'BDF' )
        or
        ( Value = 'BLANK' )
        or
        ( Value = 'B' )
        or
        ( Value = 'NOT_BLANK' )
        or
        ( Value = 'NB' )
      ) then
    begin
        Result := True ;
    end ;
end ;


function Create_COM_String_Stream( const S : string ) : TCOM_String_Stream ;

begin
    Result := TCOM_String_Stream.Create ;
    Result.Set_String( PChar( S ) ) ;
end ;


function TCEF_Assembler.Handle_Directives( var Value : string ;
    var Any_Code : boolean ; Status : TAssembler_Status ) : boolean ;

var Have_Value : boolean ;
    Remaining_Tokens : string ;
    Returned_Token : string ;

    function Get_Token : string ;

    var Dummy, Dummy1 : integer ;

    begin
        if( length( Returned_Token ) > 0 ) then
        begin
            Result := Returned_Token ;
            Returned_Token := '' ;
            exit ;
        end ;
        Remaining_Tokens := Edit( Remaining_Tokens, 8 ) ; // Strip leading spaces
//        if( length( Remaining_Tokens ) = 0 ) then
        if( not Have_Value ) then
        begin
            Result := _Get_Token ;
            exit ;
        end ;
        if( copy( Remaining_Tokens, 1, 1 ) = ',' ) then
        begin
            Result := ',' ;
            Remaining_Tokens := copy( Remaining_Tokens, 2, length( Remaining_Tokens ) ) ;
            exit ;
        end ;
        if(
            ( copy( Remaining_Tokens, 1, 1 ) = '"' )
            or
            ( copy( Remaining_Tokens, 1, 1 ) = #39 )
          ) then
        begin
	    Dummy := instr( 2, Remaining_Tokens, copy( Remaining_Tokens, 1, 1 ) ) ;
            if( Dummy = 0 ) then // Unterminated quoted literal
            begin
                Result := Remaining_Tokens ; // Entire rest of line is the token
                Remaining_Tokens := '' ;
            end else
            begin
                Result := copy( Remaining_Tokens, 1, Dummy ) ;
                Remaining_Tokens := copy( Remaining_Tokens, Dummy + 1, length( Remaining_Tokens ) ) ;
            end ;
            exit ;
        end ;
        Dummy := pos( ' ', Remaining_Tokens + ' ' ) ;
        Dummy1 := pos( ',', Remaining_Tokens + ',' ) ;
        if( Dummy1 < Dummy ) then
        begin
            Dummy := Dummy1 ;
        end ;
        Result := copy( Remaining_Tokens, 1, Dummy - 1 ) ;
        Remaining_Tokens := copy( Remaining_Tokens, Dummy, length( Remaining_Tokens ) ) ;
    end ; // .Get_Token


    procedure Put_Token( const S : string ) ;

    begin
        Returned_Token := S ;
    end ;


    function Grab_This_Line : string ;

    begin
        if( Have_Value ) then
        begin
            Result := '' ;
        end else
        begin
            Result := Grab_Current_Line ;
        end ;
        if( length( Remaining_Tokens ) > 0 ) then
        begin
            if( length( Result ) > 0 ) then
            begin
                Result := ' ' + Result ;
            end ;
            Result := Remaining_Tokens + Result ;
            Remaining_Tokens := '' ;
        end ;
        if( length( Returned_Token ) > 0 ) then
        begin
            if( length( Result ) > 0 ) then
            begin
                Result := ' ' + Result ;
            end ;
            Result := Returned_Token + Result ;
            Returned_Token := '' ;
        end ;
    end ;


    function Peek_Token : string ;

    begin
        if( ( length( Remaining_Tokens ) = 0 ) and ( Returned_Token <> '' ) ) then
        begin
            Result := '' ;
            exit ;
        end ;
        Result := Get_Token ;
        if( length( Result ) > 0 ) then
        begin
            Put_Token( Result ) ;
        end ;
    end ;


var A : integer ;
    Align : int64 ;
    B : longint ;
    BB : string ;
    C, Dummy : integer ;
    D : double ;
    Entry : TState ;
    Ex : int64 ;
    Found : boolean ;
    Name, Arguments : string ;
    Nest_Level : integer ;
    Next : string ;
    P : pSymbol_Record ;
    S : string ;
    SL, SL1 : TStringList ;
    Si : single ;
    Stream : TCOM_Stream ;
    T, Original_T : boolean ;
    Temp, Work : string ;
    UEC : TUEC ;

label ASCII_Loop, ASCIC_Loop, ASCIZ_Loop ;

begin
    // Setup...
    Have_Value := ( length( Value ) > 0 ) ;
    Handle_Directives := False ;
    Remaining_Tokens := Edit( Value, 16 or 32 or 128 or 256 ) ;
    Value := Get_Token ;
    Next := Peek_Token ; // Peek at the next token on the line
    if( ( copy( Value, length( Value ), 1 ) <> ':' )
        and
        (
          ( Next = 'DB' ) or ( Next = '.DB' ) or ( Next = '.BYTE' ) or
          ( Next = 'DEFB' ) or ( Next = '.DEFB' ) or ( Next = '.SIGNED_BYTE' )
          or
          ( Next = '.ADDRESS' ) or ( Next = '.BLKA' )
          or
          ( copy( Next, 1, 4 ) = '.REF' ) or ( Next = '.ALIGN' ) or
          ( Next = '.ASCII' ) or ( Next = '.ASCIC' ) or ( Next = '.ASCIZ' ) or
          ( Next = '.BCD' )
          or
          ( Next = 'DW' ) or ( Next = '.DW' ) or ( Next = '.WORD' ) or
          ( Next = 'DEFW' ) or ( Next = '.DEFW' ) or ( Next = '.SIGNED_WORD' )
          or
          ( Next = 'DS' ) or ( Next = '.DS' ) or
          ( Next = 'DEFS' ) or ( Next = '.DEFS' ) or
          ( Next = '.BLKW' ) or ( Next = '.BLKB' ) or
          ( Next = '.BLKL' ) or ( Next = '.BLKQ' ) or ( Value = '.BLKO' ) or
          ( Next = '.BLKD' ) or ( Next = '.BLKF' )
          or
          ( Next = 'DEFM' ) or ( Next = '.DEFM' )
          or
          ( Next = '.DOUBLE' ) or ( Next = '.D_FLOATING' )
          or
          ( Next = '.LONG' )
          or
          ( Next = '.PACKED' )
          or
          ( Next = '.QUAD' )
          or
          ( Next = '.SINGLE' )
          or
          ( Next = '.S_FLOATING' )
          or
          ( Next = '.FLOAT' )
          or
          ( Next = '.S_FLOATING' )
        )
      ) then // Data directive
    begin
        new( P ) ;
        fillchar( P^, sizeof( P^ ), 0 ) ;
        P^.Address := PC ;
        P^.Typ := 1 ; // Integer
        P^.Flags := 1 ; // Constant
        P^.Data := PC ;
        UEC := Add_Symbol( PChar( Value ), P ) ;
        if( UEC.Code <> 0 ) then
        begin
            Status.Log_Error( PChar( Assembler_Context.Symbol_Table.ERT( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
        end ;
        Value := Get_Token ;
    end else
    if( length( Next ) > 0 ) then
    begin
        Put_Token( Next ) ;
    end ;

    if( ( Value = 'DB' ) or ( Value = '.DB' ) ) then
    begin
        S := Get_Token ;
	if( length( S ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'Missing expression' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        if(
            ( ( S[ 1 ] = '"' ) or ( S[ 1 ] = #39 ) )
            and
            ( length( S ) = 3 ) // Single quoted character
          ) then
        begin
            Put_Token( inttostr( ord( S[ 2 ] ) ) ) ; // Convert to number and push the token
        end else
        begin
            Put_Token( S ) ;
            if(
                ( ( S[ 1 ] = '"' ) or ( S[ 1 ] = #39 ) )
                and
                ( pos( '*', S ) = 0 )
                and
                ( pos( '/', S ) = 0 )
                and
                ( pos( '-', S ) = 0 )
                and
                ( pos( '+', S ) = 0 )
              ) then
            begin
                Value := 'DEFM' ;
            end ;
        end ;
    end ; // if( ( Value = 'DB' ) or ( Value = '.DB' ) ) 
    if( Value = '.ADDRESS' ) then
    begin
        if( _CPU = nil ) then
        begin
            Value := '.WORD' ;
        end else
        begin
            Dummy := XBits_For_Value( _CPU.CPU.Get_High_Virtual_Memory( 0 ) ) ;
            case Dummy of
                0..8 : Value := '.BYTE' ;
                9..16 : Value := '.WORD' ;
                17..32 : Value := '.LONG' ;
                else Value := '.QUAD' ;
            end ;
        end ;
    end ;
    if( Value = '.BLKA' ) then
    begin
        if( _CPU = nil ) then
        begin
            Value := '.BLKW' ;
        end else
        begin
            Dummy := XBits_For_Value( _CPU.CPU.Get_High_Virtual_Memory( 0 ) ) ;
            case Dummy of
                0..8 : Value := '.BLKB' ;
                9..16 : Value := '.BLKW' ;
                17..32 : Value := '.BLKL' ;
                else Value := '.BLKQ' ;
            end ;
        end ;
    end ;

    if( copy( Value, 1, 4 ) = '.REF' ) then
    begin
        Map( PC ) ;
        Value := copy( Value, 5, length( Value ) ) ;
        if( Value = '1' ) then
        begin
            Value := ',BYTE' ;
        end else
        if( Value = '2' ) then
        begin
            Value := '.WORD' ;
        end else
        if( Value = '4' ) then
        begin
            Value := '.LONG' ;
        end else
        if( Value = '8' ) then
        begin
            Value := '.QUAD' ;
        end else
        if( Value = '16' ) then
        begin
            Value := '.OCTA' ;
        end else
        begin
            Value := '.REF' + Value ; // So error reports actual directive
        end ;
    end ;

    // Process standard directives...
    if( Value = '.' ) then
    begin
        Value := Get_Token ;
        if( Value <> '=' ) then
        begin
            Put_Token( Value ) ;
            Value := '' ;
            Handle_Directives := False ;
            exit ;
        end ;
        if( _Token_EOL ) then
        begin
            Value := '' ;
        end else
        begin
            Value := _Grab_Line( True ) ; // Get value
        end ;
        Value := Edit( Value, 8 or 16 or 128 or 256 ) ;
        if( length( Value ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'Missing expression' ), nil, Tokens.Token_Line, Severity_Error ) ;
            Handle_Directives := False ;
            exit ;
        end ;
        Dummy := pos( '.', Value ) ;
        while( Dummy > 0 ) do
        begin
            Value := copy( Value, 1, Dummy - 1 ) +
                CvtB( 10, 16, num1( PC ) ) + 'H ' +
                copy( Value, Dummy + 1, length( Value ) ) ;
            Dummy := pos( '.', Value ) ;
        end ;
        UEC := Evaluate( PChar( Value ), Ex ) ;
        PC := Ex ;
        Value := '' ;
	Handle_Directives := True ;
        Altered_PC := True ;
        Altered_Address := PC ;
    end else
    if( Value = '.ALIGN' ) then
    begin
        Value := uppercase( Get_Token ) ;
        if( Value = 'BYTE' ) then
        begin
            Value := '0' ;
        end else
        if( Value = 'WORD' ) then
        begin
            Value := '1' ;
        end else
        if( Value = 'LONG' ) then
        begin
            Value := '2' ;
        end else
        if( Value = 'QUAD' ) then
        begin
            Value := '3' ;
        end else
        if( ( Value = 'OCTA' ) or ( Value = 'PARAGRAPH' ) ) then
        begin
            Value := '4' ;
        end else
        if( Value = 'PAGE' ) then
        begin
            if( _CPU <> nil ) then
            begin
                if( _CPU.CPU.Page_Size = 0 ) then
                begin
                    Value := '9' ;
                end else
                begin
                    Value := '' ;
                end ;
            end else
            begin
                Value := '9' ;
            end ;
        end ;
        if( Value = '' ) then
        begin
            Align := _CPU.CPU.Page_Size ;
        end else
        begin
            try
                Align := strtoint( Value ) ;
            except
                Status.Log_Error( PChar( 'Invalid alignment' ), nil, Tokens.Token_Line, Severity_Error ) ;
                Handle_Directives := False ;
                exit ;
            end ;
            if( ( Align < 0 ) or ( Align > 63 ) ) then
            begin
                Status.Log_Error( PChar( 'Invalid alignment' ), nil, Tokens.Token_Line, Severity_Error ) ;
                Handle_Directives := False ;
                exit ;
            end ;
            Align := Xbit_Values[ Align ] ;
            if( ( _CPU <> nil ) and ( Align > _CPU.CPU.Get_High_Virtual_Memory( 0 ) ) ) then
            begin
                Status.Log_Error( PChar( 'Invalid alignment' ), nil, Tokens.Token_Line, Severity_Error ) ;
                Handle_Directives := False ;
                exit ;
            end ;
        end ;
        Ex := PC div Align ;
        if( PC <> Ex * Align ) then // Adjustment is needed
        begin
            PC := ( Ex + 1 ) * Align ;
        end ;
        Value := '' ;
	Handle_Directives := True ;
        Altered_PC := True ;
        Altered_Address := PC ;
    end else
    if( Value = '.ASCII' ) then
    begin
        Map( PC ) ;
	Handle_Directives := True ;
        Value := '' ;
ASCII_Loop:
        S := Get_Token ;
        if( Pos( S[ 1 ], Copy( S, 2, Length( S ) ) ) = 0 ) then
	begin
            Status.Log_Error( PChar( 'Unterminated literal' ), nil, Tokens.Token_Line, Severity_Error ) ;
            Handle_Directives := False ;
            exit ;
        end ;
        for A := 2 to Length( S ) do
        begin
            if S[ A ] = S[ 1 ] then
            begin
                if ( A = Length( S ) ) or ( S[ A + 1 ] = ';' ) then
		begin
		    exit ;
		end ;
                S := Copy( S, A + 1, Length( S ) - A ) ;
                goto ASCII_Loop ;
            end ;
            Value := Value + S[ A ] ;
            Pc := Pc + 1 ;
        end ;
    end else
    if( Value = '.ASCIC' ) then
    begin
        Map( PC ) ;
	Handle_Directives := True ;
        Value := '' ;
ASCIC_Loop:
        S := Get_Token ;
        if( Pos( S[ 1 ], Copy( S, 2, Length( S ) ) ) = 0 ) then
	begin
            Status.Log_Error( PChar( 'Unterminated literal' ), nil, Tokens.Token_Line, Severity_Error ) ;
            Handle_Directives := False ;
            exit ;
        end ;
        for A := 2 to Length( S ) do
        begin
            if S[ A ] = S[ 1 ] then
            begin
                if ( A = Length( S ) ) or ( S[ A + 1 ] = ';' ) then
		begin
		    exit ;
		end ;
                S := Copy( S, A + 1, Length( S ) - A ) ;
                goto ASCII_Loop ;
            end ;
            Value := Value + S[ A ] ;
            Pc := Pc + 1 ;
        end ;
        Value := chr( length( Value ) ) + Value ;
        Pc := Pc + 1 ;
        if( length( Value ) > 256 ) then
        begin
            Status.Log_Error( PChar( 'Counted string is too long' ), nil, Tokens.Token_Line, Severity_Error ) ;
            Handle_Directives := False ;
            exit ;
        end ;
    end else
    if( Value = '.ASCIZ' ) then
    begin
        Map( PC ) ;
	Handle_Directives := True ;
        Value := '' ;
ASCIZ_Loop:
        S := Get_Token ;
        if Pos( S[ 1 ], Copy( S, 2, Length( S ) ) ) = 0 then
        begin
            Status.Log_Error( PChar( 'Unterminated literal' ), nil, Tokens.Token_Line, Severity_Error ) ;
            Handle_Directives := False ;
            exit ;
        end ;
        for A := 2 to Length( S ) do
        begin
            if S[ A ] = S[ 1 ] then
            begin
                if ( A = Length( S ) ) or ( S[ A + 1 ] = ';' ) then
		begin
		    exit ;
		end ;
                S := Copy( S, A + 1, Length( S ) - A ) ;
                goto ASCIZ_Loop ;
            end ;
            Value := Value + S[ A ] ;
            Pc := Pc + 1 ;
        end ;
        Value := Value + chr( 0 ) ;
        Pc := Pc + 1 ;
    end else
    if( Value = '.BCD' ) then // BCD
    begin
        Map( PC ) ;
        Work := Grab_This_Line ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'missing value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Value := '' ;
        while( length( Work ) > 0 ) do
        begin
            S := Parse_Parameter( ',', Work ) ;
            try
                if( pos( '.', S ) + pos( 'E', S ) + pos( 'e', S ) > 0 ) then
                begin
                    D := strtofloat( S ) ;
                    S := Num1( D ) ;
                end else
                begin
                    Ex := strtoint( S ) ;
                    S := Num1( Ex ) ;
                end ;
            except
                Status.Log_Error( PChar( 'Invalid number' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            if( pos( '.', S ) > 0 ) then
            begin
                Status.Log_Error( PChar( 'Invalid BCD value' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            Value := Value + String_To_BCD( S )
        end ;
    end else
    if( Value = '.DEFINE' ) then
    begin
        Value := Get_Token ;
	if( length( Value ) = 0 ) then
	begin
            Status.Log_Error( PChar( 'Missing identifier' ), nil, Tokens.Token_Line, Severity_Error ) ;
	    exit ;
	end ;
        new( P ) ;
        fillchar( P^, sizeof( P^ ), 0 ) ;
        P^.Address := Ex ;
        P^.Typ := 4 ; // Boolean
        P^.Flags := 1 ; // Constant
        UEC := Add_Symbol( PChar( Value ), P ) ;
        if( UEC.Code <> 0 ) then
        begin
            Status.Log_Error( PChar( Assembler_Context.Symbol_Table.ERT( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
        end ;
	Handle_Directives := ( UEC.Code = 0 ) ;
        Value := '' ;
    end else
    if(
        ( Value = 'DW' ) or ( Value = '.DW' ) or ( Value = '.WORD' ) or
        ( Value = 'DEFW' ) or ( Value = '.DEFW' ) or ( Value = '.SIGNED_WORD' )
      ) then
    begin
        Work := Grab_This_Line ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'missing value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Value := '' ;
        Map( PC ) ;
        while( length( Work ) > 0 ) do
        begin
            S := Parse_Parameter( ',', Work ) ;
            UEC := Evaluate( PChar( S ), Ex ) ;
            if( UEC.Code = CEFAssemblerErr_Undefined_Symbol ) then
            begin
                Add_Reference( PChar( S ), 2, PC ) ;
            end else
            if( UEC.Code <> 0 ) then
            begin
                Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            if( ( Ex > $FFFF ) or ( -Ex > $FFFF ) ) then
            begin
                Status.Log_Error( PChar( 'Illegal word value' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            setlength( S, 2 ) ;
            move( Ex, S[ 1 ], 2 ) ;
            Value := Value + S ;
        end ;
        Pc := Pc + length( Value ) ;
	Handle_Directives := True ;
    end else
    if( ( Value = 'DB' ) or ( Value = '.DB' ) or ( Value = '.BYTE' ) or
        ( Value = 'DEFB' ) or ( Value = '.DEFB' ) or ( Value = '.SIGNED_BYTE' ) ) then
    begin
        Work := Grab_This_Line ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'missing value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Value := '' ;
        Map( PC ) ;
        while( length( Work ) > 0 ) do
        begin
            S := Parse_Parameter( ',', Work ) ;
            UEC := Evaluate( PChar( S ), Ex ) ;
            if( UEC.Code = CEFAssemblerErr_Undefined_Symbol ) then
            begin
                Add_Reference( PChar( S ), 1, PC ) ;
            end else
            if( UEC.Code <> 0 ) then
            begin
                Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            if( ( Ex > 255 ) or ( Ex < -128 ) ) then
            begin
                Status.Log_Error( PChar( 'Illegal word value' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            Value := Value + chr( Ex ) ;
        end ;
        Pc := Pc + length( Value ) ;
	Handle_Directives := True ;
    end else
    if( ( Value = 'DS' ) or ( Value = '.DS' ) or
        ( Value = 'DEFS' ) or ( Value = '.DEFS' ) or
        ( Value = '.BLKW' ) or ( Value = '.BLKB' ) or
        ( Value = '.BLKL' ) or ( Value = '.BLKQ' ) or ( Value = '.BLKO' ) or
        ( Value = '.BLKD' ) or ( Value = '.BLKF' )
      ) then
    begin
        Map( PC ) ;
        UEC := Evaluate( PChar( Get_Token ), Ex ) ; //Should be Grab_Line
        if( UEC.Code <> 0 ) then
        begin
            exit ;
        end ;
        Pc := Pc + Ex ;
        if( Value = '.BLKW' ) then
        begin
            Pc := Pc + Ex ; {Twice as many bytes if BLKW}
        end else
        if( ( Value = '.BLKL' ) or ( Value = '.BLKF' ) ) then
        begin
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
        end else
        if( ( Value = '.BLKQ' ) or ( Value = '.BLKD' ) ) then
        begin
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
        end else
        if( Value = '.BLKO' ) then
        begin
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
            Pc := Pc + Ex ;
        end ;
	Handle_Directives := True ;
        Altered_PC := True ;
        Altered_Address := PC ;
        Value := '' ;
    end else
    if( ( Value = 'DEFM' ) or ( Value = '.DEFM' ) ) then
    begin
        Map( PC ) ;
        BB := Get_Token ;
        if( ( BB[ 1 ] <> '"' ) and ( BB[ 1 ] <> #39 ) ) then
        begin
            Status.Log_Error( PChar( 'Illegal value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        if( BB[ Length( BB ) ] <> BB[ 1 ] ) then
	begin
            Status.Log_Error( PChar( 'No terminating delimeter' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Value := '' ;
        for A := 2 to Length( Bb ) - 1 do
        begin
            if Bb[ A ] = Bb[ 1 ] then
            begin
                if( ( A = Length( Bb ) ) or ( Bb[ A + 1 ] = ';' ) ) then
                begin
                    exit ;
                end ;
                Bb := Copy( Bb, A + 1, Length( Bb ) - A ) ;
                goto ASCII_Loop ;
            end ;
            Value := Value + Bb[ A ] ;
            Pc := Pc + 1 ;
        end ;
	Handle_Directives := True ;
    end else
    if( ( Value = '.DOUBLE' ) or ( Value = '.D_FLOATING' ) ) then // Double-precision floating point
    begin
        Map( PC ) ;
        Work := Grab_This_Line ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'missing value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Value := '' ;
        while( length( Work ) > 0 ) do
        begin
            S := Parse_Parameter( ',', Work ) ;
            try
                D := strtofloat( S ) ;
            except
                Status.Log_Error( PChar( 'Illegal double value' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            setlength( S, sizeof( D ) ) ;
            move( D, S[ 1 ], sizeof( D ) ) ;
            Value := Value + S ;
        end ;
	Handle_Directives := True ;
    end else
    if Value = '.ELSE' then
    begin
        Status.Log_Error( '.ELSE outside of .IF block', nil, Tokens.Token_Line, Severity_Error ) ;
        exit ;
    end else
    if( ( Value = '.END' ) or ( Value = 'END' ) ) then
    begin
        Terminated := True ;
        if( not _Token_EOL ) then
        begin
            Value := Get_Token ;
            UEC := Evaluate( PChar( Value ), Ex ) ;
            if( UEC.Code <> 0 ) then
            begin
                Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            Start_Address := Ex ;
            Have_Start_Address := True ;
        end ;
	Handle_Directives := True ;
        Value := Get_Token ;
        while( Value <> '' ) do
        begin
            if( Value = ';' ) then
            begin
                Grab_Current_Line ;
                Value := Get_Token ;
            end else
            begin
                Status.Log_Error( PChar( 'Lines after .END ignored' ), nil, Tokens.Token_Line, Severity_Warning ) ;
                Value := '' ;
            end ;
        end ;
        Value := '' ;
    end else
    if Value = '.ENDC' then
    begin
        Status.Log_Error( '.ENDC outside of .IF block', nil, Tokens.Token_Line, Severity_Error ) ;
        exit ;
    end else
    if Value = '.ENDM' then
    begin
        Status.Log_Error( '.ENDM outside of .MACRO definition', nil, Tokens.Token_Line, Severity_Error ) ;
        exit ;
    end else
    if Value = '.ENDR' then
    begin
        Status.Log_Error( '.ENDR outside of .REPEAT block', nil, Tokens.Token_Line, Severity_Error ) ;
        exit ;
    end else
    if Value = '.EOT' then // Null directive
    begin
	Handle_Directives := True ;
        Value := '' ;
    end else
    if Value = '.ERROR' then
    begin
        Value := Grab_This_Line ;
        Work := Parse_Parameter( ';', Value ) ;
        if( length( Work ) > 0 ) then
        begin
            UEC := Evaluate( PChar( Work ), Ex ) ;
            if( UEC.Code <> 0 ) then
            begin
                Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            Work := inttostr( Ex ) ;
        end ;
        Status.Log_Error( PChar( Work + ' ' + Value ), nil, Tokens.Token_Line, Severity_Error ) ;
        exit ;
    end else
    if( Value = '.EVEN' ) then
    begin
        if Odd( Pc ) then
        begin
            Pc := Pc + 1 ;
            Altered_PC := True ;
            Altered_Address := PC ;
        end ;
	Handle_Directives := True ;
        Value := '' ;
    end else
    if( Value = '.IF' ) then
    begin
        Value := Get_Token ;
        S := Grab_This_Line ;
        Work := Parse_Parameter( ';', S ) ;
        if( not Valid_Condition( Value ) ) then
        begin
            Status.Log_Error( PChar( 'Invalid condition' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        T := Process_If( Value, Work, Status ) ;
        Original_T := T ;
        Have_Value := False ;
        SL := TStringList.Create ;
        Work := uppercase( Get_Token ) ;
        Nest_Level := 0 ;
        Found := False ; // .ELSE not found
        while( Work <> '' ) do
        begin
            if( Work = '.IF' ) then
            begin
                inc( Nest_Level ) ;
            end else
            if( ( ( Work = '.IF_TRUE' ) or ( Work = '.IFT' ) ) and ( Nest_Level = 0 ) ) then
            begin
                T := Original_T ;
                Work := uppercase( Get_Token ) ;
                continue ;
            end else
            if( ( ( Work = '.IF_FALSE' ) or ( Work = '.IFF' ) ) and ( Nest_Level = 0 ) ) then
            begin
                T := not Original_T ;
                Work := uppercase( Get_Token ) ;
                continue ;
            end else
            if( ( ( Work = '.IF_TRUE_FALSE' ) or ( Work = '.IFTF' ) ) and ( Nest_Level = 0 ) ) then
            begin
                T := True ;
                Work := uppercase( Get_Token ) ;
                continue ;
            end else
            if( ( Work = '.ELSE' ) and ( Nest_Level = 0 ) ) then
            begin
                T := not Original_T ;
                Work := uppercase( Get_Token ) ;
                if( Found ) then // Already an ELSE at this level
                begin
                    Status.Log_Error( PChar( 'Multiple .ELSE directives' ), nil, Tokens.Token_Line, Severity_Warning ) ;
                end ;
                Found := True ;
                continue ;
            end else
            if( Work = '.ENDC' ) then
            begin
                if( Nest_Level > 0 ) then
                begin
                    dec( Nest_Level ) ;
                end else
                begin
                    break ;
                end ;
            end ;
            if( T ) then
            begin
                SL.Add( Work + ' ' + Grab_This_Line ) ;
            end else
            begin
                Grab_This_Line ;
            end ;
            Work := uppercase( Get_Token ) ;
        end ; // while( Work <> '' )
	Handle_Directives := True ;
        Value := '' ;
        if( SL.Count > 0 ) then
        begin
            Enter_State( State_If, '', SL, nil ) ;
        end else
        begin
            SL.Free ;
        end ;
    end else
    if( copy( Value, 1, 4 ) = '.IF_' ) then
    begin
        Status.Log_Error( 'Subconditional outside of .IF block', nil, Tokens.Token_Line, Severity_Error ) ;
        exit ;
    end else
    if( Value = '.IIF' ) then
    begin
        Value := uppercase( Get_Token ) ;
        S := Get_Token ;
        if( S = ',' ) then
        begin
            S := Get_Token ;
        end ;
        if( not Valid_Condition( Value ) ) then
        begin
            Status.Log_Error( PChar( 'Invalid condition' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        if( not Process_If( Value, S, Status ) ) then
        begin
            Grab_This_Line ; // Ignore the line
        end else
        begin
            S := Get_Token ;
            if( S <> ',' ) then
            begin
                Status.Log_Error( PChar( 'Missing comma' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            S := Grab_This_Line ;
            In_Line( Create_COM_String_Stream( S ) ) ;
        end ;
	Handle_Directives := True ;
        Value := '' ;
    end else
    if( Value = '.INCLUDE' ) then
    begin
        Work := Grab_This_Line ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'Missing repeat value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Stream := _UI.Get_File_Stream( PChar( Work ) ) ;
        if( Stream = nil ) then
        begin
            exit ;
        end ;
        Enter_State( State_Include, Work, nil, Stream ) ;
	Handle_Directives := True ;
        Value := '' ;
    end else
    if( ( Value = '.IRP' ) or ( Value = '.IRPC' ) ) then
    begin
        Map( PC ) ;
        Work := Grab_This_Line ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'Missing parameter(s)' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        S := Parse_Parameter( ',', Work ) ;
        if( copy( Work, 1, 1 ) = '<' ) then
        begin
            Work := copy( Work, 2, length( Work ) ) ;
            if( copy( Work, length( Work ), 1 ) <> '>' ) then
            begin
                Status.Log_Error( PChar( 'Syntax error' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            setlength( Work, length( Work ) - 1 ) ;
        end else
        if( copy( Work, 1, 1 ) = '^' ) then
        begin
            Work := copy( Work, 2, length( Work ) ) ;
            if( copy( Work, length( Work ), 1 ) <> copy( Work, 1, 1 ) ) then
            begin
                Status.Log_Error( PChar( 'Syntax error' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            Work := copy( Work, 2, length( Work ) ) ;
            setlength( Work, length( Work ) - 1 ) ;
        end ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'Invalid parameter' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        S := S + ',' + Work ;
        Have_Value := False ;

        // Read entire block into stringlist SL...
        SL := TStringList.Create ;
        Work := uppercase( Get_Token ) ;
        Nest_Level := 0 ;
        while( Work <> '' ) do
        begin
            if( ( Work = '.REPEAT' ) or ( Work = '.REPT' ) or ( Work = '.IRP' ) or ( Work = '.IRPC' ) ) then
            begin
                inc( Nest_Level ) ;
            end ;
            if( Work = '.ENDR' ) then
            begin
                if( Nest_Level > 0 ) then
                begin
                    dec( Nest_Level ) ;
                end else
                begin
                    break ;
                end ;
            end ;
            SL.Add( Work + ' ' + Grab_This_Line ) ;
            Work := uppercase( Get_Token ) ;
        end ; // while( not Token.EOF )

        // Create string list from repeat...
        SL1 := TStringList.Create ;
        Work := Parse_Parameter( ',', S ) ; // Work is symbol
        if( Value = '.IRP' ) then
        begin
            // IRP...
            while( length( S ) > 0 ) do
            begin
                Temp := Parse_Parameter( ',', S ) ; // This parameter
                Process_IRP( Work, Temp, SL, SL1 ) ;
            end ;
        end else
        begin
            // IRPC...
            for Dummy := 1 to length( S ) do
            begin
                Process_IRP( Work, #39 + S[ Dummy ] + #39, SL, SL1 ) ;
            end ;
        end ;
        Enter_State( State_IRP, '', SL1, nil ) ;

        // Cleanup...
        SL.Free ;
	Handle_Directives := True ;
        Value := '' ;
    end else
    if( Value = '.LONG' ) then
    begin
        Work := Grab_This_Line ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'missing value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Value := '' ;
        Map( PC ) ;
        while( length( Work ) > 0 ) do
        begin
            S := Parse_Parameter( ',', Work ) ;
            UEC := Evaluate( PChar( S ), Ex ) ;
            if( UEC.Code = CEFAssemblerErr_Undefined_Symbol ) then
            begin
                Add_Reference( PChar( S ), 4, PC ) ;
            end else
            if( UEC.Code <> 0 ) then
            begin
                Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            if( ( Ex > $FFFFFFFF ) or ( -Ex > $FFFFFFFF ) ) then
            begin
                Status.Log_Error( PChar( 'Illegal long value' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            setlength( S, 4 ) ;
            move( Ex, S[ 1 ], 4 ) ;
            Value := Value + S ;
        end ;
        Pc := Pc + length( Value ) ;
	Handle_Directives := True ;
    end else
    if( Value = '.MACRO' ) then
    begin
        Arguments := Grab_This_Line ;
        if( length( Arguments ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'Missing macro name' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Name := Parse_Parameter( ' ', Arguments ) ;
        Have_Value := False ;
        SL := TStringList.Create ;
        Work := uppercase( Get_Token ) ;
        Nest_Level := 0 ;
        while( Work <> '' ) do
        begin
            if( Work = '.MACRO' ) then
            begin
                inc( Nest_Level ) ;
            end ;
            if( Work = '.ENDM' ) then
            begin
                if( Nest_Level > 0 ) then
                begin
                    dec( Nest_Level ) ;
                end else
                begin
                    break ;
                end ;
            end ;
            SL.Add( Work + ' ' + Grab_This_Line ) ;
            Work := uppercase( Get_Token ) ;
        end ; // while( Work <> '' )
        Define_Macro( Name, Arguments, SL ) ;
	Handle_Directives := True ;
        Value := '' ;
    end else
    if( Value = '.MDELETE' ) then
    begin
        Value := uppercase( Grab_This_Line ) ;
        if( Macros.IndexOf( Value ) <> -1 ) then
        begin
            Delete_Macro( Value ) ;
            Handle_Directives := True ;
        end else
        begin
            Status.Log_Error( PChar( 'Undefined macro' ), nil, Tokens.Token_Line, Severity_Error ) ;
        end ;
        Value := '' ;
    end else
    if( Value = '.MEXIT' ) then
    begin
        // See if we are inside any macro...
        Found := False ;
        for A := 0 to States.Count - 1 do
        begin
            Entry := TState( States[ States.Count - 1 ] ) ;
            if( Entry.State = State_Macro ) then
            begin
                Found := True ;
                break ;
            end ;
        end ;
        if( not Found ) then
        begin
            Status.Log_Error( PChar( 'Exit when not in macro' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;

        // Now pop all states back to, and including, the last macro
        while( True ) do
        begin
            Entry := TState( States[ States.Count - 1 ] ) ;
            if( Entry.State = State_Macro ) then
            begin
                Exit_State ;
                break ;
            end ;
            Exit_State ;
        end ;
    end else
    if( Value = '.NARG' ) then
    begin
        Status.Log_Error( PChar( '.NARG when not in macro' ), nil, Tokens.Token_Line, Severity_Error ) ;
        exit ;
    end else
    if( Value = '.NCHR' ) then
    begin
        Value := Edit( Grab_This_Line, 8 or 128 ) ;
        if( ( length( Value ) = 0 ) or ( Value[ 1 ] = ',' ) ) then
        begin
            Status.Log_Error( 'Missing symbol', nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Dummy := pos( ',', Value ) ;
        if( Dummy = 0 ) then
        begin
            Status.Log_Error( 'Missing symbol', nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        S := Edit( copy( Value, Dummy + 1, length( Value ) ), 8 or 128 ) ; // String
        Value := Edit( copy( Value, 1, Dummy - 1 ), 8 or 128 ) ; // Symbol name
        if( copy( S, 1, 1 ) = '<' ) then
        begin
            S := copy( S, 2, length( S ) ) ;
            if( copy( S, length( S ), 1 ) <> '>' ) then
            begin
                Status.Log_Error( 'Unterminated string literal', nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            S := copy( S, 1, length( S ) - 1 ) ;
        end else
        if( copy( S, 1, 1 ) = '^' ) then
        begin
            S := copy( S, 2, length( S ) ) ;
            if( copy( S, length( S ), 1 ) <> copy( S, 1, 1 ) ) then
            begin
                Status.Log_Error( PChar( 'Syntax error' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            S := copy( S, 2, length( S ) ) ;
            setlength( S, length( S ) - 1 ) ;
        end ;
        new( P ) ;
        fillchar( P^, sizeof( P^ ), 0 ) ;
        P^.Address := length( S ) ;
        P^.Typ := 1 ; // Integer
        P^.Flags := 1 ; // Constant
        P^.Data := length( S ) ; // Value
        UEC := Add_Symbol( PChar( Value ), P ) ;
        if( UEC.Code <> 0 ) then
        begin
            Status.Log_Error( PChar( Assembler_Context.Symbol_Table.ERT( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
	Handle_Directives := True ;
        Value := '' ;
    end else
    if( Value = '.ODD' ) then
    begin
        if not Odd( Pc ) then
        begin
            Pc := Pc + 1 ;
            Altered_PC := True ;
            Altered_Address := PC ;
        end ;
	Handle_Directives := True ;
        Value := '' ;
    end else
    if( Value = '.PACKED' ) then // Packed BCD
    begin
        Map( PC ) ;
        Work := Grab_This_Line ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'missing value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Value := '' ;
        while( length( Work ) > 0 ) do
        begin
            S := Parse_Parameter( ',', Work ) ;
            try
                if( pos( '.', S ) + pos( 'E', S ) + pos( 'e', S ) > 0 ) then
                begin
                    D := strtofloat( S ) ;
                    S := Num1( D ) ;
                end else
                begin
                    Ex := strtoint( S ) ;
                    S := Num1( Ex ) ;
                end ;
            except
                Status.Log_Error( PChar( 'Invalid number' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            if( pos( '.', S ) > 0 ) then
            begin
                Status.Log_Error( PChar( 'Invalid BCD value' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            Value := Value + String_To_Packed_BCD( S )
        end ;
    end else
    if( Value = '.QUAD' ) then
    begin
        Work := Grab_This_Line ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'missing value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Value := '' ;
        Map( PC ) ;
        while( length( Work ) > 0 ) do
        begin
            S := Parse_Parameter( ',', Work ) ;
            UEC := Evaluate( PChar( S ), Ex ) ;
            if( UEC.Code = CEFAssemblerErr_Undefined_Symbol ) then
            begin
                Add_Reference( PChar( S ), 8, PC ) ;
            end else
            if( UEC.Code <> 0 ) then
            begin
                Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            setlength( S, 8 ) ;
            move( Ex, S[ 1 ], 8 ) ;
            Value := Value + S ;
        end ;
        Pc := Pc + length( Value ) ;
	Handle_Directives := True ;
    end else
    if Value = '.RADIX' then // Base conversion
    begin
        Value := Get_Token ;
        Val( Value, B, C ) ;
        Value := '' ;
        if( ( C <> 0 ) or ( B < 2 ) or ( B > 49 ) ) then
	begin
            Status.Log_Error( PChar( 'Invalid base' ), nil, Tokens.Token_Line, Severity_Error ) ;
	    Err := True ;
	    exit ;
	end ;
        Base := B ;
	Handle_Directives := True ;
    end else
    if( Value = '.RADIX50' ) then
    begin
        Map( PC ) ;
        Work := Grab_This_Line ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'missing value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Value := '' ;
        while( length( Work ) > 0 ) do
        begin
            S := Parse_Parameter( ',', Work ) ;
            if( not Valid_Radix50( S ) ) then
            begin
                Status.Log_Error( PChar( 'Illegal Radix-50 value' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            Value := Value + ASCII_To_RAD( S ) ;
        end ;
	Handle_Directives := True ;
    end else
    if( ( Value = '.REPEAT' ) or ( Value = '.REPT' ) ) then
    begin
        Map( PC ) ;
        Work := Grab_This_Line ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'Missing repeat value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        UEC := Evaluate( PChar( Work ), Ex ) ;
        if( UEC.Code <> 0 ) then
        begin
            Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Have_Value := False ;
        SL := TStringList.Create ;
        Work := uppercase( Get_Token ) ;
        Nest_Level := 0 ;
        while( Work <> '' ) do
        begin
            if( ( Work = '.REPEAT' ) or ( Work = '.REPT' ) or ( Work = '.IRP' ) or ( Work = '.IRPC' ) ) then
            begin
                inc( Nest_Level ) ;
            end ;
            if( Work = '.ENDR' ) then
            begin
                if( Nest_Level > 0 ) then
                begin
                    dec( Nest_Level ) ;
                end else
                begin
                    break ;
                end ;
            end ;
            SL.Add( Work + ' ' + Grab_This_Line ) ;
            Work := uppercase( Get_Token ) ;
        end ; // while( not Token.EOF )
        Enter_State( State_Repeat, inttostr( Ex ), SL, nil ) ;
	Handle_Directives := True ;
        Value := '' ;
    end else
    if(
        ( Value = '.SINGLE' )
        or
        ( Value = '.S_FLOATING' )
        or
        ( Value = '.FLOAT' )
        or
        ( Value = '.S_FLOATING' )
      ) then // Single-precision floating point
    begin
        Map( PC ) ;
        Work := Grab_This_Line ;
        if( length( Work ) = 0 ) then
        begin
            Status.Log_Error( PChar( 'missing value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
        Value := '' ;
        while( length( Work ) > 0 ) do
        begin
            S := Parse_Parameter( ',', Work ) ;
            try
                Si := strtofloat( S ) ;
            except
                Status.Log_Error( PChar( 'Illegal double value' ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            setlength( S, sizeof( Si ) ) ;
            move( Si, S[ 1 ], sizeof( Si ) ) ;
            Value := Value + S ;
        end ;
	Handle_Directives := True ;
    end else
    if( Value = '.UNDEFINE' ) then
    begin
        Value := Get_Token ;
	if( length( Value ) = 0 ) then
	begin
            Status.Log_Error( PChar( 'Missing identifier' ), nil, Tokens.Token_Line, Severity_Error ) ;
	    exit ;
	end ;
        Assembler_Context.Symbol_Table.Delete( Value ) ;
	Handle_Directives := True ;
        Value := '' ;
    end else
    if( ( Value = '.WARN' ) or ( Value = '.WARNING' ) ) then
    begin
        Value := Grab_This_Line ;
        Work := Parse_Parameter( ';', Value ) ;
        if( length( Work ) > 0 ) then
        begin
            UEC := Evaluate( PChar( Work ), Ex ) ;
            if( UEC.Code <> 0 ) then
            begin
                Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
                exit ;
            end ;
            Work := inttostr( Ex ) ;
        end ;
        Status.Log_Error( PChar( Work + ' ' + Value ), nil, Tokens.Token_Line, Severity_Error ) ;
        exit ;
    end else
    if( ( Value[ 1 ] = '.' ) or ( Value[ 1 ] = '$' ) ) then { PC adjustment }
    begin
        Fedit( Value, Value, 511 ) ; { Format it }
        if( copy( Value, 2, 1 ) <> '=' ) then
        begin
            exit ;
        end ;
        Value := Copy( Value, 3, Length( Value ) - 2 ) ;
        UEC := Evaluate( PChar( Value ), Ex ) ;
        Value := '' ;
        if( UEC.Code <> 0 ) then
	begin
            Status.Log_Error( PChar( ET( UEC.Code ) + Value ), nil, Tokens.Token_Line, Severity_Error ) ;
	    Err := True ;
	    exit ;
        end else
        if( UEC.Code <> 0 ) then
	begin
            Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
	    Err := True ;
	    exit ;
	end ;
        PC := Ex ;
        Altered_PC := True ;
        Altered_Address := PC ;
	Handle_Directives := True ;
    end else
    if( pos( ':', Value ) = length( Value ) ) then // Label
    begin
        Value := copy( Value, 1, length( Value ) - 1 ) ;
        new( P ) ;
        fillchar( P^, sizeof( P^ ), 0 ) ;
        P^.Address := PC ;
        P^.Typ := 1 ; // Integer
        P^.Flags := 1 ; // Constant
        P^.Data := PC ;
        UEC := Add_Symbol( PChar( Value ), P ) ;
        Value := '' ;
        if( UEC.Code <> 0 ) then
        begin
            Status.Log_Error( PChar( Assembler_Context.Symbol_Table.ERT( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
        end ;
	Handle_Directives := ( UEC.Code = 0 ) ;
    end ;
    if( Result ) then
    begin
        Status.Data := Status.Data + length( Value ) ;
    end ;
end ; { TCEF_Assembler.Handle_Directives }


procedure TCEF_Assembler.Enter_State( State : integer ; const Context : string ;
    List : TStringList ; Stream : TCom_Stream ) ;

var Entry : TState ;
    Stream_File : PStreamer_File ;

begin
    Entry := TState.Create ;
    States.Add( Entry ) ;
    Entry.State := State ;
    Entry.Context := Context ;
    Entry.List := List ;
    case State of
        State_Repeat, State_Macro, State_IRP, State_If :
            Entry.Tokens := TStringList_Parser.Create( List ) ;
        State_Include, State_Inline :
            begin
                Entry.Tokens := TFile_Parser.Create ;
                new( Stream_File, Init( nil, False, 1 ) ) ;
                Stream_File^.Stream := Stream ;
                TFile_Parser( Entry.Tokens ).Set_Source( Stream_File ) ;
            end ;
    end ; // case State
end ;


function TCEF_Assembler._Get_Token : string ;

var Entry : TState ;
    I : integer ;

begin
    if( States.Count > 0 ) then
    begin
        Entry := TState( States[ States.Count - 1 ] ) ;
        Result := Entry.Tokens.Token ;
        if( Result = '' ) then
        begin
            if( Entry.State = State_Repeat ) then
            begin
                I := strtoint( Entry.Context ) ;
                I := I - 1 ;
                if( I > 0 ) then
                begin
                    Entry.Context := inttostr( I ) ;
                    TStringList_Parser( Entry.Tokens ).Reset ;
                    Result := _Get_Token ;
                    exit ;
                end ;
            end ;
            Exit_State ;
            Result := _Get_Token ;
        end else
        if( Result = '.' ) then
        begin
            Result := '.' + Entry.Tokens.Token ;
        end ;
        exit ;
    end ;
    try
        Result := Tokens.Token ;
    except
        Result := '' ;
    end ;
end ;


procedure TCEF_Assembler.Exit_State ;

var Index : integer ;

begin
    Index := States.Count - 1 ;
    TState( States[ Index ] ).Free ;
    States.Delete( Index ) ;
end ;


function TCEF_Assembler._Grab_Line( Current : boolean ) : string ;

var Entry : TState ;
    I : integer ;

begin
    if( States.Count > 0 ) then
    begin
        Entry := TState( States[ States.Count - 1 ] ) ;
        Result := Entry.Tokens.Grab_Line ;
        if( Result = '' ) then
        begin
            if( Entry.State = State_Repeat ) then
            begin
                I := strtoint( Entry.Context ) ;
                I := I - 1 ;
                if( I = 0 ) then
                begin
                    Exit_State ;
                end else
                begin
                    Entry.Context := inttostr( I ) ;
                    TStringList_Parser( Entry.Tokens ).Reset ;
                end ;
            end else
            begin
                Exit_State ;
            end ;
            Result := _Grab_Line( Current ) ;
        end ;
        exit ;
    end ; // if( States.Count > 0 )

    if( Tokens.Token_EOL and Current ) then
    begin
        Result := '' ;
    end else
    begin
        Result := Tokens.Grab_Line ; // Get value
    end ;
end ; // TCEF_Assembler._Grab_Line


procedure TCEF_Assembler._Put_Token( const Token : string ) ;

var Entry : TState ;

begin
    if( States.Count > 0 ) then
    begin
        Entry := TState( States[ States.Count - 1 ] ) ;
        Entry.Tokens.Put_Token( Token ) ;
        exit ;
    end ;
    Tokens.Put_Token( Token ) ;
end ;


function TCEF_Assembler._Token_EOL : boolean ;

var Entry : TState ;

begin
    if( States.Count > 0 ) then
    begin
        Entry := TState( States[ States.Count - 1 ] ) ;
        Result := Entry.Tokens.Token_EOL ;
    end else
    begin
        Result := Tokens.Token_EOL ;
    end ;
end ;


function TCEF_Assembler.Process_If( Condition, Argument : string ;
    Status : TAssembler_Status ) : boolean ;

var Negate : boolean ;
    Symbol_D_T, Symbol_Siz, Symbol_Flg : integer ;
    Symbol_Addr : int64 ;
    Symbol_Data : pointer ;
    UEC : TUEC ;
    Value : int64 ;

begin
    Result := False ;
    Negate := False ;
    Condition := Edit( Condition, 8 or 32 or 128 ) ;
    if( ( Condition = 'NOT_EQUAL' ) or ( Condition = 'NE' ) ) then
    begin
        Negate := True ;
        Condition := 'EQUAL' ;
    end else
    if( ( Condition = 'LESS_EQUAL' ) or ( Condition = 'LE' ) ) then
    begin
        Negate := True ;
        Condition := 'GREATER' ;
    end else
    if( ( Condition = 'GREATER_EQUAL' ) or ( Condition = 'GE' ) ) then
    begin
        Negate := True ;
        Condition := 'LESS_THAN' ;
    end else
    if( ( Condition = 'NOT_DEFINED' ) or ( Condition = 'NDF' ) ) then
    begin
        Negate := True ;
        Condition := 'DEFINED' ;
    end else
    if( ( Condition = 'NOT_BLANK' ) or ( Condition = 'NB' ) ) then
    begin
        Negate := True ;
        Condition := 'BLANK' ;
    end ;

    if( ( Condition = 'EQUAL' ) or ( Condition = 'EQ' ) ) then
    begin
        UEC := Evaluate( PChar( Argument ), Value ) ;
        if( UEC.Code = 0 ) then
        begin
            Result := ( Value = 0 ) ;
        end else
        begin
            Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
    end else
    if( ( Condition = 'GREATER' ) or ( Condition = 'GT' ) ) then
    begin
        UEC := Evaluate( PChar( Argument ), Value ) ;
        if( UEC.Code = 0 ) then
        begin
            Result := ( Value > 0 ) ;
        end else
        begin
            Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
    end else
    if( ( Condition = 'LESS_THAN' ) or ( Condition = 'LT' ) ) then
    begin
        UEC := Evaluate( PChar( Argument ), Value ) ;
        if( UEC.Code = 0 ) then
        begin
            Result := ( Value < 0 ) ;
        end else
        begin
            Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
    end else
    if( ( Condition = 'DEFINED' ) or ( Condition = 'DF' ) ) then
    begin
        Result := ( Assembler_Context.Symbol_Table.Find( Argument, Symbol_Addr,
            Symbol_Flg, Symbol_D_T, Symbol_Siz, Symbol_Data ) = 0 ) ;
    end else
    if( ( Condition = 'BLANK' ) or ( Condition = 'B' ) ) then
    begin
        Result := ( Edit( Argument, -1 ) = '' ) ;
    end ;
    if( Negate ) then
    begin
        Result := not Result ;
    end ;
end ; // TCEF_Assembler.Process_If


procedure TCEF_Assembler.Process_IRP( Symbol, Value : string ;
    Source, Target : TStringList ) ;

var Dummy : integer ;
    S : string ; // Argument values
    Parser : TString_Parser ;
    Work_S : string ; // Current argument being processed

begin
    // Setup...
    Symbol := Edit( Symbol, 32 or 128 ) ;
    Parser := TString_Parser.Create ;

    // Make copy of macro string list...
    for Dummy := 0 to Source.Count - 1 do // For each line in macro
    begin
        Work_S := Source[ Dummy ] ;
        S := Parse_Parameter( ';', Work_S ) ;
        Work_S := '' ;
        if( length( S ) > 0 ) then // Non-blank line
        begin
            Parser.Set_Source( S ) ;
            S := Parser.Token ;
            while( length( S ) > 0 ) do // Parse through all tokens in this line
            begin
                if( uppercase( S ) = Symbol ) then // An argument found
                begin
                    S := Value ;
                end ;
                if( length( Work_S ) > 0 ) then
                begin
                    Work_S := Work_S + ' ' ; // Space to separate tokens
                end ;
                Work_S := Work_S + S ;
                S := Parser.Token ;
            end ; // while( length( S ) > 0 )
        end ; // if( length( S ) > 0 )
        Target.Add( Work_S ) ;
    end ; // for Dummy := 0 to Source.Count - 1
    Parser.Free ;
end ; // TCEF_Assembler.Process_IRP


// API...

function TCEF_Assembler.Get_Assembler_Context : TAssembler_Context ;

begin
    if( _Assembler_Context = nil ) then
    begin
        _Assembler_Context := TAssembler_Context.Create ;
    end ;
    Result := _Assembler_Context ;
end ;


// API...

procedure TCEF_Assembler.Define_Macro( Name : string ;
    const Arguments : string ; SL : TStringList ) ;

var Index : integer ;

begin
    Name := uppercase( Name ) ;
    Index := Macros.IndexOf( Name ) ;
    if( Index = -1 ) then // New definition
    begin
        Macros.AddObject( Name, SL ) ;
        Macro_Arguments.Add( Arguments ) ;
    end else
    begin
        TStringList( Macros.Objects[ Index ] ).Free ;
        Macros.Objects[ Index ] := SL ;
        Macro_Arguments[ Index ] := Arguments ;
    end ;
end ;


procedure TCEF_Assembler.Delete_Macro( Name : string ) ;

var Index : integer ;

begin
    Name := uppercase( Name ) ;
    Index := Macros.IndexOf( Name ) ;
    if( Index <> -1 ) then
    begin
        TStringList( Macros.Objects[ Index ] ).Free ;
        Macros.Delete( Index ) ;
        Macro_Arguments.Delete( Index ) ;
    end ;
end ;


function TCEF_Assembler.Process_Macro( const Name : string ;
    Status : TAssembler_Status ) : boolean ;

var Dummy, I, Idx, Index : integer ;
    Args, Default, S, Upper_S : string ; // Argument values
    Parser : TString_Parser ;
    SL, Temp, Arg_Names, Defaults, XLate : TStringList ;
    Work, Work_Arg, Work_S : string ; // Current argument being processed
    UEC : TUEC ;
    Value : int64 ;

begin
    // Setup...
    Result := False ; // Assume no errors
    Index := Macros.IndexOf( Name ) ;
    if( Index = -1 ) then
    begin
        Status.Log_Error( 'Unknown macro', nil, Tokens.Token_Line, Severity_Error ) ;
        Result := True ;
        exit ;
    end ;

    XLate := TStringList.Create ; // Used to substitute arguments in macro
    Arg_Names := TStringList.Create ;
    Defaults := TStringList.Create ;
    Parser := TString_Parser.Create ;

    // Validate arguments...
    try
        S := _Grab_Line( True ) ;
        Args := Macro_Arguments[ Index ] ;
        while( ( length( S ) > 0 ) and ( length( Args ) > 0 ) ) do
        begin
            // Parse arguments...
            Work_Arg := uppercase( Parse_Parameter( ',', Args ) ) ;
            Dummy := pos( '=', Work_Arg + '=' ) ;
            Default := Edit( copy( Work_Arg, Dummy + 1, length( Work_Arg ) ), 8 or 128 ) ; // Formal argument default value
            Work_Arg := Edit( copy( Work_Arg, 1, Dummy - 1 ), 8 or 128 ) ; // Formal argument name
            Work_S := Edit( Parse_Parameter( ',', S ), 8 or 128 ) ; // Next actual argument
            if( Arg_Names.IndexOf( Work_Arg ) <> -1 ) then
            begin
                Status.Log_Error( 'Duplicate formal arguments', nil, Tokens.Token_Line, Severity_Error ) ;
                Result := True ;
                exit ;
            end ;

            // Handle symbol numeric value translations
            if( copy( Work_S, 1, 1 ) = '\' ) then
            begin
                Work_S := copy( Work_S, 2, length( Work_S ) ) ;
                UEC := Evaluate( PChar( Work_S ), Value ) ;
                if( UEC.Code <> 0 ) then
                begin
                    Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
                    Result := True ;
                    exit ;
                end ;
                Work_S := inttostr( Value ) ;
            end ;

            // Handle created local labels
            if( copy( Work_Arg, 1, 1 ) = '?' ) then
            begin
                Work_Arg := copy( Work_Arg, 2, length( Work_Arg ) ) ;
                if( length( Default ) = 0 ) then
                begin
                    Default := inttostr( Current_Created_Local_Label ) + '$' ;
                    inc( Current_Created_Local_Label ) ;
                end ;
            end ;

            // Store arguments in lists...
            Arg_Names.Add( Work_Arg ) ; // Formal argument name
            XLate.Add( Work_S ) ; // Actual argument
            Defaults.Add( Default ) ; // Formal argument default value
        end ; // while( ( length( S ) > 0 ) and ( length( Args ) > 0 ) )
        if( S <> '' ) then // User supplied more arguments than macro has
        begin
            Status.Log_Error( 'Too many macro arguments', nil, Tokens.Token_Line, Severity_Error ) ;
            Result := True ;
            exit ;
        end ;
        if( Args <> '' ) then // Not all arguments supplied
        begin
            Status.Log_Error( 'Not enough macro arguments', nil, Tokens.Token_Line, Severity_Error ) ;
            Result := True ;
            exit ;
        end ;

        // Default blank actual arguments...
        for I := 0 to XLate.Count - 1 do
        begin
            if( ( XLate[ I ] = '' ) and ( Defaults.Count > I  ) ) then
            begin
                XLate[ I ] := Defaults[ I ] ;
            end ;
        end ;

        // Make copy of macro string list...
        SL := TStringList.Create ;
        Temp := TStringList( Macros.Objects[ Index ] ) ;
        for Dummy := 0 to Temp.Count - 1 do // For each line in macro
        begin
            Work_S := Temp[ Dummy ] ;
            if( XLate.Count > 0 ) then
            begin
                S := Parse_Parameter( ';', Work_S ) ;
                Work_S := '' ;
                if( length( S ) > 0 ) then
                begin
                    // Do argument concatenation...
                    I := pos( #39, S ) ;
                    while( I <> 0 ) do
                    begin
                        Index := instr( I + 1, S + #39, #39 ) ; // Find ending apostrophe (or end of line)
                        Work := copy( S, I + 1, Index - I - 1 ) ;
                        Idx := Arg_Names.IndexOf( uppercase( Work ) ) ;
                        if( Idx = -1 ) then // Not a formal argument
                        begin
                            I := Index + 1 ; // Skip past ending apostrophe
                        end else
                        begin
                            S := copy( S, 1, I - 1 ) + XLate[ Idx ] + copy( S, Index + 1, length( S ) ) ;
                            I := I + length( XLate[ Idx ] ) ;
                        end ;
                        I := instr( I, S, #39 ) ; // Find next apostrophe
                    end ;

                    // Do argument substitution
                    Parser.Set_Source( S ) ;
                    S := Parser.Token ;
                    while( length( S ) > 0 ) do // Parse through all tokens in this line
                    begin
                        if( S = '.' ) then
                        begin
                            S := S + Parser.Token ;
                        end ;
                        Upper_S := uppercase( S ) ;
                        I := Arg_Names.Indexof( Upper_S ) ;
                        if( I <> -1 ) then // An argument found
                        begin
                            S := XLate[ I ] ;
                        end else
                        if( Upper_S = '.NARG' ) then
                        begin
                            Arg_Names.Add( uppercase( Parser.Token ) ) ;
                            XLate.Add( inttostr( Defaults.Count ) ) ;
                            S := Parser.Token ;
                            continue ;
                        end ;
                        if( length( Work_S ) > 0 ) then
                        begin
                            Work_S := Work_S + ' ' ; // Space to separate tokens
                        end ;
                        Work_S := Work_S + S ;
                        S := Parser.Token ;
                    end ; // while( length( S ) > 0 )
                end ; // if( XLate.Count > 0 )
            end ; // if( length( S ) > 0 )
            SL.Add( Work_S ) ;
        end ; // for Dummy := 0 to Temp.Count - 1
    finally
        XLate.Free ;
        Arg_Names.Free ;
        Defaults.Free ;
        Parser.Free ;
    end ;

    // Enter macro state...
    Enter_State( State_Macro, '', SL, nil ) ;
end ; // TCEF_Assembler.Process_Macro


// API (overrides)...

{ Assembles code from the Input stream, directing the image data to the
  Output stream, and placing any list output to the Listing stream.
  Note that Listing can be NULL if no listing is desired.  Output can be
  NULL to do a syntax-check-only assembly.  Status can be used by the assembler
  to provide assembly statistics. }
function TCEF_Assembler.Assemble( Input, Output, Listing : TCOM_Stream ;
    Status : TAssembler_Status ) : TUEC ;

var A : string ;
    Address : int64 ;
    _Assembler : TAssembler ;
    Ignore, Machine : PChar ;
    Index, Len : integer ;
    Stream_File : TStreamer_File ;
    P : PChar ;
    PL : longint ;
    S : string ;
    UEC : TUEC ;

begin
    // Setup...
    Current_Created_Local_Label := 30000 ;
    Terminated := False ;
    A := '' ;
    Address := 0 ;
    Err := False ; // No errors so far
    Stream_File.Init( nil, True, 1 ) ;
    Stream_File.Stream := Input ;
    Tokens.Set_Source( @Stream_File ) ;
    In_Assembly_Status := Status ;

    // Default symbol values...
    Initial_Symbols := 'ABCDEFGHIJKLMNOPQRSTUVWXYZ' ;
    Valid_Symbols := 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789' ;
    if( _CPU <> nil ) then
    begin
        _Assembler := _CPU.CPU.Get_Assembler( Self ) ;
        P := _Assembler.Valid_Symbol_Initial ;
        if( P <> nil ) then
        begin
            Initial_Symbols := string( P ) ;
        end ;
        P := _Assembler.Valid_Symbol_After ;
        if( P <> nil ) then
        begin
            Valid_Symbols := string( P ) ;
        end ;
        Address := _CPU.CPU.Get_Current_Address( 0, True ) ;
    end ;
    Have_Start_Address := False ;

    // Do the assembly...
    while( not Terminated ) do
    begin
        Status.Set_Line( Tokens.Token_Line ) ;
        Application.ProcessMessages ;
	if( Input.At_End and ( States.Count = 0 ) ) then
        begin
            Status.Log_Error( 'Unexpected end of source', nil, Tokens.Token_Line, Severity_Warning ) ;
            break ;
        end ;
        S := uppercase( _Get_Token ) ;
        Index := Macros.IndexOf( S ) ;
        if( Index = -1 ) then
        begin
            _Put_Token( S ) ;
        end else
        begin
            if( Process_Macro( S, Status ) ) then
            begin
                break ;
            end ;
            continue ;
        end ;
        if( _CPU <> nil ) then
        begin
            PC := Address ;
            if( Status.Aborted ) then
            begin
                break ;
            end ;
            UEC := _Assembler.Assemble( nil, Ignore, Machine, Len, PC, Segment, Status ) ;
            if( UEC.Code = 0 ) then
            begin
                if( Altered_PC ) then
                begin
                    Address := Altered_Address ;
                    PC := Address ;
                    _CPU.CPU.Set_Current_Address( 0, True, Address ) ;
                end else
                begin
                    Address := _CPU.CPU.Translate( 0, PC ) ;
                end ;
                if( ( Output <> nil ) and ( Len > 0 ) ) then
                begin
                    Output.Seek( Address ) ;
                    Output.Write( Machine[ 0 ], Len ) ;
                    Address := Address + Len ;
                end ;
            end ;
            Altered_PC := False ;
        end else
        begin
            UEC := Expand( nil, P, PL, Status ) ;
            if( UEC.Code <> 0 ) then
            begin
                Status.Log_Error( PChar( 'Illegal instruction' ), nil, Tokens.Token_Line, Severity_Error ) ;
                break ;
            end else
            begin
                Output.Write( P[ 0 ], PL ) ;
            end ;
        end ;
    end ; // while( True )
    Stream_File.Done ;
    In_Assembly_Status := nil ;
    if( ( _CPU <> nil ) and Have_Start_Address ) then
    begin
        _CPU.CPU.Set_Current_Address( 0, True, Start_Address ) ;
    end ;
end ; // TCEF_Assembler.Assemble


procedure TCEF_Assembler.Add_Reference( Name : PChar ; Size : longint ;
    Address : int64 ) ;

var P : PSymbol_Record ;

begin
    if( Backpatch_Table = nil ) then
    begin
        Backpatch_Table := TStringList.Create ;
    end ;
    new( P ) ;
    P^.Address := Address ;
    P^.Size := Size ;
    Backpatch_Table.AddObject( string( Name ), TObject( P ) ) ;
end ;


procedure TCEF_Assembler.Backpatch( Status : TAssembler_Status ;
    Stream : TCOM_Stream ) ;

    function In_Range( Size : longint ; Value : int64 ) : boolean ;

    var H : int64 ;

    begin
       case Size of
           1 : H := 255 ;
           2 : H := 65535 ;
           3 : H := $FFFFFF ;
           4 : H := $FFFFFFFF ;
           5 : H := $FFFFFFFFFF ;
           6 : H := $FFFFFFFFFFFF ;
           7 : H := $FFFFFFFFFFFFFF ;
           else
               begin
                   Result := True ;
                   exit ;
               end ;
       end ;
       if( Value < 0 ) then
       begin
           Value := Value + ( H div 2 ) + 1 ;
       end ;
       In_Range := Value <= H ;
    end ;

var Loop : integer ;
    P : PSymbol_Record ;
    Resolved : boolean ;
    UEC : TUEC ;
    Value : int64 ;

begin // TCEF_Assembler.Backpatch
    // Resolve forward-declared symbols
    if( Forwards_Names <> nil ) then
    begin
        while( Forwards_Names.Count > 0 ) do
        begin
            Resolved := False ;
            for Loop := Forwards_Names.Count - 1 downto 0 do
            begin
                UEC := Evaluate( PChar( Forwards_Values[ Loop ] ), Value ) ;
                if( UEC.Code = 0 ) then
                begin
                    Resolved := True ;
                    new( P ) ;
                    fillchar( P^, sizeof( P^ ), 0 ) ;
                    P^.Address := Value ;
                    P^.Typ := 1 ; // Integer
                    P^.Flags := 1 ; // Constant
                    P^.Data := Value ; // Value
                    Add_Symbol( PChar( Forwards_Values[ Loop ] ), P ) ;
                    Forwards_Names.Delete( Loop ) ;
                    Forwards_Values.Delete( Loop ) ;
                end ;
            end ;
            if( not Resolved ) then // Nothing resolved in this loop
            begin
                for Loop := 0 to Forwards_Names.Count - 1 do
                begin
                    Status.Log_Error( PChar( 'Could not resolve symbol ' + Forwards_Names[ Loop ] ), nil, Tokens.Token_Line, Severity_Error ) ;
                end ;
                Forwards_Names.Clear ;
                Forwards_Values.Clear ;
            end ;
        end ; // while( Forwards_Names.Count > 0 )
    end ;

    // Now backpatch
    if( Backpatch_Table <> nil ) then
    begin
        for Loop := 0 to Backpatch_Table.Count - 1 do
        begin
            UEC := Evaluate( PChar( Backpatch_Table[ Loop ] ), Value ) ;
            if( UEC.Code <> 0 ) then
            begin
                P := new( PSymbol_Record ) ;
                Add_Symbol( PChar( Backpatch_Table[ Loop ] ), P ) ; // So we don't report this symbol more than once
                Status.Log_Error( PChar( 'Undefined symbol ' + Backpatch_Table[ Loop ] + ' referenced' ), nil, Tokens.Token_Line, Severity_Error ) ;
            end ;
            P := PSymbol_Record( Backpatch_Table.Objects[ Loop ] ) ;
            Stream.Seek( P^.Address ) ;
            Stream.Write( Value, P^.Size ) ;
            if( not In_Range( P^.Size, Value ) ) then
            begin
                Status.Log_Error( PChar( 'Value truncated for symbol ' + Backpatch_Table[ Loop ] ), nil, Tokens.Token_Line, Severity_Warning ) ;
            end ;
            freemem( P ) ;
            Backpatch_Table.Objects[ Loop ] := nil ;
        end ;
        Backpatch_Table.Free ;
        Backpatch_Table := nil ;
    end ;
end ; // TCEF_Assembler.Backpatch


{ This method is used during assmebly by a TAssembler to insert assembly
  source in-place.  For instance, translating a directive into a
  standard directive, or doing macro expansions, etc. }
function TCEF_Assembler.Expand( Source : PChar ; var Res : PChar ;
    var Res_Length : longint ; Status : TAssembler_Status ) : TUEC ;

var Input_Buffer : string ;

    function Get_Token : string ;

    var C : char ;
        Dummy : integer ;

    begin
        while( copy( Input_Buffer, 1, 1 ) = ' ' ) do
        begin
            Input_Buffer := copy( Input_Buffer, 2, length( Input_Buffer ) ) ;
        end ;
        if( length( Input_Buffer ) > 0 ) then
        begin
            C := Input_Buffer[ 1 ] ;
        end else
        begin
            C := ' ' ;
        end ;
        if( ( C = '"' ) or ( C = #39 ) ) then
        begin
            Dummy := Instr( 2, Input_Buffer + C, C ) ;
            Result := copy( Input_Buffer, 1, Dummy ) ;
            Input_Buffer := copy( Input_Buffer, Dummy + 1, length( Input_Buffer ) ) ;
            exit ;
        end ; // if( ( C = '"' ) or ( C = #39 ) )
        for Dummy := 1 to length( Input_Buffer ) do
        begin
            if( pos( Input_Buffer[ Dummy ], 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789$_.' ) = 0 ) then
            begin
                if( Dummy = 1 ) then
                begin
                    Result := Input_Buffer[ 1 ] ;
                    Input_Buffer := copy( Input_Buffer, 2, length( Input_Buffer ) ) ;
                end else
                begin
                    Result := copy( Input_Buffer, 1, Dummy - 1 ) ;
                    Input_Buffer := Edit( copy( Input_Buffer, Dummy, length( Input_Buffer ) ), 32 ) ;
                end ;
                exit ;
            end ;
        end ; // for Dummy := 1 to length( Input_Buffer )
        Result := Input_Buffer ;
        Input_Buffer := '' ;
    end ; // Get_Token


    function Peek_Token : string ;

    begin
        Result := Get_Token ;
        if( length( Result ) > 0 ) then
        begin
            Input_Buffer := Result + ' ' + Input_Buffer ;
        end ;
    end ;


var AA : string ;
    Any_Code : boolean ;
    B : integer ;
    Ex : int64 ;
    Next, S, Value : string ;
    P : pSymbol_Record ;
    UEC : TUEC ;

begin // TCEF_Assembler.Expand
    // Setup...
    Res := nil ;
    Res_Length := 0 ;
    fillchar( Result, sizeof( Result ), 0 ) ;

    if( Source = nil ) then
    begin
        Input_Buffer := _Grab_Line( True ) ;
    end else
    begin
        Input_Buffer := Source ;
    end ;
    Value := Input_Buffer ;
    Input_Buffer := Edit( Parse_Parameter( ';', Value ), 16 or 128 or 256 ) ;
    if( length( Input_Buffer ) = 0 ) then
    begin
	exit ;
    end ;
    Value := Get_Token ;
    Next := Peek_Token ; // Peek at the next token on the line

    // Handle standard directives...
    if( ( Next = 'EQU' ) or ( Next = '=' ) ) then
    begin
        S := Get_Token ; // Get "EQU"
        if( length( Input_Buffer ) = 0 ) then
        begin
            Result.Facility := CEFAssemblerErr_Facility ;
            Result.Code := CEFAssemblerErr_Missing_Value ;
            Status.Log_Error( PChar( 'Missing value' ), nil, Tokens.Token_Line, Severity_Error ) ;
            exit ;
        end ;
	UEC := Evaluate( PChar( Input_Buffer ), Ex ) ; // Convert to numeric value
        if( UEC.Code = CEFAssemblerErr_Undefined_Symbol ) then
	begin
            if( Forward_Declare( Value, Input_Buffer ) ) then
            begin
                Status.Log_Error( PChar( 'Symbol redeclared: ' + Value ), nil, Tokens.Token_Line, Severity_Error ) ;
                Result.Facility := CEFAssemblerErr_Facility ;
                Result.Code := CEFAssemblerErr_Multiply_Defined ;
            end ;
	    exit ;
        end else
	if( UEC.Code <> 0 ) then
	begin
            Result := UEC ;
            Status.Log_Error( PChar( ET( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
	    exit ;
	end ;
        new( P ) ;
        fillchar( P^, sizeof( P^ ), 0 ) ;
        P^.Address := Ex ;
        P^.Typ := 1 ; // Integer
        P^.Flags := 1 ; // Constant
        P^.Data := Ex ; // Value
        UEC := Add_Symbol( PChar( Value ), P ) ;
        if( UEC.Code <> 0 ) then
        begin
            Result := UEC ;
            Status.Log_Error( PChar( Assembler_Context.Symbol_Table.ERT( UEC.Code ) ), nil, Tokens.Token_Line, Severity_Error ) ;
        end ;
	exit ;
    end ; // if( Next = 'EQU' )
    Any_Code := False ;

    { Handle directives with parameters }
    if( Value = 'ORG' ) then { PC placement }
    begin
        Value := Input_Buffer ;
        B := Pos( ' ', Value + ' ' ) ; { Find end of op-code }
        Aa := Copy( Value, 1, B - 1 ) ; { Get operand }
        Fedit( AA, AA, 511 ) ;
        if( Length( AA ) = 0 ) then
        begin
            Err := True ;
            Value := '' ;
            exit ;
        end ;
        UEC := Evaluate( PChar( AA ), Ex ) ;
        if( UEC.Code <> 0 ) then
        begin
            exit ;
        end ;
        Pc := Ex ;
        Altered_PC := True ;
        Altered_Address := PC ;
        exit ;
    end ;

    if( Next = ':' ) then
    begin
        Value := Value + Input_Buffer ;
    end else
    begin
        Value := Value + ' ' + Input_Buffer ;
    end ;
    if( Handle_Directives( Value, Any_Code, Status ) ) then { If a directive }
    begin
        Temp_Expand := Value ;
        Res := PChar( Temp_Expand ) ;
        Res_Length := length( Value ) ;
        exit ;
    end ;
    Result.Facility := CEFAssemblerErr_Facility ;
    Result.Code := CEFAssemblerErr_Illegal_Instruction ;
end ; // TCEF_Assembler.Expand


function TCEF_Assembler.Facility_Code : longint ;

begin
    Result := CEFAssemblerErr_Facility ;
end ;


procedure TCEF_Assembler.In_Line( Input : TCOM_Stream ) ;

begin
    Enter_State( State_Inline, '', nil, Input ) ;
end ;


{ Logs an assembly error.  Severity indicates the seriousness of the
  error:
                0 = Informational
                1 = Warning
                2 = Error
                3 = Fatal error (abort assembly) }
procedure TCEF_Assembler.Log_Error( Text : PChar ; Severity : longint ) ;

begin
    if( In_Assembly_Status = nil ) then
    begin
        ShowMessage( Text + ' at line ' + Num1( Tokens.Token_Line ) ) ;
    end else
    begin
        In_Assembly_Status.Log_Error( PChar( Text ), nil, Tokens.Token_Line, Severity ) ;
    end ;
end ;


var Static_Get_Token : string ;

function TCEF_Assembler.Get_Token : PChar ;

begin
    Static_Get_Token := _Get_Token ;
    Get_Token := pchar( Static_Get_Token ) ;
end ;


procedure TCEF_Assembler.Put_Token( Token : PChar ) ;

begin
    _Put_Token( string( Token ) ) ;
end ;


function TCEF_Assembler.Peek_Token( Same_Line : boolean ) : PChar ;

begin
    if( Same_Line and _Token_EOL ) then
    begin
        Static_Get_Token := ''
    end else
    begin
        Static_Get_Token := _Get_Token ;
        _Put_Token( Static_Get_Token ) ;
    end ;
    Peek_Token := pchar( Static_Get_Token ) ;
end ;


{ Begins a sub-scope for identifiers within the current scope. }
procedure TCEF_Assembler.Push_Scope ;

begin
    Assembler_Context.Symbol_Table.Push_Level ;
end ;


{ Ends the current sub-scope for identifiers. }
procedure TCEF_Assembler.Pop_Scope ;

begin
    Assembler_Context.Symbol_Table.Pop_Level ;
end ;


{ Returns the symbol record for the identifier in Name.  Returns NULL,
  if Name is undeclared. }
function TCEF_Assembler.Get_Symbol( Name : PChar ) : PSymbol_Record ;

var Addr : int64 ;
    Flag, DT, Size : integer ;
    P : pointer ;

begin
    if( Assembler_Context.Symbol_Table.Find( Name, Addr, Flag, DT, Size, P ) = 0 ) then
    begin
        Result := P ;
    end else
    begin
        Result := nil ;
    end ;
end ;


{ Adds a symbol to the current scope, with the name in Name and the
  specified information. }
function TCEF_Assembler.Add_Symbol( Name : PChar ; P : pSymbol_Record ) : TUEC ;

begin
    Assembler_Context.Symbol_Table.Add( Name, 0, 0, 0, 0, P ) ;
end ;


{ Sets case sensitivity on identifier names. }
procedure TCEF_Assembler.Set_Case_Sensitive( Value : boolean ) ;

begin
    Assembler_Context.Symbol_Table.Case_Sensitive := Value ;
end ;


{ Returns True if case sensitivity to identifier names is on. }
function TCEF_Assembler.Get_Case_Sensitive : boolean ;

begin
    Result := Assembler_Context.Symbol_Table.Case_Sensitive ;
end ;


function TCEF_Assembler.Grab_Line( Current : boolean ) : PChar ;

begin
    Static_Get_Token := _Grab_Line( Current ) ; // Get value
    Result := PChar( Static_Get_Token ) ;
end ;


procedure TCEF_Assembler.Map( Address : int64 ) ;

var P : PChar ;
    S : string ;

begin
    if( In_Assembly_Status = nil ) then
    begin
        exit ;
    end ;
    P := In_Assembly_Status.Filename ;
    if( P = nil ) then
    begin
        exit ;
    end ;
    S := string( P ) ;
    if( length( S ) > 0 ) then
    begin
        Last_Mapping_Index := Assembler_Context.Add_Mapping( S, Address, Tokens.Token_Line ) ;
    end else
    begin
        Last_Mapping_Index := -1 ;
    end ;
end ;


procedure TCEF_Assembler.UnMap ;

begin
    if( Last_Mapping_Index <> -1 ) then
    begin
        Assembler_Context.Delete_Mapping( Last_Mapping_Index ) ;
        Last_Mapping_Index := -1 ;
    end ;
end ;



// TState methods...

destructor TState.Destroy ;

begin
    List.Free ;
    List := nil ;
    Tokens.Free ;
    Tokens := nil ;

    inherited Destroy ;
end ;


end.
