{$N+}
{
        Program Name : VT_05
        Package Name : CEF
        Purpose      : Simple video terminal CEF component
        Institution  : 
        Date Written : 10-Dec-2006
        Written By   : Alan Conroy
        Version      : 1.0A

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        Released to the public domain.

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        * 20-Jan-2007    EAC         Handle shutdown better.        *
        *                            Add flow-control.  Emulate     *
        *                            DEV VT05 and VT50.             *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit implements the Processor Technology SOL-20 component for
        CEF.
}

unit VT_05 ;

interface

{$I EDEFINES.INC}

uses // Borland...
     Classes,

     // C&C...
     Compatib,
     Collect, { TCollection }
     _DebugIn, // TDebug_Interface
     DebugInt, // TText_Debugger
     _Streams, // TCOM_Stream
     _UEHDefs, { TUEC }

     // CEF...
     CEF, { TMemory }
     _CEFUtil, // TCEF_Watchpoint
     Serial_Cable, // TSerial_Cable

     // VT05...
     VT05_Main ; // TMain_Form

const VT05_Facility = 16 ;
      VT05_Success = 0 ;
      VT05_Invalid_Range = 1 ;
      VT05_Component_Not_Found = 2 ;
      VT05_Address_Out_Of_Range = 3 ;
      VT05_Invalid_Component = 4 ;
      VT05_No_Matching_Watchpoint = 5 ;
      VT05_Access_Violation = 6 ;

type TVT05_Cable = class ;

     TVT05 = class( TSerial_Cable )
	private // Instance data...
            Main_Form : TMain_Form ;
            _Notifications : TList ;

        private // Internal utility routines...
            function Set_Error( Code : integer ) : TUEC ;

        public { API... }
            _Cable : TVT05_Cable ;

            function Facility_Code : longint ; override ;

            function Initialize( UI : TUI_Interface ) : TUEC ; override ;

            function Terminate : TUEC ; override ;

            function Add_Notification( Component : TComponent ) : TUEC ;
                override ;

            function Cable : TCable ; override ;

            procedure Child_Notification( Child : TComponent ;
                var Notice : longint ; var Params : int64 ) ; override ;

            function Clear_Watchpoint( Address : int64 ; Memory : boolean ;
                Access : longint ) : TUEC ; override ;

            function Component_Type : longint ; override ;

            function Connect_Input( Component : TComponent ) : TUEC ; override ;

            function Connect_Output( Component : TComponent ) : TUEC ; override ;

            function Debugger : TDebug_Interface ; override ;

            function Delete_Notification( Component : TComponent ) : TUEC ;
                override ;

            function Deposit( Address : int64 ; Size : longint ; Buffer : pointer ;
                Memory : boolean ) : TUEC ; override ;

            function Examine( Address : int64 ; var Size : longint ;
                Buffer : pointer ; Memory : boolean ) : TUEC ; override ;

            function Get_Access_Mode( Address : int64 ;
                Memory : boolean ) : longint ; override ;

            function Get_Profiling : boolean ; override ;

            function Get_Read_Latency : longint ; override ;

            function Get_Write_Latency : longint ; override ;

            function Name : PChar ; override ;

            function Read( Address : int64 ; Size : longint ;
                IO_Type : longint ) : boolean ; override ;

            function Restore_Contents( Stream : TCOM_Stream ) : TUEC ; override ;

            function Restore_State( Stream : TCOM_Stream ) : TUEC ; override ;

            function Save_Contents( Stream : TCOM_Stream ) : TUEC ; override ;

            function Save_State( Stream : TCOM_Stream ) : TUEC ; override ;

            function Set_Access_Mode( Low, High : int64 ; Memory : boolean ;
                Typ : longint ) : TUEC ; override ;

            procedure Set_Profiling( _On, Children : boolean ) ;  override ;

            procedure Set_Read_Latency( Value : longint ) ; override ;

            function Set_Watchpoint( Address : int64 ; Memory : boolean ;
                Access : longint ) : TUEC ; override ;

            procedure Set_Write_Latency( Value : longint ) ; override ;

            procedure Show_Status ; override ;

            function Write( Address : int64 ; Value, Size : longint ;
                IO_Type : longint ) : TUEC ; override ;

            procedure Set_Up( P : PChar ) ; override ;
     end ; // TVT05


     TVT05_Cable = class( TSerial_Cable_Cable )
                       private // Instance data...
                           Main_Form : TMain_Form ;

                       public // API...
                            procedure Receive( Source : TComponent ; Speed : int64 ;
                                Value, Data_Size, Stop_Bits : longint ) ;
                                override ;
                   end ; // TVT05_Cable

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses // Borland...
     Windows, // MessageBeep
     Forms, // Application
     Sysutils, // Allocmem

     // C&C...
     CommonUt, // Edit
     CVT, // CVTB
     ErrMastr, // Register_Error_Interface
     Num1s,
     Parse ; // TString_Parser


{ TVT05_Debugger methods... }

type TVT05_Debugger = class( TText_Debugger )
                                private
                                    _Memory : TVT05 ;

                                public { API... }
                                    function Child( Ordinal : longint ) : PDebug_Interface ;
                                        override ;

                                    function Count : longint ; override ;

                                    property Memory : TVT05
                                                 read _Memory
                                                 write _Memory ;
                            end ;

function TVT05_Debugger.Child( Ordinal : longint ) : PDebug_Interface ;

    function Pointer_To_String( P : pointer ) : string ;

    begin
        if( P = nil ) then
        begin
            Result := 'nil' ;
        end else
        begin
            Result := Num1( longint( P ) ) ;
        end ;
    end ;


    function Boolean_To_String( B : boolean ) : string ;

    begin
        if( B ) then
        begin
            Result := 'True' ;
        end else
        begin
            Result := 'False' ;
        end ;
    end ;


    function Access_Mode_To_String( AM : integer ) : string ;

    begin
        if( ( AM and Access_All ) = Access_All ) then
        begin
            Result := 'Access_All' ;
            exit ;
        end ;
        if( ( AM and Access_RW ) = Access_RW ) then
        begin
            Result := 'Access_RW' ;
            exit ;
        end ;
        if( AM = 0 ) then
        begin
            Result := 'Access_None' ;
            exit ;
        end ;
        if( ( AM and Access_Read ) = Access_Read ) then
        begin
            Result := 'Access_Read' ;
        end ;
        if( ( AM and Access_Write ) = Access_Write ) then
        begin
            if( Result <> '' ) then
            begin
                Result := Result + ', ' ;
            end ;
            Result := Result + 'Access_Write' ;
        end ;
        if( ( AM and Access_Execute ) = Access_Execute ) then
        begin
            if( Result <> '' ) then
            begin
                Result := Result + ', ' ;
            end ;
            Result := Result + 'Access_Execute' ;
        end ;
        if( Result = '' ) then
        begin
            Result := '* unknown *' ;
        end ;
    end ;


var I : TText_Debugger ;

begin
    I := TText_Debugger.Create ;
    Result := PDebug_Interface( I ) ;
end ;


function TVT05_Debugger.Count : longint ;

begin
    Result := 0 ;
end ;


{ TVT05 methods... }

function TVT05.Set_Error( Code : integer ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := Code ;
end ;


{ API... }

function TVT05.Facility_Code : longint ;

begin
    Facility_Code := VT05_Facility ;
end ;


function TVT05.Initialize( UI : TUI_Interface ) : TUEC ;

begin
    if( Main_Form = nil ) then // Not already initialized
    begin
        Result := inherited Initialize( UI ) ;

        { General setup... }
        Application.CreateForm( TMain_Form, Main_Form ) ;
        Main_Form._VT05 := self ;
        Main_Form._UI := UI ;
        try
            Main_Form.Visible := True ;
        except
        end ;
        Set_Signal( 'DTR', True ) ;

        _Cable := TVT05_Cable.Create ;
        _Cable.Parent := self ;
        _Cable.Main_Form := Main_Form ;

        _Notifications := TList.Create ;
    end ;
    Initialize := Set_Error( VT05_Success ) ;
end ; { TVT05.Initialize }


function TVT05.Terminate : TUEC ;

var Loop : integer ;
    Notice : longint ;
    P : int64 ;

begin
    if( Parent <> nil ) then
    begin
        Notice := Child_Notice_Request_Terminate ;
        Parent.Child_Notification( self, Notice, P ) ;
        if( Notice = 0 ) then // Parent denied termination
        begin
            exit ;
        end ;
    end ;
    if( Parent <> nil ) then
    begin
        Notice := Child_Notice_Terminating ;
        Parent.Child_Notification( self, Notice, P ) ;
    end ;
    for Loop := 0 to _Notifications.Count - 1 do
    begin
        Notice := Child_Notice_Terminating ;
        TComponent( _Notifications[ Loop ] ).Child_Notification( self, Notice, P ) ;
    end ;
    _Notifications.Free ;
    _Notifications := nil ;
    Main_Form.ParentWindow := 0 ;
    Main_Form.Timer.Enabled := False ;
    Main_Form.Close ;
    Main_Form := nil ;
    Result := inherited Terminate ;
end ; // TVT05.Terminate


function TVT05.Add_Notification( Component : TComponent ) : TUEC ;

begin
    if( Component = nil ) then
    begin
        Result := Set_Error( VT05_Invalid_Component ) ;
        exit ;
    end ;
    if( _Notifications.Indexof( Component ) = -1 ) then
    begin
        _Notifications.Add( Component ) ;
    end ;
end ;


function TVT05.Cable : TCable ;

begin
    Result := _Cable ;
end ;


procedure TVT05.Child_Notification( Child : TComponent ; var Notice : longint ;
    var Params : int64 ) ;

begin
    if( Notice = Child_Notice_Terminating ) then
    begin
        if( Child = Main_Form.Printer_Component ) then
        begin
            Main_Form.Printer_Component.Disconnect_Input( Child ) ;
            Main_Form.Printer_Component := nil ;
        end else
        if( Child = Main_Form.Cable_Component ) then
        begin
            Disconnect_Input( Child ) ;
            Main_Form.Cable_Component := nil ;
        end ;
    end ;
end ;


function TVT05.Clear_Watchpoint( Address : int64 ; Memory : boolean ;
            Access : longint ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


function TVT05.Component_Type : longint ;

begin
    Component_Type := Component_Type_Cable ;
end ;


function TVT05.Connect_Input( Component : TComponent ) : TUEC ;

begin
    if( Component = nil ) then
    begin
        Connect_Input := Set_Error( VT05_Invalid_Component ) ;
        exit ;
    end ;
    inherited Connect_Input( Component ) ;
    Set_Signal( 'DTR', True ) ;
    Connect_Input := Set_Error( VT05_Success ) ;
end ;


function TVT05.Connect_Output( Component : TComponent ) : TUEC ;

begin
    if( Component = nil ) then
    begin
        Connect_Output := Set_Error( VT05_Invalid_Component ) ;
        exit ;
    end ;
    inherited Connect_Output( Component ) ;
    Set_Signal( 'DTR', True ) ;
    Connect_Output := Set_Error( VT05_Success ) ;
end ;


function TVT05.Debugger : TDebug_Interface ;

begin
    Result := TVT05_Debugger.Create ;
    TVT05_Debugger( Result ).Memory := Self ;
end ;


function TVT05.Delete_Notification( Component : TComponent ) : TUEC ;

begin
    if( Component = nil ) then
    begin
        Result := Set_Error( VT05_Invalid_Component ) ;
        exit ;
    end ;
    if( _Notifications.Indexof( Component ) = -1 ) then
    begin
        Result.Facility := Facility_Code ;
        Result.Code := ComponentErr_Component_Not_Found ;
    end else
    begin
        _Notifications.Remove( Component ) ;
    end ;
end ;


function TVT05.Deposit( Address : int64 ; Size : longint ;
    Buffer : pointer ; Memory : boolean ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


function TVT05.Examine( Address : int64 ; var Size : longint ;
    Buffer : pointer ; Memory : boolean ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


function TVT05.Get_Access_Mode( Address : int64 ;
            Memory : boolean ) : longint ;

begin
    Result := 0 ;
end ;


function TVT05.Get_Profiling : boolean ;

begin
    Get_Profiling := False ;
end ;


function TVT05.Get_Read_Latency : longint ;

begin
    Get_Read_Latency := 0 ;
end ;


function TVT05.Get_Write_Latency : longint ;

begin
    Get_Write_Latency := 0 ;
end ;


const _Name : PChar = 'VT05'#0 ;

function TVT05.Name : PChar ;

begin
    Name := _Name ;
end ;


function TVT05.Read( Address : int64 ; Size : longint ;
    IO_Type : longint ) : boolean ;

begin
    Result := False ;
end ; { TVT05.Read }


function TVT05.Restore_Contents( Stream : TCOM_Stream ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


function TVT05.Restore_State( Stream : TCOM_Stream ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


function TVT05.Save_Contents( Stream : TCOM_Stream ): TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


function TVT05.Save_State( Stream : TCOM_Stream ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


function TVT05.Set_Access_Mode( Low, High : int64 ; Memory : boolean ;
    Typ : longint ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


procedure TVT05.Set_Profiling( _On, Children : boolean ) ;

begin
    // This routine left intentionally blank
end ;


procedure TVT05.Set_Read_Latency( Value : longint ) ;

begin
    { This routine intentionally left blank - no status to show }
end ;


function TVT05.Set_Watchpoint( Address : int64 ; Memory : boolean ;
            Access : longint ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


procedure TVT05.Set_Write_Latency( Value : longint ) ;

begin
    { This routine intentionally left blank - no status to show }
end ;


procedure TVT05.Show_Status ;

begin
    { This routine intentionally left blank - no status to show }
end ;


function TVT05.Write( Address : int64 ; Value, Size : longint ;
    IO_Type : longint ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ; // TVT05.Write


procedure TVT05.Set_Up( P : PChar ) ;

var Dummy : integer ;
    Parser : TString_Parser ;
    S, S1 : string ;

begin
    Parser := TString_Parser.Create ;
    Parser.Set_Source( string( P ) ) ;
    S := uppercase( Parser.Token ) ;
    while( S <> '' ) do
    begin
        if( S = '8_BIT' ) then
        begin
            Main_Form._8_Bit_Mode := True ;
            Main_Form.N8bitdisplay1.Checked := True ;
        end else
        if( S = 'AUTO_COPY' ) then
        begin
            Main_Form.Auto_Copy := True ;
            Main_Form.Autocopy1.Checked := True
        end else
        if( S = 'BAUD' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = 'AUTO' ) then
            begin
                S := '0' ;
            end ;
            try
                Main_Form.Baud := strtoint( S ) ;
                case strtoint( S ) of
                    0 : Main_Form.Auto1.Checked := True ;
                    75 : Main_Form.N751.Checked := True ;
                    110 : Main_Form.N1101.Checked := True ;
                    150 : Main_Form.N1501.Checked := True ;
                    300 : Main_Form.N3001.Checked := True ;
                    600 : Main_Form.N6001.Checked := True ;
                    1200 : Main_Form.N12001.Checked := True ;
                    2400 : Main_Form.N24001.Checked := True ;
                    4800 : Main_Form.N48001.Checked := True ;
                    9600 : Main_Form.N96001.Checked := True ;
                end ;
            except
            end ;
        end else
        if( S = 'COLUMNS' ) then
        begin
            S := uppercase( Parser.Token ) ;
            try
                Dummy := strtoint( S ) ;
                if( Dummy > 0 ) then
                begin
                    Main_Form.Columns := Dummy ;
                end ;
            except
            end ;
        end else
        if( S = 'COLUMN' ) then
        begin
            S := uppercase( Parser.Token ) ;
            try
                Main_Form.Cursor_Column := strtoint( S ) ;
            except
            end ;
        end else
        if( S = 'CONTROL' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = 'BLANK' ) then
            begin
                Main_Form._Blank_Control := True ;
            end else
            if( S = 'IGNORE' ) then
            begin
                Main_Form._Ignore_Control := True ;
            end else
            if( S = 'NOBLANK' ) then
            begin
                Main_Form._Blank_Control := False ;
            end else
            if( S = 'NOIGNORE' ) then
            begin
                Main_Form._Ignore_Control := False ;
            end ;
        end else
        if( S = 'CURSOR' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = 'NONE' ) then
            begin
                Main_Form._Cursor_Mode := 0 ;
            end else
            if( S = 'BLINK' ) then
            begin
                Main_Form._Cursor_Mode := 1 ;
            end else
            begin
                Main_Form._Cursor_Mode := 2 ;
            end ;
        end else
        if( S = 'DATA_BITS' ) then
        begin
            S := uppercase( Parser.Token ) ;
            try
                Main_Form.Data_Bits := strtoint( S ) ;
            except
            end ;
        end else
        if( S = 'EMULATION' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = 'TTY' ) then
            begin
                Main_Form.Set_Mode( VTM_TTY ) ;
            end else
            if( S = 'VT05' ) then
            begin
                Main_Form.Set_Mode( VTM_VT05 ) ;
            end else
            if( S = 'VT50' ) then
            begin
                Main_Form.Set_Mode( VTM_VT50 ) ;
            end ;
        end else
        if( S = 'FLOW' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = 'NONE' ) then
            begin
                Main_Form.None1.Checked := True ;
            end else
            if( S = 'XOFF' ) then
            begin
                Main_Form.XOnXOff1.Checked := True ;
            end else
            if( S = 'CTS' ) then
            begin
                Main_Form.CTSRTS1.Checked := True ;
            end ;
        end else
        if( S = 'FONT' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = '/WINDOWS' ) then
            begin
                Main_Form.Load_Character_Set( '' ) ;
                S := uppercase( Parser.Token ) ;
                Main_Form.Default_Font := S ;
            end else
            begin
                Main_Form.Load_Character_Set( S ) ;
            end ;
        end else
        if( S = 'KEYBOARD' ) then
        begin
            S := uppercase( Parser.Token ) ;
            Main_Form.Load_Keyboard( S, True ) ;
        end else
        if( S = 'LOG' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = '/APPEND' ) then
            begin
                S := uppercase( Parser.Token ) ;
                assignfile( Main_Form.Log_File, S ) ;
                {$I-}
                System.reset( Main_Form.Log_File, 1 ) ;
                {$I-}
                if( IOResult = 0 ) then
                begin
                    {$I-}
                    seek( Main_Form.Log_File, filesize( Main_Form.Log_File ) ) ;
                    {$I-}
                end ;
            end else
            begin
                assignfile( Main_Form.Log_File, S ) ;
                {$I-}
                rewrite( Main_Form.Log_File, 1 ) ;
                {$I-}
                IOResult ;
            end ;
        end else
        if( S = 'MAP' ) then
        begin
            S := uppercase( Parser.Token ) ;
            S1 := uppercase( Parser.Token ) ;
            if( S1 = '""' ) then
            begin
                S1 := '' ;
            end ;
            Main_Form.Key_Mapper.Set_Mapping( PChar( S ), PChar( S1 ) ) ;
        end else
        if( S = 'MARGIN' ) then
        begin
            S := uppercase( Parser.Token ) ;
            try
                Main_Form.Margin := strtoint( S ) ;
            except
            end ;
        end else
        if( S = 'MARGIN_BELL' ) then
        begin
            Main_Form._Margin_Bell := True ;
        end else
        if( S = 'MIXEDCASE' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = 'IN' ) then
            begin
                Main_Form._Uppercase_In := False ;
            end else
            begin
                Main_Form._Uppercase_Out := False ;
            end ;
        end ;
        if( S = 'NO8_BIT' ) then
        begin
            Main_Form._8_Bit_Mode := False ;
            Main_Form.N8bitdisplay1.Checked := False ;
        end else
        if( S = 'NOAUTO_COPY' ) then
        begin
            Main_Form.Auto_Copy := False ;
            Main_Form.Autocopy1.Checked := False ;
        end else
        if( S = 'NOKEYBOARD' ) then
        begin
            if( Main_Form.Panel2.Visible ) then
            begin
                Main_Form.Keyboard1Click( nil ) ;
            end ;
        end else
        if( S = 'NORMAL' ) then
        begin
            Main_Form._Invert := False ;
        end else
        if( S = 'NOMARGIN_BELL' ) then
        begin
            Main_Form._Margin_Bell := False ;
        end else
        if( S = 'NOSCROLL' ) then
        begin
            S := uppercase( Parser.Token ) ;
            Main_Form._Scroll := False ;
        end else
        if( S = 'NOWRAP' ) then
        begin
            Main_Form._Line_Wrap := False ;
        end else
        if( S = 'PPBAUD' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = 'AUTO' ) then
            begin
                S := '0' ;
            end ;
            try
                Main_Form.Printer_Baud := strtoint( S ) ;
                case strtoint( S ) of
                    0 : Main_Form.Auto2.Checked := True ;
                    75 : Main_Form.N752.Checked := True ;
                    110 : Main_Form.N1102.Checked := True ;
                    150 : Main_Form.N1502.Checked := True ;
                    300 : Main_Form.N3002.Checked := True ;
                    600 : Main_Form.N6002.Checked := True ;
                    1200 : Main_Form.N12002.Checked := True ;
                    2400 : Main_Form.N24002.Checked := True ;
                    4800 : Main_Form.N48002.Checked := True ;
                    9600 : Main_Form.N96002.Checked := True ;
                end ;
            except
            end ;
        end else
        if( S = 'PPDATA_BITS' ) then
        begin
            S := uppercase( Parser.Token ) ;
            try
                Main_Form.Printer_Data_Bits := strtoint( S ) ;
            except
            end ;
        end else
        if( S = 'PPFLOW' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = 'NONE' ) then
            begin
                Main_Form.None2.Checked := True ;
            end else
            if( S = 'XOFF' ) then
            begin
                Main_Form.XOnXOff2.Checked := True ;
            end else
            if( S = 'CTS' ) then
            begin
                Main_Form.CTSRTS2.Checked := True ;
            end ;
        end else
        if( S = 'PPSTOP' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = '1' ) then
            begin
                Main_Form.N2stopbits2.Checked := False ;
            end else
            begin
                Main_Form.N2stopbits2.Checked := True ;
            end ;
        end else
        if( S = 'PRINTER' ) then
        begin
            S := uppercase( Parser.Token ) ;
            Main_Form.Load_Printer( S, False ) ;
        end else
        if( S = 'RATE' ) then
        begin
            S := uppercase( Parser.Token ) ;
            try
                Main_Form.Cursor_Rate := strtoint( S ) ;
            except
            end ;
        end else
        if( S = 'ROW' ) then
        begin
            S := uppercase( Parser.Token ) ;
            try
                Main_Form.Cursor_Row := strtoint( S ) ;
            except
            end ;
        end else
        if( S = 'ROWS' ) then
        begin
            S := uppercase( Parser.Token ) ;
            try
                Dummy := strtoint( S ) ;
                if( Dummy > 0 ) then
                begin
                    Main_Form.Rows := Dummy ;
                end ;
            except
            end ;
        end else
        if( S = 'SCROLL' ) then
        begin
            S := uppercase( Parser.Token ) ;
            Main_Form._Scroll := True ;
        end else
        if( S = 'STOP' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = '1' ) then
            begin
                Main_Form.N2stopbits1.Checked := False ;
            end else
            begin
                Main_Form.N2stopbits1.Checked := True ;
            end ;
        end else
        if( S = 'UPPERCASE' ) then
        begin
            S := uppercase( Parser.Token ) ;
            if( S = 'IN' ) then
            begin
                Main_Form._Uppercase_In := True ;
            end else
            begin
                Main_Form._Uppercase_Out := True ;
            end ;
        end else
        if( S = 'WRAP' ) then
        begin
            Main_Form._Line_Wrap := True ;
        end ;

        // If we get here it is a unknown token - just ignore it and continue on
        S := uppercase( Parser.Token ) ; // Get next token
    end ; // while( S <> '' )
    Parser.Free ;
    if( Main_Form.Character_Set <> nil ) then
    begin
        Main_Form.Character_Set.Invert := Main_Form._Invert ;
    end ;
    while( Main_Form.Columns * Main_Form.Rows > 32768 ) do
    begin
        Main_Form.Rows := Main_Form.Rows - 1 ;
    end ;
    if( Main_Form.Cursor_Row > Main_Form.Rows ) then
    begin
        Main_Form.Cursor_Row := Main_Form.Rows ;
    end ;
    if( Main_Form.Cursor_Column > Main_Form.Columns ) then
    begin
        Main_Form.Cursor_Column := Main_Form.Columns ;
    end ;
end ; // TVT05.Set_Up


procedure TVT05_Cable.Receive( Source : TComponent ; Speed : int64 ;
    Value, Data_Size, Stop_Bits : longint ) ;

begin
    if( Source = Main_Form.Printer_Component ) then
    begin
        Main_Form.Receive_Printer_Char( Value ) ;
    end else
    begin
        Main_Form.Receive_Char( Value ) ;
    end ;
end ;


type tVT05_EI = class( tError_Interface )
                            public
                                function Get_Facility : longint ;
                                    override ; stdcall ;
                                function Get_Name : PChar ; override ; stdcall ;
                                function Translate_Error( Code : longint ) : PChar ;
                                    override ; stdcall ;
                                procedure Done ; override ; stdcall ;
                                function Facility_Version : longint ; override ;
                                function Version : longint ; override ;
                                function Severity( Code : longint ) : longint ;
                                    override ; stdcall ;
                                function Translate_Error_Ex( Code : longint ;
                                    var Size, Typ : longint ) : PChar ;
                                    override ; stdcall ;
                  end ;
     pVT05_EI = tVT05_EI ;


{ tVT05_EI methods... }

function tVT05_EI.Get_Facility : longint ;

begin
    Get_Facility := VT05_Facility ;
end ;


const VT05_Facility_Name : PChar = 'SOL-20' ;

function tVT05_EI.Get_Name : PChar ;

begin
    Get_Name := VT05_Facility_Name ;
end ;


var Memory_Error_Text : array[ 0..127 ] of char ;

function tVT05_EI.Translate_Error( Code : longint ) : PChar ;

var _Error : string ;

begin
    case Code of
        VT05_Success: _Error := 'Success' ;
        VT05_Invalid_Range : _Error := 'Invalid range' ;
        VT05_Component_Not_Found : _Error := 'Component not found' ;
        VT05_Address_Out_Of_Range : _Error := 'Address out of range' ;
        VT05_Invalid_Component : _Error := 'Invalid component' ;
        VT05_No_Matching_Watchpoint : _Error := 'No Matching Watchpoint' ;
        VT05_Access_Violation : _Error := 'Access Violation' ;
	else _Error := 'Unknown error number ' + Num1( Code ) ;
    end ;
    _Error := _Error + #0 ;
    move( _Error[ 1 ], Memory_Error_Text[ 0 ], length( _Error ) ) ;
    Translate_Error := @Memory_Error_Text ;
end ; { tFS_EI.Translate_Error }


procedure tVT05_EI.Done ;

begin
    { Done by unit }
end ;


function tVT05_EI.Facility_Version : longint ;

begin
    Facility_Version := 0 ;
end ;


function tVT05_EI.Version : longint ;

begin
    Version := 11 ; { Compatible with UEH V1.1 }
end ;


function tVT05_EI.Severity( Code : longint ) : longint ;

begin
    Severity := 2 ;
end ;


function tVT05_EI.Translate_Error_Ex( Code : longint ;
    var Size, Typ : longint ) : PChar ;

begin
    Translate_Error_Ex := Translate_Error( Code ) ;
end ;


var MEI : pVT05_EI ;

initialization
    { Set up the memory error interface... }
    MEI := tVT05_EI.Create ;
//    Register_Error_Interface( MEI ) ;
{$IFDEF Test}
    Test_Unit ;
{$ENDIF}

finalization
    Unregister_Error_Interface( MEI.Get_Facility ) ;
    MEI.Free ;
end.

